(PackageInfo named: 'Worlds') preamble: '"code to be run before the loading of this package"

"unload Worlds2 package if the image has it"
(Smalltalk includesKey: #WObject2)
  ifTrue: [ (MCPackage named: #Worlds2) workingCopy unload ].

"add currentWWorld IV to Process if not already there"
(Process instVarNames includes: ''currentWWorld'')
  ifFalse: [
	Link subclass: #Process
		instanceVariableNames: (Process instanceVariablesString, '' currentWWorld'')
		classVariableNames: (Process classVariablesString)
		poolDictionaries: ''''
		category: ''Kernel-Processes''
	].
'!
SystemOrganization addCategory: #Worlds!
SystemOrganization addCategory: #'Worlds-Compiler'!
SystemOrganization addCategory: #'Worlds-Tools'!
SystemOrganization addCategory: #'Worlds-CollectionsTests'!
(PackageInfo named: 'Worlds') postscript: '"below, add code to be run after the loading of this package"

WObject withAllSubclassesDo: [ :c | c compileAll ]; doneCompiling.

WWorld initialize.
'!

Inspector subclass: #WCollectionInspector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Worlds-Tools'!

!WCollectionInspector methodsFor: 'as yet unclassified' stamp: 'ael 6/27/2011 14:52'!
fieldList
	"don't log access to the collection's size"
	object ifNil: [ ^ OrderedCollection new].
	^ self baseFieldList ,
		(object unloggedSize <= (self i1 + self i2)
			ifTrue: [(1 to: object size)
						collect: [:i | i printString]]
			ifFalse: [(1 to: self i1) , (object size-(self i2-1) to: object size)
						collect: [:i | i printString]])
"
WCollection2 new inspect
(WCollection2 newFrom: #(3 5 7 123)) inspect
(WCollection2 newFrom: (1 to: 1000)) inspect
"! !

!WCollectionInspector methodsFor: 'as yet unclassified' stamp: 'ael 6/27/2011 14:52'!
replaceSelectionValue: anObject 
	"The receiver has a list of variables of its inspected object. One of these 
	is selected. The value of the selected variable is set to the value, anObject."

	(selectionIndex - 2) <= object class instSize
		ifTrue: [^ super replaceSelectionValue: anObject].
	object at: self selectedObjectIndex put: anObject! !

!WCollectionInspector methodsFor: 'as yet unclassified' stamp: 'ael 6/27/2011 14:53'!
selectedObjectIndex
	"Answer the index of the inspectee's collection that the current selection refers to."
	"Don't log access to the collection's size"
	| basicIndex |
	basicIndex := selectionIndex - 2 - object class instSize.
	^ (object unloggedSize <= (self i1 + self i2)  or: [basicIndex <= self i1])
		ifTrue: [basicIndex]
		ifFalse: [object size - (self i1 + self i2) + basicIndex]! !

!WCollectionInspector methodsFor: 'as yet unclassified' stamp: 'ael 6/27/2011 14:55'!
selection
	"The receiver has a list of variables of its inspected object.
	One of these is selected. Answer the value of the selected variable."

	(selectionIndex - 2) <= object class instSize
		ifTrue: [^ super selection].
	^ object unloggedAt: self selectedObjectIndex! !

!ClassDescription methodsFor: '*worlds-compiler' stamp: 'yo 3/31/2009 00:44'!
reallyBasicCompile: text classified: category withStamp: changeStamp notifying: requestor logSource: logSource
	| methodAndNode |
	methodAndNode _ self reallyBasicCompile: text asString notifying: requestor 
							trailer: self defaultMethodTrailer ifFail: [^nil].
	methodAndNode method: (methodAndNode method copyWithTempNames: methodAndNode node tempNames).
	logSource ifTrue: [
		self logMethodSource: text forMethodWithNode: methodAndNode 
			inCategory: category withStamp: changeStamp notifying: requestor.
	].
	self addSelectorSilently: methodAndNode selector withMethod: methodAndNode method.
	self theNonMetaClass noteCompilationOf: methodAndNode selector meta: self isMeta.
	^ methodAndNode selector! !

!ClassDescription methodsFor: '*worlds-compiler' stamp: 'yo 3/12/2010 19:57'!
removeCategorySilently: aString 
	"Remove each of the messages categorized under aString in the method 
	dictionary of the receiver. Then remove the category aString."
	| categoryName |
	categoryName _ aString asSymbol.
	(self organization listAtCategoryNamed: categoryName) do:
		[:sel | self removeSelectorSilently: sel].
! !

Parser subclass: #WParser
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Worlds-Compiler'!

!WParser methodsFor: 'as yet unclassified' stamp: 'ael 10/4/2011 15:45'!
assignment: varNode
	" var '_' expression => AssignmentNode."
	| loc start isInstVar isPoolVar |
	isInstVar := (varNode isMemberOf: MessageNode) and: [varNode selector key beginsWith: #wiv666].
	isPoolVar := (varNode isMemberOf: MessageNode) and: [varNode selector key = #nsAt:].
	isInstVar ifFalse: [
		(loc _ varNode assignmentCheck: encoder at: prevMark + requestorOffset) >= 0
			ifTrue: [^self notify: 'Cannot store into' at: loc].
	].
	start _ self startOfNextToken.
	self advance.
	self expression ifFalse: [^self expected: 'Expression'].
	isPoolVar ifTrue: [
		self error: 'you are not supposed to assign namespace variable this way.'].
	isInstVar ifTrue: [
		parseNode _ MessageNode new
				receiver: (encoder encodeVariable: 'self')
				selector: (varNode selector key, #:) asSymbol
				arguments: {parseNode}
				precedence: 3
				from: encoder
				sourceRange: (start to: hereEnd).
	] ifFalse: [
		varNode nowHasDef.
		parseNode _ AssignmentNode new
				variable: varNode
				value: parseNode
				from: encoder
				sourceRange: (start to: self endOfLastToken).
	].
	^true! !

!WParser methodsFor: 'as yet unclassified' stamp: 'sn 9/10/2013 13:24'!
parse: sourceStream class: class category: aCategory noPattern: noPattern context: aContext notifying: req ifFail: aBlock 
	| c |
	c := CompilationCue
			source: sourceStream
			context: aContext
			class: class
			category: aCategory
			requestor: req.
	^ self 
		parse: sourceStream 
		cue: c 
		noPattern: noPattern 
		ifFail: aBlock
		for: nil! !

!WParser methodsFor: 'as yet unclassified' stamp: 'sn 9/10/2013 13:25'!
parse: sourceStream class: class noPattern: noPattern context: ctxt notifying: req ifFail: aBlock 

	^ self parse: sourceStream class: class noPattern: noPattern context: ctxt notifying: req ifFail: aBlock for: nil
	
	"^self parse: sourceStream class: class category: nil noPattern: noPattern context: ctxt notifying: req ifFail: aBlock "
! !

!WParser methodsFor: 'as yet unclassified' stamp: 'ael 10/4/2011 15:45'!
parse: sourceStream class: class noPattern: noPattern context: ctxt notifying: req ifFail: aBlock for: namespaceObject

	^ self parse: sourceStream class: class category: nil noPattern: noPattern context: ctxt notifying: req ifFail: aBlock for: namespaceObject! !

!WParser methodsFor: 'as yet unclassified' stamp: 'sn 9/9/2013 13:32'!
parse: sourceStream cue: aCue noPattern: noPattern ifFail: aBlock 
	"Answer a MethodNode for the argument, sourceStream, that is the root of
	 a parse tree. Parsing is done with respect to the CompilationCue to 
	 resolve variables. Errors in parsing are reported to the cue's requestor; 
	 otherwise aBlock is evaluated. The argument noPattern is a Boolean that is
	 true if the the sourceStream does not contain a method header (i.e., for DoIts)."

	| methNode repeatNeeded myStream s p subSelection |
	myStream := sourceStream.
	[repeatNeeded := false.
	 p := myStream position.
	 s := myStream upToEnd.
	 myStream position: p.
	 subSelection := aCue requestor notNil and: [aCue requestor selectionInterval = (p + 1 to: p + s size)].
	 encoder ifNil: [self encoderClass: WEncoder. self encoder namespace: nil].
	 self encoder init: aCue notifying: self.
	 self init: myStream cue: aCue failBlock: [^ aBlock value].
	 doitFlag := noPattern.
	 failBlock:= aBlock.
	 [methNode := self method: noPattern context: cue context] 
		on: ReparseAfterSourceEditing 
		do:	[ :ex |
			repeatNeeded := true.
			myStream := subSelection
							ifTrue:
								[ReadStream
									on: cue requestor text string
									from: cue requestor selectionInterval first
									to: cue requestor selectionInterval last]
							ifFalse:
								[ReadStream on: cue requestor text string]].
	 repeatNeeded] whileTrue:
		[encoder := self encoder class new].
	methNode sourceText: s.
	^methNode! !

!WParser methodsFor: 'as yet unclassified' stamp: 'sn 9/9/2013 17:24'!
parse: sourceStream cue: aCue noPattern: noPattern ifFail: aBlock for: namespaceObject
	"Answer a MethodNode for the argument, sourceStream, that is the root of
	 a parse tree. Parsing is done with respect to the CompilationCue to 
	 resolve variables. Errors in parsing are reported to the cue's requestor; 
	 otherwise aBlock is evaluated. The argument noPattern is a Boolean that is
	 true if the the sourceStream does not contain a method header (i.e., for DoIts)."

	| methNode repeatNeeded myStream s p subSelection |
	myStream := sourceStream.
	[repeatNeeded := false.
	 p := myStream position.
	 s := myStream upToEnd.
	 myStream position: p.
	 subSelection := aCue requestor notNil and: [aCue requestor selectionInterval = (p + 1 to: p + s size)].
	 encoder ifNil: [self encoderClass: WEncoder. self encoder namespace: namespaceObject].
	 self encoder init: aCue notifying: self.
	 self init: myStream cue: aCue failBlock: [^ aBlock value].
	 doitFlag := noPattern.
	 failBlock:= aBlock.
	 [methNode := self method: noPattern context: cue context] 
		on: ReparseAfterSourceEditing 
		do:	[ :ex |
			repeatNeeded := true.
			myStream := subSelection
							ifTrue:
								[ReadStream
									on: cue requestor text string
									from: cue requestor selectionInterval first
									to: cue requestor selectionInterval last]
							ifFalse:
								[ReadStream on: cue requestor text string]].
	 repeatNeeded] whileTrue:
		[encoder := self encoder class new].
	methNode sourceText: s.
	^methNode! !

!ObjectExplorer class methodsFor: '*worlds' stamp: 'ael 5/3/2012 14:26'!
new
	"hijack all explorer creation"
	^WObjectExplorer basicNew initialize

! !

ObjectExplorer subclass: #WObjectExplorer
	instanceVariableNames: 'wworld'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Worlds-Tools'!

!WObjectExplorer methodsFor: 'worldiness' stamp: 'ael 5/3/2012 14:04'!
doItWWorld
	^wworld ifNil: [ WTopLevelWorld default ]
! !

!WObjectExplorer methodsFor: 'worldiness' stamp: 'ael 1/10/2013 17:05'!
getList

	^Array with: (ObjectExplorerWrapper with: rootObject name: 'root' model: self parent: nil)
! !

!WObjectExplorer methodsFor: 'worldiness' stamp: 'ael 5/3/2012 14:04'!
getWWorld
	"when asked by an ObjectExplorerWrapper"
	^wworld ifNil: [ WTopLevelWorld default ]
! !

!WObjectExplorer methodsFor: 'worldiness' stamp: 'ael 5/3/2012 14:05'!
label

	| base |
	base := self worldEval: [ rootObject printStringLimitedTo: 32 ].
	^wworld
	  ifNil: [ base ]
	  ifNotNil: [ '[',wworld asString,'] ', base ]

	! !

!WObjectExplorer methodsFor: 'worldiness' stamp: 'ael 5/3/2012 14:16'!
openExplorerFor: anObject
"
WObjectExplorer new openExplorerFor: Smalltalk
"

	| win thisWorld |
	(thisWorld := WWorld thisWorld) isTopLevel ifFalse: [ wworld := thisWorld ].
	win := (self explorerFor: anObject) openInWorld.
	Cursor wait showWhile:
		[win submorphs do:
			[:sm|
			(sm respondsTo: #expandRoots) ifTrue:
				[sm expandRoots]]].
	^self
! !

!WObjectExplorer methodsFor: 'as yet unclassified' stamp: 'ael 5/3/2012 14:17'!
openExplorerFor: anObject withLabel: label 
     "ObjectExplorer new openExplorerFor: Smalltalk withLabel: 'Smalltalk'"

	| thisWorld |
	(thisWorld := WWorld thisWorld) isTopLevel ifFalse: [ wworld := thisWorld ].
	(self explorerFor: anObject withLabel: label)
openInWorld! !

!WObjectExplorer methodsFor: 'worldiness' stamp: 'ael 5/3/2012 14:05'!
worldEval: aBlock
	^(wworld ifNil: [ WTopLevelWorld default ]) eval: aBlock! !

!Integer class methodsFor: '*worlds-utilities' stamp: 'ael 9/15/2011 17:01'!
bitSequence: low to: high
	"Answer an Integer with 1s in all bits from position low to high (where LSB is bit 1).
	Return zero if the result is not a meaningful positive number"
	(low < 1 or: [ high < 1 or: [ high < low ] ]) ifTrue: [ ^0 ].
	
	low = high ifTrue: [ ^1 bitShift: low - 1 ].
	
	^(1 bitShift: high) - (1 bitShift: low - 1)
	
"  (Integer bitSequence: 4 to: 8) radix: 2 "
"  (Integer bitSequence: 7 to: 7) radix: 2 "! !

!IdentityDictionary methodsFor: '*worlds' stamp: 'ael 11/14/2011 20:18'!
primTestAelAt: key ifNotFound: notFound
	<primitive: 'primLookup' module: 'WorldsPlugin'>
	^self primitiveFailed
"	^WorldsPlugin doPrimitive: 'primLookup'"
! !

!IdentityDictionary methodsFor: '*worlds' stamp: 'ael 11/11/2011 13:48'!
primTestAelAt: key put: object assocClass: assocClass
	<primitive: 'primStore' module: 'WorldsPlugin'>
	^self primitiveFailed
"	^WorldsPlugin doPrimitive: 'primStore'"
! !

!IdentityDictionary methodsFor: '*worlds' stamp: 'ael 11/11/2011 18:06'!
testAelAt: key
	^self testAelAt: key ifAbsent: [ self errorNotFound: key ]
! !

!IdentityDictionary methodsFor: '*worlds' stamp: 'ael 11/14/2011 20:20'!
testAelAt: key ifAbsent: aBlock
	| notFound result |
	notFound := #aelNotFound.
	result := self primTestAelAt: key ifNotFound: notFound.
	result == notFound ifTrue: [ ^aBlock value ].
	^result

! !

!IdentityDictionary methodsFor: '*worlds' stamp: 'ael 11/11/2011 01:07'!
testAelAt: key put: object
	^self primTestAelAt: key put: object assocClass: Association
! !

!IdentityDictionary methodsFor: '*worlds' stamp: 'ael 11/14/2011 21:37'!
testAelGoodPrime: lowerLimit
	<primitive: 'primGoodPrime' module: 'WorldsPlugin'>
	^self primitiveFailed
"	^WorldsPlugin doPrimitive: 'primGoodPrime'"
! !

!IdentityDictionary methodsFor: '*worlds' stamp: 'ael 11/14/2011 21:37'!
testAelGrow
	<primitive: 'primGrowDictionary' module: 'WorldsPlugin'>
	^self primitiveFailed
"	^WorldsPlugin doPrimitive: 'primGrowDictionary'"
! !

!Behavior methodsFor: '*worlds-compiler' stamp: 'yo 2/27/2009 16:00'!
reallyBasicCompile: code notifying: requestor trailer: bytes ifFail: failBlock
	"Compile code without logging the source in the changes file"

	| methodNode |
	methodNode _ Compiler new
				compile: code
				in: self
				notifying: requestor
				ifFail: failBlock.
	methodNode encoder requestor: requestor.
	^ CompiledMethodWithNode generateMethodFromNode: methodNode trailer: bytes.! !

!DictionaryTest methodsFor: '*worlds-private-override' stamp: 'ael 9/25/2011 22:08'!
commitIfNeeded
	"ignore"
! !

!DictionaryTest methodsFor: '*worlds-private-override' stamp: 'ael 6/21/2011 23:19'!
targetClass
  |className|

  className := self class name asText copyFrom: 0 to: self class name size - 4.
  ^ Smalltalk at: (className asString asSymbol).
! !

!DictionaryTest methodsFor: '*worlds-tests-override' stamp: 'ael 9/25/2011 22:19'!
testAdd
	
	| dict |
	dict := self targetClass new.
	dict add: #a -> 1.
	dict add: #b -> 2.
	
	self commitIfNeeded.

	self assert: (dict at: #a) = 1.
	self assert: (dict at: #b) = 2! !

!DictionaryTest methodsFor: '*worlds-tests-override' stamp: 'ael 9/25/2011 22:19'!
testAddAll
	
	| dict1 dict2 |
	dict1 := self targetClass new.
	dict1 at: #a put:1 ; at: #b put: 2. 
	dict2 := self targetClass new.
	dict2 at: #a put: 3 ; at: #c put: 4.
	dict1 addAll: dict2.
	
	self commitIfNeeded.

	self assert: (dict1 at: #a) = 3.
	self assert: (dict1 at: #b) = 2.
	self assert: (dict1 at: #c) = 4.

	dict1 := self targetClass new.
	dict1 at: #a put:1 ; at: #b put: 2. 
	dict1 addAll: {#a -> 3. #c -> 4}.
	self assert: (dict1 at: #a) = 3.
	self assert: (dict1 at: #b) = 2.
	self assert: (dict1 at: #c) = 4.! !

!DictionaryTest methodsFor: '*worlds-association tests-override' stamp: 'ael 9/25/2011 22:18'!
testAddAssociation
	"self run:#testAddAssociation"
	"self debug:#testAddAssociation"
	
	| dict |
	dict := self targetClass new.
	dict at: #a put: 1.
	dict at: #b put: 2.

	self assert: (dict at: #a) = 1.
	self assert: (dict at: #b) = 2.
	
	dict at: #a put: 10.
	dict at: #c put: 2.
	
	self commitIfNeeded.

	self assert: (dict at: #a) = 10.
	self assert: (dict at: #b) = 2.
	self assert: (dict at: #c) = 2
	
	! !

!DictionaryTest methodsFor: '*worlds-tests-override' stamp: 'ael 9/25/2011 22:17'!
testAsSet
	"Non regression test for http://bugs.squeak.org/view.php?id=7258"
	
	| aDictionary aSet assoc0 assoc1 |
	
	"Create a dictionary"
	aDictionary := self targetClass new.
	
	"Convert it to a Set"
	aSet := aDictionary asSet.

	"Add two associations to it"	
	assoc0 := #first -> 0.
	assoc1 := #first -> 1.
	aSet add: assoc0 copy; add: assoc1.

	self commitIfNeeded.
	
	"Check if the two associations were added (that should happen if they are different)"
	self
		assert: (assoc0 copy ~= assoc1) ==> (aSet size > 1)
		description:
  			'When adding two different elements, the set size should be greater than one'! !

!DictionaryTest methodsFor: '*worlds-association tests-override' stamp: 'ael 9/25/2011 22:20'!
testAssociationsSelect
	| answer d |
	d := self targetClass new.
	d at: (Array with: #hello with: #world)
	  put: #fooBar.
	d at: Smalltalk put: #'Smalltalk is the key'.
	d at: #Smalltalk put: Smalltalk.

	self commitIfNeeded.

	answer := d
				associationsSelect: [:assoc | assoc key == #Smalltalk
						and: [assoc value == Smalltalk]].
	self
		should: [answer isKindOf: self targetClass].
	self
		should: [answer size = 1].
	self
		should: [(answer at: #Smalltalk)
				== Smalltalk].
	answer := d
				associationsSelect: [:assoc | assoc key == #NoSuchKey
						and: [assoc value == #NoSuchValue]].
	self
		should: [answer isKindOf: self targetClass].
	self
		should: [answer size = 0]! !

!DictionaryTest methodsFor: '*worlds-basic tests-override' stamp: 'ael 9/25/2011 22:20'!
testAtError
	"self run: #testAtError"
	
	| dict |
	dict := self targetClass new.
	dict at: #a put: 666.
	
	self commitIfNeeded.

	self shouldnt: [ dict at: #a ] raise: KeyNotFound.
	self should: [ dict at: #b ] raise: KeyNotFound.
	
	! !

!DictionaryTest methodsFor: '*worlds-basic tests-override' stamp: 'ael 9/25/2011 22:20'!
testAtIfAbsent
	"self run: #testAtIfAbsent"
	
	| dict |
	dict := self targetClass new.
	dict at: #a put: 666.

	self commitIfNeeded.

	self assert: (dict at: #a ifAbsent: [nil]) = 666.
	
	self assert: (dict at: #b ifAbsent: [nil]) isNil.

	! !

!DictionaryTest methodsFor: '*worlds-basic tests-override' stamp: 'ael 9/25/2011 22:21'!
testAtIfPresentIfAbsent
	"Test at:ifPresent:ifAbsent:"
	| dict present absent |
	dict := self targetClass new.
	present := absent := false.
	dict at: #foo ifPresent:[:v| present := true] ifAbsent:[absent := true].
	self deny: present.
	self assert: absent.

	dict at: #foo put: #bar.
	present := absent := false.
	dict at: #foo ifPresent:[:v| present := true] ifAbsent:[absent := true].
	self assert: present.
	self deny: absent.

	self commitIfNeeded.

	present := absent := false.
	dict at: #foo ifPresent:[:v| present := true. nil] ifAbsent:[absent := true].
	self assert: present.
	self deny: absent.
! !

!DictionaryTest methodsFor: '*worlds-implementation tests-override' stamp: 'ael 9/25/2011 22:21'!
testAtNil
	"(self run: #testAtNil)"
	"nil is a valid key in squeak. In VW nil is not a valid key"
	"Ansi 1.9 p, 168
    		5.7.2.5 Message: at: key put: newElement
    		Synopsis
    			Store newElement at key in the receiver. Answer newElement.
    		Definition: <abstractDictionary>
    		If lookup succeeds for key, then newElement replaces the element previously stored at key.
    		Otherwise, the newElement is stored at the new key. In either case, subsequent successful
    		lookups for key will answer newElement.  Answer newElement.

    		The result is undefined if the key is nil.

		This clearly indicates that different smalltalks where doing different assumptions."
	
	
	
	| dict1  |
	dict1 := self targetClass new.
	self shouldnt: [ dict1 at: nil put: #none] raise: Error.

	self commitIfNeeded.

	self assert: (dict1 at: nil) = #none. 
	! !

!DictionaryTest methodsFor: '*worlds-basic tests-override' stamp: 'ael 9/25/2011 22:21'!
testAtPut
	"self run: #testAtPut"
	"self debug: #testAtPut"
	
	| adictionary |
	adictionary := self targetClass new.
	adictionary at: #a put: 3.
	self assert: (adictionary at: #a) = 3.
	adictionary at: #a put: 3.
	adictionary at: #a put: 4.
	self assert: (adictionary at: #a) = 4.
	
	self commitIfNeeded.

	adictionary at: nil put: 666.
	self assert: (adictionary at: nil) = 666! !

!DictionaryTest methodsFor: '*worlds-basic tests-override' stamp: 'ael 9/25/2011 22:21'!
testAtPutNil
	"self run: #testAtPut"
	"self debug: #testAtPut"
	
	| dict |
	dict := self targetClass new.
	dict at: nil put: 1.
	self assert: (dict at: nil) = 1.
	dict at: #a put: nil.
	self assert: (dict at: #a) = nil.
	
	self commitIfNeeded.

	dict at: nil put: nil.
	self assert: (dict at: nil) = nil.
	
	
	! !

!DictionaryTest methodsFor: '*worlds-testing-override' stamp: 'ael 9/25/2011 22:22'!
testCollect
	"Ensure that Dictionary>>collect: answers a dictionary not something else"
	| dict expected result |
	dict := self targetClass newFromPairs:{
		#first.		1.
		#second.	2.
		#third.		3.
		#fourth.	4.
		#fifth.		5.
	}.
	
	self commitIfNeeded.

	result := dict collect:[:each| each asWords].
	expected := self targetClass newFromPairs:{
		#first.		'one'.
		#second.	'two'.
		#third.		'three'.
		#fourth.	'four'.
		#fifth.		'five'.
	}.
	self assert: result = expected.! !

!DictionaryTest methodsFor: '*worlds-tests-override' stamp: 'ael 9/25/2011 22:17'!
testComma
	
	| dict1 dict2 dict3 |
	dict1 := self targetClass new.
	dict1 at: #a put:1 ; at: #b put: 2. 
	dict2 := self targetClass new.
	dict2 at: #a put: 3 ; at: #c put: 4.
	dict3 := dict1, dict2.

	self commitIfNeeded.

	self assert: (dict3 at: #a) = 3.
	self assert: (dict3 at: #b) = 2.
	self assert: (dict3 at: #c) = 4.! !

!DictionaryTest methodsFor: '*worlds-keys and value tests-override' stamp: 'ael 9/25/2011 22:22'!
testDictionaryConcatenation
	"self run: #testDictionaryConcatenation"
	
	
	| dict1 dict2 dict3 |
	dict1 := self targetClass new.
	dict1 at: #a put: 'Nicolas' ; at: #b put: 'Damien'. 
	
	dict2 := self targetClass new.
	dict2 at: #a put: 'Christophe' ; at: #c put: 'Anthony'.
	dict3 := dict1, dict2.
	
	self commitIfNeeded.
	
	self assert: (dict3 at: #a) = 'Christophe'.
	self assert: (dict3 at: #b) = 'Damien'.
	self assert: (dict3 at: #c) = 'Anthony'.
	

	! !

!DictionaryTest methodsFor: '*worlds-association tests-override' stamp: 'ael 9/25/2011 22:22'!
testIncludesAssociation
	"self run:#testIncludesAssociation"
	
	| dict |
	dict := self targetClass new.
	dict at: #a put: 1.
	dict at: #b put: 2.
	
	self commitIfNeeded.
	
	self assert: (dict includesAssociation: (#a -> 1)).
	self assert: (dict includesAssociation: (#b -> 2)).
	
	! !

!DictionaryTest methodsFor: '*worlds-association tests-override' stamp: 'ael 9/25/2011 22:22'!
testIncludesAssociationNoValue
	"self run:#testIncludesAssociationNoValue"
	"self debug:#testIncludesAssociationNoValue"
	
	| dict a1 a3 |
	a1 := Association key: #Italie.
	a3 := Association key: #France value: 'Paris'.

	self commitIfNeeded.
	
	self assert: (a1 key = #Italie).
	self assert: (a1 value isNil).
	
	dict := self targetClass new.
	dict add: a1.
	dict add: a3.
	self assert: (dict includesKey: #France).
	self assert: (dict includesKey: #Italie).
	self assert: (dict at: #Italie) isNil.
	self assert: (dict at: #France) = 'Paris'

	
	
	! !

!DictionaryTest methodsFor: '*worlds-basic tests-override' stamp: 'ael 9/25/2011 22:23'!
testIncludesKey
	"self run:#testIncludesKey"
	"self debug:#testIncludesKey"
	
	| dict a1 a2 a3 usesIdentity |
	a1 := Association key: 'Italie'.
	a2 := Association new.
	a3 := Association key: 'France' value: 'Paris'.
	
	dict := self targetClass new.
	dict add: a1 .
	dict add: a2.
	dict add: a3.

	self commitIfNeeded.

	usesIdentity := dict class name asString includesSubString: 'Identity'. 
	self assert: (usesIdentity | (dict includesKey: #France)).
	self assert: (dict includesKey: 'France').
	self assert: (usesIdentity | (dict includesKey: #Italie)).
	self assert: (dict includesKey: nil).	
		
	self assert: (dict at: 'France' ) = 'Paris'.
! !

!DictionaryTest methodsFor: '*worlds-keys and value tests-override' stamp: 'ael 9/25/2011 22:23'!
testKeyAtValue
	"self run: #testKeyAtValue"
	"self debug: #testKeyAtValue"
	
	| dict |
	dict := self targetClass new.
	dict at: #a put: 1.
	dict at: #b put: 2.
	dict at: #c put: 1.

	self commitIfNeeded.
	
	self assert: (dict keyAtValue: 2) = #b.
	self assert: ((dict keyAtValue: 1) = #a | ((dict keyAtValue: 1) = #c)).		"ael change"
	
	self should: [dict keyAtValue: 0] raise: Error
	
	
! !

!DictionaryTest methodsFor: '*worlds-keys and value tests-override' stamp: 'ael 9/25/2011 22:23'!
testKeys

	"self run:#testKeys "
	
	| a1 a2  dict | 
	a1 := Association key: 'France' value: 'Paris'.
	a2 := Association key: 'Italie' value: 'Rome'.
	dict := self targetClass new.
	dict add: a1.
	dict add: a2.
	
	self commitIfNeeded.
	 		
	self assert: (dict keys size) = 2.
	
	self assert: (dict keys includes: #France)
	
	


	
	! !

!DictionaryTest methodsFor: '*worlds-keys and value tests-override' stamp: 'ael 9/25/2011 22:23'!
testKeysDo
	"self run: #testKeysDo"
	"self debug: #testKeysDo"
	
	| dict res |
	dict := self targetClass new.
	
	dict at: #a put: 33.
	dict at: #b put: 66.

	self commitIfNeeded.
	
	res := OrderedCollection new.
	dict keysDo: [ :each | res add: each].
	
	self assert: res asSet = #(a b) asSet.


	
	! !

!DictionaryTest methodsFor: '*worlds-tests-override' stamp: 'ael 9/25/2011 22:23'!
testNilHashCollision
	"Ensures that fixCollisionsFrom: does the right thing in the presence of a nil key"
	| dict key |
	dict := self targetClass new.
	key := nil hash. "any key with same hash as nil"
	dict at: key hash put: 1.
	dict at: nil put: 2.
	self assert: (dict includesKey: nil).

	self commitIfNeeded.

	dict removeKey: key.
	self assert: (dict includesKey: nil).
! !

!DictionaryTest methodsFor: '*worlds-basic tests-override' stamp: 'ael 9/25/2011 22:23'!
testOccurrencesOf
	"self run:#testOccurrencesOf"
	
	| dict |
	dict := self targetClass new.
	dict at: #a put: 1.
	dict at: #b put: 2.
	dict at: #c put: 1.
	dict at: #d put: 3.
	dict at: nil put: nil.
	dict at: #z put: nil.
	
	self commitIfNeeded.
	
	self assert: (dict occurrencesOf: 1 ) = 2.
	self assert: (dict occurrencesOf: nil ) = 2.
	
	
	
	! !

!DictionaryTest methodsFor: '*worlds-tests-override' stamp: 'ael 9/25/2011 22:23'!
testPseudo
	"(self run: #testPseudo)"
	"true and false are valid keys"
	
	| dict1  |
	dict1 := self targetClass new.
	self shouldnt: [dict1 at: true put: #true] raise: Error.
	self assert: (dict1 at: true) = #true.

	self commitIfNeeded.

	self shouldnt: [dict1 at: false put: #false] raise: Error.
	self assert: (dict1 at: false) = #false.! !

!DictionaryTest methodsFor: '*worlds-implementation tests-override' stamp: 'ael 9/25/2011 22:23'!
testPseudoVariablesAreValidKeys
	"(self run: #testPseudoVariablesAreValidKeys)"
	"true and false are valid keys"
	
	| dict1  |
	dict1 := self targetClass new.
	self shouldnt: [dict1 at: true put: #true] raise: Error.
	self assert: (dict1 at: true) = #true.

	self commitIfNeeded.
		
	self shouldnt: [dict1 at: false put: #false] raise: Error.
	self assert: (dict1 at: false) = #false.! !

!DictionaryTest methodsFor: '*worlds-testing-override' stamp: 'ael 9/25/2011 22:24'!
testReject
	"Ensure that Dictionary>>reject: answers a dictionary not something else"
	| dict expected result |
	dict := self targetClass newFromPairs:{
		#first.		1.
		#second.	2.
		#third.		3.
		#fourth.	4.
		#fifth.		5.
	}.
	
	self commitIfNeeded.

	result := dict reject:[:each| each odd].
	expected := self targetClass newFromPairs:{
		#second.	2.
		#fourth.	4.
	}.
	self assert: result = expected.! !

!DictionaryTest methodsFor: '*worlds-tests-override' stamp: 'ael 9/25/2011 22:24'!
testRemoveAll
	"Allows one to remove all elements of a collection" 
	
	| dict1 dict2 s2 |
	dict1 := self targetClass new.
	dict1 at: #a put:1 ; at: #b put: 2. 
	dict2 := dict1 copy.
	s2 := dict2 size.
	
	dict1 removeAll.
	
	self commitIfNeeded.

	self assert: dict1 size = 0.
	self assert: dict2 size = s2 description: 'the copy has not been modified'.! !

!DictionaryTest methodsFor: '*worlds-keys and value tests-override' stamp: 'ael 9/25/2011 22:24'!
testRemoveKey
		"self run:#testRemoveKey "

	| dict | 
	dict := self targetClass new.
	dict at: #a put: 1.
	dict at: #b put: 2.
	 
	self assert: (dict keys size) = 2.
	dict removeKey: #a.
	self assert: dict keys size  = 1.

	self commitIfNeeded.

	self should: [dict at: #a] raise: Error.
	self assert: (dict at: #b) = 2


	
	! !

!DictionaryTest methodsFor: '*worlds-testing-override' stamp: 'ael 9/25/2011 22:24'!
testSelect
	"Ensure that Dictionary>>select: answers a dictionary not something else"
	| dict expected result |
	dict := self targetClass newFromPairs:{
		#first.		1.
		#second.	2.
		#third.		3.
		#fourth.	4.
		#fifth.		5.
	}.

	self commitIfNeeded.

	result := dict select:[:each| each odd].
	expected := self targetClass newFromPairs:{
		#first.		1.
		#third.		3.
		#fifth.		5.
	}.
	self assert: result = expected.
	result at: #first put: 0.
	self assert: (dict at: #first) = 1 "No side effects"! !

!DictionaryTest methodsFor: '*worlds-keys and value tests-override' stamp: 'ael 9/25/2011 22:24'!
testValues
	"self run:#testValues "
	
	| a1 a2 a3 dict | 
	a1 := Association key: 'France' value: 'Paris'.
	a2 := Association key: 'Italie' value: 'Rome'.
	dict := self targetClass new.
	dict add: a1.
	dict add: a2.

	self commitIfNeeded.
	 
	self assert: (dict values size ) = 2.
	self assert: (dict values includes: 'Paris').
	
	a3 := Association new.
	dict add: a3.
	self assert: (dict values size ) = 3.
	self assert: (dict values includes: nil).
	
	
	
	
	
	


	
	! !

!DictionaryTest methodsFor: '*worlds-tests-override' stamp: 'ael 9/25/2011 22:25'!
testWithAll
	
	| dict1 dict2 |
	dict1 := self targetClass withAll: {#a -> 1. #b -> 2}.

	self commitIfNeeded.

	self assert: (dict1 at: #a) = 1.
	self assert: (dict1 at: #b) = 2.

	dict2 := self targetClass withAll: dict1.
	self assert: (dict2 at: #a) = 1.
	self assert: (dict2 at: #b) = 2.
! !

DictionaryTest subclass: #WIdentityDictionaryTest
	instanceVariableNames: ''
	classVariableNames: 'CommitLevels'
	poolDictionaries: ''
	category: 'Worlds-CollectionsTests'!

!WIdentityDictionaryTest class methodsFor: 'as yet unclassified' stamp: 'ael 9/25/2011 22:34'!
addToSuiteFromSelectors: suite
	CommitLevels := 0.
	super addToSuiteFromSelectors: suite
! !

!WIdentityDictionaryTest class methodsFor: 'as yet unclassified' stamp: 'ael 9/25/2011 22:39'!
initialize
	CommitLevels := 0
	! !

!WIdentityDictionaryTest class methodsFor: 'as yet unclassified' stamp: 'ael 7/12/2011 16:51'!
shouldInheritSelectors
	^true
! !

!WIdentityDictionaryTest class methodsFor: 'as yet unclassified' stamp: 'ael 11/10/2011 13:46'!
testPrimitiveLookup
	"WIdentityDictionaryTest testPrimitiveLookup"
	IdentityDictionary allInstancesDo: [ :d |
		d keysDo: [ :k |
			(d at: k) == (d testAelAt: k) ifFalse: [ self halt ].
			]
		]
! !

!WIdentityDictionaryTest methodsFor: 'accessing' stamp: 'ael 9/25/2011 22:39'!
commitIfNeeded
	"commit by the chosen number of levels"
	| w |
	w := self thisWorld.
	w isTopLevel
	  ifFalse: [
		CommitLevels timesRepeat: [ w commit.  w := w parent ].
		Processor activeProcess currentWWorld: w.
		].! !

!WIdentityDictionaryTest methodsFor: 'tests' stamp: 'ael 12/6/2011 15:23'!
testAllRecords
	| dict w1 w2 w2Contents w2Reads deletion w2SeenAll w1Contents w1Reads w2Flattened dnk |
	dict := WIdentityDictionary new.
	dict
		at: #a put: #a0;
		at: #b put: #b0;
		at: #c put: #c0;
		at: #e put: #e0;
		at: #f put: #f0;
		at: #h put: #h0.

	deletion := dict deletionToken.
	dnk := dict dnkToken.

	w1 := self thisWorld sprout.
	w2 := w1 sprout.
	self assert: (w2 eval: [ dict worldFlattened includesKey: #f ]).

	dict removeKey: #f.
	self assert: (w2 eval: [ dict worldFlattened includesKey: #f ]) not.

	dict at: #h put: #h0a.				"even though #h has been read, we can change it"
	self assert: (w2 eval: [ (dict worldFlattened at: #h) == #h0a ]).

	w2 eval: [ dict at: #a; includesKey: #b; includesKey: #d ].	"now record some reads"
	dict at: #a put: #a0a; at: #b put: #b0a; at: #d put: #d0.

	w2Flattened := w2 eval: [ dict worldFlattened ].		
	self assert: (w2Flattened size = 5).
	self assert: ((w2Flattened at: #a) == #a0).
	self assert: ((w2Flattened at: #b) == #b0a).
	self assert: ((w2Flattened at: #c) == #c0).
	self assert: ((w2Flattened at: #e) == #e0).
	self assert: ((w2Flattened at: #h) == #h0a).
	"#d cannot appear, because its absence was recorded in w2 before it was added" 
	self should: [w2Flattened associationAt: #d ] raise: KeyNotFound.

	w1 eval: [
		dict
			at: #a put: #a1;		"overwrite without reading"
			at: #g put: #g1;
			removeKey: #h.
		].
	self assert: (w1 eval: [ dict worldFlattened keys asIdentitySet ]) = #(a b c d e g) asIdentitySet.
	
	w2 eval: [
		dict at: #c put: #c2.							"overwriting previously read value"
		w2Contents := IdentityDictionary new addAll: dict worldFlattened associations; yourself.
		w2Reads := dict debugUpdateTrackerState at: 1.
		w2SeenAll := dict debugUpdateTrackerState at: 2.
		].

	self assert: (w2Contents size = 5).
	self assert: ((w2Contents at: #a) == #a0).
	self assert: ((w2Contents at: #b) == #b0a).
	self assert: ((w2Contents at: #c) == #c2).
	self assert: ((w2Contents at: #e) == #e0).
	self assert: ((w2Contents at: #g) == #g1).

	self assert: (w2Reads size = 3).
	self assert: ((w2Reads at: #a) == #a0).			"it was read before a1 was set"
	self assert: ((w2Reads at: #b) == dnk).
	self assert: ((w2Reads at: #d) == deletion).

	self assert: (w2SeenAll == nil).

	self should: [ w2 eval: [ dict worldFlattened at: #d ] ] raise: KeyNotFound.

	self should: [ w2 commit ] raise: WorldSerializationError.	"clash between a0 and a1, and the appearance of d (but we don't know which will be found first)"

	"first take away the clash on #d"
	dict removeKey: #d.
	self should: [ w2 commit ] raise: WorldValueChanged.		"should now be the clash between a0 and a1"

	"fix #a entry"
	w1 eval: [ dict at: #a put: #a0 ].
	self shouldnt: [ w2 commit ] raise: WorldSerializationError.

	w1 eval: [
		w1Contents := IdentityDictionary new addAll: dict associations; yourself.
		w1Reads := dict debugUpdateTrackerState at: 1.
		].

	self assert: (w1Contents size = 5).
	self assert: ((w1Contents at: #a) == #a0).
	self assert: ((w1Contents at: #b) == #b0a).
	self assert: ((w1Contents at: #c) == #c2).			"result of the commit"
	self assert: ((w1Contents at: #e) == #e0).
	self assert: ((w1Contents at: #g) == #g1).

	self assert: (w1Reads size = 4).
	self assert: ((w1Reads at: #b) == #b0a).
	self assert: ((w1Reads at: #e) == #e0).
	self assert: ((w1Reads at: #d) == deletion).			"result of the commit"
	self assert: ((w1Reads at: #h) == #h0a).			"read when being removed"
	
	self assert: (dict size = 5). 
	self assert: ((dict at: #a) == #a0a). 
	self assert: ((dict at: #b) == #b0a). 
	self assert: ((dict at: #c) == #c0). 
	self assert: ((dict at: #e) == #e0). 
	self assert: ((dict at: #h) == #h0a). 

	self shouldnt: [ w1 commit ] raise: WorldSerializationError.

	self assert: (dict size = 5). 
	self assert: ((dict at: #a) == #a0). 
	self assert: ((dict at: #b) == #b0a). 
	self assert: ((dict at: #c) == #c2). 
	self assert: ((dict at: #e) == #e0). 
	self assert: ((dict at: #g) == #g1). 
! !

!WIdentityDictionaryTest methodsFor: 'tests' stamp: 'ael 9/25/2011 23:20'!
testDepthOneNonrecording
	| w |
	CommitLevels := 1.
	
	DictionaryTest allTestSelectors do: [ :selector |
		w := self thisWorld sprout.
		w tracksReads: false.
		w eval: [ self perform: selector ]
		].

	CommitLevels := 0.
	DictionaryTest allTestSelectors do: [ :selector |
		w := self thisWorld sprout.
		w tracksReads: false.
		w eval: [ self perform: selector ]
		].

! !

!WIdentityDictionaryTest methodsFor: 'tests' stamp: 'ael 9/25/2011 23:21'!
testDepthOneWorld
	CommitLevels := 1.
	DictionaryTest allTestSelectors do: [ :selector |
		self thisWorld sprout eval: [ self perform: selector ]
		].

	CommitLevels := 0.
	DictionaryTest allTestSelectors do: [ :selector |
		self thisWorld sprout eval: [ self perform: selector ]
		].

! !

!WIdentityDictionaryTest methodsFor: 'tests' stamp: 'ael 9/25/2011 22:38'!
testDepthTwoNonrecording
	| w |
	CommitLevels := 2.
	DictionaryTest allTestSelectors do: [ :selector |
		w := self thisWorld sprout sprout.
		w tracksReads: false.
		w eval: [ self perform: selector ]
		].

	CommitLevels := 1.
	DictionaryTest allTestSelectors do: [ :selector |
		w := self thisWorld sprout sprout.
		w tracksReads: false.
		w eval: [ self perform: selector ]
		].

	CommitLevels := 0.
	DictionaryTest allTestSelectors do: [ :selector |
		w := self thisWorld sprout sprout.
		w tracksReads: false.
		w eval: [ self perform: selector ]
		].

! !

!WIdentityDictionaryTest methodsFor: 'tests' stamp: 'ael 9/25/2011 23:21'!
testDepthTwoWorld
	CommitLevels := 2.
	DictionaryTest allTestSelectors do: [ :selector |
		self thisWorld sprout sprout eval: [ self perform: selector ]
		].

	CommitLevels := 1.
	DictionaryTest allTestSelectors do: [ :selector |
		self thisWorld sprout sprout eval: [ self perform: selector ]
		].

	CommitLevels := 0.
	DictionaryTest allTestSelectors do: [ :selector |
		self thisWorld sprout sprout eval: [ self perform: selector ]
		].
! !

!WIdentityDictionaryTest methodsFor: 'tests' stamp: 'ael 2/25/2012 00:39'!
testThreeLevels
	| dict w1 w2 w2Contents w2Reads deletion dnk w2SeenAll w1Contents w1Reads |
	dict := WIdentityDictionary new.
	dict
		at: #a put: #a0;
		at: #b put: #b0;
		at: #c put: #c0;
		at: #e put: #e0;
		at: #f put: #f0;
		at: #h put: #h0.

	dnk := dict dnkToken.
	deletion := dict deletionToken.

	w1 := self thisWorld sprout.
	w2 := w1 sprout.
	self assert: (w2 eval: [ dict includesKey: #f ]).

	dict removeKey: #f.
	self assert: (w2 eval: [ dict includesKey: #f ]).		"should not be affected"

	w1 eval: [ dict keys ].
	self assert: (w1 eval: [ dict debugUpdateTrackerState at: 2 ]) == #keys.
	self assert: (w1 eval: [ dict size ]) = 5.
	w1 eval: [ dict associations ].
	self assert: (w1 eval: [ dict debugUpdateTrackerState at: 2 ]) == #values.
	self assert: (w1 eval: [ dict size ]) = 5.

	dict at: #h put: #h0a.				"now that #h has been read, change it"

	w2 eval: [
		dict at: #a; includesKey: #b; includesKey: #d.
		w2SeenAll := dict debugUpdateTrackerState at: 2.
		w2Reads := dict debugUpdateTrackerState at: 1.
		].
	self assert: (w2SeenAll isNil).
	self assert: (w2Reads size = 4).
	self assert: ((w2Reads at: #a) == #a0).			"we've looked at #a"
	self assert: ((w2Reads at: #b) == dnk).			"only checked for key #b"
	self assert: ((w2Reads at: #f) == dnk).			"ditto for #f"
	self assert: ((w2Reads at: #d) == deletion).		"there is no #d"

	self should: [ w2 eval: [ dict associationAt: #f ] ] raise: KeyNotFound.

	w1 eval: [
		dict
			at: #a put: #a1;
			at: #g put: #g1;
			removeKey: #h.
		].
	self assert: (w1 eval: [ dict keys asIdentitySet ]) = #(a b c e g) asIdentitySet.
	self assert: (w1 eval: [ dict size ]) = 5.
	
	w2 eval: [
		dict at: #c put: #c2.							"NB: overwriting without reading"
		w2Contents := IdentityDictionary new addAll: dict associations; yourself.
		w2Reads := dict debugUpdateTrackerState at: 1.
		w2SeenAll := dict debugUpdateTrackerState at: 2.
		].

	self assert: (w2Contents size = 5).
	self assert: ((w2Contents at: #a) == #a0).
	self assert: ((w2Contents at: #b) == #b0).
	self assert: ((w2Contents at: #c) == #c2).
	self assert: ((w2Contents at: #e) == #e0).
	self assert: ((w2Contents at: #g) == #g1).

	self assert: (w2Reads size = 7).
	self assert: ((w2Reads at: #a) == #a0).			"it was read before a1 was set"
	self assert: ((w2Reads at: #b) == #b0).
	self assert: ((w2Reads at: #d) == deletion).
	self assert: ((w2Reads at: #e) == #e0).
	self assert: ((w2Reads at: #f) == deletion).		"we've already seen that f disappeared"
	self assert: ((w2Reads at: #g) == #g1).
	self assert: ((w2Reads at: #h) == deletion).

	self assert: (w2SeenAll == #values).

	self assert: (w2 eval: [ dict size ]) = 5.

	self should: [ w2 eval: [ dict at: #d ] ] raise: KeyNotFound.

	self should: [ w2 commit ] raise: WorldValueChanged.	"clash between a0 and a1"

	w1 eval: [ dict at: #a put: #a0 ].					"restore to the state seen in w2"

	self shouldnt: [ w2 commit ] raise: WorldSerializationError.

	w1 eval: [
		w1Contents := IdentityDictionary new addAll: dict associations; yourself.
		w1Reads := dict debugUpdateTrackerState at: 1.
		].

	self assert: (w1Contents size = 5).
	self assert: ((w1Contents at: #a) == #a0).
	self assert: ((w1Contents at: #b) == #b0).
	self assert: ((w1Contents at: #c) == #c2).			"result of the commit"
	self assert: ((w1Contents at: #e) == #e0).
	self assert: ((w1Contents at: #g) == #g1).

	self assert: (w1Reads size = 7).
	self assert: ((w1Reads at: #a) == #a0).
	self assert: ((w1Reads at: #b) == #b0).
	self assert: ((w1Reads at: #c) == #c0).
	self assert: ((w1Reads at: #e) == #e0).
	self assert: ((w1Reads at: #d) == deletion).			"result of the commit"
	self assert: ((w1Reads at: #f) == deletion).			"result of the commit"
	self assert: ((w1Reads at: #h) == #h0).
	
	self assert: (dict size = 5). 
	self assert: ((dict at: #a) == #a0). 
	self assert: ((dict at: #b) == #b0). 
	self assert: ((dict at: #c) == #c0). 
	self assert: ((dict at: #e) == #e0). 
	self assert: ((dict at: #h) == #h0a). 

	self should: [ w1 commit ] raise: WorldValueChanged.	"clash between h0 and h0a"

	dict at: #h put: #h0.

	self shouldnt: [ w1 commit ] raise: WorldSerializationError.

	self assert: (dict size = 5). 
	self assert: ((dict at: #a) == #a0). 
	self assert: ((dict at: #b) == #b0). 
	self assert: ((dict at: #c) == #c2). 
	self assert: ((dict at: #e) == #e0). 
	self assert: ((dict at: #g) == #g1). 
! !

!WIdentityDictionaryTest methodsFor: 'accessing' stamp: 'ael 9/25/2011 21:43'!
thisWorld
	^WWorld thisWorld
! !

EncoderForV3PlusClosures subclass: #WEncoder
	instanceVariableNames: 'namespace usedObjects'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Worlds-Compiler'!

!WEncoder methodsFor: 'as yet unclassified' stamp: 'ael 10/4/2011 15:45'!
autoBind: name 
	"Declare a block argument as a temp if not already declared."
	| node |
	node _ scopeTable 
			at: name
			ifAbsent: 
				[(self basicLookupInPools: name ifFound: [:assoc | assoc])
					ifTrue: [self notify: 'Name already used in a Pool or Global'].
				^ (self reallyBind: name) nowHasDef nowHasRef scope: 1].
	node isTemp
		ifTrue: [node scope >= 0 ifTrue:
					[^ self notify: 'Name already used in this method'].
				node nowHasDef nowHasRef scope: 1]
		ifFalse: [^ self notify: 'Name already used in this class'].
	^node! !

!WEncoder methodsFor: 'as yet unclassified' stamp: 'ael 10/4/2011 15:45'!
basicLookupInPools: varName ifFound: assocBlock

	^  super lookupInPools: varName ifFound: assocBlock.! !

!WEncoder methodsFor: 'as yet unclassified' stamp: 'ael 10/4/2011 15:45'!
global: assocOrName name: name

	| nsAccess |
	assocOrName isVariableBinding ifFalse: [
		usedObjects add: assocOrName.
		nsAccess := MessageNode new
			receiver: (self encodeVariable: 'self')
			selector: #namespace
			arguments: #() precedence: #namespace precedence from: self. 
		^ MessageNode new
			receiver: nsAccess
			selector: #nsAt:
			arguments: {self encodeLiteral: name asSymbol} precedence: #at: precedence from: self.
	] ifTrue: [
		^ super global: assocOrName name: name]! !

!WEncoder methodsFor: 'as yet unclassified' stamp: 'ael 10/4/2011 15:45'!
init: aClass context: aContext notifying: req
	| node n homeNode indexNode |
	usedObjects := Set new.
	requestor _ req.
	class _ aClass.
	nTemps _ 0.
	supered _ false.
	self initScopeAndLiteralTables.
	sourceRanges _ Dictionary new: 32.
	globalSourceRanges _ OrderedCollection new: 32.
	n _ -1.
	class allInstVarNames withIndexDo: 
		[:variable :offset |
			offset < class firstRedirectedIV
				ifTrue: [
					node := InstanceVariableNode new
								name: variable index: offset]
				ifFalse: [node := self variableNodeFor: variable].	
			scopeTable at: variable put: node].
	aContext == nil
		ifFalse: 
			[homeNode _ self bindTemp: 'homeContext'.
			"first temp = aContext passed as arg"
			n _ 0.
			aContext tempNames do: 
				[:variable | 
				indexNode _ self encodeLiteral: (n _ n + 1).
				node _ MessageAsTempNode new
							receiver: homeNode
							selector: #tempAt:
							arguments: (Array with: indexNode)
							precedence: 3
							from: self.
				scopeTable at: variable put: node]].
! !

!WEncoder methodsFor: 'as yet unclassified' stamp: 'sn 9/10/2013 03:09'!
init: aCue notifying: anObject
	"The use of the variable requestor is a bit confusing here. This is
	*not* the original requestor, which is available through the cue.
	It's the Parser instance that is using the encoder."

	|n node homeNode indexNode|

	self setCue: aCue.
	requestor := anObject.
	nTemps := 0.
	supered := false.
	self initScopeAndLiteralTables.
	sourceRanges _ Dictionary new: 32.
	globalSourceRanges _ OrderedCollection new: 32.
	n _ -1.
	
	cue getClass allInstVarNames withIndexDo: 
		[:variable :offset |
			offset < cue getClass firstRedirectedIV
				ifTrue: [
					node := InstanceVariableNode new
								name: variable index: offset]
				ifFalse: [node := self variableNodeFor: variable].	
			scopeTable at: variable put: node].
	cue context == nil
		ifFalse: 
			[homeNode _ self bindTemp: 'homeContext'.
			"first temp = aContext passed as arg"
			n _ 0.
			cue context tempNames do: 
				[:variable | 
				indexNode _ self encodeLiteral: (n _ n + 1).
				node _ MessageAsTempNode new
							receiver: homeNode
							selector: #tempAt:
							arguments: (Array with: indexNode)
							precedence: 3
							from: self.
				scopeTable at: variable put: node]].
! !

!WEncoder methodsFor: 'as yet unclassified' stamp: 'sn 9/10/2013 02:57'!
lookupInPools: varName ifFound: assocBlock

	(namespace isNil or: [namespace objectsPool isNil]) ifTrue: [
		^  super lookupInPools: varName ifFound: assocBlock].
	(cue bindingOf: varName) ifNotNilDo: [:assoc |
		assocBlock value: assoc.
		^true].
	assocBlock value: varName.
	^ true.
! !

!WEncoder methodsFor: 'as yet unclassified' stamp: 'ael 10/4/2011 15:45'!
lookupInPoolsRecursively: varName in: ns ifFound: assocBlock

	| pool |
	ns ifNotNil: [
		pool := ns objectsPool.
		(pool bindingOf: varName asSymbol) ifNotNilDo:[:assoc|
			usedObjects add: assoc value.
			assocBlock value: assoc.
			^true
		] ifNil: [
			^ self lookupInPoolsRecursively: varName in: (pool at: #parentNamespace ifAbsent: []) ifFound: assocBlock]].
	^false
! !

!WEncoder methodsFor: 'as yet unclassified' stamp: 'ael 10/4/2011 15:45'!
namespace: anObject

	namespace _ anObject.
! !

!WEncoder methodsFor: 'as yet unclassified' stamp: 'ael 10/4/2011 15:45'!
undeclared: name
	"hmm..."

	| sym |
	(requestor notNil and: [requestor interactive]) ifTrue: 
		[requestor requestor == #error: ifTrue: [requestor error: 'Undeclared' translated].
		^ self notify: 'Undeclared'].
	Transcript show: ' (' , name , ' is Undeclared) '.
	sym _ name asSymbol.
	Undeclared at: sym put: nil.
	^ self global: (Undeclared associationAt: sym) name: sym! !

!WEncoder methodsFor: 'as yet unclassified' stamp: 'ael 10/4/2011 15:45'!
usedObjects

	^ usedObjects.
! !

!WEncoder methodsFor: 'as yet unclassified' stamp: 'ael 10/4/2011 15:45'!
variableNodeFor: name
	^ MessageNode new
			receiver: (self encodeVariable: 'self')
			selector: (#wiv666, name) asSymbol
			arguments: {}
			precedence: 3
			from: self
			sourceRange: (1 to: 0).
! !

AbstractObjectsAsMethod subclass: #WorldLookupMethod
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Worlds'!

!WorldLookupMethod methodsFor: 'all' stamp: 'ael 10/12/2011 15:25'!
run: oldSelector with: arguments in: aReceiver
	| lookupClass |

	"The receiver is a uniclass with the myMethodDict override mechanism, but it can also have superclasses that are themselves such uniclasses.  If we don't find an entry in the myMethodDict for the receiver's class, we work our way up the superclass chain.  Once we reach a superclass that isn't a uniclass, we can dispatch through the perform:...inSuperclass: mechanism."

	lookupClass := aReceiver class.
	[ lookupClass isUniClass ]
	  whileTrue: [
		(lookupClass myMethodDict at: oldSelector ifAbsent: [ nil ])
		  ifNotNil: [ :m | ^ m valueWithReceiver: aReceiver arguments: arguments].
		lookupClass := lookupClass superclass.
		].

	^ aReceiver perform: oldSelector withArguments: arguments inSuperclass: lookupClass.
! !

!Object methodsFor: '*worlds' stamp: 'ael 10/31/2011 22:22'!
deeplyWorldFlattenedIn: aWorld with: doneSoFar
	^self
! !

!Object methodsFor: '*worlds' stamp: 'ael 9/18/2011 16:04'!
isWorldDeletion
	"is this a deletion token for use in Worlds?"
	^false
! !

!Object methodsFor: '*worlds' stamp: 'ael 9/18/2011 16:05'!
isWorldOverwrite
	"is this an overwrite token for use in Worlds?"
	^false
! !

!Object methodsFor: '*worlds' stamp: 'ael 11/1/2011 11:00'!
slangInstVarAt: index 
	"Primitive. Answer a fixed variable in an object. The numbering of the 
	variables corresponds to the named instance variables. Fail if the index 
	is not an Integer or is not the index of a fixed variable. Essential. See 
	Object documentation whatIsAPrimitive."

	<primitive: 73>
	"Access beyond fixed variables."
	^self basicAt: index - self class instSize		! !

!Object methodsFor: '*worlds' stamp: 'ael 11/11/2011 09:18'!
slangInstVarAt: index put: v

	^ self instVarAt: index put: v
! !

Object subclass: #WObject
	instanceVariableNames: ''
	classVariableNames: 'DelegateLookup DeletionToken EqualityClasses OverwriteToken'
	poolDictionaries: ''
	category: 'Worlds'!

!WObject commentStamp: 'ael 6/16/2011 00:34' prior: 0!
Initialising special values:

	Smalltalk at: #W2DelegateLookup put: Object new.
	Smalltalk at: #W2DeletionToken put: Object new.
	Smalltalk at: #W2OverwriteToken put: Object new.

!

WObject variableWordSubclass: #WBitmap
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Worlds'!

!WBitmap methodsFor: 'as yet unclassified' stamp: 'aw 3/26/2010 13:58'!
at: index

	^ self instVarAt: index! !

!WBitmap methodsFor: 'as yet unclassified' stamp: 'aw 3/26/2010 13:58'!
at: index put: value

	^ self instVarAt: index put: value! !

!WBitmap methodsFor: 'as yet unclassified' stamp: 'aw 3/29/2010 12:18'!
basicInstVarAt: index put: value

	^ super basicAt: index put: value! !

!WBitmap methodsFor: 'as yet unclassified' stamp: 'ael 10/27/2011 15:34'!
basicValueAt: index
	"obtain a value from the top-level world"
	^ super basicAt: index! !

!WBitmap methodsFor: 'as yet unclassified' stamp: 'ael 10/2/2011 11:51'!
basicValueAt: index put: value
	"sent in top-level world only"
	^ super basicAt: index put: value! !

!WBitmap methodsFor: 'as yet unclassified' stamp: 'ael 9/26/2011 08:01'!
dnkToken
	"special case"
	^ 1! !

!WBitmap methodsFor: 'as yet unclassified' stamp: 'aw 3/29/2010 11:45'!
do: aBlock

	1 to: self size do: [:ii | aBlock value: (self at: ii)]! !

!WBitmap methodsFor: 'as yet unclassified' stamp: 'ael 10/27/2011 15:15'!
iMakeDeltaOrOrig

	^Bitmap new: self iSize withAll: self dnkToken! !

!WBitmap methodsFor: 'as yet unclassified' stamp: 'aw 3/26/2010 15:19'!
iSize

	^ self basicSize! !

!WBitmap methodsFor: 'as yet unclassified' stamp: 'ael 11/12/2011 11:35'!
initializedClone
	"return an instance that's ready to record values read from the original"
	^self class basicNew: self size.
! !

!WBitmap methodsFor: 'as yet unclassified' stamp: 'aw 3/29/2010 11:45'!
isEmpty

	^ self size == 0! !

!WBitmap methodsFor: 'as yet unclassified' stamp: 'ael 11/12/2011 13:13'!
nutterConvertTo: newClass
	"this is nuts"
	| nutter w writes size thisDnk v |
	size := self size.
	nutter := newClass new: size.
	"set up in the top world, to fix the size"
	1 to: self size do: [ :index | nutter add: (self basicValueAt: index) ].
	thisDnk := self dnkToken.
	w := self thisWorld.
	[ w isTopLevel ] whileFalse: [
		writes := w writesRecordFor: self.
		writes ifNotNil: [
			w eval: [
				1 to: size do: [ :index |
					v := writes at: index.
					v == thisDnk ifFalse: [ nutter at: index put: v ].
					]
				]
			].
		w := w parent.
		].
	^nutter
! !

!WBitmap methodsFor: 'as yet unclassified' stamp: 'ael 11/12/2011 17:35'!
nutterConvertToSlowCollection
	"this is nuts"
	| nutter w writes size thisDnk v |
	size := self size.
	nutter :=" WCollection WArray "WIdentityDictionary" WSimpleOrderedCollection" new: size.
	"set up in the top world, to fix the size"
	WTopLevelWorld default eval: [ 1 to: self size do: [ :index | nutter uncheckedAt: index put: (self basicValueAt: index) ] ].
	thisDnk := self dnkToken.
	w := self thisWorld.
	[ w isTopLevel ] whileFalse: [
		writes := w writesRecordFor: self.
		writes ifNotNil: [
			w eval: [
				1 to: size do: [ :index |
					v := writes at: index.
					v == thisDnk ifFalse: [ nutter uncheckedAt: index put: v ].
					]
				]
			].
		w := w parent.
		].
	^nutter
! !

!WBitmap methodsFor: 'as yet unclassified' stamp: 'ael 11/12/2011 15:57'!
nutterWorldFlattenIn: w
	"completely nuts.  for testing only."
	| flatNut nut |
	nut := w eval: [ self nutterConvertToSlowCollection ].
	flatNut := WBitmap new: self size.
	(nut respondsTo: #withIndexDo:)
	  ifTrue: [ (nut worldFlattenIn: w) withIndexDo: [ :v :i | flatNut at: i put: v ] ]
	  ifFalse: [ (nut worldFlattenIn: w) keysAndValuesDo: [ :i :v | flatNut at: i put: v ] ].
	^flatNut
! !

!WBitmap methodsFor: 'as yet unclassified' stamp: 'aw 3/29/2010 11:44'!
printElementsOn: aStream

	aStream nextPut: $(.
	self do: [:ee | aStream print: ee; space].
	self isEmpty ifFalse: [aStream skip: -1].
	aStream nextPut: $)! !

!WBitmap methodsFor: 'as yet unclassified' stamp: 'aw 3/29/2010 11:44'!
printOn: aStream

	aStream nextPutAll: self class name.
	self printElementsOn: aStream! !

!WBitmap methodsFor: 'as yet unclassified' stamp: 'ael 11/3/2011 14:16'!
worldFlattenIn: aWorld
	| dest |
	dest := self class basicNew: self size.
	aWorld flatten: self into: dest.
	^dest
! !

!WObject class methodsFor: 'compiling' stamp: 'ael 10/4/2011 15:46'!
compilerClass

	^ WCompiler! !

!WObject class methodsFor: 'compiling' stamp: 'ael 10/24/2011 15:55'!
doneCompiling
	| bb ans setter size useUpdatedVector vectorIVString |
	self removeCategorySilently: #'*generated-accessors'.
	bb := self superclass allInstVarNames size.
	vectorIVString := self firstRedirectedIV asString.	"assume 1st is update vector"
	useUpdatedVector := Smalltalk globals at: #LObject ifPresent: [ :lobjClass | self includesBehavior: lobjClass ] ifAbsent: [ false ]. 
	setter := useUpdatedVector
		ifTrue: [ [:vNameAndIndex |
				 ((vNameAndIndex at: 2) = vectorIVString
					ifTrue: ['wiv666{1}: v ^self instVarAt: {2} put: v']
					ifFalse: ['wiv666{1}: v | vv upvec | vv := self instVarAt: {2} put: v. (upvec := self instVarAt: ', vectorIVString, ') ifNotNil: [ upvec basicAt: {2} put: 1]. ^ vv'])
				  format: vNameAndIndex]]
		ifFalse: [setter := [:vNameAndIndex | 'wiv666{1}: v ^ self instVarAt: {2} put: v' format: vNameAndIndex]].
	self instVarNames
		withIndexDo: [:nn :ii |
			bb + ii >= self firstRedirectedIV
			  ifTrue: [
				self
				  compileSilently: ('wiv666{1} ^ self instVarAt: {2}' format: {nn. (bb + ii) printString})
				  classified: #'*generated-accessors'.
				self
				  compileSilently: (setter value: {nn. (bb + ii) printString})
				  classified: #'*generated-accessors'.
				]
			].
	ans := super doneCompiling.
	useUpdatedVector
		ifTrue: [size := self allInstVarNames size.
			self
				allInstancesDo: [:e | e updateUpdatedVectorTo: size]].
	self
		allSubclassesDo: [:c | c doneCompiling].
	^ ans! !

!WObject class methodsFor: 'instance variables' stamp: 'ael 9/26/2011 08:17'!
firstCommittableIV
	^ self firstRedirectedIV
! !

!WObject class methodsFor: 'instance variables' stamp: 'ael 9/18/2011 15:09'!
firstRedirectedIV
	^ 1
! !

!WObject class methodsFor: 'instance variables' stamp: 'ael 9/26/2011 17:02'!
firstSerialisationIV
	^ self firstRedirectedIV
! !

!WObject class methodsFor: 'class initialization' stamp: 'ael 6/25/2011 12:43'!
initialize
	EqualityClasses := IdentitySet new.
	EqualityClasses
		addAll: #(ByteString WideString Point Rectangle Color DateAndTime Duration Time);
		addAll: (Number allSubclasses collect: [ :c | c name ])
! !

!WObject class methodsFor: 'instance creation' stamp: 'ael 9/25/2011 21:35'!
new
	^WTopLevelWorld default eval: [ super new ]
! !

!WObject class methodsFor: 'instance creation' stamp: 'ael 12/2/2011 14:29'!
newAndDo: aBlock
	| inst |
	WTopLevelWorld default eval: [
		inst := super new initialize.
		aBlock value: inst.
		].
	
	^inst
! !

!WObject class methodsFor: 'primitive support' stamp: 'ael 11/2/2011 19:00'!
readsSlot
	^1
! !

!WObject class methodsFor: 'browsing' stamp: 'ael 9/18/2011 15:49'!
whichMethodsStoreInto: instVarName 
	"Answer a collection of CompiledMethod whose methods access the argument, instVarName, as a named instance variable."
	
	| methods selectors ivIndex |
	ivIndex := self instVarIndexFor: instVarName ifAbsent: [^#()].
	ivIndex < self firstRedirectedIV ifTrue: [ ^super whichMethodsStoreInto: instVarName ].

	selectors := self whichSelectorsReferTo: (#wiv666 , instVarName , ':') asSymbol.
	methods := self methodDict.
	^selectors collect: [ :sel | methods at: sel ].

	"LObject whichMethodsStoreInto: 'whole'."
! !

!WObject class methodsFor: 'browsing' stamp: 'ael 9/18/2011 15:48'!
whichSelectorsAccess: instVarName
	| ivIndex |
	ivIndex := self instVarIndexFor: instVarName ifAbsent: [^#()].
	ivIndex < self firstRedirectedIV ifTrue: [ ^super whichSelectorsAccess: instVarName ].

	^(self whichSelectorsStoreInto: instVarName),
		(self whichSelectorsReferTo: (#wiv666, instVarName) asSymbol)
! !

!WObject class methodsFor: 'browsing' stamp: 'ael 9/18/2011 15:49'!
whichSelectorsStoreInto: instVarName
	| ivIndex |
	ivIndex := self instVarIndexFor: instVarName ifAbsent: [^#()].
	ivIndex < self firstRedirectedIV ifTrue: [ ^super whichSelectorsStoreInto: instVarName ].

	^self whichSelectorsReferTo: (#wiv666, instVarName, ':') asSymbol
! !

!WObject class methodsFor: 'primitive support' stamp: 'ael 11/2/2011 19:00'!
writesSlot
	^2
! !

!WObject methodsFor: 'testing' stamp: 'ael 6/20/2011 14:57'!
areEqualForCommit: a and: b
	a class == b class ifFalse: [ ^false ].
	
	^(EqualityClasses includes: a class name)
	  ifTrue: [ a = b ]
	  ifFalse: [ a == b ]
! !

!WObject methodsFor: 'accessing' stamp: 'ael 10/27/2011 15:34'!
basicValueAt: index
	"obtain a value from the top-level world"
	^ super instVarAt: index! !

!WObject methodsFor: 'accessing' stamp: 'ael 10/2/2011 11:51'!
basicValueAt: index put: value
	"sent in top-level world only"
	^ super instVarAt: index put: value! !

!WObject methodsFor: 'world awareness' stamp: 'ael 10/27/2011 14:35'!
commitInto: parentWorld using: updateTrackers
	"updateTrackers is a tuple { orig. delta }"
	| delta val iSize dnk |
	delta := updateTrackers at: 2.
	delta ifNotNil: [
		iSize := self iSize.
		dnk := self dnkToken.
		parentWorld eval: [
			self class firstCommittableIV to: iSize do: [:ii |
				val := delta at: ii.
				val == dnk ifFalse: [ parentWorld valueAt: ii put: val for: self ]
				]
			]
		]
! !

!WObject methodsFor: 'copying' stamp: 'ael 11/12/2011 11:51'!
copy
	^self worldFlattenedCopy
! !

!WObject methodsFor: 'world awareness' stamp: 'ael 3/2/2012 01:11'!
copyMissingOrigValuesInto: parentWorld using: updateTrackers
	"updateTrackers is a tuple { orig. delta }.
	as part of commit, copy into the parent world's reads record any values that were read by the child-level object but came from an ancestor beyond the parent (i.e., the parent has recorded neither a write nor a read for this slot).

	This is important for maintaining the validity of serialisability checks.  Suppose world A has a child B, that has a child C.  In world C some value V is read, being delivered from A.  When C's changes are committed into B, the read record for V is copied to B so that if B is later committed into A the serialisability check will look for consistency in values such as V, from which some values in C - and therefore now in B - may have been calculated.

	This method is invoked in the child world."
	| parentReads childReads val parentTrackers dnk parentWrites |
	parentWorld tracksReads ifFalse: [ ^self ].		"parent not interested"
	childReads := updateTrackers at: 1.
	childReads ifNil: [ ^self ].							"nothing to tell it anyway"

	"we can ask for assured update trackers in the parent world, because either we'll be writing into the reads record or we'll be prevented from doing so by the parent's writes record (in which case the trackers exist already)"
	parentTrackers := parentWorld ensureUpdateTrackersFor: self.
	
	"there may or may not be a writes record.  if not, that makes our checking simpler"
	parentWrites := self writesRecordIn: parentTrackers.	"may be nil"

	"there may be a reads record already.  if not, don't create one until we're sure we need it."
	parentReads := self readsRecordIn: parentTrackers.	"may be nil"
	dnk := self dnkToken.
	
	self class firstCommittableIV to: self iSize do: [:ii |
		val := childReads at: ii.
		(val ~~ dnk
		  and: [ (parentWrites isNil or: [ (parentWrites at: ii) == dnk ])
				 and: [ ((parentReads ifNil: [ parentReads := self iEnsureReadsRecordIn: parentTrackers ])
						at: ii) == dnk ] ])
		  ifTrue: [ parentReads at: ii put: val ]
		]
! !

!WObject methodsFor: 'debugging' stamp: 'ael 5/21/2012 21:28'!
debugUpdateTrackerState
	^self debugUpdateTrackerStateIn: self thisWorld
! !

!WObject methodsFor: 'debugging' stamp: 'ael 5/21/2012 21:28'!
debugUpdateTrackerStateIn: aWorld
	| trackers |
	trackers := aWorld updateTrackersFor: self.
	^trackers ifNotNil: [ trackers collect: [ :x | x clone ] ].
! !

!WObject methodsFor: 'world awareness' stamp: 'ael 10/31/2011 22:28'!
deeplyWorldFlattenedIn: aWorld with: doneSoFar
	| flatMe |
	doneSoFar at: self ifPresent: [ :flat | ^flat ].
	
	WTopLevelWorld default eval: [
		doneSoFar at: self put: (flatMe := self worldFlattenIn: aWorld).
	
		self class firstCommittableIV to: self iSize do: [ :index |
			flatMe basicValueAt: index put: ((flatMe basicValueAt: index) deeplyWorldFlattenedIn: aWorld with: doneSoFar).
			]
		].
	
	^flatMe

! !

!WObject methodsFor: 'constants' stamp: 'ael 6/25/2011 12:34'!
deletionToken
	"a special value that can be recorded as having been written to a slot, to mean that the slot was removed"
	^ WorldDeletionToken new
! !

!WObject methodsFor: 'constants' stamp: 'ael 9/26/2011 07:57'!
dnkToken

	^ WorldDNKToken new! !

!WObject methodsFor: 'world awareness' stamp: 'ael 6/22/2011 15:58'!
evalInParentWorld: aBlock
	^self evalInParentWorld: aBlock ifNone: [ nil ]
! !

!WObject methodsFor: 'world awareness' stamp: 'ael 9/25/2011 21:40'!
evalInParentWorld: aBlock ifNone: failBlock
	| w |
	w := self thisWorld parent.
	w ifNil: [ ^failBlock value ].
	
	^w eval: aBlock
! !

!WObject methodsFor: 'world awareness' stamp: 'ael 5/16/2012 12:47'!
evalInTopWorld: aBlock
	^WTopLevelWorld default eval: aBlock
! !

!WObject methodsFor: 'world awareness' stamp: 'ael 10/27/2011 14:36'!
iChangedOrNewInWorld: world
	| trackers writesTarget dnk |
	trackers := (world updateTrackersFor: self) ifNil: [ ^false ].
	
	"updateTrackers for a standard WObject is a tuple { reads. writes }"
	writesTarget := (trackers at: 2) ifNil: [ ^false ].		"nothing written yet"
	dnk := self dnkToken.
	^writesTarget anySatisfy: [ :val | val ~~ dnk ]
! !

!WObject methodsFor: 'world awareness' stamp: 'ael 9/23/2011 16:08'!
iEnsureReadsRecordIn: updateTrackers
	"updateTrackers for a standard WObject is a tuple { reads. writes }
	
	This is only sent if the world is set to tracksReads=true."
	| reads |
	(reads := updateTrackers at: 1) ifNotNil: [ ^reads ].

	^updateTrackers at: 1 put: self iMakeDeltaOrOrig
! !

!WObject methodsFor: 'world awareness' stamp: 'ael 9/23/2011 16:08'!
iEnsureWriteTargetIn: updateTrackers
	"updateTrackers for a standard WObject is a tuple { reads. writes }"
	| writes |
	(writes := updateTrackers at: 2) ifNotNil: [ ^writes ].

	^updateTrackers at: 2 put: self iMakeDeltaOrOrig
! !

!WObject methodsFor: 'world awareness' stamp: 'ael 9/26/2011 08:01'!
iMakeDeltaOrOrig

	^ Array new: self iSize withAll: self dnkToken
! !

!WObject methodsFor: 'world awareness' stamp: 'ael 1/13/2012 21:15'!
iMergeUpdatesFrom: otherTrackers into: updateTrackers
	"updateTrackers for a standard WObject is a tuple { reads. writes }"
	| dnk oldWrites otherWrites val |
	oldWrites := updateTrackers at: 2.
	otherWrites := otherTrackers at: 2.
	
	oldWrites ifNil: [ ^updateTrackers at: 2 put: otherWrites ].

	dnk := self dnkToken.
	self class firstCommittableIV to: self iSize do: [ :ii |
		(val := otherWrites at: ii) == dnk
		  ifFalse: [ oldWrites at: ii put: val ]
		]
! !

!WObject methodsFor: 'world awareness' stamp: 'aw 3/26/2010 15:22'!
iSize

	^ self class instSize! !

!WObject methodsFor: 'world awareness' stamp: 'ael 6/23/2011 10:18'!
iUpdateTrackerSize
	"updateTrackers for a standard WObject is a tuple { orig. delta }"
	^2
! !

!WObject methodsFor: 'world awareness' stamp: 'ael 1/13/2012 19:28'!
iUpdatesIn: updateTrackers
	"updateTrackers for a standard WObject is a tuple { reads. writes }"
	| writes |
	writes := updateTrackers at: 2.
	^writes ifNotNil: [ { nil. writes } ].
! !

!WObject methodsFor: 'world awareness' stamp: 'ael 10/27/2011 14:36'!
iVarsChangedInWorld: world
	| indices ivNames trackers writesTarget dnk |
	trackers := (world updateTrackersFor: self) ifNil: [ ^#() ].
	
	"updateTrackers for a standard WObject is a tuple { reads. writes }"
	writesTarget := (trackers at: 2) ifNil: [ ^#() ].		"nothing written yet"

	dnk := self dnkToken.
	indices := (self class firstRedirectedIV to: writesTarget iSize) reject: [ :ii |
		(writesTarget at: ii) == dnk.
		].

	ivNames := self class allInstVarNames.
	^indices collect: [ :i | ivNames at: i ]
! !

!WObject methodsFor: 'copying' stamp: 'ael 11/12/2011 11:33'!
initializedClone
	"return an instance that's ready to record values read from the original"
	^self class new
! !

!WObject methodsFor: 'accessing' stamp: 'ael 9/30/2011 21:35'!
instVarAt: index
	"if the world already has update trackers for this object, use that. 
	but don't force creation of trackers: in a world that does not cache reads;
	we may never need them."
	| thisWorld |
	(thisWorld := self thisWorld) isTopLevel ifTrue: [ ^super instVarAt: index ].

	^thisWorld valueAt: index for: self using: (thisWorld updateTrackersFor: self)
! !

!WObject methodsFor: 'accessing' stamp: 'ael 10/2/2011 11:47'!
instVarAt: index put: value
	| thisWorld |
	(thisWorld := self thisWorld) isTopLevel ifTrue: [ ^self basicValueAt: index put: value ].

	^thisWorld valueAt: index put: value for: self
! !

!WObject methodsFor: 'world awareness' stamp: 'ael 12/6/2011 14:24'!
isSerializableIn: parentWorld using: updateTrackers
	"updateTrackers is a tuple { orig. delta }.
	make sure that, for all non-delegating keys in orig objects, the value stored in there == that value, if looked up in the parent world "
	| orig dnk currentVal |
	orig := updateTrackers at: 1.
	orig ifNotNil: [		"nil if nothing was read from parent, or this world has tracksReads=false" 
		dnk := self dnkToken.
		self class firstSerialisationIV to: self iSize do: [:ii |
			| val |
			val := orig at: ii.
			(val ~~ dnk
			  and: [ (self areEqualForCommit: val and: (currentVal := self nonCachingInstVarAt: ii in: parentWorld)) not ])
				ifTrue: [
					self reportSerializationErrorAt: ii oldValue: val newValue: currentVal.
					^false
					]
			]
		].
	^true
! !

!WObject methodsFor: 'accessing' stamp: 'ael 9/26/2011 17:16'!
nonCachingInstVarAt: index in: aWorld
	^ aWorld nonCachingInstVarAt: index for: self writesSlot: 2 readsSlot: 1
! !

!WObject methodsFor: 'accessing' stamp: 'ael 9/26/2011 17:18'!
nonCachingInstVarNamed: ivName in: aWorld

	| ivIndex |
	ivIndex := self class instVarIndexFor: ivName asString
						ifAbsent: [self error: 'no such inst var'].
	^self nonCachingInstVarAt: ivIndex in: aWorld
! !

!WObject methodsFor: 'constants' stamp: 'ael 6/25/2011 12:35'!
overwriteToken
	"a special value that can be recorded as having been written to a slot, to mean that the slot was removed"
	^ WorldOverwriteToken new
! !

!WObject methodsFor: 'copying' stamp: 'ael 11/12/2011 11:10'!
postCopy
	"nothing to do"! !

!WObject methodsFor: 'copying' stamp: 'ael 9/23/2011 15:27'!
postCopyTrackers: oldUTs into: newUTs
	"updateTrackers is a tuple { reads. writes }"
	1 to: oldUTs size do: [ :i | newUTs at: i put: (oldUTs at: i) copy ].
! !

!WObject methodsFor: 'copying' stamp: 'ael 11/12/2011 11:28'!
postFlattenedCopy
	"by default, do normal postCopy"
	self postCopy
! !

!WObject methodsFor: 'primitives' stamp: 'ael 11/9/2011 18:12'!
primDepthFirstInBlocksFlattenInWorlds: worlds to: dst startingAt: startIndex blockSize: blockSize writesSlot: writesSlot delegate: dnk
	<primitive: 'primDepthFirstInBlocksFlattenObjInto' module: 'WorldsPlugin'>
	worlds last nonPluginFlatten: self into: dst startingAt: startIndex delegate: dnk
	"WorldsPlugin doPrimitive: 'primDepthFirstInBlocksFlattenObjInto'".
! !

!WObject methodsFor: 'primitives' stamp: 'yo 3/26/2010 23:16'!
primFillWith: val

	<primitive: 'fillObject' module: 'W2Plugin'>
	1 to: self iSize do: [:i | super instVarAt: i put: val].
! !

!WObject methodsFor: 'primitives' stamp: 'ael 11/9/2011 18:12'!
primSparseFlattenInWorlds: worlds to: dst startingAt: startIndex writesSlot: writesSlot delegate: dnk
	<primitive: 'primSparseFlattenObjInto' module: 'WorldsPlugin'>
	worlds last nonPluginFlatten: self into: dst startingAt: startIndex delegate: dnk
	"WorldsPlugin doPrimitive: 'primSparseFlattenObjInto'."
! !

!WObject methodsFor: 'world awareness' stamp: 'ael 12/5/2011 22:20'!
readsRecordIn: updateTrackers
	"updateTrackers for a standard WObject is a tuple { orig. delta }"
	^ updateTrackers at: 1
! !

!WObject methodsFor: 'accessing' stamp: 'ael 9/30/2011 21:39'!
realWorldInstVarAt: index
	"for objects that positively need to avoid the world mechanisms"
	^super instVarAt: index
! !

!WObject methodsFor: 'accessing' stamp: 'ael 11/9/2011 18:13'!
realWorldInstVarAt: index put: value
	"for objects that positively have to avoid the world mechanisms"
	^super instVarAt: index put: value
! !

!WObject methodsFor: 'exceptions' stamp: 'ael 12/6/2011 15:02'!
reportSerializationErrorAt: index oldValue: readValue newValue: currentValue
	^(WorldValueChanged new indexPrefix: 'index' index: index readValue: readValue currentValue: currentValue) signal! !

!WObject methodsFor: 'world awareness' stamp: 'ael 9/20/2011 16:36'!
thisWorld

	^ Processor activeProcess currentWWorld! !

!WObject methodsFor: 'copying' stamp: 'ael 10/31/2011 22:28'!
veryDeepCopyWith: deepCopier
	"Copy me and the entire tree of objects I point to.  An object in the tree twice is copied once, and both references point to him.  deepCopier holds a dictionary of objects we have seen.  Some classes refuse to be copied.  Some classes are picky about which fields get deep copied."
	| class index sub subAss new uc sup has mine thisWorld |
	deepCopier references at: self ifPresent: [:newer | ^ newer]. 	"already did him"

	(thisWorld := self thisWorld) isTopLevel
	  ifTrue: [ ^super veryDeepCopyWith: deepCopier ].
	
  WTopLevelWorld default eval: [
	class := self class.
	(class isSystemDefined not and: [deepCopier newUniClasses "allowed"])
	  ifTrue: [
		uc := deepCopier uniClasses at: class ifAbsent: [nil].
		uc ifNil: [
			deepCopier uniClasses at: class put: (uc := self copyUniClassWith: deepCopier).
			deepCopier references at: class put: uc].	"remember"
		new := uc new.
		new copyFrom: self]	"copy inst vars in case any are weak"
	  ifFalse: [ new := self worldFlattenIn: thisWorld ].			"creates a shallow copy, with everything in the top world.  However, the copying of each slot value - potentially a world-aware object - must be done in the world in which this message was sent."

	deepCopier references at: self put: new.	"remember"
	(class isVariable and: [class isPointers]) ifTrue: 
		[index := self basicSize.
		[index > 0] whileTrue: 
			[sub := new basicAt: index.
			(subAss := deepCopier references associationAt: sub ifAbsent: [nil])
				ifNil: [new basicAt: index put: (thisWorld eval: [ sub veryDeepCopyWith: deepCopier ])]
				ifNotNil: [new basicAt: index put: subAss value].
			index := index - 1]].
	"Ask each superclass if it wants to share (weak copy) any inst vars"
	new veryDeepInner: deepCopier.		"does super a lot"

	"other superclasses want all inst vars deep copied"
	sup := class.  index := class instSize.
	[has := sup compiledMethodAt: #veryDeepInner: ifAbsent: [nil].
	has := has ifNil: [class isSystemDefined not "is a uniClass"] ifNotNil: [true].
	mine := sup instVarNames.
	has ifTrue: [index := index - mine size]	"skip inst vars"
		ifFalse: [1 to: mine size do: [:xx |
				sub := new instVarAt: index.
				(subAss := deepCopier references associationAt: sub ifAbsent: [nil])
						"use association, not value, so nil is an exceptional value"
					ifNil: [new instVarAt: index put: (thisWorld eval: [ sub veryDeepCopyWith: deepCopier ])]
					ifNotNil: [new instVarAt: index put: subAss value].
				index := index - 1]].
	(sup := sup superclass) == nil] whileFalse.
	new rehash.	"force Sets and Dictionaries to rehash"
  ].
	^ new
! !

!WObject methodsFor: 'world awareness' stamp: 'ael 11/3/2011 14:17'!
worldFlattenIn: aWorld
	| dest |
	dest := self class basicNew.
	aWorld flatten: self into: dest.
	^dest
! !

!WObject methodsFor: 'world awareness' stamp: 'ael 11/11/2011 20:18'!
worldFlattened
	^self worldFlattenIn: self thisWorld
! !

!WObject methodsFor: 'copying' stamp: 'ael 11/12/2011 11:53'!
worldFlattenedCopy
	"get a flattened copy of the receiver by looking up each value in turn - thus keeping appropriate records of reads from parent worlds"
	| clone thisWorld trackers |
	(thisWorld := self thisWorld) isTopLevel ifTrue: [ ^self clone postCopy ].

	clone := self initializedClone.
	trackers := thisWorld updateTrackersFor: self.
	WTopLevelWorld default eval: [
		1 to: self iSize do: [ :index | clone basicValueAt: index put: (thisWorld valueAt: index for: self using: trackers) ].
		clone postFlattenedCopy.
		].
	^clone
! !

!WObject methodsFor: 'world awareness' stamp: 'ael 11/1/2011 19:29'!
writesRecordIn: updateTrackers
	"updateTrackers for a standard WObject is a tuple { orig. delta }"
	^ updateTrackers at: 2
! !

WObject subclass: #WObjectWithCachedLookup
	instanceVariableNames: 'lookupCache'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Worlds'!

WObjectWithCachedLookup subclass: #WArray
	instanceVariableNames: 'contents'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Worlds'!

!WArray commentStamp: 'ael 1/13/2012 21:11' prior: 0!
In WArray the contents IV is non-redirected, and is therefore accessible from all worlds (e.g., providing a fast way to discover the collection's size).  The values it holds apply only to the top-level world.

Array elements are accessed using instVarAt:, which in the top world calls on basicValueAt: and in child worlds calls on WWorld>>valueAt:for:using:, managing Reads and Writes records in the ancestor chain.
!

!WArray class methodsFor: 'as yet unclassified' stamp: 'aw 9/4/2009 13:48'!
braceStream: nElements
	"This method is used in compilation of brace constructs.
	It MUST NOT be deleted or altered."

	^ WriteStream basicNew braceArray: (self new: nElements)
! !

!WArray class methodsFor: 'as yet unclassified' stamp: 'aw 9/4/2009 13:48'!
braceWith: a
	"This method is used in compilation of brace constructs.
	It MUST NOT be deleted or altered."

	| array |
	array := self new: 1.
	array at: 1 put: a.
	^ array! !

!WArray class methodsFor: 'as yet unclassified' stamp: 'aw 9/4/2009 13:47'!
braceWith: a with: b 
	"This method is used in compilation of brace constructs.
	It MUST NOT be deleted or altered."

	| array |
	array := self new: 2.
	array at: 1 put: a.
	array at: 2 put: b.
	^ array! !

!WArray class methodsFor: 'as yet unclassified' stamp: 'aw 9/4/2009 13:47'!
braceWith: a with: b with: c 
	"This method is used in compilation of brace constructs.
	It MUST NOT be deleted or altered."

	| array |
	array := self new: 3.
	array at: 1 put: a.
	array at: 2 put: b.
	array at: 3 put: c.
	^ array! !

!WArray class methodsFor: 'as yet unclassified' stamp: 'aw 9/4/2009 13:47'!
braceWith: a with: b with: c with: d
	"This method is used in compilation of brace constructs.
	It MUST NOT be deleted or altered."

	| array |
	array := self new: 4.
	array at: 1 put: a.
	array at: 2 put: b.
	array at: 3 put: c.
	array at: 4 put: d.
	^ array! !

!WArray class methodsFor: 'as yet unclassified' stamp: 'aw 9/4/2009 13:48'!
braceWithNone
	"This method is used in compilation of brace constructs.
	It MUST NOT be deleted or altered."

	^ self new: 0! !

!WArray class methodsFor: 'as yet unclassified' stamp: 'ael 10/24/2011 16:52'!
firstCommittableIV
	^1				"first index in the array"
! !

!WArray class methodsFor: 'as yet unclassified' stamp: 'ael 9/24/2011 12:06'!
firstRedirectedIV
	^99				"we don't want contents IV to be redirected"! !

!WArray class methodsFor: 'as yet unclassified' stamp: 'ael 10/24/2011 16:52'!
firstSerialisationIV
	^1				"first index in the array"
! !

!WArray class methodsFor: 'as yet unclassified' stamp: 'ael 12/1/2011 22:21'!
new
	^self new: 0
! !

!WArray class methodsFor: 'as yet unclassified' stamp: 'ael 12/1/2011 22:21'!
new: size

	^ super new initialize: size! !

!WArray class methodsFor: 'as yet unclassified' stamp: 'aw 3/23/2010 16:54'!
new: size withAll: val

	| r |
	r := self new: size.
	1 to: size do: [:idx | r at: idx put: val].
	^ r! !

!WArray class methodsFor: 'as yet unclassified' stamp: 'ael 9/30/2011 20:13'!
ofSize: n
	"Create a new collection of size n with nil as its elements.
	This method exists because OrderedCollection new: n creates an
	empty collection,  not one of size n."
	^ self new: n! !

!WArray class methodsFor: 'as yet unclassified' stamp: 'ael 11/2/2011 12:33'!
simpleBenchmark
	"WArray simpleBenchmark"
	"WWorld runWithoutPlugin: [ WArray simpleBenchmark ]"
	| a numWorlds worlds arraySize worldIndex arrayIndex |
	worlds := OrderedCollection new.
	worlds add: WTopLevelWorld default.
	numWorlds := 10.
	numWorlds - 1 timesRepeat: [ worlds add: (worlds last sprout tracksReads: false) ].
	worlds := worlds asArray.

	arraySize := 1000.
	a := WArray new: arraySize withAll: 1.
	1 to: arraySize*numWorlds//4 do: [:x |
		worldIndex := numWorlds atRandom.
		arrayIndex := arraySize atRandom.
		(worlds at: worldIndex) eval: [ a at: arrayIndex put: numWorlds atRandom ].
		].

	worldIndex := 10.

	QSystemProfiler spyOn: 
	[
	10000 timesRepeat: [
		1 to: arraySize do: [ :ai |
			worldIndex := (worlds at: worldIndex) eval: [ a at: ai ].
			].
		].
	]
! !

!WArray class methodsFor: 'lobjects-siss-private' stamp: 'tak 9/9/2010 10:28'!
sissCreateInstanceFromSexp: sexp idref: idref from: from to: to 
	| inst i childInst |
	inst := self sissNew: self withSexp: sexp.
	idref ifNotNil: [to at: idref put: inst].
	sexp elements do: [:elem | 
			i := Integer readFromString: (elem attributeAt: #name ifAbsent: []).
			childInst := self fromSexp: elem from: from to: to.
			inst at: i put: childInst].
	^ inst! !

!WArray class methodsFor: 'as yet unclassified' stamp: 'aw 9/4/2009 12:19'!
with: e1

	^ (self new: 1)
		at: 1 put: e1;
		yourself! !

!WArray class methodsFor: 'as yet unclassified' stamp: 'aw 9/4/2009 12:19'!
with: e1 with: e2

	^ (self new: 2)
		at: 1 put: e1;
		at: 2 put: e2;
		yourself! !

!WArray class methodsFor: 'as yet unclassified' stamp: 'aw 9/4/2009 12:19'!
with: e1 with: e2 with: e3

	^ (self new: 3)
		at: 1 put: e1;
		at: 2 put: e2;
		at: 3 put: e3;
		yourself! !

!WArray class methodsFor: 'as yet unclassified' stamp: 'aw 9/4/2009 12:20'!
with: e1 with: e2 with: e3 with: e4

	^ (self new: 4)
		at: 1 put: e1;
		at: 2 put: e2;
		at: 3 put: e3;
		at: 4 put: e4;
		yourself! !

!WArray class methodsFor: 'as yet unclassified' stamp: 'aw 9/4/2009 12:20'!
with: e1 with: e2 with: e3 with: e4 with: e5

	^ (self new: 4)
		at: 1 put: e1;
		at: 2 put: e2;
		at: 3 put: e3;
		at: 4 put: e4;
		at: 5 put: e5;
		yourself! !

!WArray class methodsFor: 'as yet unclassified' stamp: 'aw 9/4/2009 12:20'!
with: e1 with: e2 with: e3 with: e4 with: e5 with: e6

	^ (self new: 4)
		at: 1 put: e1;
		at: 2 put: e2;
		at: 3 put: e3;
		at: 4 put: e4;
		at: 5 put: e5;
		at: 6 put: e6;
		yourself! !

!WArray class methodsFor: 'as yet unclassified' stamp: 'aw 9/4/2009 13:46'!
withAll: xs

	| r |
	r := self new: xs size.
	1 to: xs size do: [:idx | r at: idx put: (xs at: idx)].
	^ r! !

!WArray methodsFor: 'copying' stamp: 'ael 9/20/2011 17:32'!
, otherCollection
	| size |
	size := self size.
	^ self copyReplaceFrom: size + 1
		  to: size
		  with: otherCollection! !

!WArray methodsFor: 'comparing' stamp: 'ael 9/30/2011 16:17'!
= otherCollection 
	"Answer true if the receiver is equivalent to the otherCollection.
	First test for identity, then rule out different species and sizes of
	collections. As a last resort, examine each element of the receiver
	and the otherCollection."

	self == otherCollection ifTrue: [^ true].
	self species == otherCollection species ifFalse: [^ false].
	^ self hasEqualElements: otherCollection! !

!WArray methodsFor: 'adding' stamp: 'ael 9/27/2011 12:13'!
add: newObject
	self shouldNotImplement! !

!WArray methodsFor: 'adding' stamp: 'ael 9/27/2011 12:15'!
add: newObject withOccurrences: anInteger
	self shouldNotImplement! !

!WArray methodsFor: 'adding' stamp: 'ael 9/27/2011 12:15'!
addAll: aCollection 
	self shouldNotImplement! !

!WArray methodsFor: 'adding' stamp: 'ael 9/27/2011 12:15'!
addIfNotPresent: anObject
	self shouldNotImplement! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/27/2011 12:15'!
after: target
	"Answer the element after target.  Raise an error if target is not
	in the receiver, or if there are no elements after it."

	^ self after: target ifAbsent: [self errorNotFound: target]! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/27/2011 12:16'!
after: target ifAbsent: exceptionBlock
	"Answer the element after target.  Answer the result of evaluation
	the exceptionBlock if target is not in the receiver, or if there are 
	no elements after it."

	| index |
	index := self indexOf: target.
	^ (index = 0 or: [index = self size])
		ifTrue: [exceptionBlock value]
		ifFalse: [self at: index + 1]! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/27/2011 12:16'!
allButFirst
	"Answer a copy of the receiver containing all but the first
	element. Raise an error if there are not enough elements."

	^ self allButFirst: 1! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/27/2011 12:17'!
allButFirst: n
	"Answer a copy of the receiver containing all but the first n
	elements. Raise an error if there are not enough elements."

	^ self copyFrom: n + 1 to: self size! !

!WArray methodsFor: 'enumerating' stamp: 'ael 9/27/2011 12:17'!
allButFirstDo: block

	2 to: self size do:
		[:index | block value: (self at: index)]! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/27/2011 12:17'!
allButLast
	"Answer a copy of the receiver containing all but the last
	element. Raise an error if there are not enough elements."

	^ self allButLast: 1! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/27/2011 12:18'!
allButLast: n
	"Answer a copy of the receiver containing all but the last n
	elements. Raise an error if there are not enough elements."

	^ self copyFrom: 1 to: self size - n! !

!WArray methodsFor: 'enumerating' stamp: 'ael 9/27/2011 12:17'!
allButLastDo: block

	1 to: self size - 1 do:
		[:index | block value: (self at: index)]! !

!WArray methodsFor: 'enumerating' stamp: 'aw 9/9/2009 12:01'!
allSatisfy: aBlock

	self do: [:x | (aBlock value: x) ifFalse: [^ false]].
	^ true! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/27/2011 12:18'!
anyOne
	^ self first! !

!WArray methodsFor: 'enumerating' stamp: 'aw 9/9/2009 12:00'!
anySatisfy: aBlock

	self do: [:x | (aBlock value: x) ifTrue: [^ true]].
	^ false! !

!WArray methodsFor: 'converting' stamp: 'yo 9/28/2010 18:01'!
asArray

! !

!WArray methodsFor: 'converting' stamp: 'ael 9/27/2011 12:19'!
asBag
	"Answer a Bag whose elements are the elements of the receiver."

	^ Bag withAll: self! !

!WArray methodsFor: 'converting' stamp: 'ael 9/27/2011 12:19'!
asByteArray
	"Answer a ByteArray whose elements are the elements of the receiver."

	^ ByteArray withAll: self! !

!WArray methodsFor: 'converting' stamp: 'ael 9/27/2011 12:19'!
asCharacterSet
	"Answer a CharacterSet whose elements are the unique elements of the receiver.
	The reciever should only contain characters."

	^ CharacterSet newFrom: self! !

!WArray methodsFor: 'converting' stamp: 'ael 9/27/2011 12:19'!
asColorArray
	^ColorArray withAll: self! !

!WArray methodsFor: 'converting' stamp: 'ael 9/27/2011 12:20'!
asCommaString
	"Return collection printed as 'a, b, c' "

	^String streamContents: [:s | self asStringOn: s delimiter: ', ']
		! !

!WArray methodsFor: 'converting' stamp: 'ael 9/27/2011 12:20'!
asCommaStringAnd
	"Return collection printed as 'a, b and c' "

	^String streamContents: [:s | self asStringOn: s delimiter: ', ' last: ' and ']
		! !

!WArray methodsFor: 'converting' stamp: 'ael 9/27/2011 12:21'!
asFloatArray
	"Answer a FloatArray whose elements are the elements of the receiver, in 
	the same order."

	| floatArray |
	floatArray := FloatArray new: self size.
	1 to: self size do:[:i| floatArray at: i put: (self at: i) asFloat ].
	^floatArray! !

!WArray methodsFor: 'converting' stamp: 'ael 9/27/2011 12:21'!
asIdentitySet
	^(IdentitySet new: self size) addAll: self; yourself! !

!WArray methodsFor: 'converting' stamp: 'ael 9/27/2011 12:21'!
asIntegerArray
	"Answer an IntegerArray whose elements are the elements of the receiver, in 
	the same order."

	| intArray |
	intArray := IntegerArray new: self size.
	1 to: self size do:[:i| intArray at: i put: (self at: i)].
	^intArray! !

!WArray methodsFor: 'converting' stamp: 'ael 9/25/2011 21:24'!
asOrderedCollection
	^WCollection new addAll: self; yourself
! !

!WArray methodsFor: 'converting' stamp: 'ael 9/27/2011 12:22'!
asSet
	"Answer a Set whose elements are the unique elements of the receiver."

	^ Set withAll: self! !

!WArray methodsFor: 'converting' stamp: 'ael 9/30/2011 19:14'!
asSortedArray
	self isSorted ifTrue: [^ self asArray].
	^ WArray withAll: self asSortedCollection! !

!WArray methodsFor: 'converting' stamp: 'ael 9/30/2011 19:16'!
asSortedCollection
	"Answer a SortedCollection whose elements are the elements of the 
	receiver. The sort order is the default less than or equal. 
	Use #sorted: if you don't really need a SortedCollection, but a sorted collection!!"
	
	^SortedCollection withAll: self! !

!WArray methodsFor: 'converting' stamp: 'ael 9/27/2011 12:23'!
asSortedCollection: aSortBlock 
	"Answer a SortedCollection whose elements are the elements of the 
	receiver. The sort order is defined by the argument, aSortBlock.
	Use #sorted: if you don't really need a SortedCollection, but a sorted collection!!"

	| aSortedCollection |
	aSortedCollection := SortedCollection new: self size.
	aSortedCollection sortBlock: aSortBlock.
	aSortedCollection addAll: self.
	^ aSortedCollection! !

!WArray methodsFor: 'printing' stamp: 'ael 9/27/2011 12:24'!
asStringOn: aStream delimiter: delimString
	"Print elements on a stream separated
	with a delimiter String like: 'a, b, c'
	Uses #asString instead of #print:."

	self do: [:elem | aStream nextPutAll: elem asString]
		separatedBy: [aStream nextPutAll: delimString]! !

!WArray methodsFor: 'printing' stamp: 'ael 9/27/2011 12:25'!
asStringOn: aStream delimiter: delimString last: lastDelimString
	"Print elements on a stream separated
	with a delimiter between all the elements and with
	a special one before the last like: 'a, b and c'.
	Uses #asString instead of #print:

	Note: Feel free to improve the code to detect the last element."

	| n sz |
	n := 1.
	sz := self size.
	self do: [:elem |
		n := n + 1.
		aStream nextPutAll: elem asString]
	separatedBy: [
		aStream nextPutAll: (n = sz ifTrue: [lastDelimString] ifFalse: [delimString])]! !

!WArray methodsFor: 'converting' stamp: 'ael 9/27/2011 12:26'!
asStringWithCr
	"Convert to a string with returns between items.  Elements are
usually strings.
	 Useful for labels for PopUpMenus."
	| labelStream |
	labelStream := WriteStream on: (String new: 200).
	self do: [:each |
		each isString
			ifTrue: [labelStream nextPutAll: each; cr]
			ifFalse: [each printOn: labelStream. labelStream cr]].
	self size > 0 ifTrue: [labelStream skip: -1].
	^ labelStream contents! !

!WArray methodsFor: 'converting' stamp: 'ael 9/27/2011 12:26'!
asWordArray
	"Answer a WordArray whose elements are the elements of the receiver, in 
	the same order."

	| wordArray |
	wordArray := WordArray new: self size.
	1 to: self size do:[:i| wordArray at: i put: (self at: i)].
	^wordArray! !

!WArray methodsFor: 'enumerating' stamp: 'ael 9/27/2011 12:26'!
associationsDo: aBlock
	"Evaluate aBlock for each of the receiver's elements (key/value 
	associations).  If any non-association is within, the error is not caught now,
	but later, when a key or value message is sent to it."

	self do: aBlock! !

!WArray methodsFor: 'accessing' stamp: 'ael 10/2/2011 12:30'!
at: index
	| thisWorld trackers |
	(thisWorld := self thisWorld) isTopLevel ifTrue: [ ^contents at: index ].

	"if we already have update trackers for this world in our cache, use that. 
	but don't force creation of trackers: if the world doesn't cache reads, 
	we may never need them."
	trackers := self updateTrackersForWorld: thisWorld createIfAbsent: false caching: true.
	^thisWorld valueAt: index for: self using: trackers.
! !

!WArray methodsFor: 'accessing' stamp: 'tk 7/8/2010 14:37'!
at: index ifAbsent: exceptionBlock 
	"Answer the element at my position index. If I do not contain an element at index, answer the result of evaluating the argument, exceptionBlock."

	(index between: 1 and: self size) ifTrue: [^ self at: index].
	^ exceptionBlock value! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/27/2011 12:27'!
at: index incrementBy: value
	^self at: index put: (self at: index) + value! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/30/2011 21:46'!
at: index put: value
	| thisWorld trackers |
	(thisWorld := self thisWorld) isTopLevel ifTrue: [ ^contents at: index put: value ].

	trackers := self updateTrackersForWorld: thisWorld createIfAbsent: true caching: true.
	^thisWorld valueAt: index put: value for: self using: trackers.
! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/27/2011 12:27'!
atAll: indexArray
	"Answer a new collection like the receiver which contains all elements
	of the receiver at the indices of indexArray."
	"#('one' 'two' 'three' 'four') atAll: #(3 2 4)"

	| newCollection |
	newCollection := self species ofSize: indexArray size.
	1 to: indexArray size do:
		[:index |
		newCollection at: index put: (self at: (indexArray at: index))].
	^ newCollection! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/27/2011 12:27'!
atAll: aCollection put: anObject 
	"Put anObject at every index specified by the elements of aCollection."

	aCollection do: [:index | self at: index put: anObject].
	^ anObject! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/27/2011 12:27'!
atAll: indexArray putAll: valueArray
	"Store the elements of valueArray into the slots
	of this collection selected by indexArray."

	indexArray with: valueArray do: [:index :value | self at: index put: value].
	^ valueArray! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/27/2011 12:28'!
atAllPut: anObject 
	"Put anObject at every one of the receiver's indices."

	| size |
	(size := self size) > 26 "first method faster from 27 accesses and on"
		ifTrue: [self from: 1 to: size put: anObject]
		ifFalse: [1 to: size do: [:index | self at: index put: anObject]]! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/27/2011 12:28'!
atLast: indexFromEnd
	"Return element at indexFromEnd from the last position.
	 atLast: 1, returns the last element"

	^ self atLast: indexFromEnd ifAbsent: [self error: 'index out of range']! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/27/2011 12:28'!
atLast: indexFromEnd ifAbsent: block
	"Return element at indexFromEnd from the last position.
	 atLast: 1 ifAbsent: [] returns the last element"

	^ self at: self size + 1 - indexFromEnd ifAbsent: block! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/27/2011 12:28'!
atLast: indexFromEnd put: obj
	"Set the element at indexFromEnd from the last position.
	 atLast: 1 put: obj, sets the last element"

	^ self at: self size + 1 - indexFromEnd put: obj! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/27/2011 12:28'!
atPin: index 
	"Return the index'th element of me if possible.
	Return the first or last element if index is out of bounds."

	index < 1 ifTrue: [^ self first].
	index > self size ifTrue: [^ self last].
	^ self at: index! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/30/2011 20:22'!
atRandom
	"Answer a random element of the receiver.  Uses a shared random 
	number generator owned by class Collection.  If you use this a lot, 
	define your own instance of Random and use #atRandom:.  Causes 
	an error if self has no elements."

	^ Collection mutexForPicking critical: [
		self atRandom: Collection randomForPicking ]

"Examples:
	#('one' 'or' 'the' 'other') atRandom
	(1 to: 10) atRandom
	'Just pick one of these letters at random' atRandom
	#(3 7 4 9 21) asSet atRandom		(just to show it also works for Sets)
"! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/27/2011 12:29'!
atRandom: aGenerator
	"Answer a random element of the receiver.  Uses aGenerator which
	should be kept by the user in a variable and used every time. Use
	this instead of #atRandom for better uniformity of random numbers 
	because only you use the generator.  Causes an error if self has no 
	elements."

	^ self at: (aGenerator nextInt: self size)! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/27/2011 12:29'!
atWrap: index 
	^ self at: index - 1 \\ self size + 1! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/27/2011 12:29'!
atWrap: index put: anObject
	^ self at: index - 1 \\ self size + 1 put: anObject! !

!WArray methodsFor: 'math functions' stamp: 'ael 9/27/2011 12:30'!
average
	^ self sum / self size! !

!WArray methodsFor: 'world awareness' stamp: 'ael 10/27/2011 15:34'!
basicValueAt: index
	"obtain a value from the top-level world"
	^ contents at: index! !

!WArray methodsFor: 'world awareness' stamp: 'ael 10/2/2011 11:51'!
basicValueAt: index put: value
	"sent in top-level world only"
	^ contents at: index put: value
! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/30/2011 15:30'!
before: target
	"Answer the receiver's element immediately before target. Raise an
	error if target is not an element of the receiver, or if there are no 
	elements before it (i.e. it is the first element)."

	^ self before: target ifAbsent: [self errorNotFound: target]! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/30/2011 15:30'!
before: target ifAbsent: exceptionBlock
	"Answer the receiver's element immediately before target. Answer
	the result of evaluating the exceptionBlock if target is not an element
	of the receiver, or if there are no elements before it."

	| index |
	index := self indexOf: target.
	^ (index = 0 or: [index = 1])
		ifTrue: [exceptionBlock value]
		ifFalse: [self at: index - 1]! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/30/2011 20:29'!
byteEncode:aStream
self flag: #aelToDo.			"needed??"
	aStream writeArray:self.
! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/30/2011 15:36'!
capacity
	"Answer the current capacity of the receiver."

	^ self size! !

!WArray methodsFor: 'enumerating' stamp: 'aw 9/9/2009 13:29'!
collect: aBlock

	^ self collect: aBlock from: 1 to: self size! !

!WArray methodsFor: 'enumerating' stamp: 'aw 9/9/2009 13:29'!
collect: aBlock from: start to: end

	| r |
	r := self class new: end - start + 1.
	start to: end do: [:i |
		r at: i - start + 1 put: (aBlock value: (self at: i))
	].
	^ r! !

!WArray methodsFor: 'enumerating' stamp: 'ael 9/30/2011 15:39'!
collect: collectBlock thenSelect: selectBlock
	"Utility method to improve readability."

	^ (self collect: collectBlock) select: selectBlock! !

!WArray methodsFor: 'filter streaming' stamp: 'ael 9/30/2011 15:40'!
contents
	^ self! !

!WArray methodsFor: 'copying' stamp: 'ael 9/30/2011 15:40'!
copyAfter: anElement
	"Answer a copy of the receiver from after the first occurence
	of anElement up to the end. If no such element exists, answer 
	an empty copy."

	^ self allButFirst: (self indexOf: anElement ifAbsent: [^ self copyEmpty])! !

!WArray methodsFor: 'copying' stamp: 'ael 9/30/2011 15:40'!
copyAfterLast: anElement
	"Answer a copy of the receiver from after the last occurence
	of anElement up to the end. If no such element exists, answer 
	an empty copy."

	^ self allButFirst: (self lastIndexOf: anElement ifAbsent: [^ self copyEmpty])! !

!WArray methodsFor: 'copying' stamp: 'ael 9/30/2011 15:41'!
copyEmpty
	^ self species new: 0! !

!WArray methodsFor: 'copying' stamp: 'ael 9/30/2011 15:41'!
copyFrom: start to: stop 
	"Answer a copy of a subset of the receiver, starting from element at 
	index start until element at index stop."

	| newSize |
	newSize := stop - start + 1.
	^(self species new: newSize)
		replaceFrom: 1
		to: newSize
		with: self
		startingAt: start! !

!WArray methodsFor: 'copying' stamp: 'bf 10/18/2010 13:52'!
copyReplaceFrom: start to: stop with: replacementCollection 
	"Answer a copy of the receiver satisfying the following conditions: If 
	stop is less than start, then this is an insertion; stop should be exactly 
	start-1, start = 1 means insert before the first character, start = size+1 
	means append after last character. Otherwise, this is a replacement; start 
	and stop have to be within the receiver's bounds."

	| newCollection newSize endReplacement |
	newSize _ self size - (stop - start + 1) + replacementCollection size.
	endReplacement _ start - 1 + replacementCollection size.
	newCollection _ self class new: newSize.
	start > 1 ifTrue:[
		newCollection
			replaceFrom: 1
			to: start - 1
			with: self
			startingAt: 1].
	start <= endReplacement ifTrue:[
		newCollection
			replaceFrom: start
			to: endReplacement
			with: replacementCollection
			startingAt: 1].
	endReplacement < newSize ifTrue:[
		newCollection
			replaceFrom: endReplacement + 1
			to: newSize
			with: self
			startingAt: stop + 1].
	^newCollection! !

!WArray methodsFor: 'copying' stamp: 'ael 9/30/2011 15:43'!
copyUpTo: anElement 
	"Answer all elements up to but not including anObject. If there
	is no such object, answer a copy of the receiver."

	^ self first: (self indexOf: anElement ifAbsent: [^ self copy]) - 1! !

!WArray methodsFor: 'copying' stamp: 'ael 9/30/2011 15:43'!
copyUpToLast: anElement
	"Answer a copy of the receiver from index 1 to the last occurrence of 
	anElement, not including anElement."

	^ self first: (self lastIndexOf: anElement ifAbsent: [^ self copy]) - 1! !

!WArray methodsFor: 'copying' stamp: 'ael 9/30/2011 15:43'!
copyWith: newElement 
	"Answer a copy of the receiver that is 1 bigger than the receiver and has 
	newElement at the last element."

	| newIC |
	newIC := self species new: self size + 1.
	newIC 
		replaceFrom: 1
		to: self size
		with: self
		startingAt: 1.
	newIC at: newIC size put: newElement.
	^newIC! !

!WArray methodsFor: 'copying' stamp: 'ael 9/30/2011 15:45'!
copyWithout: oldElement 
	"Answer a copy of the receiver that does not contain any
	elements equal to oldElement."

	^ self reject: [:each | each = oldElement]

"Examples:
	'fred the bear' copyWithout: $e
	#(2 3 4 5 5 6) copyWithout: 5
"! !

!WArray methodsFor: 'copying' stamp: 'ael 9/30/2011 15:45'!
copyWithoutAll: aCollection
	"Answer a copy of the receiver that does not contain any elements 
	equal to those in aCollection."

	^ self reject: [:each | aCollection includes: each]! !

!WArray methodsFor: 'enumerating' stamp: 'ael 9/30/2011 15:45'!
count: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument.  
	Answer the number of elements that answered true."

	| sum |
	sum := 0.
	self do: [:each | (aBlock value: each) ifTrue: [sum := sum + 1]].
	^ sum! !

!WArray methodsFor: 'enumerating' stamp: 'ael 9/30/2011 15:46'!
detect: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument. 
	Answer the first element for which aBlock evaluates to true."

	^ self detect: aBlock ifNone: [self errorNotFound: aBlock]! !

!WArray methodsFor: 'enumerating' stamp: 'ael 9/30/2011 15:46'!
detect: aBlock ifFound: foundBlock ifNone: exceptionBlock 
	"foundBlock takes one argument, the found object."
	self 
		do: [ :element | (aBlock value: element) ifTrue: [ ^foundBlock value: element ] ].
	^exceptionBlock value
! !

!WArray methodsFor: 'enumerating' stamp: 'ael 9/30/2011 15:46'!
detect: aBlock ifNone: exceptionBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument.  
	Answer the first element for which aBlock evaluates to true. If none  
	evaluate to true, then evaluate the argument, exceptionBlock."

	self do: [:each | (aBlock value: each) ifTrue: [^ each]].
	^ exceptionBlock value! !

!WArray methodsFor: 'enumerating' stamp: 'aw 8/24/2009 14:01'!
do: aBlock

	1 to: self size do: [:ii | aBlock value: (self at: ii)]! !

!WArray methodsFor: 'enumerating' stamp: 'ael 9/30/2011 15:47'!
do: elementBlock separatedBy: separatorBlock
	"Evaluate the elementBlock for all elements in the receiver,
	and evaluate the separatorBlock between."

	1 to: self size do:
		[:index |
		index = 1 ifFalse: [separatorBlock value].
		elementBlock value: (self at: index)]! !

!WArray methodsFor: 'enumerating' stamp: 'ael 9/30/2011 15:47'!
do: aBlock without: anItem
	"Enumerate all elements in the receiver.
	Execute aBlock for those elements that are not equal to the given item"
	"Refer to the comment in Collection|do:."
	1 to: self size do:
		[:index | anItem = (self at: index) ifFalse:[aBlock value: (self at: index)]]! !

!WArray methodsFor: 'lobjects' stamp: 'bf 10/18/2010 15:19'!
doArgs: aBlock

	1 to: self size do: [:ii | aBlock valueWithPossibleArgs: (self at: ii)]! !

!WArray methodsFor: 'enumerating' stamp: 'ael 9/30/2011 15:47'!
doWithIndex: elementAndIndexBlock
	"Use the new version with consistent naming"
	^ self withIndexDo: elementAndIndexBlock! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/30/2011 15:48'!
eighth
	"Answer the eighth element of the receiver.
	Raise an error if there are not enough elements."

	^ self at: 8! !

!WArray methodsFor: 'private' stamp: 'ael 9/30/2011 15:48'!
emptyCheck

	self isEmpty ifTrue: [self errorEmptyCollection]! !

!WArray methodsFor: 'private' stamp: 'ael 9/30/2011 15:49'!
errorEmptyCollection

	self error: 'this collection is empty'! !

!WArray methodsFor: 'private' stamp: 'ael 9/30/2011 15:49'!
errorNoMatch

	self error: 'collection sizes do not match'! !

!WArray methodsFor: 'private' stamp: 'ael 9/30/2011 15:49'!
errorNotFound: anObject
	"Actually, this should raise a special Exception not just an error."

	self error: 'Object is not in the collection.'! !

!WArray methodsFor: 'private' stamp: 'ael 9/30/2011 15:50'!
errorOutOfBounds

	self error: 'indices are out of bounds'! !

!WArray methodsFor: 'Morphic-Explorer' stamp: 'ael 9/30/2011 15:51'!
explorerContents

	^Array new: self size streamContents: [ :stream |
		1 to: self size do: [ :index |
			stream nextPut: (
				ObjectExplorerWrapper
					with: (self at: index)
					name: index printString
					model: self) ] ]! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/30/2011 15:52'!
fifth
	"Answer the fifth element of the receiver.
	Raise an error if there are not enough elements."

	^ self at: 5! !

!WArray methodsFor: 'private' stamp: 'ael 9/30/2011 15:52'!
fillFrom: aCollection with: aBlock
	"Evaluate aBlock with each of aCollections's elements as the argument.  
	Collect the resulting values into self. Answer self."

	| index |
	index := 0.
	aCollection do: [ :each |
		self at: (index := index + 1) put: (aBlock value: each) ]! !

!WArray methodsFor: 'enumerating' stamp: 'ael 9/30/2011 15:53'!
findFirst: aBlock
	"Return the index of my first element for which aBlock evaluates as true."

	| index |
	index := 0.
	[(index := index + 1) <= self size] whileTrue:
		[(aBlock value: (self at: index)) ifTrue: [^index]].
	^ 0! !

!WArray methodsFor: 'enumerating' stamp: 'ael 9/30/2011 15:54'!
findLast: aBlock
	"Return the index of my last element for which aBlock evaluates as true."

	| index |
	index := self size + 1.
	[(index := index - 1) >= 1] whileTrue:
		[(aBlock value: (self at: index)) ifTrue: [^index]].
	^ 0! !

!WArray methodsFor: 'accessing' stamp: 'yo 11/4/2009 06:20'!
first

	^ self at: 1.
! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/30/2011 16:12'!
first: n
	"Answer the first n elements of the receiver.
	Raise an error if there are not enough elements."

	^ self copyFrom: 1 to: n! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/30/2011 16:15'!
fourth
	"Answer the fourth element of the receiver.
	Raise an error if there are not enough elements."

	^ self at: 4! !

!WArray methodsFor: 'enumerating' stamp: 'ael 9/30/2011 16:15'!
from: start to: stop do: aBlock
	"Evaluate aBlock for all elements between start and stop (inclusive)."

	start to: stop do: [:index | aBlock value: (self at: index)]! !

!WArray methodsFor: 'accessing' stamp: 'yo 9/17/2010 11:17'!
from: startIndex to: endIndex put: anObject
	"Put anObject in all indexes between startIndex 
	and endIndex. Very fast. Faster than to:do: for
	more than 26 positions. Answer anObject"

	| written toWrite thisWrite |

	startIndex > endIndex ifTrue: [^self].
	self at: startIndex put: anObject.
	written _ 1.
	toWrite _ endIndex - startIndex + 1.
	[written < toWrite] whileTrue:
		[
			thisWrite _ written min: toWrite - written.
			self 
				replaceFrom: startIndex + written
				to: startIndex + written + thisWrite - 1
				with: self startingAt: startIndex.
			written _ written + thisWrite
		].
	^anObject! !

!WArray methodsFor: 'comparing' stamp: 'ael 9/30/2011 19:01'!
hasEqualElements: otherCollection
	"Answer whether the receiver's size is the same as otherCollection's
	size, and each of the receiver's elements equal the corresponding 
	element of otherCollection.
	This should probably replace the current definition of #= ."

	| size |
	otherCollection isSequenceable ifFalse: [^ false].
	(size := self size) = otherCollection size ifFalse: [^ false].
	1 to: size do:
		[:index |
		(self at: index) = (otherCollection at: index) ifFalse: [^ false]].
	^ true! !

!WArray methodsFor: 'comparing' stamp: 'ael 9/30/2011 16:18'!
hash
	| hash |

	hash := self species hash.
	1 to: self size do: [:i | hash := (hash + (self at: i) hash) hashMultiply].
	^hash! !

!WArray methodsFor: 'world awareness' stamp: 'ael 9/20/2011 18:15'!
iSize
	"for world-aware operations where we need to know the number of values"
	^ self size! !

!WArray methodsFor: 'testing' stamp: 'aw 9/9/2009 12:02'!
ifEmpty: aBlock

	^ self isEmpty ifTrue: aBlock! !

!WArray methodsFor: 'testing' stamp: 'ael 9/30/2011 16:18'!
ifEmpty: emptyBlock ifNotEmpty: notEmptyBlock
	"Evaluate emptyBlock if I'm empty, notEmptyBlock otherwise"
	" If the notEmptyBlock has an argument, eval with the receiver as its argument"

	self isEmpty ifTrue: [ ^emptyBlock value ].
	^notEmptyBlock valueWithPossibleArgument: self! !

!WArray methodsFor: 'testing' stamp: 'aw 9/9/2009 12:02'!
ifNotEmpty: aBlock

	^ self isEmpty ifFalse: aBlock! !

!WArray methodsFor: 'testing' stamp: 'ael 9/30/2011 16:19'!
ifNotEmpty: notEmptyBlock ifEmpty: emptyBlock
	"Evaluate emptyBlock if I'm empty, notEmptyBlock otherwise
	 If the notEmptyBlock has an argument, eval with the receiver as its argument"

	self isEmpty ifFalse: [ ^notEmptyBlock valueWithPossibleArgument: self ].
	^emptyBlock value! !

!WArray methodsFor: 'testing' stamp: 'aw 9/9/2009 12:02'!
includes: x

	^ self anySatisfy: [:y | x = y]! !

!WArray methodsFor: 'testing' stamp: 'ael 9/30/2011 16:19'!
includesAllOf: aCollection 
	"Answer whether all the elements of aCollection are in the receiver."
	aCollection do: [:elem | (self includes: elem) ifFalse: [^ false]].
	^ true! !

!WArray methodsFor: 'testing' stamp: 'ael 9/30/2011 16:19'!
includesAnyOf: aCollection 
	"Answer whether any element of aCollection is one of the receiver's elements."
	aCollection do: [:elem | (self includes: elem) ifTrue: [^ true]].
	^ false! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/30/2011 16:20'!
indexOf: anElement
	"Answer the index of the first occurence of anElement within the  
	receiver. If the receiver does not contain anElement, answer 0."

	^ self indexOf: anElement ifAbsent: [0]! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/30/2011 16:20'!
indexOf: anElement ifAbsent: exceptionBlock
	"Answer the index of the first occurence of anElement within the  
	receiver. If the receiver does not contain anElement, answer the 
	result of evaluating the argument, exceptionBlock."

	^ self indexOf: anElement startingAt: 1 ifAbsent: exceptionBlock! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/30/2011 16:20'!
indexOf: anElement startingAt: start
	"Answer the index of the first occurence of anElement after start
	within the receiver. If the receiver does not contain anElement, 
	answer 0."

	^self indexOf: anElement startingAt: start ifAbsent: 0! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/30/2011 16:20'!
indexOf: anElement startingAt: start ifAbsent: exceptionBlock
	"Answer the index of the first occurence of anElement after start
	within the receiver. If the receiver does not contain anElement, 
	answer the 	result of evaluating the argument, exceptionBlock."

	start to: self size do:
		[:index |
		(self at: index) = anElement ifTrue: [^ index]].
	^ exceptionBlock value! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/30/2011 16:20'!
indexOfAnyOf: aCollection
	"Answer the index of the first occurence of any element included in aCollection within the receiver.
	If the receiver does not contain anElement, answer zero, which is an invalid index."

	^self indexOfAnyOf: aCollection startingAt: 1 ifAbsent: [0]! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/30/2011 16:20'!
indexOfAnyOf: aCollection ifAbsent: exceptionBlock
	"Answer the index of the first occurence of any element included in aCollection within the receiver.
	If the receiver does not contain anElement, answer the result of evaluating the argument, exceptionBlock."

	^self indexOfAnyOf: aCollection startingAt: 1 ifAbsent: exceptionBlock! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/30/2011 16:20'!
indexOfAnyOf: aCollection startingAt: start
	"Answer the index of the first occurence of any element included in aCollection after start within the receiver.
	If the receiver does not contain anElement, answer zero, which is an invalid index."

	^self indexOfAnyOf: aCollection startingAt: start ifAbsent: [0]! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/30/2011 16:20'!
indexOfAnyOf: aCollection startingAt: start ifAbsent: exceptionBlock
	"Answer the index of the first occurence of any element included in aCollection after start within the receiver.
	If the receiver does not contain anElement, answer the result of evaluating the argument, exceptionBlock.
	Note: it is user responsibility to provide aCollection that behaves relatevily fast when asked for includes: (like a Set)"

	start to: self size do:
		[:index |
		(aCollection includes: (self at: index)) ifTrue: [^ index]].
	^ exceptionBlock value! !

!WArray methodsFor: 'private' stamp: 'ael 9/20/2011 18:11'!
initialize: size
	contents := Array new: size! !

!WArray methodsFor: 'copying' stamp: 'ael 11/12/2011 11:33'!
initializedClone
	"return an instance that's ready to record values read from the original"
	^self class new: self size
! !

!WArray methodsFor: 'enumerating' stamp: 'ael 9/30/2011 20:51'!
inject: anObject into: aBlock
	| result |
	result := anObject.
	1 to: self size do: [:i | result := aBlock value: result value: (self at: i)].
	^result! !

!WArray methodsFor: 'world awareness' stamp: 'ael 9/30/2011 21:58'!
instVarAt: index
	"to support debugging, inspecting etc"
	"WArray doesn't redirect any of its IVs"
	^self realWorldInstVarAt: index
! !

!WArray methodsFor: 'world awareness' stamp: 'ael 9/30/2011 21:39'!
instVarAt: index put: value
	"WArray doesn't redirect any of its IVs"
	^self realWorldInstVarAt: index put: value
! !

!WArray methodsFor: 'testing' stamp: 'sw 7/20/2010 17:35'!
isArray
	"Confirm that the receiver is an array."

	^ true! !

!WArray methodsFor: 'testing' stamp: 'ael 9/30/2011 16:22'!
isCollection
	"Return true if the receiver is some sort of Collection and responds to basic collection messages such as #size and #do:"
	^true! !

!WArray methodsFor: 'testing' stamp: 'aw 8/24/2009 14:00'!
isEmpty

	^ self size == 0! !

!WArray methodsFor: 'testing' stamp: 'tk 4/28/2010 05:12'!
isEmptyOrNil

	^ self isEmpty! !

!WArray methodsFor: 'testing' stamp: 'ael 9/30/2011 16:23'!
isSequenceable
	^ true! !

!WArray methodsFor: 'sorting' stamp: 'ael 9/30/2011 16:23'!
isSorted
	"Return true if the receiver is sorted by the given criterion.
	Optimization for isSortedBy: [:a :b | a <= b]."

	| lastElm elm |
	self isEmpty ifTrue: [^ true].
	lastElm := self first.
	2 to: self size do: 
		[:index | 
		elm := self at: index.
		lastElm <= elm ifFalse: [^ false].
		lastElm := elm].
	^ true! !

!WArray methodsFor: 'sorting' stamp: 'ael 9/30/2011 16:23'!
isSortedBy: aBlock
	"Return true if the receiver is sorted by the given criterion."

	| lastElm elm |
	self isEmpty ifTrue: [^ true].
	lastElm := self first.
	2 to: self size do: 
		[:index | 
		elm := self at: index.
		(aBlock value: lastElm value: elm) ifFalse: [^ false].
		lastElm := elm].
	^ true! !

!WArray methodsFor: 'testing' stamp: 'ael 9/30/2011 16:24'!
isZero
	"Answer whether the receiver is zero"
	^ false! !

!WArray methodsFor: 'enumerating' stamp: 'ael 9/30/2011 16:25'!
keysAndValuesDo: aBlock 
	"Enumerate the receiver with all the keys (aka indices) and values."

	1 to: self size do: [:index | aBlock value: index value: (self at: index)]! !

!WArray methodsFor: 'kscript' stamp: 'yo 7/17/2012 10:25'!
ksAt: key
	^ self at: key + 1
! !

!WArray methodsFor: 'kscript' stamp: 'yo 7/6/2012 10:49'!
ksIsValidIndex: key
	^ key isInteger and: [key >= 0 and: [key < self size]]! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/30/2011 16:26'!
last
	"Answer the last element of the receiver"

	^ self at: self size! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/30/2011 16:26'!
last: n
	"Answer the last n elements of the receiver.  
	Raise an error if there are not enough elements."

	| size |
	size := self size.
	^ self copyFrom: size - n + 1 to: size! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/30/2011 16:26'!
lastIndexOf: anElement
	"Answer the index of the last occurence of anElement within the 
	receiver. If the receiver does not contain anElement, answer 0."

	^ self lastIndexOf: anElement startingAt: self size ifAbsent: [0]! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/30/2011 16:26'!
lastIndexOf: anElement ifAbsent: exceptionBlock
	"Answer the index of the last occurence of anElement within the  
	receiver. If the receiver does not contain anElement, answer the
	result of evaluating the argument, exceptionBlock."
	^self lastIndexOf: anElement startingAt: self size ifAbsent: exceptionBlock! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/30/2011 16:26'!
lastIndexOf: anElement startingAt: lastIndex ifAbsent: exceptionBlock
	"Answer the index of the last occurence of anElement within the  
	receiver. If the receiver does not contain anElement, answer the
	result of evaluating the argument, exceptionBlock."

	lastIndex to: 1 by: -1 do:
		[:index |
		(self at: index) = anElement ifTrue: [^ index]].
	^ exceptionBlock value! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/30/2011 16:26'!
lastIndexOfAnyOf: aCollection startingAt: lastIndex ifAbsent: exceptionBlock
	"Answer the index of the last occurence of anElement within the  
	receiver. If the receiver does not contain anElement, answer the
	result of evaluating the argument, exceptionBlock."

	lastIndex to: 1 by: -1 do:
		[:index |
		(aCollection includes: (self at: index)) ifTrue: [^ index]].
	^ exceptionBlock value! !

!WArray methodsFor: 'math functions' stamp: 'ael 9/30/2011 16:27'!
max
	"Answer the maximum value in the collection.  The collection must be non-empty and contain 'compatible' Magnitudes (eg: don't try this with a collection containing both Dates and Characters)."
	^ self inject: self anyOne into: [:max :each | max max: each]! !

!WArray methodsFor: 'sorting' stamp: 'ael 9/30/2011 20:55'!
mergeFirst: first middle: middle last: last into: dst by: aBlock
	"Private. Merge the sorted ranges [first..middle] and [middle+1..last] 
	of the receiver into the range [first..last] of dst."

	| i1 i2 val1 val2 out |
	i1 := first.
	i2 := middle + 1.
	val1 := self at: i1.
	val2 := self at: i2.
	out := first - 1.  "will be pre-incremented"

	"select 'lower' half of the elements based on comparator"
	[ (i1 <= middle) and: [ i2 <= last ] ] whileTrue: 	[
		(aBlock 
			ifNil: [ val1 <= val2 ]
			ifNotNil: [ aBlock value: val1 value: val2 ])
				ifTrue: [
					dst at: (out := out + 1) put: val1.
					val1 := self at: (i1 := i1 + 1)]
				ifFalse: [
					dst at: (out := out + 1) put: val2.
					(i2 := i2 + 1) <= last ifTrue: [
						val2 := self at: i2 ] ] ].

	"copy the remaining elements"
	i1 <= middle
		ifTrue: [dst replaceFrom: out + 1 to: last with: self startingAt: i1]
		ifFalse: [dst replaceFrom: out + 1 to: last with: self startingAt: i2]! !

!WArray methodsFor: 'sorting' stamp: 'ael 9/30/2011 21:51'!
mergeSortFrom: startIndex to: stopIndex by: aBlock
	"Sort the given range of indices using the mergesort algorithm.
	Mergesort is a worst-case O(N log N) sorting algorithm that usually
	does only half as many comparisons as heapsort or quicksort."

	"Details: recursively split the range to be sorted into two halves,
	mergesort each half, then merge the two halves together. An extra 
	copy of the data is used as temporary storage and successive merge 
	phases copy data back and forth between the receiver and this copy.
	The recursion is set up so that the final merge is performed into the
	receiver, resulting in the receiver being completely sorted."

	self size <= 1 ifTrue: [^ self].  "nothing to do"
	startIndex = stopIndex ifTrue: [^ self].
	self assert: [startIndex >= 1 and: [startIndex < stopIndex]]. "bad start index"
	self assert: [stopIndex <= self size]. "bad stop index"
	self
		mergeSortFrom: startIndex
		to: stopIndex 
		src: self copy		"Array uses self clone, but that wouldn't be enough here" 
		dst: self 
		by: aBlock! !

!WArray methodsFor: 'sorting' stamp: 'ael 9/30/2011 18:50'!
mergeSortFrom: first to: last src: src dst: dst by: aBlock
	"Private. Split the range to be sorted in half, sort each half, and 
	merge the two half-ranges into dst."

	| middle |
	first = last ifTrue: [^ self].
	middle := (first + last) // 2.
	self mergeSortFrom: first to: middle src: dst dst: src by: aBlock.
	self mergeSortFrom: middle + 1 to: last src: dst dst: src by: aBlock.
	src mergeFirst: first middle: middle last: last into: dst by: aBlock! !

!WArray methodsFor: 'math functions' stamp: 'ael 9/30/2011 16:27'!
min
	"Answer the minimum value in the collection.  The collection must be non-empty and contain 'compatible' Magnitudes (eg: don't try this with a collection containing both Dates and Characters)."
	^ self inject: self anyOne into: [:min :each | min min: each]! !

!WArray methodsFor: 'math functions' stamp: 'ael 9/30/2011 16:27'!
minMax
	"Scans for minimum and maximum in one pass returning the results as a two-element array"
	| min max |
	min := max := self anyOne.
	self do: [ :each |
		min := min min: each.
		max := max max: each ].
	^ Array with: min with: max! !

!WArray methodsFor: 'math functions' stamp: 'ael 9/30/2011 16:28'!
negated
	"Negated value of all elements in the collection"
	^ self collect: [:a | a negated]! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/30/2011 16:31'!
ninth
	"Answer the ninth element of the receiver.
	Raise an error if there are not enough elements."

	^ self at: 9! !

!WArray methodsFor: 'enumerating' stamp: 'ael 9/30/2011 16:31'!
noneSatisfy: aBlock
	"Evaluate aBlock with the elements of the receiver.
	If aBlock returns false for all elements return true.
	Otherwise return false"

	self do: [:item | (aBlock value: item) ifTrue: [^ false]].
	^ true! !

!WArray methodsFor: 'testing' stamp: 'ael 8/10/2011 05:43'!
notEmpty
	"Answer whether the receiver contains any elements."

	^ self isEmpty not! !

!WArray methodsFor: 'testing' stamp: 'ael 9/30/2011 16:31'!
occurrencesOf: anObject 
	"Answer how many of the receiver's elements are equal to anObject."

	| tally |
	tally := 0.
	self do: [:each | anObject = each ifTrue: [tally := tally + 1]].
	^tally! !

!WArray methodsFor: 'enumerating' stamp: 'bf 9/29/2010 17:38'!
pairsDo: aBlock 
	1 to: self size // 2 do:
		[:index | aBlock value: (self at: 2 * index - 1) value: (self at: 2 * index)]! !

!WArray methodsFor: 'copying' stamp: 'ael 9/30/2011 20:17'!
postCopy
	super postCopy.
	contents := contents copy.
! !

!WArray methodsFor: 'copying' stamp: 'ael 11/12/2011 11:35'!
postFlattenedCopy
	"suppress normal postCopy"
! !

!WArray methodsFor: 'printing' stamp: 'aw 8/24/2009 13:58'!
printElementsOn: aStream

	aStream nextPut: $(.
	self do: [:ee | aStream print: ee; space].
	self isEmpty ifFalse: [aStream skip: -1].
	aStream nextPut: $)! !

!WArray methodsFor: 'printing' stamp: 'aw 8/24/2009 13:59'!
printOn: aStream

	aStream nextPutAll: self class name.
	self printElementsOn: aStream! !

!WArray methodsFor: 'filter streaming' stamp: 'ael 9/30/2011 16:33'!
putOn: aStream

	self do: [ :each | each putOn: aStream ]! !

!WArray methodsFor: 'converting' stamp: 'ael 9/30/2011 16:33'!
readStream
	^ ReadStream on: self! !

!WArray methodsFor: 'enumerating' stamp: 'ael 9/30/2011 16:34'!
reduce: binaryBlock
	"Apply the argument, binaryBlock cumulatively to the elements of the receiver.
	For sequenceable collections the elements will be used in order, for unordered
	collections the order is unspecified."

	| first nextValue |
	first := true.
	self do: [ :each |
		first
			ifTrue: [ nextValue := each. first := false ]
			ifFalse: [ nextValue := binaryBlock value: nextValue value: each ] ].
	first ifTrue: [ self errorEmptyCollection ].
	^nextValue! !

!WArray methodsFor: 'enumerating' stamp: 'ael 9/30/2011 18:39'!
reject: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument. 
	Collect into a new collection like the receiver only those elements for 
	which aBlock evaluates to false. Answer the new collection."

	^self select: [:element | (aBlock value: element) == false]! !

!WArray methodsFor: 'removing' stamp: 'ael 9/30/2011 18:41'!
remove: oldObject 
	self shouldNotImplement! !

!WArray methodsFor: 'removing' stamp: 'ael 9/30/2011 18:41'!
remove: oldObject ifAbsent: anExceptionBlock 
	"SequencableCollections cannot implement removing."

	self shouldNotImplement! !

!WArray methodsFor: 'removing' stamp: 'ael 9/30/2011 18:41'!
removeAll

	self shouldNotImplement! !

!WArray methodsFor: 'removing' stamp: 'ael 9/30/2011 18:41'!
removeAll: aCollection

	self shouldNotImplement! !

!WArray methodsFor: 'removing' stamp: 'ael 9/30/2011 18:41'!
removeAllFoundIn: aCollection

	self shouldNotImplement! !

!WArray methodsFor: 'removing' stamp: 'ael 9/30/2011 18:42'!
removeAllSuchThat: aBlock 
	self shouldNotImplement! !

!WArray methodsFor: 'enumerating' stamp: 'ael 9/30/2011 18:42'!
replace: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument.  
	Collect the resulting values into self."

	1 to: self size do: [ :index |
		self at: index put: (aBlock value: (self at: index)) ]! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/30/2011 18:42'!
replaceAll: oldObject with: newObject 
	"Replace all occurences of oldObject with newObject"
	| index |
	index := self
				indexOf: oldObject
				startingAt: 1
				ifAbsent: [0].
	[index = 0]
		whileFalse: 
			[self at: index put: newObject.
			index := self
						indexOf: oldObject
						startingAt: index + 1
						ifAbsent: [0]]! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/30/2011 18:42'!
replaceFrom: start to: stop with: replacement 
	"This destructively replaces elements from start to stop in the receiver. 
	Answer the receiver itself. Use copyReplaceFrom:to:with: for 
	insertion/deletion which may alter the size of the result."

	replacement size = (stop - start + 1)
		ifFalse: [self error: 'Size of replacement doesnt match'].
	^self replaceFrom: start to: stop with: replacement startingAt: 1! !

!WArray methodsFor: 'private' stamp: 'aw 9/4/2009 12:24'!
replaceFrom: start to: stop with: replacement startingAt: repStart

	| srcIdx |
	srcIdx := repStart.
	start to: stop do: [:dstIdx |
		self at: dstIdx put: (replacement at: srcIdx).
		srcIdx := srcIdx + 1
	]! !

!WArray methodsFor: 'exceptions' stamp: 'ael 12/6/2011 15:00'!
reportSerializationErrorAt: index oldValue: readValue newValue: currentValue
	^(WorldValueChanged new indexPrefix: 'index' index: index readValue: readValue currentValue: currentValue) signal! !

!WArray methodsFor: 'converting' stamp: 'ael 9/30/2011 18:43'!
reverse
	"The ANSI standard (5.7.8.26) requires this method to return a copy of the receiver. If you want to reverse the collection in place, then use #reverseInPlace."
	
	^self reversed! !

!WArray methodsFor: 'enumerating' stamp: 'ael 9/20/2011 18:16'!
reverseDo: aBlock
	| size |
	size := self size.
	1 to: size do: [:i |
		aBlock value: (self at: size - i + 1)
	]! !

!WArray methodsFor: 'converting' stamp: 'ael 9/30/2011 18:44'!
reverseInPlace
	"Reverse this collection in place."
	
	| start end |
	start := 1.
	end := self size.
	[ start < end ] whileTrue: [
		| temp |
		temp := self at: start.
		self
			at: start put: (self at: end);
			at: end put: temp.
		start := start + 1.
		end := end - 1 ]
! !

!WArray methodsFor: 'converting' stamp: 'ael 9/30/2011 18:44'!
reversed
	"Answer a copy of the receiver with element order reversed."
	"Example: 'frog' reversed"

	| n result src |
	n := self size.
	result := self species ofSize: n.
	src := n + 1.
	1 to: n do: [:i | result at: i put: (self at: (src := src - 1))].
	^ result
! !

!WArray methodsFor: 'lobjects' stamp: 'ael 9/30/2011 18:44'!
safeDo: aBlock

	^ self do: aBlock
! !

!WArray methodsFor: 'accessing' stamp: 'yo 11/4/2009 06:20'!
second

	^ self at: 2.
! !

!WArray methodsFor: 'enumerating' stamp: 'aw 9/9/2009 13:35'!
select: aBlock

	| r n |
	r := self class new: self size.
	n := 0.
	self do: [:x |
		(aBlock value: x) ifTrue: [
			n := n + 1.
			r at: n put: x
		]
	].
	^ r copyFrom: 1 to: n! !

!WArray methodsFor: 'enumerating' stamp: 'aw 9/9/2009 13:36'!
select: aBlock thenCollect: anotherBlock

	^ (self select: aBlock) collect: anotherBlock! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/30/2011 18:45'!
seventh
	"Answer the seventh element of the receiver.
	Raise an error if there are not enough elements."

	^ self at: 7! !

!WArray methodsFor: 'lobjects-siss-private' stamp: 'ael 9/30/2011 18:46'!
sissContentsInto: sexp context: dictionary 
	"It skips unnecessary tests (it is not needed to override like this actually).
	But I will revisit this later to see more straightforward way"
self flag: #aelToDo.
	^ self sissIndexFieldsWithBasicAtInto: sexp context: dictionary

! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/30/2011 18:47'!
sixth
	"Answer the sixth element of the receiver.
	Raise an error if there are not enough elements."

	^ self at: 6! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/20/2011 18:17'!
size
	^contents size		"NB: contents is not redirected"! !

!WArray methodsFor: 'sorting' stamp: 'ael 9/30/2011 18:48'!
sort
	"Sort this array into ascending order using the '<=' operator."

	self sort: nil! !

!WArray methodsFor: 'sorting' stamp: 'ael 9/30/2011 18:48'!
sort: aSortBlock 
	"Sort this array using aSortBlock. The block should take two arguments
	and return true if the first element should preceed the second one.
	If aSortBlock is nil then <= is used for comparison."

	self
		mergeSortFrom: 1
		to: self size
		by: aSortBlock! !

!WArray methodsFor: 'sorting' stamp: 'ael 9/30/2011 18:51'!
sorted
	"Return a new sequenceable collection which contains the same elements as self but its elements are sorted in ascending order using the #'<=' operator."

	^self sorted: nil! !

!WArray methodsFor: 'sorting' stamp: 'ael 9/30/2011 18:52'!
sorted: aSortBlockOrNil
	"Return a new sequenceable collection which contains the same elements as self but its elements are sorted by aSortBlockOrNil. The block should take two arguments and return true if the first element should preceed the second one. If aSortBlock is nil then <= is used for comparison."
	
	^self copy sort: aSortBlockOrNil! !

!WArray methodsFor: 'printing' stamp: 'ael 9/30/2011 18:54'!
storeOn: aStream

	aStream nextPutAll: '(('.
	aStream nextPutAll: self class name.
	aStream nextPutAll: ' new: '.
	aStream store: self size.
	aStream nextPut: $).
	(self storeElementsFrom: 1 to: self size on: aStream)
		ifFalse: [aStream nextPutAll: '; yourself'].
	aStream nextPut: $)! !

!WArray methodsFor: 'filter streaming' stamp: 'ael 9/30/2011 18:55'!
storeOnStream:aStream
	 aStream writeCollection: self ! !

!WArray methodsFor: 'math functions' stamp: 'ael 9/30/2011 20:24'!
sum
	"Compute the sum of all the elements in the receiver"

	^self reduce:[:a :b| a + b]! !

!WArray methodsFor: 'accessing' stamp: 'ael 9/30/2011 18:56'!
swap: oneIndex with: anotherIndex 
	"Move the element at oneIndex to anotherIndex, and vice-versa."

	| element |
	element := self at: oneIndex.
	self at: oneIndex put: (self at: anotherIndex).
	self at: anotherIndex put: element! !

!WArray methodsFor: 'accessing' stamp: 'sw 6/3/2010 07:58'!
third
	"Answer the receiver's third element."

	^ self at: 3
! !

!WArray methodsFor: 'accessing' stamp: 'ael 11/12/2011 14:44'!
uncheckedAt: index put: value
	"for testing"
	| thisWorld trackers |
	(thisWorld := self thisWorld) isTopLevel ifTrue: [ ^contents at: index put: value ].

	trackers := self updateTrackersForWorld: thisWorld createIfAbsent: true caching: true.
	^thisWorld valueAt: index put: value for: self using: trackers.
! !

!WArray methodsFor: 'enumerating' stamp: 'ael 9/30/2011 18:57'!
with: otherCollection collect: twoArgBlock 
	"Collect and return the result of evaluating twoArgBlock with corresponding elements from this collection and otherCollection."
	| result |
	otherCollection size = self size ifFalse: [self error: 'otherCollection must be the same size'].
	result := self species new: self size.
	1 to: self size do:
		[:index | result at: index put:
		(twoArgBlock
			value: (self at: index)
			value: (otherCollection at: index))].
	^ result! !

!WArray methodsFor: 'enumerating' stamp: 'aw 9/9/2009 11:54'!
with: aCollection do: aBlock

	self withIndexDo: [:x :i |
		aBlock value: x value: (aCollection at: i)
	]! !

!WArray methodsFor: 'enumerating' stamp: 'aw 9/9/2009 11:52'!
withIndexCollect: aBlock

	| r |
	r := self class new: self size.
	self withIndexDo: [:x :i |
		r at: i put: (aBlock value: x value: i)
	].
	^ r! !

!WArray methodsFor: 'enumerating' stamp: 'aw 9/4/2009 11:59'!
withIndexDo: aBlock

	1 to: self size do: [:i |
		aBlock value: (self at: i) value: i
	]! !

!WArray methodsFor: 'world awareness' stamp: 'ael 11/2/2011 16:46'!
worldFlattenIn: aWorld
	| dest flatMe |
	flatMe := self class new: self size.
	dest := flatMe instVarAt: 2.

	aWorld flatten: self into: dest.
	^flatMe
! !

!WArray methodsFor: 'filter streaming' stamp: 'ael 10/1/2011 12:35'!
write: anObject 
	self shouldNotImplement! !

!WArray methodsFor: 'converting' stamp: 'ael 9/30/2011 18:57'!
writeStream
	^ WriteStream on: self! !

WObjectWithCachedLookup subclass: #WCollection
	instanceVariableNames: 'contents'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Worlds'!

!WCollection commentStamp: 'ael 9/20/2011 23:52' prior: 0!
Because the relationship between indices and values in this kind of object is so fluid, WCollection has special parent-world lookup semantics:

i. When the collection is first accessed in a child world, a snapshot of its contents in the nearest ancestor world is stored in the child.

ii. Only when an element is actually read (from the snapshot) in a child world is a read record taken.

iii. Any operation that queries or changes the size of the collection (notably, adding or removing) records that it has taken note of the original size - i.e., the size of the snapshot.  Any attempt thereafter to commit changes into a world where the collection now has a different size will fail the serialisability check.

!

!WCollection class methodsFor: 'instance creation' stamp: 'ael 6/17/2011 15:13'!
new
	^ self new: 10! !

!WCollection class methodsFor: 'instance creation' stamp: 'ael 11/1/2011 20:50'!
new: anInteger
	^self new: anInteger andDo: [ :x | ]
! !

!WCollection class methodsFor: 'instance creation' stamp: 'ael 11/1/2011 20:34'!
new: anInteger andDo: aBlock
	| inst |
	WTopLevelWorld default eval: [
		inst := super basicNew initialize.
		inst setContentsSize: anInteger.
		aBlock value: inst.
		].

	^inst
! !

!WCollection class methodsFor: 'stream creation' stamp: 'sn 9/10/2013 01:49'!
new: newSize streamContents: blockWithArg

	| stream inst result |
	result := WTopLevelWorld default eval: [
		inst := super basicNew initialize.
		inst setContentsSize: newSize.

		stream := WriteStream on: inst.
		blockWithArg value: stream.
		stream position = newSize
			ifTrue: [ stream originalContents ]
			ifFalse: [ stream contents ]
		].
	^result
! !

!WCollection class methodsFor: 'instance creation' stamp: 'ael 11/1/2011 20:34'!
new: anInteger withAll: anObject
	^ self new: anInteger andDo: [ :inst | inst setContents: (Array new: anInteger withAll: anObject) ]! !

!WCollection class methodsFor: 'instance creation' stamp: 'ael 12/2/2011 14:28'!
newAndDo: aBlock
	^self new: 10 andDo: aBlock
! !

!WCollection class methodsFor: 'instance creation' stamp: 'ael 11/1/2011 20:33'!
newFrom: aCollection 
	"Answer an instance of me containing the same elements as aCollection."

	^ self new: aCollection size andDo: [ :inst | inst addAll: aCollection ]! !

!WCollection class methodsFor: 'instance creation' stamp: 'ael 9/21/2011 01:35'!
ofSize: n
	"Create a new collection of size n with nil as its elements.
	This method exists because OrderedCollection new: n creates an
	empty collection,  not one of size n."
	^ self new: n withAll: nil.
! !

!WCollection class methodsFor: 'as yet unclassified' stamp: 'ael 11/2/2011 08:51'!
simpleBenchmark
	"WCollection simpleBenchmark"
	"WWorld runWithoutPlugin: [ WCollection simpleBenchmark ]"
	| l |
	MessageTally spyOn: 
	[
	WWorld thisWorld sprout eval: [
	l := WCollection new.
	1 to: 100000 do: [:x |
		0 to: 9 do: [:i| l addFirst: i ].
		[ l isEmpty ] whileFalse: [ l removeFirst]
		].
	 ]
	]
! !

!WCollection class methodsFor: 'as yet unclassified' stamp: 'ael 5/14/2013 23:42'!
simpleBenchmark: n
	"WCollection simpleBenchmark: 100"
	| l runs |
	runs := 10000.
	^(Time millisecondsToRun: [
		WWorld thisWorld sprout eval: [
			runs timesRepeat: [
				l := self new..
				1 to: n do: [:i| l addFirst: i ].
				[ l isEmpty ] whileFalse: [ l removeLast ]
				].
			 ]
		]) asFloat / runs.
! !

!WCollection class methodsFor: 'instance creation' stamp: 'ael 6/17/2011 18:01'!
streamContents: blockWithArg
	
	^self new: 100 streamContents: blockWithArg! !

!WCollection class methodsFor: 'instance creation' stamp: 'ael 11/1/2011 20:29'!
with: anObject 
	"Answer an instance of me containing anObject."

	^ self newAndDo: [ :inst | inst add: anObject ].
! !

!WCollection class methodsFor: 'instance creation' stamp: 'ael 11/1/2011 20:30'!
with: firstObject with: secondObject 
	"Answer an instance of me containing the two arguments as elements."

	^ self newAndDo: [ :inst | inst add: firstObject; add: secondObject ].
	! !

!WCollection class methodsFor: 'instance creation' stamp: 'ael 11/1/2011 20:30'!
with: firstObject with: secondObject with: thirdObject 
	"Answer an instance of me containing the three arguments as elements."

	^ self newAndDo: [ :inst | inst add: firstObject; add: secondObject; add: thirdObject ]		! !

!WCollection class methodsFor: 'instance creation' stamp: 'ael 11/1/2011 20:31'!
with: firstObject with: secondObject with: thirdObject with: fourthObject 
	"Answer an instance of me, containing the four arguments as the elements."

	^ self newAndDo: [ :inst | inst add: firstObject; add: secondObject; add: thirdObject; add: fourthObject ]
! !

!WCollection class methodsFor: 'instance creation' stamp: 'ael 11/1/2011 20:31'!
with: firstObject with: secondObject with: thirdObject with: fourthObject with: fifthObject
	"Answer an instance of me, containing the five arguments as the elements."

	^ self newAndDo: [ :inst | inst add: firstObject; add: secondObject; add: thirdObject; add: fourthObject; add: fifthObject ]
! !

!WCollection class methodsFor: 'instance creation' stamp: 'ael 11/1/2011 20:32'!
with: firstObject with: secondObject with: thirdObject with: fourthObject with: fifthObject with: sixthObject
	"Answer an instance of me, containing the six arguments as the elements."

	^ self newAndDo: [ :inst | inst add: firstObject; add: secondObject; add: thirdObject; add: fourthObject; add: fifthObject; add: sixthObject ]
! !

!WCollection class methodsFor: 'instance creation' stamp: 'ael 11/1/2011 20:36'!
withAll: aCollection
	"Create a new collection containing all the elements from aCollection."

	^ self new: aCollection size andDo: [ :inst | inst addAll: aCollection ]
! !

!WCollection methodsFor: 'copying' stamp: 'ael 6/27/2011 13:38'!
, otherCollection 
	"Concatenate two Strings or Collections."
	
	| size |
	size := self size.		"log access to size, because this determines which elements are accessed"
	^ self copyReplaceFrom: size + 1
		  to: size
		  with: otherCollection
"
#(2 4 6 8) , #(who do we appreciate)
((2989 storeStringBase: 16) copyFrom: 4 to: 6) , ' boy!!'
"! !

!WCollection methodsFor: 'comparing' stamp: 'ael 6/17/2011 16:47'!
= otherCollection 
	"Answer true if the receiver is equivalent to the otherCollection.
	First test for identity, then rule out different species and sizes of
	collections. As a last resort, examine each element of the receiver
	and the otherCollection."

	self == otherCollection ifTrue: [^ true].
	self species == otherCollection species ifFalse: [^ false].
	^ self hasEqualElements: otherCollection! !

!WCollection methodsFor: 'adding' stamp: 'ael 6/17/2011 15:13'!
add: newObject

	^self addLast: newObject! !

!WCollection methodsFor: 'adding' stamp: 'ael 9/21/2011 01:57'!
add: newObject after: oldObject 
	"Add the argument, newObject, as an element of the receiver. Put it in 
	the sequence just succeeding oldObject. Answer newObject."
	
	| index |
	index := self indexOf: oldObject ifAbsent: [ ^self errorNotFound: oldObject ].
	self recordReadsFrom: 1 to: index.
	self add: newObject afterIndex: index.
	^newObject! !

!WCollection methodsFor: 'adding' stamp: 'ael 9/21/2011 01:56'!
add: newObject afterIndex: index
	"Add the argument, newObject, as an element of the receiver. Put it in 
	the sequence just after index. Answer newObject."
	(index between: 0 and: self unloggedSize) ifFalse:[^self errorSubscriptBounds: index].
	self recordReadOfSize.
	contents add: newObject afterIndex: index.
	self recordWritesFrom: index+1 to: self unloggedSize.
	^ newObject! !

!WCollection methodsFor: 'adding' stamp: 'ael 9/21/2011 01:57'!
add: newObject before: oldObject 
	"Add the argument, newObject, as an element of the receiver. Put it in 
	the sequence just preceding oldObject. Answer newObject."
	
	| index |
	index := self indexOf: oldObject ifAbsent: [ ^self errorNotFound: oldObject ].
	self recordReadsFrom: 1 to: index.
	self add: newObject afterIndex: index-1.
	^newObject
! !

!WCollection methodsFor: 'adding' stamp: 'ael 6/17/2011 15:13'!
add: newObject beforeIndex: index 
	"Add the argument, newObject, as an element of the receiver. Put it in 
	the sequence just before index. Answer newObject."

	self add: newObject afterIndex: index - 1.
	^ newObject! !

!WCollection methodsFor: 'adding' stamp: 'ael 6/17/2011 18:08'!
add: newObject withOccurrences: anInteger
	"Add newObject anInteger times to the receiver. Do nothing if anInteger is less than one. Answer newObject."

	anInteger timesRepeat: [self add: newObject].
	^ newObject! !

!WCollection methodsFor: 'adding' stamp: 'ael 9/20/2011 18:43'!
addAll: aCollection

	^self addAllLast: aCollection
! !

!WCollection methodsFor: 'adding' stamp: 'ael 9/20/2011 18:41'!
addAllFirst: anOrderedCollection 
	"Add each element of anOrderedCollection at the beginning of the 
	receiver. Answer anOrderedCollection."

	anOrderedCollection
	  ifNotEmpty: [
		self recordReadOfSize.		"ensure a read record for the collection size"

		"record a read of all items in their initial positions"
		self recordReadsFrom: 1 to: contents size.

		contents addAllFirst: anOrderedCollection.
	
		"record a write of all items in their new positions"
		self recordWritesFrom: 1 to: contents size.
		].

	^anOrderedCollection
! !

!WCollection methodsFor: 'adding' stamp: 'ael 6/17/2011 17:41'!
addAllFirstUnlessAlreadyPresent: anOrderedCollection 
	"Add each element of anOrderedCollection at the beginning of the receiver, preserving the order, but do not add any items that are already in the receiver.  Answer anOrderedCollection."

	anOrderedCollection reverseDo:
		[:each | (self includes: each) ifFalse: [self addFirst: each]].
	^ anOrderedCollection! !

!WCollection methodsFor: 'adding' stamp: 'ael 9/20/2011 18:46'!
addAllLast: col
	"Add a collection of objects to the end of the receiver. Answer the collection."
	"Records updates for worlds"
	"If the supplied collection is empty (so nothing is added), don't even record a read of size - on the basis that if the caller ran the add: loop it wouldn't even be called once."
	| oldSize |
	col ifNotEmpty: [
		oldSize := self size.	"record, because size determines where we'll be adding"

		contents addAllLast: col.

		self recordWritesFrom: oldSize+1 to: oldSize+col size.
		].

	^ col! !

!WCollection methodsFor: 'adding' stamp: 'ael 9/20/2011 18:33'!
addFirst: newObject 
	"Add newObject to the beginning of the receiver. Answer newObject."
	"Records updates for worlds"

	self recordReadOfSize.		"ensure a read record for the collection size"

	"record a read of all items in their current positions"
	self recordReadsFrom: 1 to: contents size.

	contents addFirst: newObject.
	
	"record a write of all items in their new positions"
	self recordWritesFrom: 1 to: contents size.

	^ newObject! !

!WCollection methodsFor: 'adding' stamp: 'ael 6/17/2011 18:09'!
addIfNotPresent: anObject
	"Include anObject as one of the receiver's elements, but only if there
	is no such element already. Anwser anObject."

	(self includes: anObject) ifFalse: [self add: anObject].
	^ anObject! !

!WCollection methodsFor: 'adding' stamp: 'ael 9/20/2011 18:46'!
addLast: newObject 
	"Add newObject to the end of the receiver. Answer newObject."
	"Records updates for worlds"

	self recordReadOfSize.		"because it's changing"

	contents addLast: newObject.
	
	self recordWriteAt: contents size.

	^ newObject
! !

!WCollection methodsFor: 'accessing' stamp: 'ael 6/17/2011 18:10'!
after: target
	"Answer the element after target.  Raise an error if target is not
	in the receiver, or if there are no elements after it."

	^ self after: target ifAbsent: [self errorNotFound: target]! !

!WCollection methodsFor: 'accessing' stamp: 'ael 6/25/2011 13:03'!
after: target ifAbsent: exceptionBlock
	"Answer the element after target.  Answer the result of evaluation
	the exceptionBlock if target is not in the receiver, or if there are 
	no elements after it."

	| index |
	index := self indexOf: target.
	^ (index = 0 or: [index = self unloggedSize])
		ifTrue: [exceptionBlock value]
		ifFalse: [self at: index + 1]! !

!WCollection methodsFor: 'accessing' stamp: 'ael 6/17/2011 18:11'!
allButFirst
	"Answer a copy of the receiver containing all but the first
	element. Raise an error if there are not enough elements."

	^ self allButFirst: 1! !

!WCollection methodsFor: 'accessing' stamp: 'ael 6/25/2011 03:25'!
allButFirst: n
	"Answer a copy of the receiver containing all but the first n
	elements. Raise an error if there are not enough elements."

	^ self copyFrom: n + 1 to: self size.		"result depends on size, so show that we've read it"! !

!WCollection methodsFor: 'enumerating' stamp: 'ael 9/20/2011 18:47'!
allButFirstDo: block
	"World-aware behaviour: only log access to size if we get to the end of the collection"

	2 to: self unloggedSize do:		"if the caller stops the loop early, size won't have any impact"
		[:index | block value: (self at: index)].
	
	self recordReadOfSize.			"now that we've reached the end"
! !

!WCollection methodsFor: 'accessing' stamp: 'ael 6/17/2011 18:13'!
allButLast
	"Answer a copy of the receiver containing all but the last
	element. Raise an error if there are not enough elements."

	^ self allButLast: 1! !

!WCollection methodsFor: 'accessing' stamp: 'ael 6/27/2011 13:22'!
allButLast: n
	"Answer a copy of the receiver containing all but the last n
	elements. Raise an error if there are not enough elements."

	^ self copyFrom: 1 to: self size - n.		"which indices are accessed depends on size, so show that we've read it"! !

!WCollection methodsFor: 'enumerating' stamp: 'ael 6/27/2011 13:50'!
allButLastDo: block

	1 to: self size - 1 do:				"log access to size, because it clearly influences what happens"
		[:index | block value: (self at: index)]! !

!WCollection methodsFor: 'enumerating' stamp: 'ael 6/17/2011 18:14'!
allSatisfy: aBlock
	"Evaluate aBlock with the elements of the receiver.
	If aBlock returns false for any element return false.
	Otherwise return true."

	self do: [:each | (aBlock value: each) ifFalse: [^ false]].
	^ true! !

!WCollection methodsFor: 'accessing' stamp: 'ael 6/17/2011 18:15'!
anyOne
	^ self first! !

!WCollection methodsFor: 'enumerating' stamp: 'ael 6/17/2011 18:15'!
anySatisfy: aBlock
	"Evaluate aBlock with the elements of the receiver.
	If aBlock returns true for any element return true.
	Otherwise return false."

	self do: [:each | (aBlock value: each) ifTrue: [^ true]].
	^ false! !

!WCollection methodsFor: 'data classes' stamp: 'ael 9/25/2011 21:23'!
arrayClass
	"the class to be used for arrays returned to the user"
	^WArray
! !

!WCollection methodsFor: 'converting' stamp: 'ael 9/20/2011 18:52'!
asArray
	"Records updates for worlds
		- by marking all values as having been read"
	self recordReadsFrom: 1 to: self size.		"size is also accessed, of course"

	^self arrayClass withAll: contents.
! !

!WCollection methodsFor: 'converting' stamp: 'ael 9/20/2011 18:53'!
asBag
	self recordReadOfSize.
	^contents asBag
! !

!WCollection methodsFor: 'converting' stamp: 'ael 9/20/2011 18:54'!
asIdentitySet
	^(IdentitySet new: self size) addAll: contents; yourself! !

!WCollection methodsFor: 'converting' stamp: 'ael 6/27/2011 11:20'!
asOrderedCollection
	"The whole point of WCollection2 is that it can be used interchangeably with OrderedCollections in environments where worlds are at work.  Therefore we answer self.  If a user really wants an OrderedCollection, send 'as: OrderedCollection'"
	^self
! !

!WCollection methodsFor: 'converting' stamp: 'ael 9/20/2011 18:55'!
asSet
	self recordReadOfSize.
	^contents asSet! !

!WCollection methodsFor: 'converting' stamp: 'ael 9/20/2011 18:56'!
asSortedCollection
	"Answer a SortedCollection whose elements are the elements of the 
	receiver. The sort order is the default less than or equal. 
	Use #sorted: if you don't really need a SortedCollection, but a sorted collection!!"
	self recordReadOfSize.
	^contents as: SortedCollection! !

!WCollection methodsFor: 'converting' stamp: 'ael 9/20/2011 18:57'!
asSortedCollection: aSortBlock 
	"Answer a SortedCollection whose elements are the elements of the 
	receiver. The sort order is defined by the argument, aSortBlock.
	Use #sorted: if you don't really need a SortedCollection, but a sorted collection!!"

	| aSortedCollection |
	aSortedCollection := SortedCollection new: self size.		"record read of size"
	aSortedCollection sortBlock: aSortBlock.
	aSortedCollection addAll: contents.
	^ aSortedCollection! !

!WCollection methodsFor: 'accessing' stamp: 'ael 9/20/2011 18:59'!
at: anInteger 
	"Answer my element at index anInteger. at: is used by a knowledgeable
	client to access an existing element"
	"World awareness: only record access to size if the index was not found"

	(anInteger < 1 or: [anInteger > self unloggedSize])
		ifTrue: [ self recordReadOfSize.   self errorNoSuchElement]
		ifFalse: [
			self recordReadAt: anInteger.
			^contents at: anInteger
			]! !

!WCollection methodsFor: 'accessing' stamp: 'ael 9/20/2011 19:00'!
at: index ifAbsent: exceptionBlock 
	"Answer the element at my position index. If I do not contain an element at index, answer the result of evaluating the argument, exceptionBlock."
	"World-aware behaviour:  If an item is found, the only recorded read is at the index where it was found.  But if you get to the end of the collection, record a read of size because that determines where the search stopped."
	(index between: 1 and: self unloggedSize) ifTrue: [^ self at: index].
	
	self recordReadOfSize.
	^ exceptionBlock value! !

!WCollection methodsFor: 'adding' stamp: 'ael 9/20/2011 18:38'!
at: index ifAbsentPut: block
	"Return value at index - but if value does not exist (nil or out of bounds) then add block's value at index (growing self if necessary)"

	| v |
	index <= self unloggedSize ifTrue: [
		^ (v := self at: index)
			ifNotNil: [v]
			ifNil: [self at: index put: block value]
	].

	self recordReadOfSize.			"changing, so record access to the original value"
	[self unloggedSize < index] whileTrue: [ self add: nil ].
	^ self at: index put: block value! !

!WCollection methodsFor: 'accessing' stamp: 'ael 9/20/2011 19:02'!
at: anInteger put: anObject 
	"Put anObject at element index anInteger. at:put: cannot be used to
	append, front or back, to an ordered collection; it is used by a
	knowledgeable client to replace an element."

	"World awareness: only record access to size if the index was not found"

	| index |
	index := anInteger asInteger.
	(index < 1 or: [index > contents size])
		ifTrue: [ self recordReadOfSize.   self errorNoSuchElement]
		ifFalse: [
			contents at: anInteger put: anObject.
			self recordWriteAt: index.
			^ anObject].
! !

!WCollection methodsFor: 'accessing' stamp: 'ael 6/17/2011 18:23'!
atAll: indexArray
	"Answer a new collection like the receiver which contains all elements
	of the receiver at the indices of indexArray."
	"#('one' 'two' 'three' 'four') atAll: #(3 2 4)"

	| newCollection |
	newCollection := self species ofSize: indexArray size.
	1 to: indexArray size do:
		[:index |
		newCollection at: index put: (self at: (indexArray at: index))].
	^ newCollection! !

!WCollection methodsFor: 'accessing' stamp: 'ael 6/17/2011 18:23'!
atAll: aCollection put: anObject 
	"Put anObject at every index specified by the elements of aCollection."

	aCollection do: [:index | self at: index put: anObject].
	^ anObject! !

!WCollection methodsFor: 'accessing' stamp: 'ael 6/17/2011 18:23'!
atAll: indexArray putAll: valueArray
	"Store the elements of valueArray into the slots
	of this collection selected by indexArray."

	indexArray with: valueArray do: [:index :value | self at: index put: value].
	^ valueArray! !

!WCollection methodsFor: 'accessing' stamp: 'ael 6/27/2011 13:27'!
atAllPut: anObject 
	"Put anObject at every one of the receiver's indices."

	self from: 1 to: self size put: anObject.		"log access to size, because that determines how many elements are changed"
! !

!WCollection methodsFor: 'accessing' stamp: 'ael 6/17/2011 18:24'!
atLast: indexFromEnd
	"Return element at indexFromEnd from the last position.
	 atLast: 1, returns the last element"

	^ self atLast: indexFromEnd ifAbsent: [self error: 'index out of range']! !

!WCollection methodsFor: 'accessing' stamp: 'ael 6/27/2011 13:28'!
atLast: indexFromEnd ifAbsent: block
	"Return element at indexFromEnd from the last position.
	 atLast: 1 ifAbsent: [] returns the last element"

	^ self at: self size + 1 - indexFromEnd ifAbsent: block.	"size determines which indices we get"! !

!WCollection methodsFor: 'accessing' stamp: 'ael 6/27/2011 13:28'!
atLast: indexFromEnd put: obj
	"Set the element at indexFromEnd from the last position.
	 atLast: 1 put: obj, sets the last element"

	^ self at: self size + 1 - indexFromEnd put: obj.		"size determines which index is overwritten"! !

!WCollection methodsFor: 'accessing' stamp: 'ael 6/27/2011 13:29'!
atPin: index 
	"Return the index'th element of me if possible.
	Return the first or last element if index is out of bounds."

	"World-aware behaviour: it could be argued that when the index isn't beyond the end of the collection, the size value itself is not being used.  But I prefer consistency with the natural way for the caller to do it, such as by using (index min: collection size)."

	index < 1 ifTrue: [^ self first].
	index > self size ifTrue: [^ self last].	"size was accessed in making the decision"
	^ self at: index! !

!WCollection methodsFor: 'accessing' stamp: 'ael 12/2/2011 15:17'!
atRandom: aGenerator
	"Answer a random element of the receiver.  Uses aGenerator which
	should be kept by the user in a variable and used every time. Use
	this instead of #atRandom for better uniformity of random numbers 
	because only you use the generator.  Causes an error if self has no 
	elements."

	^ self at: (aGenerator nextInt: self size)! !

!WCollection methodsFor: 'accessing' stamp: 'ael 8/23/2011 13:20'!
atWrap: index 
	"Answer the index'th element of the receiver.  If index is out of bounds,
	let it wrap around from the end to the beginning until it is in bounds."

	"World-aware behaviour:  Not obvious what to do about size.  Some options:

		a. size is always considered to have been accessed (the logic used for atPin:, reflecting how users would implement the equivalent behaviour for themselves)

		b. size is only considered accessed if the index actually wrapped

		c. size is considered accessed only if it could have made a difference to the answer - therefore only for indices from 2 upwards.

	This code adopts behaviour a."
	^ self at: index - 1 \\ self size + 1
! !

!WCollection methodsFor: 'accessing' stamp: 'ael 8/23/2011 13:13'!
atWrap: index put: value
	"Store value into the index'th element of the receiver.  If index is out
	of bounds, let it wrap around from the end to the beginning until it 
	is in bounds. Answer value."

	"World-aware behaviour: size is always considered to have been accessed.  See comment in atWrap:"
	^ self at: index  - 1 \\ self size + 1 put: value! !

!WCollection methodsFor: 'copying' stamp: 'ael 9/21/2011 01:07'!
basicContents
	"Answer a copy of the receiver's contents without leaving any trace in the world"
	^contents copy
! !

!WCollection methodsFor: 'accessing' stamp: 'ael 6/17/2011 18:25'!
before: target
	"Answer the receiver's element immediately before target. Raise an
	error if target is not an element of the receiver, or if there are no 
	elements before it (i.e. it is the first element)."

	^ self before: target ifAbsent: [self errorNotFound: target]! !

!WCollection methodsFor: 'accessing' stamp: 'ael 6/17/2011 15:13'!
before: target ifAbsent: exceptionBlock
	"Answer the receiver's element immediately before target. Answer
	the result of evaluating the exceptionBlock if target is not an element
	of the receiver, or if there are no elements before it."

	| index |
	index _ self indexOf: target.
	^ (index == 0 or: [index == 1])
		ifTrue: [exceptionBlock value]
		ifFalse: [self at: index - 1]! !

!WCollection methodsFor: 'accessing' stamp: 'ael 11/1/2011 21:32'!
capacity
	^WTopLevelWorld default eval: [ contents capacity ]
! !

!WCollection methodsFor: 'enumerating' stamp: 'ael 9/21/2011 01:12'!
collect: aBlock 
	"Evaluate aBlock with each of my elements as the argument. Collect the 
	resulting values into a collection that is like me. Answer the new 
	collection. Override superclass in order to use addLast:, not at:put:."
	"Records updates for worlds"
	| newCollection element size |
	size := self size.			"logged"
	newCollection := self species new: size.
	1 to: size do:
		[:index |
		element := self at: index.
		newCollection addLast: (aBlock value: element)].
	^ newCollection
! !

!WCollection methodsFor: 'enumerating' stamp: 'ael 9/21/2011 01:14'!
collect: aBlock from: fromIndex to: toIndex
	"Override superclass in order to use addLast:, not at:put:."
	"Records updates for worlds"
	| result element |
	(fromIndex < 1 or: [toIndex > self unloggedSize])
		ifTrue: [^self errorNoSuchElement].
	result := self species new: toIndex - fromIndex + 1.
	fromIndex to: toIndex do:
		[:index |
			element := self at: index.
			result addLast: (aBlock value: element)].
	^ result
! !

!WCollection methodsFor: 'enumerating' stamp: 'ael 6/17/2011 18:27'!
collect: collectBlock thenSelect: selectBlock
	"Utility method to improve readability."

	^ (self collect: collectBlock) select: selectBlock! !

!WCollection methodsFor: 'world awareness' stamp: 'ael 2/29/2012 00:45'!
commitInto: parentWorld using: updateTrackers
	"this is run in the child world.
	updateTrackers is a tuple { proxy. reads. writes. sizeRead. remainingReadSlots. }
		- if there are no Writes indices, we're done
		- for each index in Writes
			- if the index is beyond the end of our collection (representing a truncation):
				* truncate the parent-world collection to one below the index
				* we're done
			- copy the value from index in our collection to index in the parent-world collection"

	| index thisSize val writes writesHolder parentSize |
	writesHolder := updateTrackers at: 3.
	writesHolder ifNil: [ ^self ].		"nothing to do"

	writes := writesHolder at: 1.	"the bit-recorded indices"

	thisSize := self unloggedSize.		"size in the committing child world"
	parentSize := parentWorld eval: [ self unloggedSize ].	"will install IV values in parent for the first time if needed"

	index := 1.
	[ writes = 0 ]
	  whileFalse: [
 		(writes bitAnd: 1) = 1
		  ifTrue: [
			index > thisSize
			  ifTrue: [
				"perform any truncation necessary to leave the parent with size (index-1)"
				parentSize > (index - 1)
				  ifTrue: [ parentWorld eval: [ parentSize - index + 1 timesRepeat: [ self removeLast ] ] ].
				^self.			"any writes to greater indices are now meaningless" 
				].

			val := self at: index.		"no need to worry about this changing the Reads record"

			"NB: the serialisation check confirms that any writes beyond the end are contiguous (and can therefore be done using add:)"
			parentWorld eval: [
				index > parentSize
				  ifTrue: [ self add: val. parentSize := parentSize + 1 ]
				  ifFalse: [ self at: index put: val ].
				].
			].
		writes := writes >> 1.
		index := index + 1.
		].
! !

!WCollection methodsFor: 'copying' stamp: 'ael 6/17/2011 15:13'!
copyEmpty
	"Answer a copy of the receiver that contains no elements."

	^self species new! !

!WCollection methodsFor: 'copying' stamp: 'ael 6/17/2011 15:13'!
copyFrom: startIndex to: endIndex 
	"Answer a copy of the receiver that contains elements from position
	startIndex to endIndex."

	| targetCollection |
	endIndex < startIndex ifTrue: [^self species new: 0].
	targetCollection _ self species new: endIndex + 1 - startIndex.
	startIndex to: endIndex do: [:index | targetCollection addLast: (self at: index)].
	^ targetCollection! !

!WCollection methodsFor: 'world awareness' stamp: 'ael 12/5/2011 22:58'!
copyMissingOrigValuesInto: parentWorld using: updateTrackers
	"As part of commit, copy into the parent world's reads record any values that were read by the child-level object but came from an ancestor beyond the parent (i.e., the parent has recorded neither a write nor a read for this slot).

	This message is sent in the child world.
	
	For WCollection this isn't needed for the basic instance variables, because as soon as the collection as a whole is accessed in any way the IVs are all copied down from an ancestor.  However, the Reads record shows which of the collection's elements have actually been looked at, and this must not be updated in the parent until a child world commits.
	
	updateTrackers is a tuple { proxy. reads. writes. sizeRead. remainingReadSlots. }
	
	NB: there is no need to propagate the fact that we read the collection's size in this world (if indeed we did).  Our commits might have involved merely overwrites in the parent, in which case as far as it is concerned it has not yet needed to read the size from its ancestors."
	
	| overwritten reads dnk |
	parentWorld tracksReads ifFalse: [ ^self ].		"parent doesn't want to know"

	reads := updateTrackers at: 2.	
	reads ifNil: [ ^self ].								"nothing to tell"

	dnk := self dnkToken.
	overwritten := self overwriteToken.
	parentWorld eval: [
		reads withIndexDo: [ :entry :ii |
			(entry ~~ dnk and: [ entry ~~ overwritten ]) ifTrue: [ self recordReadAt: ii ].		"NB: won't record a read in the parent if there was already an 'overwritten' marker"
			].
		].
! !

!WCollection methodsFor: 'copying' stamp: 'ael 9/21/2011 01:38'!
copyReplaceFrom: start to: stop with: replacementCollection 
	"Answer a copy of the receiver with replacementCollection's elements in
	place of the receiver's start'th to stop'th elements. This does not expect
	a 1-1 map from replacementCollection to the start to stop elements, so it
	will do an insert or append."

	| newOrderedCollection delta startIndex stopIndex |
	"if start is less than 1, ignore stop and assume this is inserting at the front. 
	if start greater than self size, ignore stop and assume this is appending. 
	otherwise, it is replacing part of me and start and stop have to be within my 
	bounds. "
	delta := 0.
	startIndex := start.
	stopIndex := stop.
	start < 1
		ifTrue: [startIndex := stopIndex := 0]
		ifFalse: [startIndex > self size
				ifTrue: [startIndex := stopIndex := self size + 1]
				ifFalse: 
					[(stopIndex < (startIndex - 1) or: [stopIndex > self size])
						ifTrue: [self errorOutOfBounds].
					delta := stopIndex - startIndex + 1]].
	newOrderedCollection := 
		self species new: self size + replacementCollection size - delta.
	1 to: startIndex - 1 do: [:index | newOrderedCollection add: (self at: index)].
	1 to: replacementCollection size do: 
		[:index | newOrderedCollection add: (replacementCollection at: index)].
	stopIndex + 1 to: self size do: [:index | newOrderedCollection add: (self at: index)].
	^newOrderedCollection! !

!WCollection methodsFor: 'copying' stamp: 'ael 6/17/2011 18:30'!
copyUpTo: anElement 
	"Answer all elements up to but not including anObject. If there
	is no such object, answer a copy of the receiver."

	^ self first: (self indexOf: anElement ifAbsent: [^ self copy]) - 1! !

!WCollection methodsFor: 'copying' stamp: 'ael 6/17/2011 18:31'!
copyUpToLast: anElement
	"Answer a copy of the receiver from index 1 to the last occurrence of 
	anElement, not including anElement."

	^ self first: (self lastIndexOf: anElement ifAbsent: [^ self copy]) - 1! !

!WCollection methodsFor: 'copying' stamp: 'ael 6/17/2011 15:13'!
copyWith: newElement 
	"Answer a copy of the receiver that is 1 bigger than the receiver and 
	includes the argument, newElement, at the end."

	| newCollection |
	newCollection _ self copy.
	newCollection add: newElement.
	^newCollection! !

!WCollection methodsFor: 'copying' stamp: 'ael 6/17/2011 18:31'!
copyWithout: oldElement 
	"Answer a copy of the receiver that does not contain any
	elements equal to oldElement."

	^ self reject: [:each | each = oldElement]

"Examples:
	'fred the bear' copyWithout: $e
	#(2 3 4 5 5 6) copyWithout: 5
"! !

!WCollection methodsFor: 'copying' stamp: 'ael 6/17/2011 18:31'!
copyWithoutAll: aCollection
	"Answer a copy of the receiver that does not contain any elements 
	equal to those in aCollection."

	^ self reject: [:each | aCollection includes: each]! !

!WCollection methodsFor: 'enumerating' stamp: 'ael 6/17/2011 18:32'!
count: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument.  
	Answer the number of elements that answered true."

	| sum |
	sum := 0.
	self do: [:each | (aBlock value: each) ifTrue: [sum := sum + 1]].
	^ sum! !

!WCollection methodsFor: 'enumerating' stamp: 'ael 6/17/2011 18:32'!
detect: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument. 
	Answer the first element for which aBlock evaluates to true."

	^ self detect: aBlock ifNone: [self errorNotFound: aBlock]! !

!WCollection methodsFor: 'enumerating' stamp: 'ael 6/17/2011 18:32'!
detect: aBlock ifNone: exceptionBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument.  
	Answer the first element for which aBlock evaluates to true. If none  
	evaluate to true, then evaluate the argument, exceptionBlock."

	self do: [:each | (aBlock value: each) ifTrue: [^ each]].
	^ exceptionBlock value! !

!WCollection methodsFor: 'enumerating' stamp: 'ael 9/21/2011 01:44'!
do: aBlock 
	"World-aware behaviour: only log access to size if we get to the end of the collection"
	1 to: self unloggedSize do: [ :index |
		aBlock value: (self at: index).
		].
		
	self recordReadOfSize.		"ensure a read record for the collection size"
! !

!WCollection methodsFor: 'enumerating' stamp: 'ael 6/27/2011 13:53'!
do: elementBlock separatedBy: separatorBlock
	"Evaluate the elementBlock for all elements in the receiver,
	and evaluate the separatorBlock between."

	"World-aware behaviour: only log access to size if we get to the end of the collection"
	1 to: self unloggedSize do:
		[:index |
		index = 1 ifFalse: [separatorBlock value].
		elementBlock value: (self at: index)].
	
	self size.! !

!WCollection methodsFor: 'adding' stamp: 'ael 6/17/2011 15:13'!
emptyCheck

	self isEmpty ifTrue: [self errorEmptyCollection]! !

!WCollection methodsFor: 'private' stamp: 'ael 6/17/2011 18:35'!
errorEmptyCollection

	self error: 'this collection is empty'! !

!WCollection methodsFor: 'private' stamp: 'ael 6/17/2011 15:13'!
errorNoSuchElement

	self error: 'attempt to index non-existent element in an ordered collection'! !

!WCollection methodsFor: 'private' stamp: 'ael 6/17/2011 15:13'!
errorNotFound: anObject
	"Actually, this should raise a special Exception not just an error."

	self error: 'Object is not in the collection.'! !

!WCollection methodsFor: 'private' stamp: 'ael 6/17/2011 18:36'!
errorOutOfBounds

	self error: 'indices are out of bounds'! !

!WCollection methodsFor: 'explorer' stamp: 'ael 9/21/2011 01:45'!
explorerContents
	"re-implemented to avoid recording any worlds-relevant information"
	^self basicContents explorerContents! !

!WCollection methodsFor: 'enumerating' stamp: 'ael 9/21/2011 02:10'!
findFirst: aBlock
	"Return the index of my first element for which aBlock evaluates as true."

	"World-aware behaviour: only log access to size if we get to the end of the collection"
	| index |
	index := 0.
	[(index := index + 1) <= self unloggedSize] whileTrue:
		[(aBlock value: (self at: index)) ifTrue: [^index]].

	self recordReadOfSize.
	^ 0! !

!WCollection methodsFor: 'enumerating' stamp: 'ael 6/25/2011 03:39'!
findLast: aBlock
	"Return the index of my last element for which aBlock evaluates as true."

	| index |
	index := self size + 1			"outcome depends on the size, so record that we read it".
	[(index := index - 1) >= 1] whileTrue:
		[(aBlock value: (self at: index)) ifTrue: [^index]].
	^ 0! !

!WCollection methodsFor: 'accessing' stamp: 'ael 6/17/2011 15:13'!
first

	^ self at: 1! !

!WCollection methodsFor: 'accessing' stamp: 'ael 6/17/2011 18:02'!
first: n
	"Answer the first n elements of the receiver.
	Raise an error if there are not enough elements."

	^ self copyFrom: 1 to: n! !

!WCollection methodsFor: 'copying' stamp: 'ael 6/27/2011 13:45'!
forceTo: length paddingStartWith: elem 
	"Force the length of the collection to length, padding  
	the beginning of the result if necessary with elem.  
	Note that this makes a copy."
	
	"World-aware behaviour: only log access to size if length is off the end, because then size determines how many padding elements are inserted"
	| newCollection padLen size |
	newCollection := self species ofSize: length.
	size := self unloggedSize.
	padLen := length - size max: 0.
	padLen > 0 ifTrue: [ self size ].
	newCollection
		from: 1
		to: padLen
		put: elem.
	newCollection
		replaceFrom: padLen + 1
		to: ((padLen + size) min: length)
		with: self
		startingAt:  1.
	^ newCollection! !

!WCollection methodsFor: 'copying' stamp: 'ael 9/21/2011 02:11'!
forceTo: length paddingWith: elem
	"Force the length of the collection to length, padding
	if necessary with elem.  Note that this makes a copy."

	"World-aware behaviour: only log access to size if length is off the end, because then size determines how many padding elements are appended"
	| newCollection size |
	size := self unloggedSize.
	length > size ifTrue: [ self recordReadOfSize ].		"our size affects the result"
	newCollection := self species new: length withAll: elem.
	newCollection replaceFrom: 1 to: (size min: length) with: self startingAt: 1.
	^ newCollection! !

!WCollection methodsFor: 'accessing' stamp: 'ael 6/17/2011 20:00'!
from: startIndex to: endIndex put: anObject
	"Put anObject in all indexes between startIndex 
	and endIndex. Very fast. Faster than to:do: for
	more than 26 positions. Answer anObject"

	| written toWrite thisWrite |

	startIndex > endIndex ifTrue: [^self].
	self at: startIndex put: anObject.
	written := 1.
	toWrite := endIndex - startIndex + 1.
	[written < toWrite] whileTrue:
		[
			thisWrite := written min: toWrite - written.
			self 
				replaceFrom: startIndex + written
				to: startIndex + written + thisWrite - 1
				with: self startingAt: startIndex.
			written := written + thisWrite
		].
	^anObject! !

!WCollection methodsFor: 'adding' stamp: 'ael 6/25/2011 13:04'!
grownBy: length
	"Answer a copy of receiver collection with size grown by length"

	| newCollection size |
	size := self unloggedSize.
	newCollection := self species ofSize: size + length.
	newCollection replaceFrom: 1 to: size with: self startingAt: 1.
	^ newCollection! !

!WCollection methodsFor: 'testing' stamp: 'ael 6/17/2011 15:13'!
hasContentsInExplorer

	^self isEmpty not! !

!WCollection methodsFor: 'comparing' stamp: 'ael 9/21/2011 02:14'!
hasEqualElements: otherCollection 
	"World-aware behaviour: only log access to size if we get to the end of the collection"
	| size |
	(size := self unloggedSize) = otherCollection size
		ifFalse: [^ false].
	1
		to: size
		do: [:index | (self at: index)
					= (otherCollection at: index)
				ifFalse: [^ false]].

	self recordReadOfSize.	
	^ true
! !

!WCollection methodsFor: 'comparing' stamp: 'ael 9/21/2011 02:15'!
hash
	"World-aware behaviour: in principle the hash wouldn't have to access size, so don't log access to it"
	| hash |

	hash := self species hash.
	1 to: self unloggedSize do: [:i | hash := (hash + (self at: i) hash) hashMultiply].
	^hash! !

!WCollection methodsFor: 'world awareness' stamp: 'ael 9/21/2011 13:51'!
iChangedOrNewInWorld: world
	"this object is unchanged if there are no recorded Writes"
	| trackers writes |
	trackers := (world updateTrackersFor: self) ifNil: [ ^false ].

	"updateTrackers for a WCollection is a tuple  { proxy. reads. writes etc }"
	writes := (trackers at: 3) ifNil: [ ^false ].		"nothing written yet"

	"writes is a singleton array with a bit-index collection"
	^(writes at: 1) ~= 0
! !

!WCollection methodsFor: 'world awareness' stamp: 'ael 1/13/2012 20:00'!
iEnsureProxyIn: updateTrackers
	"the 'proxy' is an array of the same size as an instance of this species, which is used just to maintain and supply the private variable values for this object when it is operating other than in the top world.  it should not be sent messages other than at:(put:)"

	"updateTrackers for an WCollection is a tuple { proxy. reads. writes. sizeRead. remainingReadSlots.}"
	| proxy |
	(proxy := updateTrackers at: 1) ifNotNil: [ ^proxy ].
	
	^updateTrackers at: 1 put: self iMakeDeltaOrOrig.
! !

!WCollection methodsFor: 'world awareness' stamp: 'ael 10/31/2011 19:03'!
iEnsureReadsRecordIn: updateTrackers
	"updateTrackers for a WCollection is a tuple { proxy. reads. writes. sizeRead. remainingReadSlots. }
	NB: this method assumes that the world tracks reads (so caller must confirm that before).
	NB: this method assumes that if Reads has not yet been set up the collection is still at the same size as when we received it down the parent chain.  Therefore every method that is going to change the size must explicitly read the size before changing it."
	| reads size |
	(reads := updateTrackers at: 2) ifNotNil: [ ^reads ].

	self iEnsureProxyIn: updateTrackers.	"make sure we can query the size"
	size := self unloggedSize.				"this includes installing the contents IV"
	reads := Array new: size withAll: self dnkToken.
	updateTrackers at: 5 put: size.		"remainingReadSlots"
	^updateTrackers at: 2 put: reads
! !

!WCollection methodsFor: 'world awareness' stamp: 'ael 9/23/2011 16:07'!
iEnsureWritesRecordIn: updateTrackers
	"updateTrackers for an LCollection is a tuple { proxy. reads. writes. sizeRead. remainingReadSlots. }"
	| writes |
	(writes := updateTrackers at: 3) ifNotNil: [ ^writes ].

	^updateTrackers at: 3 put: { 0 }.
! !

!WCollection methodsFor: 'world awareness' stamp: 'ael 1/13/2012 23:14'!
iMergeUpdatesFrom: otherTrackers into: updateTrackers
	"updateTrackers is a tuple { proxy. reads. writes. sizeRead. remainingReadSlots. }
	writes is a singleton array holding a bit-index mask of the indices that have been written.  the corresponding values will be found in the contents 'IV' of the proxy array.
	for details, see WCollection>>iUpdatesIn:

	NB: this is executed in the receiving world"
	
	"first ensure we have a snapshot of the collection"
	| contentsIVIndex index otherContents otherProxy otherSize otherWrites otherWritesHolder receivingContents receivingWrites receivingWritesHolder val |
	receivingContents := contents.					"that'll do it"

	"now merge in the writes.  first we merge the bits"
	otherWritesHolder := otherTrackers at: 3.	"won't be nil"
	otherWrites := otherWritesHolder at: 1.
	
	receivingWritesHolder := self iEnsureWritesRecordIn: updateTrackers.
	receivingWrites := receivingWritesHolder at: 1.
	receivingWritesHolder at: 1 put: (receivingWrites bitOr: otherWrites).
	
	"now take care of the values"
	otherProxy := otherTrackers at: 1.
	contentsIVIndex := self class firstRedirectedIV.		"hard-coded assumption"
	otherContents := otherProxy at: contentsIVIndex.
	otherSize := otherContents size.
	
	index := 1.
	[ otherWrites = 0 ]
	  whileFalse: [
 		(otherWrites bitAnd: 1) = 1
		  ifTrue: [
			index > otherSize
			  ifTrue: [
				"it's a deletion.  perform any truncation necessary to leave us with size no greater than (index-1)"
				receivingContents size > (index - 1)
				  ifTrue: [ receivingContents size - index + 1 timesRepeat: [ self removeLast ] ].
				^self.			"any writes to greater indices are now meaningless" 
				].

			val := otherContents at: index.

			"we may need to extend the snapshot to allow the value to be added"
			index > receivingContents size
			  ifFalse: [ self at: index put: val ]
			  ifTrue: [
				index - receivingContents size - 1 timesRepeat: [self add: nil].
				self add: val.
				].
			].
		otherWrites := otherWrites >> 1.
		index := index + 1.
		].

! !

!WCollection methodsFor: 'world awareness' stamp: 'ael 1/13/2012 20:08'!
iUpdateTrackerSize
	"updateTrackers for a WCollection is a tuple { proxy. reads. writes. sizeRead. remainingReadSlots. }"
	^5
! !

!WCollection methodsFor: 'world awareness' stamp: 'ael 1/13/2012 22:51'!
iUpdatesIn: updateTrackers
	"updateTrackers for a WCollection is a tuple { proxy. reads. writes. sizeRead. remainingReadSlots. }"

	"Because of the special snapshot semantics of WCollections in child worlds, there are many ways to interpret how we should handle transfer of writes.  For now, here are the rules we follow:
		* The semantics of merge is 'we'll take only the values that have been written explicitly in the child world; everything else will be read from the parent'.
		* Therefore in the receiving world we'll take a new snapshot of the collection from the parent, and apply the writes on top of that snapshot.  If the collection has shifted since the snapshot on which the writes were based, you may end up with a nonsense collection.  Not Our Problem.
		* If there's a write beyond the range of the new snapshot, we'll extend the snapshot (padding with nils if necessary) to ensure that the write is valid.
		* If there's a deletion within the range of the new snapshot, we'll truncate the snapshot.

	Therefore in the transfer-trackers tuple we only fill in the writes slot and the proxy (to supply the values that have been written).  Everything else will have to be worked out on the receiving side. 
	"
	| oldProxy contentsIVIndex oldContents transferProxy transferTrackers writesHolder |
	(writesHolder := updateTrackers at: 3) ifNil: [ ^nil ].
	
	transferTrackers := Array new: self iUpdateTrackerSize withAll: nil.

	transferTrackers at: 3 put: writesHolder copy.

	"we also send a proxy with the current state of the contents, in which the values for the writes will be looked up"
	oldProxy := updateTrackers at: 1.
	transferProxy := self iMakeDeltaOrOrig.
	contentsIVIndex := self class firstRedirectedIV.		"hard-coded assumption"
	oldContents := oldProxy at: contentsIVIndex.
	transferProxy at: contentsIVIndex put: oldContents copy.	"probably safe even without copying, but certainly safer with"
	transferTrackers at: 1 put: transferProxy.

	^transferTrackers
! !

!WCollection methodsFor: 'testing' stamp: 'ael 6/17/2011 18:41'!
ifEmpty: aBlock
	"Evaluate the block if I'm empty"

	self isEmpty ifTrue: [ ^aBlock value ]! !

!WCollection methodsFor: 'testing' stamp: 'ael 6/17/2011 18:42'!
ifEmpty: emptyBlock ifNotEmpty: notEmptyBlock
	"Evaluate emptyBlock if I'm empty, notEmptyBlock otherwise"
	" If the notEmptyBlock has an argument, eval with the receiver as its argument"

	self isEmpty ifTrue: [ ^emptyBlock value ].
	^notEmptyBlock valueWithPossibleArgument: self! !

!WCollection methodsFor: 'testing' stamp: 'ael 6/17/2011 15:13'!
ifNotEmpty: aBlock

	self isNotEmpty ifTrue: [aBlock value].
! !

!WCollection methodsFor: 'testing' stamp: 'ael 6/17/2011 18:42'!
ifNotEmpty: notEmptyBlock ifEmpty: emptyBlock
	"Evaluate emptyBlock if I'm empty, notEmptyBlock otherwise
	 If the notEmptyBlock has an argument, eval with the receiver as its argument"

	self isEmpty ifFalse: [ ^notEmptyBlock valueWithPossibleArgument: self ].
	^emptyBlock value! !

!WCollection methodsFor: 'testing' stamp: 'ael 9/21/2011 14:06'!
includes: anElement 
	"Return true if the element is in the collection"
	"World-awareness: if the item is found, only the element in question (the earliest such element in the list) is marked as having been read.  This guarantees that the serialisability check will only be approved if the parent world still includes at least this one instance of the sought value.  On the other hand, if the value is not found we mark all items, to avoid a situation where a decision made on the absence of the item would be incompatible with other changes in the parent world."

	| size |
	size := self unloggedSize.
	1 to: size do:
		[:index |
		(contents at: index) = anElement			"not recorded"
		  ifTrue: [self recordReadAt: index. ^ true]
		].

	"failed.  record access to all items, and to size"
	self recordReadsFrom: 1 to: size.
	self recordReadOfSize.
	^ false
! !

!WCollection methodsFor: 'accessing' stamp: 'ael 9/21/2011 01:51'!
indexOf: x
	"World-aware behaviour: only log access to size if we get to the end of the collection without finding a match"
	1 to: self unloggedSize do: [:ii | (self at: ii) = x ifTrue: [^ ii]].
	self recordReadOfSize.
	^ 0! !

!WCollection methodsFor: 'accessing' stamp: 'ael 6/17/2011 18:43'!
indexOf: anElement ifAbsent: exceptionBlock
	"Answer the index of the first occurence of anElement within the  
	receiver. If the receiver does not contain anElement, answer the 
	result of evaluating the argument, exceptionBlock."

	^ self indexOf: anElement startingAt: 1 ifAbsent: exceptionBlock! !

!WCollection methodsFor: 'accessing' stamp: 'ael 9/21/2011 14:07'!
indexOf: anElement startingAt: start ifAbsent: exceptionBlock
	"Answer the index of the first occurence of anElement after start
	within the receiver. If the receiver does not contain anElement, 
	answer the 	result of evaluating the argument, exceptionBlock."

	"World-aware behaviour: only log access to size if we get to the end of the collection without finding a match"
	start to: self unloggedSize do:
		[:index |
		(self at: index) = anElement ifTrue: [^ index]].

	self recordReadOfSize.
	^ exceptionBlock value! !

!WCollection methodsFor: 'enumerating' stamp: 'ael 6/17/2011 18:44'!
inject: thisValue into: binaryBlock 
	"Accumulate a running value associated with evaluating the argument, 
	binaryBlock, with the current value of the argument, thisValue, and the 
	receiver as block arguments. For instance, to sum the numeric elements 
	of a collection, aCollection inject: 0 into: [:subTotal :next | subTotal + 
	next]."

	| nextValue |
	nextValue := thisValue.
	self do: [:each | nextValue := binaryBlock value: nextValue value: each].
	^nextValue! !

!WCollection methodsFor: 'inspecting' stamp: 'ael 10/3/2011 15:13'!
inspectorClass 
	"Answer the class of the inspector to be used on the receiver.  Called by inspect; 
	use basicInspect to get a normal (less useful) type of inspector."

	^WCollectionInspector! !

!WCollection methodsFor: 'world awareness' stamp: 'ael 9/26/2011 12:44'!
instVarAt: index
	| thisWorld trackers |
	(thisWorld := self thisWorld) isTopLevel ifTrue: [ ^self basicValueAt: index ].
	
	"the first access to our (single) instance variable involves obtaining a copy of it for use in this world."
	trackers := self updateTrackersForWorld: thisWorld createIfAbsent: true caching: true.
	 ^thisWorld ancestorCopyingValueAt: index for: self using: trackers proxySlot: 1 cacheRead: false.
! !

!WCollection methodsFor: 'world awareness' stamp: 'ael 9/21/2011 14:11'!
instVarAt: index put: value
	self thisWorld isTopLevel
	  ifTrue: [ ^self basicValueAt: index put: value ]
	  ifFalse: [ self error: 'Attempt to store in non-toplevel world' ]
! !

!WCollection methodsFor: 'enumerating' stamp: 'ael 12/2/2011 15:18'!
intersection: aCollection
	"Answer the set theoretic intersection of two collections."

	^ self select: [:each | aCollection includes: each]! !

!WCollection methodsFor: 'testing' stamp: 'ael 6/17/2011 18:45'!
isCollection
	"Return true if the receiver is some sort of Collection and responds to basic collection messages such as #size and #do:"
	^true! !

!WCollection methodsFor: 'testing' stamp: 'ael 9/23/2011 17:05'!
isEmpty
	"World-aware behaviour: use unloggedSize instead of size, since this operation only needs to know whether size > 0"
	^ contents isEmpty
! !

!WCollection methodsFor: 'testing' stamp: 'ael 6/17/2011 18:45'!
isEmptyOrNil
	"Answer whether the receiver contains any elements, or is nil.  Useful in numerous situations where one wishes the same reaction to an empty collection or to nil"

	^ self isEmpty! !

!WCollection methodsFor: 'testing' stamp: 'ael 6/25/2011 13:05'!
isNotEmpty
	"World-aware behaviour: use unloggedSize instead of size, since this operation only needs to know whether size > 0"
	^ self unloggedSize ~= 0! !

!WCollection methodsFor: 'world awareness' stamp: 'ael 5/1/2013 21:38'!
isSerializableIn: parentWorld using: updateTrackers
	"updateTrackers is a tuple { proxy. reads. writes. sizeRead. remainingReadSlots. }
		- if the size of the parent-world collection was read, and its size is now different, FAIL
		- if there is no Reads record (meaning nothing has been read or written), return OK
 		- for each non-DNK, non-Overwritten record in Reads, return FAIL if the parent-world collection now has a different value or has no value (i.e., it's now smaller)
		- if there is no Writes record, return OK
		- if there are any writes beyond the parent's current size + 1, and any gaps in those writes, FAIL
		- return OK
	"
	| overwritten reads val pSize dnk pContents sizeRead writeOneBeyondEnd writes writesBeyondEnd writesHolder |
	sizeRead := updateTrackers at: 4.

	"this commit might be rejected, so if the parent hadn't looked at the collection yet we mustn't cause it to make any persistent record."
	pContents := self nonCachingInstVarNamed: #contents in: parentWorld.
	pSize := pContents size.

	"if the size of the collection was read, check that it's now the same as that size"
	(sizeRead notNil and: [ pSize ~= sizeRead ])
	  ifTrue: [
		(WorldSizeChanged new messageText: 'Parent collection has ', (pSize > sizeRead ifTrue: [ 'grown' ] ifFalse: [ 'shrunk' ])) signal.
		^false
		].
	
	reads := updateTrackers at: 2.
	reads ifNil: [ ^true ].			"nothing was read, or this world has tracksReads=false"

	dnk := self dnkToken.
	overwritten := self overwriteToken.
	(1 to: reads size) do: [ :ii |
		val := reads at: ii.
		(val ~~ dnk and: [ val ~~ overwritten ])
		  ifTrue: [
			ii > pSize
			  ifTrue: [
				"we read something at an index that no longer exists in the parent"
				(WorldSizeChanged new messageText: 'Parent collection has shrunk') signal.
				^false
				].
			(self areEqualForCommit: val and: (pContents at: ii))
			  ifFalse: [
				"we read something at an index that now holds something else"
				(WorldValueChanged new indexPrefix: 'index' index: ii readValue: val currentValue: (pContents at: ii)) signal.
				^false ].
			]
		].

	writesHolder := updateTrackers at: 3.
	writesHolder ifNil: [ ^true ].			"nothing was written"

	writes := writesHolder at: 1.		"the bit-recorded indices"
	writeOneBeyondEnd := 1<<pSize.			"a write in the next place"
	writes > (writeOneBeyondEnd * 2 - 1) ifFalse: [ ^true ].		"all writes are within or at most one beyond end"

	"we have some write(s) beyond the end.  ensure there are no gaps."
	writesBeyondEnd := writes bitClear: (writeOneBeyondEnd - 1).	"find all places beyond"
	"one way to test for no gaps (e.g., 11110000) is to add a further 1 at the first non-zero bit and check that the result now has just one bit (e.g., 100000000)"
	(writesBeyondEnd + writeOneBeyondEnd) isPowerOfTwo
	  ifFalse: [ WorldSerializationError signal: 'Child has non-contiguous writes beyond end of parent'. ^false ].
	
	^true			"no objections"! !

!WCollection methodsFor: 'enumerating' stamp: 'ael 9/21/2011 14:18'!
keysAndValuesDo: aBlock 
	"Enumerate the receiver with all the keys (aka indices) and values."

	"World-aware behaviour: only log access to size if we get to the end of the collection"
	1 to: self unloggedSize do: [:index | aBlock value: index value: (self at: index)].

	self recordReadOfSize.
! !

!WCollection methodsFor: 'kscript' stamp: 'yo 7/17/2012 10:25'!
ksAt: key
	^ self at: key + 1
! !

!WCollection methodsFor: 'kscript' stamp: 'yo 7/6/2012 10:49'!
ksIsValidIndex: key
	^ key isInteger and: [key >= 0 and: [key < self size]]! !

!WCollection methodsFor: 'accessing' stamp: 'ael 6/27/2011 13:36'!
last
	| s |
	s := self size.		"record access, because it determines which index we look at"
	s = 0 ifTrue: [^ self errorNoSuchElement].
	^ self at: s! !

!WCollection methodsFor: 'accessing' stamp: 'ael 6/27/2011 13:36'!
last: n
	"Answer the last n elements of the receiver.  
	Raise an error if there are not enough elements."

	| size |
	size := self size.		"record access, because it determines which indices we look at"
	^ self copyFrom: size - n + 1 to: size! !

!WCollection methodsFor: 'world awareness' stamp: 'ael 10/31/2011 21:45'!
nonCachingInstVarAt: index in: aWorld
	^ aWorld nonCachingInstVarAt: index for: self writesSlot: 1 readsSlot: 0
! !

!WCollection methodsFor: 'testing' stamp: 'ael 6/17/2011 18:47'!
notEmpty
	"Answer whether the receiver contains any elements."

	^ self isEmpty not! !

!WCollection methodsFor: 'testing' stamp: 'ael 6/27/2011 14:03'!
occurrencesOf: anObject 
	"Answer how many of the receiver's elements are equal to anObject."
	| tally |
	tally := 0.
	self do: [:each | anObject = each ifTrue: [tally := tally + 1]].
	^tally! !

!WCollection methodsFor: 'world awareness' stamp: 'ael 10/31/2011 22:28'!
oldCopyBasedflattenIn: aWorld

	| copy trackers w |
	copy := aWorld eval: [ self unloggedCopy ].
	w := aWorld.
	[ w isTopLevel ]
	  whileFalse: [
		(trackers := copy updateTrackersForWorld: w createIfAbsent: false caching: false)
		  ifNotNil: [
			w eval: [ copy commitInto: w parent using: trackers ].
			w removeUpdateTrackersFor: copy.
			].
		w := w parent.
		].

	^copy collect: [ :each | each worldFlattenIn: aWorld ].
! !

!WCollection methodsFor: 'copying' stamp: 'ael 9/23/2011 15:51'!
postCopy
	super postCopy.
	contents := contents copy.
! !

!WCollection methodsFor: 'copying' stamp: 'ael 9/23/2011 15:34'!
postCopyTrackers: oldUTs into: newUTs
	"updateTrackers is a tuple { proxy. reads. writes. sizeRead. remainingReadSlots. }.
	a single-level copy is sufficient for all but the proxy, which must have a copy taken of the contents IV"
	| new old |
	1 to: oldUTs size
	  do: [ :i |
		old := oldUTs at: i.
		new := i = 1
				  ifFalse: [ old copy ]
				  ifTrue: [ old copy at: 2 put: (old at: 2) copy; yourself ].
		newUTs at: i put: new.
		]! !

!WCollection methodsFor: 'world awareness' stamp: 'ael 11/12/2011 11:36'!
postFlattenedCopy
	"suppress normal postCopy"
! !

!WCollection methodsFor: 'printing' stamp: 'ael 9/21/2011 14:20'!
printElementsOn: aStream
	"On the assumption that this is not - usually - an application-level operation, do *not* record updates for worlds"
	aStream nextPut: $(.
	(1 to: self unloggedSize)
	  do: [:index |
		aStream print: (contents at: index)]
	  separatedBy: [aStream space].
	aStream nextPut: $)! !

!WCollection methodsFor: 'printing' stamp: 'ael 6/17/2011 15:13'!
printNameOn: aStream
	super printOn: aStream! !

!WCollection methodsFor: 'printing' stamp: 'ael 6/17/2011 15:13'!
printOn: aStream 
	"Append a sequence of characters that identify the receiver to aStream."

	self printNameOn: aStream.
	self printElementsOn: aStream! !

!WCollection methodsFor: 'world awareness' stamp: 'ael 10/27/2011 14:40'!
privateRecord: valueOrNil inReads: reads from: start to: finish remaining: remaining
	"nil value means get the element from the current contents"
	| collection index remainingReads realFinish dnk |
	collection := nil.
	remainingReads := remaining.
	dnk := self dnkToken.

	index := start.
	realFinish := finish min: reads size.
	[ remainingReads > 0 and: [ index <= realFinish ] ]
	  whileTrue: [
		(reads at: index) == dnk
		  ifTrue: [
			reads at: index put: (valueOrNil ifNil: [ (collection ifNil: [ collection := contents ]) at: index ]).
			remainingReads := remainingReads - 1.
			].
		index := index + 1.
		].

	^remainingReads.
! !

!WCollection methodsFor: 'filter streaming' stamp: 'ael 6/17/2011 18:49'!
putOn: aStream

	self do: [ :each | each putOn: aStream ]! !

!WCollection methodsFor: 'world awareness' stamp: 'ael 9/20/2011 18:59'!
recordReadAt: index
	self recordReadsFrom: index to: index
! !

!WCollection methodsFor: 'world awareness' stamp: 'ael 9/25/2011 15:07'!
recordReadOfSize
	"Records updates for worlds
		- updateTrackers is a tuple { proxy. reads. writes. sizeRead. remainingReadSlots. }"

	| thisWorld trackers |
	(thisWorld := self thisWorld) isTopLevel
	  ifTrue: [ ^self ]
	  ifFalse: [
		trackers := self updateTrackersForWorld: thisWorld createIfAbsent: true caching: true.
		(trackers at: 4) ifNotNil: [ ^self ].
		trackers at: 4 put: self unloggedSize.
		].
! !

!WCollection methodsFor: 'world awareness' stamp: 'ael 10/28/2011 23:47'!
recordReadsFrom: start to: finish
	"update the Reads record, if needed
		* if this world doesn't record reads, do nothing
		* if the reads record is fully marked up, do nothing
		* if the index is outside the range of the Reads array, do nothing
		* if the existing Reads record is anything other than DNK, do nothing
		* record the value that was the read
		
	updateTrackers is a tuple { proxy. reads. writes. sizeRead. remainingReadSlots.}"

	| reads remainingReads thisWorld trackers finalRemaining |
	(thisWorld := self thisWorld) isTopLevel ifTrue: [ ^self ].

	thisWorld tracksReads ifFalse: [ ^self ].
	
	trackers := self updateTrackersForWorld: thisWorld createIfAbsent: true caching: true.

	(remainingReads := trackers at: 5) = 0 ifTrue: [ ^self ].		"could still be nil"
	
	reads := self iEnsureReadsRecordIn: trackers.
	remainingReads ifNil: [ remainingReads := trackers at: 5 ].	"put there when reads record is set up by previous call"

	finalRemaining := self privateRecord: nil inReads: reads from: start to: finish remaining: remainingReads.
		
	trackers at: 5 put: finalRemaining

! !

!WCollection methodsFor: 'world awareness' stamp: 'ael 9/20/2011 18:37'!
recordWriteAt: position
	self recordWritesFrom: position to: position
! !

!WCollection methodsFor: 'world awareness' stamp: 'ael 9/25/2011 15:08'!
recordWritesFrom: start to: finish
	"when a value is written:
		- add the index to Writes (value can be found by looking at the collection again)
		- if there is a DNK record in Reads for this index, replace it with Overwritten"

	"writes is a singleton Array holding a number used as a bit mask, where the LSB corresponds to index 1"
	| reads writes finalRemaining remainingReads thisWorld trackers |
	(thisWorld := self thisWorld) isTopLevel ifTrue: [ ^self ].

	trackers := self updateTrackersForWorld: thisWorld createIfAbsent: true caching: true.
	(writes := self iEnsureWritesRecordIn: trackers) ifNil: [ ^self ].
	writes at: 1 put: ((writes at: 1) bitOr: (Integer bitSequence: start to: finish)).

	"now record overwrites in the reads collection, if there is one"
	thisWorld tracksReads ifFalse: [ ^self ].		"nothing else to do"
	
	reads := self iEnsureReadsRecordIn: trackers.
	remainingReads := trackers at: 5.
	finalRemaining := self privateRecord: self overwriteToken inReads: reads from: start to: finish remaining: remainingReads.
		
	trackers at: 5 put: finalRemaining
! !

!WCollection methodsFor: 'enumerating' stamp: 'ael 6/17/2011 18:49'!
reject: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument. 
	Collect into a new collection like the receiver only those elements for 
	which aBlock evaluates to false. Answer the new collection."

	^self select: [:element | (aBlock value: element) == false]! !

!WCollection methodsFor: 'removing' stamp: 'ael 6/17/2011 15:13'!
remove: oldObject 
	"Remove oldObject from the receiver's elements. Answer oldObject 
	unless no element is equal to oldObject, in which case, raise an error.
	ArrayedCollections cannot respond to this message."

	^ self remove: oldObject ifAbsent: [self errorNotFound: oldObject]! !

!WCollection methodsFor: 'removing' stamp: 'ael 9/22/2011 01:08'!
remove: oldObject ifAbsent: absentBlock
	"Records updates for worlds
		- here we just record the elements that were seen when looking for oldObject
		- in the removeAt: method we'll record the deletion and any shifting"

	| index |
	index := self indexOf: oldObject ifAbsent: [ ^absentBlock value ].

	^self removeAt: index
! !

!WCollection methodsFor: 'removing' stamp: 'ael 9/22/2011 01:10'!
removeAll
	"World awareness: there's no need to log any reads, but we must log writes to every deleted position"
	| size |
	size := self unloggedSize.
	contents removeAll.
	self recordWritesFrom: 1 to: size.
! !

!WCollection methodsFor: 'removing' stamp: 'ael 6/17/2011 18:50'!
removeAll: aCollection 
	"Remove each element of aCollection from the receiver. If successful for 
	each, answer aCollection. Otherwise create an error notification.
	ArrayedCollections cannot respond to this message."

	aCollection == self ifTrue: [^self removeAll].
	aCollection do: [:each | self remove: each].
	^ aCollection! !

!WCollection methodsFor: 'removing' stamp: 'ael 9/22/2011 01:22'!
removeAllSuchThat: aBlock 
	"Remove each element of the receiver for which aBlock evaluates to true.

	NB: if the block bails out in the middle of this enumeration, the collection will be in an inconsistent state"

	| firstRemoved index oldSize |
	oldSize := self size.	"record, because size will either change or will determine where we stop"

	self recordReadsFrom: 1 to: oldSize.	"everything will be examined"
	
	firstRemoved := nil.
	index := 0.
	contents removeAllSuchThat: [ :element |
		index := index + 1.
		(aBlock value: element)
		  and: [ firstRemoved ifNil: [ firstRemoved := index ]. true ].
		].

	firstRemoved ifNotNil: [ self recordWritesFrom: firstRemoved to: oldSize ].	"include writes representing deletions of the last items"
! !

!WCollection methodsFor: 'removing' stamp: 'ael 9/23/2011 16:00'!
removeAt: index
	| removed collection size |
	self recordReadOfSize.
	collection := contents.
	size := collection size.
	self recordReadsFrom: index to: size.
	removed := collection removeAt: index.
	self recordWritesFrom: index to: size.	"include deleted last entry"
	^removed! !

!WCollection methodsFor: 'removing' stamp: 'ael 9/22/2011 01:23'!
removeFirst
	"Remove the first element of the receiver and answer it. If the receiver is 
	empty, create an error notification."
	self emptyCheck.
	^self removeAt: 1
! !

!WCollection methodsFor: 'removing' stamp: 'ael 9/22/2011 01:24'!
removeFirst: n
	"Remove first n objects into an array, returned to the user"

	| list |
	list := self arrayClass new: n.
	1 to: n do: [:i |
		list at: i put: self removeFirst].
	^ list! !

!WCollection methodsFor: 'removing' stamp: 'ael 9/22/2011 01:25'!
removeLast
	"Remove the last element of the receiver and answer it. If the receiver is 
	empty, create an error notification."
	"Records updates for worlds"
	| size |
	self emptyCheck.
	size := self size.			"recorded, because size is changing"
	^self removeAt: size
! !

!WCollection methodsFor: 'removing' stamp: 'ael 9/22/2011 01:26'!
removeLast: n
	"Remove last n object into an array with last in last position"

	| list |
	list := self arrayClass new: n.
	n to: 1 by: -1 do: [:i |
		list at: i put: self removeLast].
	^ list! !

!WCollection methodsFor: 'adding' stamp: 'ael 9/21/2011 02:05'!
reorder: anElement beAt: nth
	"Move anElement so that it becomes the nth element in the collection."
	| index |
	index := self indexOf: anElement ifAbsent: [ ^self errorNotFound: anElement ].
	index = nth ifTrue: [^ self]. 	"already there"
	contents removeAt: index.
	contents add: anElement afterIndex: nth-1.
	self recordReadsFrom: 1 to: (index max: nth).
	self recordWritesFrom: (index min: nth) to: (index max: nth).! !

!WCollection methodsFor: 'accessing' stamp: 'ael 6/17/2011 19:50'!
replaceFrom: start to: stop with: replacement 
	"This destructively replaces elements from start to stop in the receiver. 
	Answer the receiver itself. Use copyReplaceFrom:to:with: for 
	insertion/deletion which may alter the size of the result."

	replacement size = (stop - start + 1)
		ifFalse: [self error: 'Size of replacement doesnt match'].
	^self replaceFrom: start to: stop with: replacement startingAt: 1! !

!WCollection methodsFor: 'accessing' stamp: 'ael 6/17/2011 19:50'!
replaceFrom: start to: stop with: replacement startingAt: repStart 
	"This destructively replaces elements from start to stop in the receiver 
	starting at index, repStart, in the sequenceable collection, 
	replacementCollection. Answer the receiver. No range checks are 
	performed."

	| index repOff |
	repOff := repStart - start.
	index := start - 1.
	[(index := index + 1) <= stop]
		whileTrue: [self at: index put: (replacement at: repOff + index)]! !

!WCollection methodsFor: 'enumerating' stamp: 'ael 9/22/2011 01:30'!
reverseDo: aBlock 
	| size |
	size := self size.		"record read of size, because that determines where the accesses start"
	size to: 1 by: -1 do: [ :index | aBlock value: (self at: index) ]
! !

!WCollection methodsFor: 'copying' stamp: 'ael 9/22/2011 01:30'!
reversed
	"Answer a copy of the receiver with element order reversed.  "
	| newCol |
	newCol := self species new.
	self reverseDo:
		[:elem | newCol addLast: elem].
	^ newCol

"#(2 3 4 'fred') reversed"! !

!WCollection methodsFor: 'accessing' stamp: 'ael 6/17/2011 15:13'!
second

	^ self at: 2! !

!WCollection methodsFor: 'enumerating' stamp: 'ael 9/22/2011 01:31'!
select: aBlock 
	"Evaluate aBlock with each of my elements as the argument. Collect into
	a new collection like the receiver, only those elements for which aBlock
	evaluates to true."
	"World-aware behaviour: only log access to size if we get to the end of the collection"
	| newCollection element |
	newCollection := self copyEmpty.
	1 to: self unloggedSize do:
		[:index |
		element := self at: index.
		(aBlock value: element) ifTrue: [newCollection addLast: element]
		].
	self recordReadOfSize.		"ensure a read record for the collection size"
	^ newCollection! !

!WCollection methodsFor: 'enumerating' stamp: 'ael 6/17/2011 18:52'!
select: selectBlock thenCollect: collectBlock
	"Utility method to improve readability."

	^ (self select: selectBlock) collect: collectBlock! !

!WCollection methodsFor: 'private' stamp: 'ael 9/21/2011 01:28'!
setContents: anArray
	"only used at initialisation, so no updates to be recorded for worlds"
	contents setContents: anArray.
! !

!WCollection methodsFor: 'private' stamp: 'ael 9/21/2011 01:29'!
setContentsSize: anInteger
	"only used at initialisation, so no updates to be recorded for worlds"
	contents := OrderedCollection new: anInteger.
! !

!WCollection methodsFor: 'siss-private' stamp: 'ael 9/22/2011 01:32'!
sissContentsInto: newElem context: dictionary
self flag: #aelToDo.
	self class isBits ifTrue: [
		self sissBitsInto: newElem context: dictionary.
		^ self.
	].
	self sissWithCollectionEntries ifTrue: [
		self sissIndexFieldsInto: newElem context: dictionary.
		^ self.
	].
	self sissInstVarsInto: newElem context: dictionary.
! !

!WCollection methodsFor: 'siss-private' stamp: 'ael 9/22/2011 01:32'!
sissInstVarNamed: varName put: childInst
self flag: #aelToDo.

	self addLast: childInst.
	self size ~= varName asNumber ifTrue: [self error: 'wrong order'].
! !

!WCollection methodsFor: 'siss-private' stamp: 'ael 9/22/2011 01:32'!
sissInstVarNamed: varName put: childInst in: context
self flag: #aelToDo.

	self addLast: childInst.
	"self size ~= varName asNumber ifTrue: [self error: 'wrong order']."
! !

!WCollection methodsFor: 'siss-private' stamp: 'ael 9/22/2011 01:32'!
sissWithBasicAt
self flag: #aelToDo.

	^ false.
! !

!WCollection methodsFor: 'siss-private' stamp: 'ael 9/22/2011 01:32'!
sissWithCollectionEntries
self flag: #aelToDo.

	^ true.
! !

!WCollection methodsFor: 'accessing' stamp: 'ael 9/21/2011 15:46'!
size
	"Answer how many elements the receiver contains."
	"Records updates for worlds
		- reading the size is recorded by putting the size seen into the last position of our Reads array.
	NB: any operation that changes the size of the collection MUST access size before making any changes, so that we capture the right value for size of the 'inherited' collection."

	| size |
	size := self unloggedSize.
	self recordReadOfSize.
	^size
! !

!WCollection methodsFor: 'sorting' stamp: 'ael 6/17/2011 18:54'!
sortBy: aBlock
	"Create a copy that is sorted.  Sort criteria is the block that accepts two arguments.
	When the block is true, the first arg goes first ([:a :b | a > b] sorts in descending
	order)."

	^ self asOrderedCollection
		sort: aBlock;
		yourself! !

!WCollection methodsFor: 'sorting' stamp: 'ael 6/17/2011 18:54'!
sorted
	"Return a new sequenceable collection which contains the same elements as self but its elements are sorted in ascending order using the #'<=' operator."

	^self sorted: nil! !

!WCollection methodsFor: 'sorting' stamp: 'ael 6/17/2011 18:56'!
sorted: aSortBlockOrNil
	"Return a new sequenceable collection which contains the same elements as self but its elements are sorted by aSortBlockOrNil. The block should take two arguments and return true if the first element should preceed the second one. If aSortBlock is nil then <= is used for comparison."
	
	^self copy sort: aSortBlockOrNil! !

!WCollection methodsFor: 'printing' stamp: 'ael 6/17/2011 18:57'!
storeOn: aStream 
	"Refer to the comment in Object|storeOn:."

	| noneYet |
	aStream nextPutAll: '(('.
	aStream nextPutAll: self class name.
	aStream nextPutAll: ' new)'.
	noneYet := true.
	self do: 
		[:each | 
		noneYet
			ifTrue: [noneYet := false]
			ifFalse: [aStream nextPut: $;].
		aStream nextPutAll: ' add: '.
		aStream store: each].
	noneYet ifFalse: [aStream nextPutAll: '; yourself'].
	aStream nextPut: $)! !

!WCollection methodsFor: 'accessing' stamp: 'ael 6/17/2011 15:13'!
third
	"Answer the third element of the receiver."

	^ self at: 3! !

!WCollection methodsFor: 'accessing' stamp: 'ael 11/12/2011 14:51'!
uncheckedAt: anInteger put: anObject 
	"for testing"
	anInteger > contents size
	  ifFalse: [ contents at: anInteger put: anObject ]
	  ifTrue: [ contents at: anInteger ifAbsentPut: [ anObject ] ].
	self recordWriteAt: anInteger.
	^ anObject.
! !

!WCollection methodsFor: 'world awareness' stamp: 'ael 9/22/2011 01:33'!
unloggedAt: anInteger
	"Answer an element of the receiver without logging the access"
	(anInteger < 1 or: [anInteger > contents size])
		ifTrue: [self errorNoSuchElement]
		ifFalse: [^contents at: anInteger ].
! !

!WCollection methodsFor: 'world awareness' stamp: 'ael 10/30/2011 22:42'!
unloggedCopy
	"now that we copy by flattening, this shouldn't be needed"
	self halt: 'No longer supported'.
	^super copy
! !

!WCollection methodsFor: 'world awareness' stamp: 'ael 9/23/2011 17:05'!
unloggedSize
	"Answer how many elements the receiver contains, without recording the fact of the lookup."
	^contents size
! !

!WCollection methodsFor: 'enumerating' stamp: 'ael 9/22/2011 01:35'!
with: otherCollection collect: twoArgBlock 
	"Collect and return the result of evaluating twoArgBlock with 
	corresponding elements from this collection and otherCollection."
	"World-aware behaviour: only log access to size if we get to the end of the collection"
	| result size |
	otherCollection size = (size := self unloggedSize) ifFalse: [self error: 'otherCollection must be the same size'].
	result := self species new: size.
	1 to: size do:
		[:index | result addLast: (twoArgBlock value: (self at: index)
									value: (otherCollection at: index))].
	self recordReadOfSize.		"record access to size"
	^ result! !

!WCollection methodsFor: 'enumerating' stamp: 'ael 9/22/2011 01:35'!
with: otherCollection do: twoArgBlock 
	"Evaluate twoArgBlock with corresponding elements from this collection and otherCollection."
	"World-aware behaviour: only log access to size if we get to the end of the collection"
	| size |
	otherCollection size = (size := self size) ifFalse: [self error: 'otherCollection must be the same size'].
	1 to: size do:
		[:index | twoArgBlock value: (self at: index)
									value: (otherCollection at: index)].
	self recordReadOfSize.		"record access to size"
	^ self! !

!WCollection methodsFor: 'enumerating' stamp: 'ael 9/22/2011 01:38'!
withIndexCollect: elementAndIndexBlock 
	"Just like with:collect: except that the iteration index supplies the second argument to the block."
	"World-aware behaviour: only log access to size if we get to the end of the collection"
	| newCollection size |
	size := self unloggedSize.
	newCollection := self species new: size.
	1 to: size do:
		[:index | newCollection addLast: (elementAndIndexBlock value: (self at: index) value: index) ].
	self recordReadOfSize.		"record access to size"
	^ newCollection
! !

!WCollection methodsFor: 'enumerating' stamp: 'ael 9/22/2011 01:39'!
withIndexDo: elementAndIndexBlock 
	"Just like do: except that the iteration index supplies the second argument to the block.  Override superclass in order to use addLast:, not at:put:."
	"World-aware behaviour: only log access to size if we get to the end of the collection"

	1 to: self unloggedSize do: [ :index |
		elementAndIndexBlock value: (self at: index) value: index.
		].
		
	self recordReadOfSize.		"ensure a read record for the collection size"
! !

!WCollection methodsFor: 'world awareness' stamp: 'ael 11/13/2011 23:34'!
worldFlattenIn: aWorld
	| w numLeft tmpDst currentSize flatSize foundSize lowestWrite max min relevantWrites trackers worldAndWrites worldStream worldsAndWrites writes writesHolder dnk |

	"for efficiency we first figure out the size of the flattened collection.
	at each commit from child to parent (in fact, to the nearest ancestor that has accessed the collection), the resulting collection size in the ancestor is determined as follows:
		if there is deletion record (a write at an index beyond the end of the collection) for index n, the maximum size is n-1
		if there is a valid write record for index m (<n), the minimum size is m
		otherwise (no writes at all) the size in the ancestor stays as is
		
		thus the size in the ancestor is (<size before commit> min: (n-1)) max: m.  therefore:
			set the estimated size S based on the collection in the leaf world
			any write > S sets a maximum
			any write <= S sets a minimum
			if minimum=maximum, we've found the size.  otherwise go to the next ancestor world, with S set as follows: if there is a maximum, S is min(parentS, max); if there is a minimum, S is max(parentS, min); if none, S is parentS.  if we get to calculate S in the top-level world, flatSize = S.
		"
	
	dnk := self dnkToken.

	"as we go along, build up a list of the worlds that have Writes records"
	worldsAndWrites := OrderedCollection new.
	max := min := nil.
	foundSize := false.
	w := aWorld.
	[ w isTopLevel ]
	  whileFalse: [
		(trackers := self updateTrackersForWorld: w createIfAbsent: false caching: false)
		  ifNotNil: [
			"a tuple { proxy. reads. writes. sizeRead. remainingReadSlots.}"
			(writesHolder := trackers at: 3)
			  ifNotNil: [
				writes := writesHolder at: 1.
				worldsAndWrites add: { w. writes }.		"for later"
				foundSize
				  ifFalse: [
					relevantWrites := writes.
					currentSize := w eval: [ self unloggedSize ].
					max ifNotNil: [
						currentSize := currentSize min: max.
						"we're only interested in writes up to max"
						relevantWrites := relevantWrites bitAnd: (1<<max)-1.
						].
					min ifNotNil: [
						currentSize := currentSize max: min.
						"we're only interested in writes above min"
						relevantWrites := relevantWrites bitClear: (1<<min)-1.
						].
					[ relevantWrites > 0 ]
					  whileTrue: [
						lowestWrite := relevantWrites lowBit.
						lowestWrite <= currentSize
						  ifTrue: [ min := lowestWrite ]
						  ifFalse: [ max := lowestWrite - 1. relevantWrites := 0 ].
						min = max
						  ifTrue: [ foundSize := true.  relevantWrites := 0 ]
						  ifFalse: [ relevantWrites := relevantWrites - (1<<(lowestWrite-1)) ].
						].
					].
				 ].
			].
		w := w parent.
		].

	"now w is the top-level world"
	worldsAndWrites add: { w. nil }.		"as a sentinel"
	flatSize :=foundSize
				ifTrue: [ min ]		"same as max"
				ifFalse: [
					currentSize := w eval: [ self unloggedSize ].
					max ifNotNil: [ currentSize := currentSize min: max ].
					min ifNotNil: [ currentSize := currentSize max: min ].
					currentSize
					].

	"for speed we build up the values in an Array then turn it into a WCollection at the end"
	tmpDst := Array new: flatSize withAll: dnk.
	numLeft := flatSize.
	worldStream := worldsAndWrites readStream.
	[numLeft > 0] whileTrue: [
		worldAndWrites := worldStream next.
		w := worldAndWrites at: 1.
		w eval: [
			w isTopLevel
			  ifFalse: [
				writes := worldAndWrites at: 2.
				1 to: (flatSize min: contents size) do: [ :index |
					(numLeft > 0 and: [ (tmpDst at: index) == dnk and: [ (writes bitAt: index) = 1 ] ])
					  ifTrue: [
						tmpDst at: index put: (contents at: index).
						numLeft := numLeft - 1
						].
					]
				]
			  ifTrue: [		"top-level world"
				1 to: flatSize do: [:index |
					(tmpDst at: index) == dnk
					  ifTrue: [ tmpDst at: index put: (contents at: index) ].
					].
				numLeft := 0.		"force us out of the loop"
				]
			]
		].

	^WTopLevelWorld default eval: [ self class withAll: tmpDst ]
! !

!WCollection methodsFor: 'converting' stamp: 'ael 11/12/2011 11:50'!
worldFlattenedCopy
	"get a flattened copy of the receiver, keeping appropriate records of reads from parent worlds"
	self thisWorld isTopLevel ifTrue: [ ^self clone postCopy ].

	self recordReadsFrom: 1 to: self size.		"size is also accessed, of course"
	^self class withAll: contents	"we have direct access to the contents in this world"
! !

WObjectWithCachedLookup subclass: #WIdentityDictionary
	instanceVariableNames: 'contents'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Worlds'!

!WIdentityDictionary commentStamp: 'ael 9/24/2011 12:05' prior: 0!
In WIdentityDictionary (like in WArray) the contents IV is used only in the top-level world, and is therefore non-redirected.!

WIdentityDictionary subclass: #WDeepIdentityDictionary
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Worlds'!

!WDeepIdentityDictionary commentStamp: 'ael 4/26/2013 13:26' prior: 0!
A WDeepIdentityDictionary (WDeepID) is an optimisation of WIdentityDictionary for use in contexts where there are very long parent-world chains (hundreds or thousands of levels) but updates are sparsely distributed through the worlds.  The normal lookup mechanics for WIdentityDictionary would visit each step on the chain for each key looked up.  Instead, we install an update tracker that is shared between all the worlds in a chain, and maintains for each key in the WDeepID a record that allows, from any world in the chain, a fast search to the value in the closest parent world in which that key was given an update.

For now, this is the only class to offer this kind of support (because it was all we needed for the experiment on WWorld-based tracing for KSObject).  In principle it could be extended to other classes.

Deep lookup is implemented using the following specialised state:

WWorld inst var: delegationDepth
	- holds the WWorld's depth below the WTopLevelWorld (i.e., number of 'parent' steps needed to get to the top level)

WWorld inst var: deepUpdateDict
	- a dictionary holding the trackers - world-specific value records - for objects that participate in deep lookup (currently only WDeepIDs, as noted above).  The dictionary entries are - as usual - updateTracker arrays keyed by the instances, but the update tracker for a WDeepID has only a single entry (described below).
	- the crucial distinction between this and the standard updateTrackerDict is that, in a non-branching chain of child WWorlds, all the worlds' deepUpdateDict IVs point to the same dictionary.  This removes the need for the normal stepwise traversal up through successive parent worlds.  Branching world chains require slightly more fiddly handling - and in fact are not yet (Apr 2013) fully supported; see WDeepID>>at:put:.

WWorld inst var: hasChildren
	- flag recording whether the world already has child worlds.  A world sprouted from one that already has children must be given a new, independent copy of the deepUpdateDict.  Note that once this flag is set it won't be reset, even if all child worlds are garbage-collected; re-implementing in the form of a weak collection of child worlds would be more general.

single-element update tracker for WDeepID (see method iEnsureDictTrackerIn:)
	- the tracker is a dictionary mapping each key in the WDeepID to a sorted OrderedCollection of { depth. value } pairs representing all the values assigned to that key in worlds on the same chain.  For example:
			#foo ->  collection( { 0. 'hello' }. )
			#bar ->  collection( { 0. 'world' }.  { 5. 'everyone' } )
		...represents a situation in which in the Top world the keys #foo and #bar have the values 'hello' and 'world' respectively, while #bar has been updated to 'everyone' in the world five levels below Top.



!

!WDeepIdentityDictionary methodsFor: 'adding' stamp: 'ael 7/2/2012 22:36'!
add: anAssociation
	self at: anAssociation key put: anAssociation value
! !

!WDeepIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 7/2/2012 22:49'!
adoptReadFromChild: assoc
	self shouldBeImplemented	"maybe"! !

!WDeepIdentityDictionary methodsFor: 'accessing' stamp: 'ael 7/3/2012 17:40'!
associationAt: key ifAbsent: aBlock 
	"Answer an association with the given key.
	If key is not found, return the result of evaluating aBlock."
	| val |
	val := self at: key ifAbsent: [ ^aBlock value ].
	^key -> val
! !

!WDeepIdentityDictionary methodsFor: 'accessing' stamp: 'ael 7/2/2012 23:17'!
associations
	"Answer a Collection containing the receiver's associations."

	| result |
	result := OrderedCollection new: self roughSize.
	self associationsDo: [ :assoc | result add: assoc ].
	^ result
! !

!WDeepIdentityDictionary methodsFor: 'enumerating' stamp: 'ael 7/9/2012 15:11'!
associationsDo: aBlock
	| thisWorld |
	(thisWorld := self thisWorld) isTopLevel ifTrue: [ ^contents associationsDo: aBlock ].

	self privateEntriesInWorld: thisWorld do: [ :key :val | aBlock value: key->val ]
! !

!WDeepIdentityDictionary methodsFor: 'accessing' stamp: 'ael 7/9/2012 21:31'!
at: key ifAbsent: aBlock 
	"Answer the value stored at the given key.
	If key is not found, return the result of evaluating aBlock."
	| thisWorld trackers index val dictTracker |
	(thisWorld := self thisWorld) isTopLevel ifTrue: [ ^contents at: key ifAbsent: aBlock ].

	trackers := self updateTrackersForWorld: thisWorld createIfAbsent: true caching: true.
	dictTracker := self iEnsureDictTrackerIn: trackers.
	dictTracker at: key ifPresent: [ :entries |
		index := self privateIndexForDepth: thisWorld delegationDepth in: entries.
		index = 0 ifFalse: [
			(val := (entries at: index) at: 2) == self deletionToken ifFalse: [ ^val ].
			]
		].
	^aBlock value
! !

!WDeepIdentityDictionary methodsFor: 'accessing' stamp: 'ael 4/26/2013 12:59'!
at: key put: anObject 
	"Set the value at key to be anObject (which may be the deletionToken).
	If key is not found, create a new entry for key and set its value to anObject.
	Answer anObject.
	NB: doesn't take account of branching world chains.  If invoked in a non-leaf world that is above a branch, this update will only be seen in worlds that use the same deepUpdateDict as that world."
	| thisWorld trackers depth dictTracker entries newEntry existing index |
self flag: #aelToDo.			"add handling for branching world chains?"
	(thisWorld := self thisWorld) isTopLevel ifTrue: [ ^contents at: key put: anObject ].

	trackers := self updateTrackersForWorld: thisWorld createIfAbsent: true caching: true.
	dictTracker := self iEnsureDictTrackerIn: trackers.
	depth := thisWorld delegationDepth.
	newEntry := { depth. anObject }.
	entries := dictTracker at: key ifAbsent: [ dictTracker at: key put: (OrderedCollection with: newEntry). ^anObject ].
	index := self privateIndexForDepth: depth in: entries.
	index = 0
		ifTrue: [ entries addLast: newEntry ]	"last entry was deeper than the requested depth"
		ifFalse: [
			existing := entries at: index.
			(existing at: 1) = depth
				ifTrue: [ existing at: 2 put: anObject ]
				ifFalse: [ entries add: newEntry beforeIndex: index ]
			].
	^anObject
! !

!WDeepIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 7/10/2012 00:27'!
commitInto: parentWorld using: updateTrackers
	"for each value that has a setting in the executing world, remove it and transfer its value into parentWorld"
	| dictTracker deletion index parentDepth thisDepth val nextEntry |
	dictTracker := self iEnsureDictTrackerIn: updateTrackers.
	thisDepth := self thisWorld delegationDepth.
	parentDepth := parentWorld delegationDepth.
	deletion := self deletionToken.
	dictTracker keysAndValuesDo: [ :key :entries |
		(self isCommittableKey: key) ifTrue: [
			index := self privateIndexForDepth: thisDepth in: entries.
			(index ~= 0 and: [ ((entries at: index) at: 1) = thisDepth ]) ifTrue: [
				val := (entries at: index) at: 2.
				"if the next entry has the right depth for the parent, remove the child world's entry and overwrite just the parent's value.  otherwise overwrite the child entry with an entry to represent the parent." 
				(index < entries size and: [ ((nextEntry := entries at: index+1) at: 1) = parentDepth ])
					ifTrue: [
						nextEntry at: 2 put: val.
						entries removeAt: index.
						]
					ifFalse: [ entries at: index put: { parentDepth. val } ].
				"if parent is the top-level world, also update the 'real' collection"
				parentDepth = 0 ifTrue: [
					self evalInTopWorld: [
						val == deletion
							ifFalse: [ self at: key put: val ]
							ifTrue: [ self removeKey: key ]
						]
					].
				].
			]
		].
! !

!WDeepIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 7/6/2012 16:40'!
copyMissingOrigValuesInto: parentWorld using: updateTrackers
	"ignore, unless we get around to implementing read tracking"
! !

!WDeepIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 7/9/2012 15:05'!
dictTrackerIn: updateTrackers
	^updateTrackers at: 1
! !

!WDeepIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 7/9/2012 23:52'!
duplicateTrackers: updateTrackers forDepth: depth
	"updateTrackers is a single-placed array"
	| dictTracker newDictTracker newEntries newTrackers |
	newTrackers := updateTrackers copy.			"shallow copy"
	dictTracker := self dictTrackerIn: updateTrackers.
	dictTracker ifNil: [ ^newTrackers ].				"nothing more to do"

	newDictTracker := newTrackers at: 1 put: (IdentityDictionary new: dictTracker size).
	dictTracker keysAndValuesDo: [ :key :entries |
		newEntries := entries select: [ :entry | (entry at: 1) <= depth ].
		newEntries ifNotEmpty: [ newDictTracker at: key put: newEntries ].
		].
	^newTrackers! !

!WDeepIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 7/2/2012 23:32'!
iChangedOrNewInWorld: world
	self shouldBeImplemented
! !

!WDeepIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 4/26/2013 13:06'!
iEnsureDictTrackerIn: updateTrackers
	| dictTracker |
	(updateTrackers at: 1) ifNotNil: [ :dict | ^dict ].

	"initialise the tracker dictionary with the values for all keys held in the top-level world"
	self evalInTopWorld: [
		dictTracker := IdentityDictionary new: self size * 4 // 3.
		self keysAndValuesDo: [ :k :v | dictTracker at: k put: (OrderedCollection with: { 0. v }) ]
		].
	^updateTrackers at: 1 put: dictTracker.
! !

!WDeepIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 7/2/2012 22:05'!
iUpdateTrackerSize
	^1
! !

!WDeepIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 7/2/2012 23:34'!
iUpdatesIn: updateTrackers
	self shouldBeImplemented
! !

!WDeepIdentityDictionary methodsFor: 'testing' stamp: 'ael 7/2/2012 23:33'!
includesKey: key 
	"Answer whether the receiver has a key equal to the argument, key."
	
	self at: key ifAbsent: [^false].
	^true! !

!WDeepIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 7/6/2012 16:39'!
isSerializableIn: parentWorld using: updateTrackers
	"for now we perform no serialisability check - which is fine for situations in which parent worlds are used as snapshots, and therefore remain unchanged after a child world is created" 
	^true

! !

!WDeepIdentityDictionary methodsFor: 'enumerating' stamp: 'ael 7/9/2012 15:11'!
keysDo: aBlock
	| thisWorld |
	(thisWorld := self thisWorld) isTopLevel ifTrue: [ ^contents keysDo: aBlock ].

	self privateEntriesInWorld: thisWorld do: [ :key :val | aBlock value: key ]
! !

!WDeepIdentityDictionary methodsFor: 'private' stamp: 'ael 7/9/2012 21:32'!
privateEntriesInWorld: world do: keyValBlock
	| deletion depth dictTracker trackers index val |
	trackers := self updateTrackersForWorld: world createIfAbsent: true caching: true.
	dictTracker := self iEnsureDictTrackerIn: trackers.
	depth := world delegationDepth.
	deletion := self deletionToken.
	dictTracker keysAndValuesDo: [ :key :entries |
		index := self privateIndexForDepth: depth in: entries.
		index = 0 ifFalse: [
			(val := (entries at: index) at: 2) == deletion ifFalse: [ keyValBlock value: key value: val ].
			]
		].
! !

!WDeepIdentityDictionary methodsFor: 'private' stamp: 'ael 7/9/2012 14:32'!
privateIndexForDepth: depth in: entries
	"the entries list is in descending order of depth (deepest first).  return the index of the first entry that is at or below (less deep than) the requested depth.  use a binary search.
	NB: any entry may have the deletionToken as its value
	NB: entries is assumed to be non-empty."
	| entry entryDepth first last probe |
	"the majority of accesses will be in leaf worlds, corresponding to the first entry"
	entry := entries first.
	(entry at: 1) <= depth ifTrue: [ ^1 ].
	
	(entries last at: 1) > depth ifTrue: [ ^0 ].		"no entry is going to match"
	
	first := 2.
	last := entries size. 
	[
		probe := first + last // 2.
		entry := entries at: probe.
		entryDepth := entry at: 1.
		(entryDepth = depth or: [ first = last ]) ifTrue: [ ^probe ].
		entryDepth < depth
			ifTrue: [ last := probe ]
			ifFalse: [ first := probe + 1 ].
	] repeat.
! !

!WDeepIdentityDictionary methodsFor: 'removing' stamp: 'ael 7/2/2012 23:46'!
removeKey: key ifAbsent: aBlock 
	"Remove key (and its associated value) from the receiver. If key is not in 
	the receiver, answer the result of evaluating aBlock. Otherwise, answer 
	the value externally named by key."

	self thisWorld isTopLevel ifTrue: [ ^contents removeKey: key ifAbsent: aBlock ].
	
	self at: key ifPresent: [ :old | self at: key put: self deletionToken. ^old ].
	^aBlock value
! !

!WDeepIdentityDictionary methodsFor: 'accessing' stamp: 'ael 7/2/2012 23:55'!
size
	| size |
	self thisWorld isTopLevel ifTrue: [ ^contents size ].

	size := 0.
	self keysDo: [ :k | size := size + 1].
	^size
! !

!WDeepIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 7/2/2012 23:50'!
syncWithTopWorld
	"nothing to do... I think"! !

!WDeepIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 7/2/2012 23:52'!
unloggedSize
	"in case someone tries putting WSimpleOrderedCollection under this class"
	^self size! !

!WDeepIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 7/9/2012 23:32'!
updateTrackersForWorld: aWorld createIfAbsent: create caching: useCache
	"in this class, we cache the trackers in the shared dictionary used by the specified world"
	| cache deepUpdateDict trackers |
	deepUpdateDict := aWorld deepUpdateDict.	"the dictionary we should use" 
	useCache ifTrue: [
		lookupCache ifNil: [ self initLookupCache ].		"we could force classes to use initialize, but..."
		cache := lookupCache.		"NB: not a redirected lookup"
		(cache at: 1) == deepUpdateDict ifTrue: [
			"double-check that the cache is non-nil (given that it's only held weakly)"
			(trackers := cache at: 2) ifNotNil: [ ^trackers ].
			].
		cache at: 1 put: nil.		"be sure to remove any previous entry"
		].
	
	trackers := deepUpdateDict at: self ifAbsent: [
		create
			ifTrue: [ deepUpdateDict at: self put: (Array new: self iUpdateTrackerSize withAll: nil) ]
		 	ifFalse: [ nil ]
		].
	trackers ifNil: [ ^nil ].

	useCache ifTrue: [
		cache at: 2 put: trackers.
		cache at: 1 put: aWorld.
		].
	^trackers
! !

!WDeepIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 7/2/2012 23:24'!
worldFlattenedAssociationsIn: aWorld
	^aWorld eval: [ self associations ]			"it's not such a big deal any more"
! !

!WIdentityDictionary class methodsFor: 'as yet unclassified' stamp: 'ael 9/24/2011 15:22'!
firstRedirectedIV
	^99				"we don't want contents IV to be redirected"! !

!WIdentityDictionary class methodsFor: 'sizing' stamp: 'ael 6/21/2011 23:23'!
goodPrimeAtLeast: lowerLimit
	"Answer the next good prime >= lowerlimit.
	If lowerLimit is larger than the largest known good prime,
	just make it odd."
	
	| primes low mid high prime |
	primes := self goodPrimes.
	low := 1.
	high := primes size.
	lowerLimit > (primes at: high) ifTrue: [
		^lowerLimit bitOr: 1 ].
	[ high - low <= 1 ] whileFalse: [
		mid := high + low // 2.
		prime := primes at: mid.
		prime = lowerLimit ifTrue: [ ^prime ].
		prime < lowerLimit
			ifTrue: [ low := mid ]
			ifFalse: [ high := mid ] ].
	(primes at: low) >= lowerLimit ifTrue: [ ^primes at: low ].
	^primes at: high! !

!WIdentityDictionary class methodsFor: 'sizing' stamp: 'ael 6/21/2011 23:24'!
goodPrimes
	"Answer a sorted array of prime numbers less than one billion that make good
	hash table sizes. Should be expanded as needed.  See comments below code"
	
	^#(
		5 11 17 23 31 43 59 79 107 149 199 269 359 479 641 857 1151 1549 2069
		2237 2423 2617 2797 2999 3167 3359 3539 3727 3911
		4441 4787 5119 5471 5801 6143 6521 6827 7177 7517 7853
		8783 9601 10243 10867 11549 12239 12919 13679 14293 15013 15731
		17569 19051 20443 21767 23159 24611 25847 27397 28571 30047 31397
		35771 38201 40841 43973 46633 48989 51631 54371 57349 60139 62969
		70589 76091 80347 85843 90697 95791 101051 106261 111143 115777 120691 126311
		140863 150523 160969 170557 181243 190717 201653 211891 221251 232591 242873 251443
		282089 300869 321949 341227 362353 383681 401411 422927 443231 464951 482033 504011
		562621 605779 647659 681607 723623 763307 808261 844709 886163 926623 967229 1014617
		1121987 1201469 1268789 1345651 1429531 1492177 1577839 1651547 1722601 1800377 1878623 1942141 2028401
		2242727 2399581 2559173 2686813 2836357 3005579 3144971 3283993 3460133 3582923 3757093 3903769 4061261
		4455361 4783837 5068529 5418079 5680243 6000023 6292981 6611497 6884641 7211599 7514189 7798313 8077189
		9031853 9612721 10226107 10745291 11338417 11939203 12567671 13212697 13816333 14337529 14938571 15595673 16147291
		17851577 18993941 20180239 21228533 22375079 23450491 24635579 25683871 26850101 27921689 29090911 30153841 31292507 32467307
		35817611 37983761 40234253 42457253 44750177 46957969 49175831 51442639 53726417 55954637 58126987 60365939 62666977 64826669
		71582779 76039231 80534381 84995153 89500331 93956777 98470819 102879613 107400389 111856841 116365721 120819287 125246581 129732203
		143163379 152076289 161031319 169981667 179000669 187913573 196826447 205826729 214748357 223713691 232679021 241591901 250504801 259470131
		285162679 301939921 318717121 335494331 352271573 369148753 385926017 402603193 419480419 436157621 453034849 469712051 486589307 503366497 520043707 
		570475349 603929813 637584271 671138659 704693081 738247541 771801929 805356457 838910803 872365267 905919671 939574117 973128521 1006682977 1040137411 
		1073741833)

"The above primes past 2069 were chosen carefully so that they do not interact badly with 1664525 (used by hashMultiply), and so that gcd(p, (256^k) +/- a) = 1, for 0<a<=32 and 0<k<=8.  See Knuth's TAOCP for details."

"The above primes also try to map the values of ((0 to: 4095) collect: [ :each | each << 18 \\ prime ]) sort to an equidistant sequence of numbers. This helps to avoid the collision of chains in identity-based hashed collections. To do that  they were chosen to return a low value when the following block is evaluated with them as argument:
 [ :prime |
	| slots cost optimalDistance previous |
	slots := Array new: 4097.
	0 to: 4095 do: [ :ea | slots at: ea + 1 put: ea *  262144 \\ prime ].
	slots at: 4097 put: prime.
	slots sort.
	cost := 0.
	optimalDistance := prime // 4096.
	2 to: 4097 do: [ :index |
		| newCost |
		newCost := optimalDistance - ((slots at: index) - (slots at: index - 1)).
		newCost > cost ifTrue: [ cost := newCost ] ].
	cost ]."! !

!WIdentityDictionary class methodsFor: 'instance creation' stamp: 'ael 9/25/2011 16:34'!
new
	^self new: 5
! !

!WIdentityDictionary class methodsFor: 'instance creation' stamp: 'ael 8/30/2012 12:24'!
new: nElements
	"Create a Set large enough to hold nElements without growing"
	| inst |
	WTopLevelWorld default eval: [
		inst := super basicNew.
		inst setContentsSize: (self sizeFor: nElements); initialize.
		].

	^inst
! !

!WIdentityDictionary class methodsFor: 'instance creation' stamp: 'ael 8/30/2012 12:25'!
new: nElements andDo: aBlock
	| inst |
	WTopLevelWorld default eval: [
		inst := super basicNew.
		inst setContentsSize: (self sizeFor: nElements); initialize.
		aBlock value: inst.
		].

	^inst
! !

!WIdentityDictionary class methodsFor: 'instance creation' stamp: 'ael 8/30/2012 12:27'!
newAndDo: aBlock
	^self new: 5 andDo: aBlock
! !

!WIdentityDictionary class methodsFor: 'instance creation' stamp: 'ael 7/26/2012 14:43'!
newFrom: aDict 
	"Answer an instance of me containing the same associations as aDict.
	 Error if any key appears twice."
	^self new: aDict size andDo: [ :inst |
		aDict associationsDo:
			[:x |
			(inst includesKey: x key)
				ifTrue: [self error: 'Duplicate key: ', x key printString]
				ifFalse: [inst add: x copy]
			].
		].


"	WIdentityDictionary newFrom: {1->#a. 2->#b. 3->#c}
	{1->#a. 2->#b. 3->#c} as: WIdentityDictionary
	WIdentityDictionary newFrom: {1->#a. 2->#b. 1->#c}
	{1->#a. 2->#b. 1->#c} as: WIdentityDictionary
"! !

!WIdentityDictionary class methodsFor: 'instance creation' stamp: 'ael 4/12/2012 01:57'!
newFromPairs: anArray 

	"Answer an instance of me associating (anArray at:i) to (anArray at: i+i)
	 for each odd i.  anArray must have an even number of entries."

	^self new: (anArray size/2) andDo: [ :inst |
		1 to: (anArray size-1) by: 2 do: [ :i|
			inst at: (anArray at: i) put: (anArray at: i+1).
			].
		]

	"  WIdentityDictionary newFromPairs: {'Red' . Color red . 'Blue' . Color blue . 'Green' . Color green}. "! !

!WIdentityDictionary class methodsFor: 'instance creation' stamp: 'ael 6/21/2011 23:00'!
ofSize: n
	"Create a new collection of size n with nil as its elements.
	This method exists because OrderedCollection new: n creates an
	empty collection,  not one of size n."
	^ self new: n! !

!WIdentityDictionary class methodsFor: 'instance creation' stamp: 'ael 11/11/2011 21:54'!
privateNewFromIDict: dict 
	"Answer an instance of me containing the same associations as dict."
	| inst |
	WTopLevelWorld default eval: [
		inst := super basicNew initialize.
		inst privateSetContents: dict copy.
		].

	^inst
! !

!WIdentityDictionary class methodsFor: 'as yet unclassified' stamp: 'ael 5/15/2013 00:22'!
simpleBenchmark
	"WIdentityDictionary simpleBenchmark"
	| l i val |
	MessageTally spyOn: 
	[
	WWorld thisWorld sprout eval: [
	l := WIdentityDictionary new. "IdentityDictionary new."
	1 to: 1000000 do: [:x |
		i := 1000 atRandom.
		val := l at: i ifAbsentPut: [0].
		val > 10
		  ifTrue: [ l removeKey: i ]
		  ifFalse: [ l at: i put: val + 1 ].
		]
	 ]
	]
! !

!WIdentityDictionary class methodsFor: 'siss-private' stamp: 'yo 7/1/2011 10:53'!
sissContents: inst into: sexp context: dictionary 
	"It is commonly used to write dictionary contents into siss."

	| keys sorted |
	keys := inst keys.
	sorted := [keys asSortedCollection]
				on: MessageNotUnderstood
				do: [:ex | keys].
	sexp
		addElements: (sorted asArray
				collect: [:k | 
					| elem slot | 
					elem := inst associationAt: k.
					slot := (Smalltalk at: #SExpElement) keyword: #slot.
					elem key -> elem value sissContentsInto: slot context: dictionary.
					slot])! !

!WIdentityDictionary class methodsFor: 'siss-private' stamp: 'ael 6/21/2011 22:15'!
sissCreateInstanceFromSexp: sexp idref: idref from: from to: to 
	^ self
		sissReadElementsFor: self
		from: sexp
		idref: idref
		from: from
		to: to! !

!WIdentityDictionary class methodsFor: 'siss-private' stamp: 'ael 6/21/2011 22:15'!
sissReadElementsFor: class from: sexp idref: idref from: from to: to 
	"It is commonly used for dictionary classes to read from siss."
	| inst |
	inst := class new: sexp elements size.
	idref ifNotNil: [to at: idref put: inst].
	sexp elements
		do: [:each | 
			| assoc | 
			assoc := Association
						sissCreateInstanceFromSexp: each
						idref: nil
						from: from
						to: to.
			inst at: assoc key put: assoc value].
	^ inst! !

!WIdentityDictionary class methodsFor: 'sizing' stamp: 'ael 6/21/2011 23:05'!
sizeFor: nElements
	"Large enough prime (or odd if too large) size to hold nElements with some slop (see fullCheck)"
	
	nElements < 4 ifTrue: [ ^5 ].
	^self goodPrimeAtLeast: nElements + 1 * 4 // 3! !

!WIdentityDictionary class methodsFor: 'testing' stamp: 'ael 7/5/2012 15:18'!
testDeepAccess1
	"QSystemProfiler spyOn: [ WIdentityDictionary testDeepAccess1 ]"
	"QSystemProfiler spyOn: [ WDeepIdentityDictionary testDeepAccess1 ]"
	| d dict dicts keys newW parent worlds numDicts numKeys numWorlds readTally writeTally |
	numDicts := 100.
	numWorlds := 200.
	numKeys := 100.
	dicts := OrderedCollection new.
	keys := ((1 to: numKeys) collect: [ :i | ('key', i asString) asSymbol ]) asArray.
	1 to: numDicts do: [ :di |
		d := self new.
		1 to: numKeys do: [ :i | d at: (keys at: i) put: i ].
		dicts add: d.
		].
	worlds := OrderedCollection new.
	parent := WTopLevelWorld default.
	1 to: numWorlds do: [ :wi | newW := parent sprout. worlds add: newW. parent := newW ].
	
	readTally := 0.
	writeTally := 0.
	1 to: numWorlds by: 3 do: [ :wi |
		(worlds at: wi) eval: [
			1 to: numDicts by: 5 do: [ :di |
				dict := dicts at: di.
				1 to: numKeys by: 3 do: [ :i | readTally := readTally + 1. ((dict at: (keys at: i)) rem: i) = 0 ifFalse: [ self halt ] ].
				1 to: numKeys by: 4 do: [ :i | writeTally := writeTally + 1. dict at: (keys at: i) put: i*wi ].
				].
			1 to: numDicts by: 4 do: [ :di |
				dict := dicts at: di.
				1 to: numKeys by: 3 do: [ :i | readTally := readTally + 1. ((dict at: (keys at: i)) rem: i) = 0 ifFalse: [ self halt ] ].
				1 to: numKeys by: 4 do: [ :i | writeTally := writeTally + 1. dict at: (keys at: i) put: i*wi*2 ].
				].
			].
		].
	1 to: numWorlds do: [ :wi |
		(worlds at: wi) eval: [
			1 to: numDicts do: [ :di |
				dict := dicts at: di.
				dict keysDo: [ :k | readTally := readTally + 1. k == #foo ].
				].
			]
		].

	Transcript cr; show: 'reads = ', readTally asString, '; writes = ', writeTally asString.
! !

!WIdentityDictionary class methodsFor: 'instance creation' stamp: 'ael 4/12/2012 01:55'!
with: anObject 
	"Answer an instance of me containing anObject."

	^ self new: 10 andDo: [ :inst | inst add: anObject ].
! !

!WIdentityDictionary class methodsFor: 'instance creation' stamp: 'ael 4/12/2012 01:55'!
with: firstObject with: secondObject 
	"Answer an instance of me containing the two arguments as elements."

	^ self new: 10 andDo: [ :inst | inst add: firstObject; add: secondObject ].
	! !

!WIdentityDictionary class methodsFor: 'instance creation' stamp: 'ael 4/12/2012 01:56'!
with: firstObject with: secondObject with: thirdObject 
	"Answer an instance of me containing the three arguments as elements."

	^ self new: 10 andDo: [ :inst | inst add: firstObject; add: secondObject; add: thirdObject ]	
	! !

!WIdentityDictionary class methodsFor: 'instance creation' stamp: 'ael 4/12/2012 01:56'!
with: firstObject with: secondObject with: thirdObject with: fourthObject 
	"Answer an instance of me, containing the four arguments as the elements."

	^ self new: 10 andDo: [ :inst | inst add: firstObject; add: secondObject; add: thirdObject; add: fourthObject ]
! !

!WIdentityDictionary class methodsFor: 'instance creation' stamp: 'ael 4/12/2012 01:56'!
with: firstObject with: secondObject with: thirdObject with: fourthObject with: fifthObject
	"Answer an instance of me, containing the five arguments as the elements."

	^ self new: 10 andDo: [ :inst | inst add: firstObject; add: secondObject; add: thirdObject; add: fourthObject; add: fifthObject ]
! !

!WIdentityDictionary class methodsFor: 'instance creation' stamp: 'ael 4/12/2012 01:56'!
with: firstObject with: secondObject with: thirdObject with: fourthObject with: fifthObject with: sixthObject
	"Answer an instance of me, containing the six arguments as the elements."

	^ self new: 10 andDo: [ :inst | inst add: firstObject; add: secondObject; add: thirdObject; add: fourthObject; add: fifthObject; add: sixthObject ]
! !

!WIdentityDictionary class methodsFor: 'instance creation' stamp: 'ael 4/12/2012 01:56'!
withAll: aCollection
	"Create a new collection containing all the elements from aCollection."

	^ self new: aCollection size andDo: [ :inst | inst addAll: aCollection ]
! !

!WIdentityDictionary methodsFor: 'copying' stamp: 'ael 6/21/2011 23:24'!
, aCollection
	^self copy addAll: aCollection; yourself! !

!WIdentityDictionary methodsFor: 'comparing' stamp: 'ael 9/25/2011 21:25'!
= aDictionary
	"Two dictionaries are equal if
	 (a) they are the same 'kind' of thing.
	 (b) they have the same set of keys.
	 (c) for each (common) key, they have the same value"

	self == aDictionary ifTrue: [ ^ true ].
	(aDictionary isKindOf: WIdentityDictionary) ifFalse: [^false].
	self size = aDictionary size ifFalse: [^false].
	self associationsDo: [:assoc|
		(aDictionary at: assoc key ifAbsent: [^false]) = assoc value
			ifFalse: [^false]].
	^true

! !

!WIdentityDictionary methodsFor: 'adding' stamp: 'ael 9/25/2011 19:11'!
add: anAssociation
	"whereas the standard ID code has to scan the whole collection to see if we already had an association with the same key, in multiple worlds we just need to store in the contents or child-world Writes"
	| thisWorld trackers |
	(thisWorld := self thisWorld) isTopLevel ifTrue: [ ^contents add: anAssociation ].
	
	trackers := self updateTrackersForWorld: thisWorld createIfAbsent: true caching: true.
	(self iEnsureWritesRecordIn: trackers) add: anAssociation.
	^anAssociation
! !

!WIdentityDictionary methodsFor: 'adding' stamp: 'ael 6/21/2011 22:18'!
add: newObject withOccurrences: anInteger
	"Add newObject anInteger times to the receiver. Do nothing if anInteger is less than one. Answer newObject."
	
	anInteger < 1 ifTrue: [ ^newObject ].
	^self add: newObject "I can only store an object once."
	! !

!WIdentityDictionary methodsFor: 'adding' stamp: 'ael 6/21/2011 22:15'!
addAll: aCollection
	"Include all the associations of aCollection as the receiver's elements. Answer 
	aCollection. Any collection containing associations can be used as argument."

	aCollection == self ifFalse: [
		aCollection associationsDo: [:assoc| self add: assoc].
	].
	^aCollection! !

!WIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 9/25/2011 16:10'!
adoptReadFromChild: assoc
	"see copyMissingOrigValuesInto:using:"

	"do nothing if this world doesn't track reads"
	| thisWorld trackers writes |
	(thisWorld := self thisWorld) tracksReads ifFalse: [ ^nil ].
	
	trackers := self updateTrackersForWorld: thisWorld createIfAbsent: true caching: true.

	writes := self iEnsureWritesRecordIn: trackers.
	(writes includesKey: assoc key)
	  ifFalse: [ self recordReadOf: assoc in: (self iEnsureReadsRecordIn: trackers) ].

! !

!WIdentityDictionary methodsFor: 'enumerating' stamp: 'ael 12/2/2011 14:04'!
allRecordsDo: aBlock includingDeletions: includeDeletes loggingReads: logToReads
	"Evaluate aBlock (single arg) for each of the receiver's elements (key/value 
	associations).
	In WID, unlike Dictionary, we don't work internally with associations.  But building them by first obtaining keys then doing lookups is inefficient (potentially involving a search through the parent-world chain for each key).  So we gather associations from the leaf and all ancestor worlds.
	NB: the block that we're passed will be evaluated in the world in which the association is found."
	| foundSoFar homeReads val doLogging dnk |

	foundSoFar := IdentityDictionary new: (self roughSize * 4 // 3 + 1).	"try to avoid growing"
	doLogging := logToReads and: [ (homeReads := self readDictIfUsable) notNil ].	"log only if we can"
	dnk := self dnkToken.
	
	self allRecordsRelevantGiven: foundSoFar inHomeWorld: true do: [ :assoc :isHomeWorld |
		(doLogging and: [ isHomeWorld not ])
		  ifTrue: [ self recordReadOf: assoc in: homeReads ].
		val := assoc value.
		(val ~~ dnk and: [ val isWorldDeletion not or: [ includeDeletes ] ])
		  ifTrue: [ aBlock value: assoc ].
		].
	
	"if we got all the way through to the end, make a note that all keys and values have been seen"
	doLogging ifTrue: [ self recordAllParentValuesSeen ].
! !

!WIdentityDictionary methodsFor: 'private' stamp: 'ael 11/13/2011 23:29'!
allRecordsRelevantGiven: foundSoFar inHomeWorld: isHomeWorld do: aBlock
	"This is the main place for iterating over associations (as opposed to looking up a single key, which is handled in lookForAssociationAt:careAboutValue:inHomeWorld:ifPresentDo:).  It's a recursive method, called once per world up the ancestor chain, that reports to the caller (through the supplied block) each association it finds."
	| dnk key processIfRelevant seenAll deletion reads seenAllKeysAlready existing sentinel val thisWorld trackers writes |
	dnk := self dnkToken.
	deletion := self deletionToken.
	sentinel := Object new.
	
	seenAllKeysAlready := foundSoFar includesKey: #xxxSeenAll.
	
	processIfRelevant := [ :a |
		key := a key.
		val := a value.
		"an association found here is relevant if
		 * it's a previously unseen key, and we are open to new ones: foundSoFar has neither a value for this key nor for #seenAll, or
		 * it's a key that has been seen earlier, but only with a don't-know value that we can now fix: foundSoFar has DNK as the value for this key, and valueHere is not DNK
			- but in this case if valueHere is Delete, there is an inconsistency.  it can be cleaned up if whichever world had the DNK record now deletes it - but we can't do that from here.  so put the deletion record into foundSoFar (thus suppressing any higher parents from supplying values), but don't evaluate the block.

		rearranging the logic:
			- if we have seen all keys already, and the value here is dnk, it tells us nothing.  ignore it.
			- otherwise look for an existing value with this key
				- if none is found, and we have not seen all keys already, record and report this value (whatever it is - even dnk)
				- if a dnk value was found, and the value here is non-dnk, record it.  but if it's a deletion, don't report.
				- otherwise do nothing
		"
		((seenAllKeysAlready and: [ val == dnk ]) not
		  and: [
			existing := foundSoFar at: key ifAbsent: [ sentinel ].
		
			((existing == sentinel and: [ seenAllKeysAlready not ])
			  or: [ existing == dnk and: [ val ~~ dnk ] ])
			])
		  ifTrue: [
			foundSoFar at: key put: val.					"record"
			(existing == dnk and: [ val == deletion ])
			  ifFalse: [ aBlock value: a copy value: isHomeWorld ].		"report, protecting the source association against being mutated"
			].
		].

	reads := seenAll := writes := nil.
	thisWorld := self thisWorld.
	thisWorld isTopLevel
	  ifTrue: [ writes := contents ]
	  ifFalse: [
		"iteration shouldn't force creation of trackers, since the collection might be empty"
		trackers := self updateTrackersForWorld: thisWorld createIfAbsent: false caching: isHomeWorld.

		trackers ifNotNil: [		"a tuple { reads. seenAllFromParent. writes. }"
			thisWorld tracksReads
			  ifTrue: [
				reads := trackers at: 1.			"may be nil"
				seenAll := trackers at: 2.
				].
			writes := trackers at: 3.				"may be nil"
			].
		].
				
	writes ifNotNil: [ writes associationsDo: [ :assoc | processIfRelevant value: assoc ] ].

	reads
	  ifNotNil: [
		"use a copy in case a sender is somehow updating Reads in this world"
		reads copy associationsDo: [ :assoc | processIfRelevant value: assoc ].

		seenAll
		  ifNotNil: [		"seen at least all the keys, and perhaps all the values too"
			(seenAll == #values					"seen all values too"
				or: [ (reads includes: dnk) not ])		"seen keys, and all (now) have non-DNK values"
 			  ifTrue: [ ^self ].	"nothing our ancestors can tell us"

			foundSoFar at: #xxxSeenAll put: seenAll.	"inform the ancestors"
			].
		].	
	
	"now recurse to get the parent's input"
	self evalInParentWorld: [ self allRecordsRelevantGiven: foundSoFar inHomeWorld: false do: aBlock ].
! !

!WIdentityDictionary methodsFor: 'enumerating' stamp: 'ael 6/21/2011 22:24'!
allSatisfy: aBlock
	"Evaluate aBlock with the elements of the receiver.
	If aBlock returns false for any element return false.
	Otherwise return true."

	self do: [:each | (aBlock value: each) ifFalse: [^ false]].
	^ true! !

!WIdentityDictionary methodsFor: 'accessing' stamp: 'ael 6/21/2011 22:24'!
anyOne
	"Answer a representative sample of the receiver. This method can
	be helpful when needing to preinfer the nature of the contents of 
	semi-homogeneous collections."

	self emptyCheck.
	self do: [:each | ^ each]! !

!WIdentityDictionary methodsFor: 'enumerating' stamp: 'ael 6/21/2011 22:24'!
anySatisfy: aBlock
	"Evaluate aBlock with the elements of the receiver.
	If aBlock returns true for any element return true.
	Otherwise return false."

	self do: [:each | (aBlock value: each) ifTrue: [^ true]].
	^ false! !

!WIdentityDictionary methodsFor: 'converting' stamp: 'ael 11/11/2011 21:38'!
asArray
	"Answer an Array whose elements are the elements of the receiver.
	Implementation note: Cannot use ''Array withAll: self'' as that only
	works for SequenceableCollections which support the replacement 
	primitive."

	| newArray index associations |
	associations := self associations.
	newArray := Array new: associations size.
	index := 0.
	associations do: [:assoc | newArray at: (index := index + 1) put: assoc value].
	^ newArray! !

!WIdentityDictionary methodsFor: 'converting' stamp: 'ael 11/11/2011 21:38'!
asBag
	"Answer a Bag whose elements are the elements of the receiver."

	^ Bag withAll: self asArray! !

!WIdentityDictionary methodsFor: 'converting' stamp: 'ael 12/2/2011 15:26'!
asIdentitySet
	^(WIdentitySet new: self size) addAll: self asArray; yourself! !

!WIdentityDictionary methodsFor: 'converting' stamp: 'ael 9/25/2011 21:24'!
asOrderedCollection
	"Answer an OrderedCollection whose elements are the elements of the
	receiver. The order in which elements are added depends on the order
	in which the receiver enumerates its elements. In the case of unordered
	collections, the ordering is not necessarily the same for multiple 
	requests for the conversion."
	"World-aware behaviour: given that we're probably operating in a multi-world
	environment, the appropriate kind of orderedCollection is a WCollection.
	A user who explicitly wants an OrderedCollection should send
	    dict as: OrderedCollection
	"
	^ self as: WCollection! !

!WIdentityDictionary methodsFor: 'converting' stamp: 'ael 11/11/2011 21:39'!
asSet
	"Answer a Set whose elements are the unique elements of the receiver."

	^ Set withAll: self asArray! !

!WIdentityDictionary methodsFor: 'converting' stamp: 'ael 6/21/2011 22:29'!
asSortedCollection
	"Answer a SortedCollection whose elements are the elements of the 
	receiver. The sort order is the default less than or equal. 
	Use #sorted: if you don't really need a SortedCollection, but a sorted collection!!"
	
	^self as: SortedCollection! !

!WIdentityDictionary methodsFor: 'converting' stamp: 'ael 11/11/2011 21:39'!
asSortedCollection: aSortBlock 
	"Answer a SortedCollection whose elements are the elements of the 
	receiver. The sort order is defined by the argument, aSortBlock.
	Use #sorted: if you don't really need a SortedCollection, but a sorted collection!!"

	| aSortedCollection array |
	array := self asArray.
	aSortedCollection := SortedCollection new: array size.
	aSortedCollection sortBlock: aSortBlock.
	aSortedCollection addAll: array.
	^ aSortedCollection! !

!WIdentityDictionary methodsFor: 'accessing' stamp: 'ael 6/21/2011 22:15'!
associationAt: key 
	^ self associationAt: key ifAbsent: [self errorKeyNotFound: key]! !

!WIdentityDictionary methodsFor: 'accessing' stamp: 'ael 9/25/2011 15:13'!
associationAt: key ifAbsent: aBlock 
	"Answer an association with the given key.
	If key is not found, return the result of evaluating aBlock.
	
	By default we log (in the leaf world) the reading of any association that is found, or a deletion record if this key turned up nothing"
	self thisWorld isTopLevel ifTrue: [ ^contents associationAt: key ifAbsent: aBlock ].

	^self associationAt: key ifAbsent: aBlock loggingReads: true
! !

!WIdentityDictionary methodsFor: 'accessing' stamp: 'ael 12/2/2011 13:49'!
associationAt: key ifAbsent: aBlock loggingReads: logToReads
	"Answer an association with the given key.
	If key is not found, return the result of evaluating aBlock."
	"world-aware handling: do a lookup that delegates through the parent-world chain until some kind of association is found (or we reach the top)"
	| homeReads val deletion dnk doLogging homeWorld |
	
	deletion := self deletionToken.
	dnk := self dnkToken.
	homeWorld := self thisWorld.
	
	doLogging := logToReads and: [ (homeReads := self readDictIfUsable) notNil ].	"log only if we can"
	
	self lookForAssociationAt: key careAboutValue: true inHomeWorld: true ifPresentDo: [ :assoc :isHomeWorld |
		"something was found - though it may be a deletion record"
		(doLogging and: [ isHomeWorld not ])
		  ifTrue: [ self recordReadOf: assoc in: homeReads ].

		(val := assoc value) == deletion ifTrue: [ ^homeWorld eval: aBlock ].

		"An association with a value of DNK has a special meaning: that some world (maybe our home) has a DNK against this key, but in one of its ancestors the key has now been deleted.  In such a case we act as if the association was not found.  In addition, if the DNK read record was here it will be replaced with a deletion record (if logToReads is true), clearing up the inconsistency."
		val ~~ dnk ifTrue: [ ^assoc ].			"it was a real value"
		].

	"The lookup found nothing, or perhaps an inconsistent DNK record.  Record a deletion token in our Reads so we don't try again"
	doLogging ifTrue: [ self recordReadOf: key->deletion in: homeReads ].
	^aBlock value		"nothing found"! !

!WIdentityDictionary methodsFor: 'accessing' stamp: 'ael 11/12/2011 02:20'!
associations
	^self associationsSnapshotTrackingReads: true

! !

!WIdentityDictionary methodsFor: 'enumerating' stamp: 'ael 12/2/2011 14:05'!
associationsDo: aBlock
	| homeWorld |
	(homeWorld := self thisWorld) isTopLevel ifTrue: [ ^contents associationsDo: aBlock ].

	self allRecordsDo: [ :assoc | homeWorld eval: [ aBlock value: assoc ] ] includingDeletions: false loggingReads: true
! !

!WIdentityDictionary methodsFor: 'enumerating' stamp: 'ael 6/22/2011 21:07'!
associationsSelect: aBlock 
	"Evaluate aBlock with each of my associations as the argument. Collect
	into a new dictionary, only those associations for which aBlock evaluates
	to true."

	| newCollection |
	newCollection := self copyEmpty.
	self associationsDo: [ :each | 
		(aBlock value: each) ifTrue: [ newCollection add: each ] ].
	^newCollection! !

!WIdentityDictionary methodsFor: 'accessing' stamp: 'ael 11/13/2011 13:24'!
associationsSnapshotTrackingReads: trackReads
	"Gather the associations that are visible in the running world.  Log reads if this world wants to.
	This was written as the model for the flattening primitive.  In theory it should be slightly faster than allRecordsDo:..., but there's not much in it."
	| foundSoFar homeReads val dnk existing homeWorld key reads seenAllKeysAlready sentinel trackers w writes seenAllHere nextW deletion seenDNKHere recordReadsHere inHomeWorld v seenDNKs seenDeletions |
	homeWorld := WWorld thisWorld.
	homeReads := trackReads ifFalse: [ nil ] ifTrue: [ self readDictIfUsable ].
	dnk := self dnkToken.
	deletion := self deletionToken.
	sentinel := Object new.

	foundSoFar := IdentityDictionary new: (self roughSize * 4 // 3 + 1).	"try to avoid growing"
	seenDNKs := seenDeletions := false.		"can save time at the end"
	seenAllKeysAlready := false.
	
	w := homeWorld.
	[ w isNil ]
	  whileFalse: [
		nextW := w parent.
		reads := writes := seenAllHere := nil.
		seenDNKHere := false.
		inHomeWorld := w == homeWorld.
		recordReadsHere := homeReads notNil and: [ inHomeWorld not ].
		w isTopLevel
		  ifTrue: [ writes := contents ]
		  ifFalse: [
			"iteration shouldn't force creation of trackers, since the collection might be empty"
			trackers := self updateTrackersForWorld: w createIfAbsent: false caching: w == homeWorld.
			trackers ifNotNil: [		"a tuple { reads. seenAllFromParent. writes. }"
				reads := trackers at: 1.			"may be nil (e.g., if tracksReads is false)"
				seenAllHere := trackers at: 2.	"ditto"
				writes := trackers at: 3.			"nil if nothing written yet"
				].
			].

		(writes ifNil: [ #() ] ifNotNil: [ writes associations ]), (reads ifNil: [ #() ] ifNotNil: [ reads associations ])
		  do: [ :a |
			"see reasoning in allRecordsRelevantGiven:inHomeWorld:do:"
			key := a key.
			val := a value.
			val == dnk ifTrue: [ seenDNKHere := true ].		"(must have been in Reads)"
			((seenAllKeysAlready and: [ val == dnk ]) not
			  and: [
				existing := foundSoFar at: key ifAbsent: [ sentinel ].
		
				((existing == sentinel and: [ seenAllKeysAlready not ])
				  or: [ existing == dnk and: [ val ~~ dnk ] ])
				])
			  ifTrue: [
				foundSoFar at: key put: val.					"record"
				val == deletion ifTrue: [ seenDeletions := true ].	"note for later"
				(recordReadsHere and: [ (existing == dnk and: [ val == deletion ]) not ])
				  ifTrue: [ self recordReadOf: a copy in: homeReads ].	"protect the source association against being mutated"
				].
			].

		seenDNKHere ifTrue: [ seenDNKs := true ].		"seen at least one"

		(reads notNil and: [ seenAllHere notNil ])
		  ifTrue: [	"seen at least all the keys, and perhaps all the values too"
			(seenAllHere == #values					"had seen all values already"
				or: [ seenDNKHere not and: [ inHomeWorld ifTrue: [ trackers at: 2 put: #values ]. true ] ]) "was only recorded as the keys, but since there are no DNKs we must have seen all the values too" 
			  ifTrue: [ nextW := nil ].	"nothing our ancestors can tell us.  bail out of the loop"
			seenAllKeysAlready := true.		"otherwise we at least know this"
			].

		w := nextW.
		].
	
	homeReads ifNotNil: [ self recordAllParentValuesSeen ].

	^seenDNKs | seenDeletions
	  ifFalse: [ foundSoFar array reject: [ :assoc | assoc isNil ] ]
	  ifTrue: [ foundSoFar array reject: [ :assoc | assoc isNil or: [ (v := assoc value) == dnk or: [ v == deletion ] ] ] ].
! !

!WIdentityDictionary methodsFor: 'accessing' stamp: 'ael 6/21/2011 22:15'!
at: key 
	"Answer the value associated with the key."

	^ self at: key ifAbsent: [self errorKeyNotFound: key]! !

!WIdentityDictionary methodsFor: 'accessing' stamp: 'ael 9/25/2011 15:54'!
at: key ifAbsent: aBlock 
	"Answer the value associated with the key or, if key isn't found,
	answer the result of evaluating aBlock."

	^(self associationAt: key ifAbsent: [ aBlock ]) value		"NB: the brackets are needed because we want to send #value either to a returned association or to an as-yet-unevaluated block"
! !

!WIdentityDictionary methodsFor: 'accessing' stamp: 'ael 6/21/2011 22:15'!
at: key ifAbsentPut: aBlock 
	"Return the value at the given key. 
	If key is not included in the receiver store the result 
	of evaluating aBlock as new value."

	^ self at: key ifAbsent: [self at: key put: aBlock value]! !

!WIdentityDictionary methodsFor: 'accessing' stamp: 'ael 6/21/2011 22:15'!
at: key ifPresent: aBlock
	"Lookup the given key in the receiver. If it is present, answer the value of evaluating the given block with the value associated with the key. Otherwise, answer nil."

	| v |
	v := self at: key ifAbsent: [^ nil].
	^ aBlock value: v
! !

!WIdentityDictionary methodsFor: 'accessing' stamp: 'ael 6/21/2011 22:15'!
at: key ifPresent: oneArgBlock ifAbsent: absentBlock
	"Lookup the given key in the receiver. If it is present, answer the value of evaluating the oneArgBlock with the value associated with the key, otherwise answer the value of absentBlock."
	self at: key ifPresent:[:v| ^oneArgBlock value: v].
	^absentBlock value! !

!WIdentityDictionary methodsFor: 'accessing' stamp: 'ael 9/25/2011 15:56'!
at: key put: anObject 
	"Set the value at key to be anObject.  If key is not found, create a
	new entry for key and set is value to anObject. Answer anObject."
	| thisWorld trackers writes |
	(thisWorld := self thisWorld) isTopLevel ifTrue: [ ^contents at: key put: anObject ].
	
	trackers := self updateTrackersForWorld: thisWorld createIfAbsent: true caching: true.
	writes := self iEnsureWritesRecordIn: trackers.
	^writes at: key put: anObject
! !

!WIdentityDictionary methodsFor: 'accessing' stamp: 'ael 12/2/2011 15:16'!
atRandom: aGenerator
	"Answer a random element of the receiver. Uses aGenerator which
	should be kept by the user in a variable and used every time. Use
	this instead of #atRandom for better uniformity of random numbers because 
	only you use the generator. Causes an error if self has no elements."

	| randomIndex index |
	self emptyCheck.
	randomIndex := aGenerator nextInt: self size.
	index := 1.
	self do: [ :each |
		index = randomIndex ifTrue: [ ^each ].
		index := index + 1 ]! !

!WIdentityDictionary methodsFor: 'enumerating' stamp: 'ael 6/22/2011 17:48'!
collect: aBlock 
	"Evaluate aBlock with each of my values as the argument.  Collect the resulting values into a collection that is like me. Answer with the new collection."
	
	| newCollection |
	newCollection := self species new: self size.
	self associationsDo: [ :each |
		newCollection at: each key put: (aBlock value: each value) ].
	^newCollection! !

!WIdentityDictionary methodsFor: 'enumerating' stamp: 'ael 6/21/2011 22:34'!
collect: collectBlock thenSelect: selectBlock
	"Utility method to improve readability."

	^ (self collect: collectBlock) select: selectBlock! !

!WIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 4/10/2012 14:07'!
commitInto: parentWorld using: updateTrackers
	"- for each entry in Writes
		- if the value is Deleted, tell the parent to remove the key (if present)
		- for any other value, tell the parent to write the value to its Writes
	"
	| deletion assocs writes key val |
	"updateTrackers is a tuple { reads. seenAllFromParent. writes. }"
	writes := updateTrackers at: 3.
	writes isEmptyOrNil ifTrue: [ ^self ].
	
	deletion := self deletionToken.
	assocs := writes associations.
	parentWorld eval: [
		assocs do: [ :assoc |
			(self isCommittableKey: (key := assoc key))
			  ifTrue: [
				(val := assoc value) == deletion
				  ifTrue: [ self privateRemoveKey: key ]
				  ifFalse: [ self privateAt: key put: val ].
				]
			].
		].
! !

!WIdentityDictionary methodsFor: 'filter streaming' stamp: 'ael 6/21/2011 22:35'!
contents
	^ self! !

!WIdentityDictionary methodsFor: 'copying' stamp: 'ael 6/21/2011 22:39'!
copyEmpty
	"Answer an empty copy of this collection"
	
	^self species new! !

!WIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 12/5/2011 23:04'!
copyMissingOrigValuesInto: parentWorld using: updateTrackers
	"as part of commit, copy into the parent world's reads record any values that were read by the child-level object but came from an ancestor beyond the parent (i.e., the parent has recorded neither a write nor a read for this slot).

	This message is sent in the child world.
	
	For WID this isn't needed for the basic instance variables, because we assign and manage them ourselves.  But the Reads record shows values read from the parent world; if a given value doesn't come directly from the parent, it must have come from a higher ancestor - so the parent now needs to adopt its own Reads record for that value.
	
	- for each entry in Reads
		- if the parent itself has no value for this key (no entry in Writes), make sure it adds the value to its Reads
 	"
	| reads assocs parentReads parentTrackers parentWrites |
	parentWorld tracksReads ifFalse: [ ^self ].		"parent doesn't want to know"

	reads := self readDictIfUsable.
	reads isEmptyOrNil ifTrue: [ ^self ].				"nothing to say"

	assocs := reads associations.
	parentTrackers := self updateTrackersForWorld: parentWorld createIfAbsent: true caching: true.
	parentWrites := self writesRecordIn: parentTrackers.			"may be nil"
	parentReads := nil.		"might not be needed"
	
	assocs do: [ :assoc |
		(parentWrites notNil and: [ parentWrites includesKey: assoc key ])
		  ifFalse: [ self recordReadOf: assoc in: (parentReads ifNil: [ parentReads := self iEnsureReadsRecordIn: parentTrackers ]) ].
		].
! !

!WIdentityDictionary methodsFor: 'copying' stamp: 'ael 6/21/2011 22:40'!
copyWith: newElement
	"Answer a new collection with newElement added (as last
	element if sequenceable)."

	^ self copy
		add: newElement;
		yourself! !

!WIdentityDictionary methodsFor: 'copying' stamp: 'ael 6/21/2011 22:40'!
copyWithout: oldElement 
	"Answer a copy of the receiver that does not contain any
	elements equal to oldElement."

	^ self reject: [:each | each = oldElement]

"Examples:
	'fred the bear' copyWithout: $e
	#(2 3 4 5 5 6) copyWithout: 5
"! !

!WIdentityDictionary methodsFor: 'copying' stamp: 'ael 6/21/2011 22:40'!
copyWithoutAll: aCollection
	"Answer a copy of the receiver that does not contain any elements 
	equal to those in aCollection."

	^ self reject: [:each | aCollection includes: each]! !

!WIdentityDictionary methodsFor: 'enumerating' stamp: 'ael 6/21/2011 22:40'!
count: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument.  
	Answer the number of elements that answered true."

	| sum |
	sum := 0.
	self do: [:each | (aBlock value: each) ifTrue: [sum := sum + 1]].
	^ sum! !

!WIdentityDictionary methodsFor: 'debugging' stamp: 'ael 11/1/2011 10:19'!
debugUpdateTrackerState
	| trackers |
	trackers := self updateTrackersForWorld: self thisWorld createIfAbsent: false caching: false.
	^trackers ifNotNil: [ trackers collect: [ :elem | elem clone ] ]
! !

!WIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 10/31/2011 22:28'!
deeplyWorldFlattenedIn: aWorld with: doneSoFar
	| flatMe |
	doneSoFar at: self ifPresent: [ :flat | ^flat ].
	
	WTopLevelWorld default eval: [
		doneSoFar at: self put: (flatMe := self worldFlattenIn: aWorld).
	
		flatMe keysAndValuesDo: [ :key :value | flatMe at: key put: (value deeplyWorldFlattenedIn: aWorld with: doneSoFar) ].
		].
	
	^flatMe
! !

!WIdentityDictionary methodsFor: 'enumerating' stamp: 'ael 6/21/2011 22:41'!
detect: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument. 
	Answer the first element for which aBlock evaluates to true."

	^ self detect: aBlock ifNone: [self errorNotFound: aBlock]! !

!WIdentityDictionary methodsFor: 'enumerating' stamp: 'ael 6/21/2011 22:41'!
detect: aBlock ifNone: exceptionBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument.  
	Answer the first element for which aBlock evaluates to true. If none  
	evaluate to true, then evaluate the argument, exceptionBlock."

	self do: [:each | (aBlock value: each) ifTrue: [^ each]].
	^ exceptionBlock value! !

!WIdentityDictionary methodsFor: 'enumerating' stamp: 'ael 6/21/2011 22:15'!
do: aBlock

	self valuesDo: aBlock! !

!WIdentityDictionary methodsFor: 'enumerating' stamp: 'ael 6/21/2011 22:41'!
do: elementBlock separatedBy: separatorBlock
	"Evaluate the elementBlock for all elements in the receiver,
	and evaluate the separatorBlock between."

	| beforeFirst | 
	beforeFirst := true.
	self do:
		[:each |
		beforeFirst
			ifTrue: [beforeFirst := false]
			ifFalse: [separatorBlock value].
		elementBlock value: each]! !

!WIdentityDictionary methodsFor: 'private' stamp: 'ael 6/21/2011 22:42'!
emptyCheck

	self isEmpty ifTrue: [self errorEmptyCollection]! !

!WIdentityDictionary methodsFor: 'private' stamp: 'ael 6/21/2011 22:42'!
errorEmptyCollection

	self error: 'this collection is empty'! !

!WIdentityDictionary methodsFor: 'private' stamp: 'ael 6/21/2011 22:42'!
errorKeyNotFound: key
	"Signal KeyNotFound error"
	^(KeyNotFound key: key) signal! !

!WIdentityDictionary methodsFor: 'private' stamp: 'ael 6/21/2011 22:42'!
errorNoFreeSpace

	self error: 'There is no free space in this collection!!'! !

!WIdentityDictionary methodsFor: 'private' stamp: 'ael 6/21/2011 22:43'!
errorNotFound: anObject
	"Actually, this should raise a special Exception not just an error."

	self error: 'Object is not in the collection.'! !

!WIdentityDictionary methodsFor: 'private' stamp: 'ael 6/21/2011 22:15'!
errorValueNotFound

	self error: 'value not found'! !

!WIdentityDictionary methodsFor: 'Morphic-Explorer' stamp: 'ael 4/26/2012 18:03'!
explorerContents

	^self keysSortedSafely replace: [ :key |
		ObjectExplorerWrapper
			with: (self at: key)
			name: (key isSymbol ifTrue: [ key ] ifFalse: [ key printString contractTo: 32 ])
			model: self ]
! !

!WIdentityDictionary methodsFor: 'private' stamp: 'ael 6/21/2011 22:15'!
fillFrom: aCollection with: aBlock
	"Evaluate aBlock with each of aCollections's elements as the argument.  
	Collect the resulting values into self. Answer self."

	aCollection keysAndValuesDo: [ :key :value |
		self at: key put: (aBlock value: value) ]! !

!WIdentityDictionary methodsFor: 'filter streaming' stamp: 'ael 6/21/2011 22:15'!
flattenOnStream:aStream
	^aStream writeDictionary:self.
! !

!WIdentityDictionary methodsFor: 'Morphic-Explorer' stamp: 'ael 6/21/2011 22:15'!
hasContentsInExplorer

	^self isEmpty not! !

!WIdentityDictionary methodsFor: 'comparing' stamp: 'ael 6/24/2011 11:52'!
hash
	"Answer an integer hash value for the receiver such that,
	  -- the hash value of an unchanged object is constant over time, and
	  -- two equal objects have equal hash values"

	| hash size |

	size := self size.
	hash := self species hash.
	size <= 10 ifTrue:
		[self do: [:elem | hash := hash bitXor: elem hash]].
	^hash bitXor: size hash! !

!WIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 9/25/2011 16:22'!
iChangedOrNewInWorld: world
	| trackers |
	trackers := (world updateTrackersFor: self) ifNil: [ ^false ].
	
	"updateTrackers for an WIdentityDict is a tuple { reads. seenAllFromParent. writes. }"
	^(trackers at: 3) isEmptyOrNil not
! !

!WIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 11/13/2011 23:26'!
iEnsureReadsRecordIn: updateTrackers
	"updateTrackers for a WIdentityDict is a tuple { reads. seenAllFromParent. writes. }
	NB: this method assumes that the world tracks reads (so caller must confirm that beforehand)."
	| reads |
	(reads := updateTrackers at: 1) ifNotNil: [ ^reads ].

	^updateTrackers at: 1 put: (IdentityDictionary new: (self roughSize // 10 max: 10)).	"if it's a big dictionary we should start off in the right ballpark, to avoid excessive use of grow:"
! !

!WIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 11/1/2011 19:30'!
iEnsureWritesRecordIn: updateTrackers
	"updateTrackers for a WIdentityDict is a tuple { reads. seenAllFromParent. writes. }"
	| writes |
	(writes := updateTrackers at: 3) ifNotNil: [ ^writes ].

	^updateTrackers at: 3 put: IdentityDictionary new
! !

!WIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 1/16/2012 14:38'!
iMergeUpdatesFrom: otherTrackers into: updateTrackers
	"updateTrackers is a tuple { reads. seenAllFromParent. writes. }
	- like commit, for each entry in Writes
		- if the value is Deleted, record a deletion (regardless of whether we currently have that key)
		- for any other value, write the value to our Writes

	NB: this	is executed in the receiving world"
		
	| deletion writes |
	writes := otherTrackers at: 3.
	
	deletion := self deletionToken.
	writes associations
	  do: [ :assoc |
		assoc value == deletion
		  ifTrue: [ self privateRemoveKey: assoc key ]
		  ifFalse: [ self privateAt: assoc key put: assoc value ].
		].
! !

!WIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 9/25/2011 16:25'!
iUpdateTrackerSize
	"updateTrackers for an WIdentityDict is a tuple { reads. seenAllFromParent. writes. }"
	^3
! !

!WIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 1/13/2012 20:15'!
iUpdatesIn: updateTrackers
	"updateTrackers for a WIdentityDict is a tuple { reads. seenAllFromParent. writes. }"
	| writes newTrackers |
	(writes := updateTrackers at: 3) ifNil: [ ^nil ].

	newTrackers := Array new: self iUpdateTrackerSize withAll: nil.
	newTrackers at: 3 put: writes copy.
	^newTrackers
! !

!WIdentityDictionary methodsFor: 'testing' stamp: 'ael 6/21/2011 22:45'!
identityIncludes: anObject 
	"Answer whether anObject is one of the receiver's elements."

	self do: [:each | anObject == each ifTrue: [^true]].
	^false! !

!WIdentityDictionary methodsFor: 'testing' stamp: 'ael 6/21/2011 22:45'!
ifEmpty: aBlock
	"Evaluate the block if I'm empty"

	self isEmpty ifTrue: [ ^aBlock value ]! !

!WIdentityDictionary methodsFor: 'testing' stamp: 'ael 6/21/2011 22:45'!
ifEmpty: emptyBlock ifNotEmpty: notEmptyBlock
	"Evaluate emptyBlock if I'm empty, notEmptyBlock otherwise"
	" If the notEmptyBlock has an argument, eval with the receiver as its argument"

	self isEmpty ifTrue: [ ^emptyBlock value ].
	^notEmptyBlock valueWithPossibleArgument: self! !

!WIdentityDictionary methodsFor: 'testing' stamp: 'ael 6/21/2011 22:45'!
ifNotEmpty: aBlock
	"Evaluate the given block unless the receiver is empty.

      If the block has an argument, eval with the receiver as its argument,
      but it might be better to use ifNotEmptyDo: to make the code easier to
      understand"

	self isEmpty ifFalse: [^ aBlock valueWithPossibleArgument: self].
! !

!WIdentityDictionary methodsFor: 'testing' stamp: 'ael 6/21/2011 22:45'!
ifNotEmpty: notEmptyBlock ifEmpty: emptyBlock
	"Evaluate emptyBlock if I'm empty, notEmptyBlock otherwise
	 If the notEmptyBlock has an argument, eval with the receiver as its argument"

	self isEmpty ifFalse: [ ^notEmptyBlock valueWithPossibleArgument: self ].
	^emptyBlock value! !

!WIdentityDictionary methodsFor: 'testing' stamp: 'ael 6/21/2011 22:15'!
includes: anObject

	self do: [:each | anObject = each ifTrue: [^true]].
	^false! !

!WIdentityDictionary methodsFor: 'testing' stamp: 'ael 6/21/2011 22:15'!
includesAssociation: anAssociation
  ^ (self   
      associationAt: anAssociation key
      ifAbsent: [ ^ false ]) value = anAssociation value
! !

!WIdentityDictionary methodsFor: 'testing' stamp: 'ael 6/21/2011 22:15'!
includesIdentity: anObject
	"Answer whether anObject is one of the values of the receiver.  Contrast #includes: in which there is only an equality check, here there is an identity check"

	self do: [:each | anObject == each ifTrue: [^ true]].
	^ false! !

!WIdentityDictionary methodsFor: 'testing' stamp: 'ael 9/26/2011 08:00'!
includesKey: key 
	"Answer whether the receiver has a key equal to the argument, key."
	
	| deletion dnk reads thisWorld trackers |
	(thisWorld := self thisWorld) isTopLevel ifTrue: [ ^contents includesKey: key ].
	
	"updateTrackers for a WIdentityDict is a tuple { reads. seenAllFromParent. writes. }"
	trackers := self updateTrackersForWorld: thisWorld createIfAbsent: true caching: true.
	reads := thisWorld tracksReads ifTrue: [ self iEnsureReadsRecordIn: trackers ] ifFalse: [ nil ].

	dnk := self dnkToken.
	deletion := self deletionToken.
	
	self lookForAssociationAt: key careAboutValue: false inHomeWorld: true ifPresentDo: [ :assoc :isHomeWorld |
		"something was found - though it may be a deletion record"
		isHomeWorld
		  ifFalse: [
			assoc value == deletion
			  ifTrue: [ self recordReadOf: key->deletion in: reads. ^false ]
			  ifFalse: [ self recordReadOf: key->dnk in: reads. ^true ].
			].
		
		"something was found in our own reads or writes"
		^assoc value ~~ deletion
		].

	"we dropped off the end of the lookup.  record a deletion token in our Reads so we don't try again"
	self recordReadOf: key->deletion in: reads.
	^false
! !

!WIdentityDictionary methodsFor: 'enumerating' stamp: 'ael 6/21/2011 22:48'!
inject: thisValue into: binaryBlock 
	"Accumulate a running value associated with evaluating the argument, 
	binaryBlock, with the current value of the argument, thisValue, and the 
	receiver as block arguments. For instance, to sum the numeric elements 
	of a collection, aCollection inject: 0 into: [:subTotal :next | subTotal + 
	next]."

	| nextValue |
	nextValue := thisValue.
	self do: [:each | nextValue := binaryBlock value: nextValue value: each].
	^nextValue! !

!WIdentityDictionary methodsFor: 'Tools-Inspector' stamp: 'ael 6/21/2011 22:15'!
inspectorClass
	"Answer the class of the inspector to be used on the receiver.  Called by inspect; 
	use basicInspect to get a normal (less useful) type of inspector."

	^ DictionaryInspector! !

!WIdentityDictionary methodsFor: 'enumerating' stamp: 'ael 12/2/2011 15:18'!
intersection: aCollection
	"Answer the set theoretic intersection of two collections."

	^ self select: [:each | aCollection includes: each]! !

!WIdentityDictionary methodsFor: 'testing' stamp: 'ael 6/21/2011 22:48'!
isCollection
	"Return true if the receiver is some sort of Collection and responds to basic collection messages such as #size and #do:"
	^true! !

!WIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 5/3/2012 15:05'!
isCommittableKey: key
	"a chance to filter out any keys we don't want committed"
	^true
! !

!WIdentityDictionary methodsFor: 'testing' stamp: 'ael 6/21/2011 22:15'!
isDictionary
	^true! !

!WIdentityDictionary methodsFor: 'testing' stamp: 'ael 6/21/2011 22:48'!
isEmpty
	"Answer whether the receiver contains any elements."

	^self size = 0! !

!WIdentityDictionary methodsFor: 'testing' stamp: 'ael 6/21/2011 22:48'!
isEmptyOrNil
	"Answer whether the receiver contains any elements, or is nil.  Useful in numerous situations where one wishes the same reaction to an empty collection or to nil"

	^ self isEmpty! !

!WIdentityDictionary methodsFor: 'testing' stamp: 'ael 6/21/2011 22:48'!
isSequenceable
	^ false! !

!WIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 4/10/2012 14:08'!
isSerializableIn: parentWorld using: updateTrackers
	"updateTrackers for a WIdentityDict is a tuple { reads. seenAllFromParent. writes. }

	- if there is no Reads collection, return OK
	- ask the parent for all keys (i.e., all the keys it can currently see).  this collection must match our Reads record in both reads and deletions.  so, for each (non-deletion) key in Reads:
		* if the value is Delete, and the parent has the key, return FAIL
		* if the value is DNK, and the parent doesn't have the key, return FAIL
		* for any other value, if the parent doesn't have the same value return FAIL

		* if we believe we've seen all keys, the parent must also not have any additional ones
	- return OK
	"
	| deletion dnk parentDict parentKeys readKeys readValue reads |
	(reads := updateTrackers at: 1) isNil ifTrue: [ ^true ].	"could be empty, but if we've seen all keys we need to check that the parent hasn't acquired any"

	dnk := self dnkToken.
	deletion := self deletionToken.
	readKeys := (reads keys select: [ :k | self isCommittableKey: k]) asOrderedCollection.	"including deletion entries"

	parentDict := IdentityDictionary newFrom: (self worldFlattenedAssociationsIn: parentWorld)..
	parentKeys := parentDict keys asIdentitySet.		"for fast lookup"

	"The next check wasn't sufficiently planned for, so it's hard to do right... and so right now we don't do it.  As a result, the serialisability check is slightly over-optimistic.

	The idea is that if we believe we've seen all the collection's keys in the ancestor worlds, the parent mustn't have any now that we didn't already know about.
	The problem is that we don't currently record a read for every key we see (for example, if it's a key that we have overwritten without reading).  The seenAllFromParent flag is just used as a way to short-cut subsequent searches.
	One day this could be done right - probably using the overwrittenToken, like in WCollection.
	
(updateTrackers at: 2) ifNotNil: [	
	keysWeveHeardOf := readKeys reject: [ :key | (reads at: key) == deletion ].
	(parentKeys copyWithoutAll: keysWeveHeardOf)
	  ifNotEmpty: [ WorldSerializationError signal: 'parent has additional key(s)'. ^false ].
	].
"
	  
	readKeys do: [ :key |
		readValue := reads at: key.
		readValue == deletion
		  ifTrue: [
			(parentKeys includes: key)
			  ifTrue: [ WorldSerializationError signal: 'Parent has additional key: ', key asString. ^false ]
			]
		  ifFalse: [
			readValue == dnk
			  ifTrue: [
				(parentKeys includes: key)
				  ifFalse: [ WorldSerializationError signal: 'Parent has lost a key: ', key asString. ^false ]
				]
			  ifFalse: [
				(self areEqualForCommit: readValue and: (parentDict at: key))
				  ifFalse: [
					(WorldValueChanged new indexPrefix: 'key' index: key readValue: readValue currentValue: (parentDict at: key)) signal.
					^false
					]
				]
			]
		].

	^true			"no objections"! !

!WIdentityDictionary methodsFor: 'testing' stamp: 'ael 6/21/2011 22:48'!
isZero
	"Answer whether the receiver is zero"
	^ false! !

!WIdentityDictionary methodsFor: 'accessing' stamp: 'ael 6/21/2011 22:15'!
keyAtIdentityValue: value 
	"Answer the key that is the external name for the argument, value. If 
	there is none, answer nil.
	Note: There can be multiple keys with the same value. Only one is returned."

	^self keyAtIdentityValue: value ifAbsent: [self errorValueNotFound]! !

!WIdentityDictionary methodsFor: 'accessing' stamp: 'ael 6/21/2011 22:15'!
keyAtIdentityValue: value ifAbsent: exceptionBlock
	"Answer the key that is the external name for the argument, value. If 
	there is none, answer the result of evaluating exceptionBlock.
	Note: There can be multiple keys with the same value. Only one is returned."
 
	self associationsDo: 
		[:association | value == association value ifTrue: [^association key]].
	^exceptionBlock value! !

!WIdentityDictionary methodsFor: 'accessing' stamp: 'ael 6/21/2011 22:15'!
keyAtValue: value 
	"Answer the key that is the external name for the argument, value. If 
	there is none, answer nil."

	^self keyAtValue: value ifAbsent: [self errorValueNotFound]! !

!WIdentityDictionary methodsFor: 'accessing' stamp: 'ael 6/21/2011 22:16'!
keyAtValue: value ifAbsent: exceptionBlock
	"Answer the key that is the external name for the argument, value. If 
	there is none, answer the result of evaluating exceptionBlock."
 
	^self keyAtIdentityValue: value ifAbsent: exceptionBlock! !

!WIdentityDictionary methodsFor: 'accessing' stamp: 'ael 12/2/2011 02:57'!
keys
	"Answer an Array containing the receiver's keys."
	
	^Array new: 10 streamContents: [:s| self keysDo: [:key| s nextPut: key]]! !

!WIdentityDictionary methodsFor: 'enumerating' stamp: 'ael 6/21/2011 22:15'!
keysAndValuesDo: aBlock
	^self associationsDo:[:assoc|
		aBlock value: assoc key value: assoc value].! !

!WIdentityDictionary methodsFor: 'removing' stamp: 'ael 6/21/2011 22:15'!
keysAndValuesRemove: keyValueBlock
	"Removes all entries for which keyValueBlock returns true."
	"When removing many items, you must not do it while iterating over the dictionary, since it may be changing.  This method takes care of tallying the removals in a first pass, and then performing all the deletions afterward.  Many places in the sytem could be simplified by using this method."

	| removals |
	removals := OrderedCollection new.
	self associationsDo:
		[:assoc | (keyValueBlock value: assoc key value: assoc value)
			ifTrue: [removals add: assoc key]].
 	removals do:
		[:aKey | self removeKey: aKey]! !

!WIdentityDictionary methodsFor: 'enumerating' stamp: 'ael 12/2/2011 14:02'!
keysDo: aBlock 
	"Evaluate aBlock for each of the receiver's keys."
	| dnk key homeWorld homeReads |
	(homeWorld := self thisWorld) isTopLevel ifTrue: [ ^contents keysDo: aBlock ].

	homeReads := self readDictIfUsable.
	
	dnk := self dnkToken.
	
	"the allRecordsDo: may pass us associations from multiple worlds.  as each association is delivered, execution will be in the world where that association was found"
	self
	  allRecordsDo: [:association |
		key := association key.
		(homeReads notNil and: [ self thisWorld ~~ homeWorld ])
		  ifTrue: [ homeReads at: key ifAbsentPut: [ dnk ] ].	"only asked for the keys, so record dnk if not seen before"

		homeWorld eval: [ aBlock value: key ].
		]
	  includingDeletions: false
	  loggingReads: false.			"because we're doing that in this method"
	
	"if we got all the way through to the end, make a note that all keys have been seen"
	self recordAllParentKeysSeen.
! !

!WIdentityDictionary methodsFor: 'accessing' stamp: 'ael 7/11/2011 17:42'!
keysSortedSafely
	"Answer a sorted Array containing the receiver's keys."
	"Remove nil keys"
	 ^(self keys copyWithout: nil) sort: [ :x :y |
		"Should really be use <obj, string, num> compareSafely..."
		((x isString and: [ y isString ])
			or: [ x isNumber and: [ y isNumber ] ])
			ifTrue: [ x < y ]
			ifFalse: [ x class == y class
				ifTrue: [ x printString < y printString ]
				ifFalse: [ x class name < y class name ] ] ].
! !

!WIdentityDictionary methodsFor: 'private' stamp: 'ael 10/27/2011 14:43'!
lookForAssociationAt: key careAboutValue: careAboutValue inHomeWorld: isHomeWorld ifPresentDo: aBlock
	"Look for an association with the given key, going up through the parent-world chain if necessary.  This method is called recursively, once for each world up the chain.
	
	NB: this method does no Reads logging: it's the responsibility of the caller"
	| tryParent seenAll candidate reads foundDNKRecord thisWorld trackers writeAssoc writes dnk |
	reads := seenAll := writes := nil.
	thisWorld := self thisWorld.
	thisWorld isTopLevel
	  ifTrue: [ writes := contents ]
	  ifFalse: [
		"iteration shouldn't force creation of trackers, since the collection might be empty"
		trackers := self updateTrackersForWorld: thisWorld createIfAbsent: false caching: isHomeWorld.

		trackers ifNotNil: [		"a tuple { reads. seenAllFromParent. writes. }"
			thisWorld tracksReads
			  ifTrue: [
				reads := trackers at: 1.		"may be nil"
				seenAll := trackers at: 2.
				].
			writes := trackers at: 3.			"may be nil"
			].
		].

	"first try the Writes"
	writes
	  ifNotNil: [
		(writeAssoc := writes associationAt: key ifAbsent: [ nil ])
		  ifNotNil: [ ^aBlock value: writeAssoc copy value: isHomeWorld ].		"copy to protect the source"
		].

	"nothing there, so try Reads"
	tryParent := true.		"default"
	foundDNKRecord := false.
	dnk := self dnkToken.
	
	reads
	  ifNotNil: [
		candidate := reads associationAt: key ifAbsent: [ nil ].
		
		"if we recorded a value of DNK this is of no interest to the sender unless careAboutValue is false"
		candidate
		  ifNotNil: [
			foundDNKRecord := careAboutValue & (candidate value == dnk).
			foundDNKRecord ifFalse: [ ^aBlock value: candidate copy value: isHomeWorld ].		"a legitimate association"
			].
		
		"if we've seen both the keys and the values, nothing in the ancestors will add to what we know.
		likewise if we've seen all keys, and the sought one was not one of them."
		(seenAll == #values or: [ seenAll == #keys and: [ candidate isNil ] ])
		  ifTrue: [ tryParent := false ].		"bail out, because our ancestors can't help"
		].	
	
	tryParent
	  ifTrue: [			"recursively ask in our parent world"
		self evalInParentWorld: [
			self lookForAssociationAt: key careAboutValue: careAboutValue inHomeWorld: false ifPresentDo: [ :parentAssoc :dummy "we know it's not the home (i.e., calling) world" |
				"Here we trap one form of inconsistency: some world has a DNK against this key, but in one of its ancestors the key has now been deleted.  This is signalled to callers by returning an association with value DNK.  Naturally this is only meaningful for callers who set careAboutValue to true."
				(foundDNKRecord and: [ parentAssoc value isWorldDeletion ]) ifTrue: [ ^aBlock value: key->dnk value: false ]. 
				^aBlock value: parentAssoc copy value: false.
				]
			]
		]
! !

!WIdentityDictionary methodsFor: 'testing' stamp: 'ael 6/21/2011 22:49'!
notEmpty
	"Answer whether the receiver contains any elements."

	^ self isEmpty not! !

!WIdentityDictionary methodsFor: 'testing' stamp: 'ael 6/21/2011 23:37'!
occurrencesOf: anObject 
	"Answer how many of the receiver's elements are equal to anObject."

	| count |
	count := 0.
	self do: [:each | anObject = each ifTrue: [count := count + 1]].
	^count! !

!WIdentityDictionary methodsFor: 'copying' stamp: 'ael 9/25/2011 17:56'!
postCopy
	super postCopy.
	contents := contents copy.
! !

!WIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 11/14/2011 21:18'!
primFlattenInWorld: world into: flattener dnk: dnk deletion: deletion notFound: notFound valuesSym: valuesSym assocClass: assocClass
	<primitive: 'primDictFlattenInto' module: 'WorldsPlugin'>
	^nil		"force the sender to run the Smalltalk version"
! !

!WIdentityDictionary methodsFor: 'printing' stamp: 'ael 6/24/2011 11:36'!
printElementsOn: aStream 
	| size |
	aStream nextPut: $(.
	size := self size.
	size > 100
		ifTrue: [aStream nextPutAll: 'size '.
			size printOn: aStream]
		ifFalse: [self keysSortedSafely
				do: [:key | aStream print: key;
						 nextPutAll: '->';				
						 print: (self at: key);
						 space]].
	aStream nextPut: $)! !

!WIdentityDictionary methodsFor: 'printing' stamp: 'ael 6/21/2011 22:50'!
printNameOn: aStream
	super printOn: aStream! !

!WIdentityDictionary methodsFor: 'printing' stamp: 'ael 6/21/2011 22:50'!
printOn: aStream 
	"Append a sequence of characters that identify the receiver to aStream."

	self printNameOn: aStream.
	self printElementsOn: aStream! !

!WIdentityDictionary methodsFor: 'private' stamp: 'ael 11/1/2011 14:54'!
privateAt: key put: anObject 
	^self at: key put: anObject
! !

!WIdentityDictionary methodsFor: 'removing' stamp: 'ael 11/1/2011 19:52'!
privateRemoveKey: key
	"record the removal of a key from the collection.  used in committing deletions from a child.  don't attempt to return a current value."
	| thisWorld trackers writes |
	(thisWorld := self thisWorld) isTopLevel ifTrue: [ ^contents removeKey: key ifAbsent: [] ].
	
	"updateTrackers is a tuple { reads. seenAllFromParent. writes. }"
	trackers := self updateTrackersForWorld: thisWorld createIfAbsent: true caching: true.	
	writes := self iEnsureWritesRecordIn: trackers.
	writes add: key->self deletionToken.
! !

!WIdentityDictionary methodsFor: 'private' stamp: 'ael 11/11/2011 21:52'!
privateSetContents: dict
	"sneaky fast way to create a WIdentityDictionary from an IdentityDictionary"
	contents := dict
! !

!WIdentityDictionary methodsFor: 'private' stamp: 'ael 9/25/2011 17:39'!
readDictIfUsable
	"if this object does not want to track reads, return nil.  otherwise ensure we have a reads dictionary"
	| thisWorld trackers |
	(thisWorld := self thisWorld) isTopLevel ifTrue: [ ^nil ].

	thisWorld tracksReads ifFalse: [ ^nil ].
	
	trackers := self updateTrackersForWorld: thisWorld createIfAbsent: true caching: true.

	^self iEnsureReadsRecordIn: trackers.
! !

!WIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 9/25/2011 16:12'!
recordAllParentKeysSeen
	self recordAllSeen: #keys
! !

!WIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 9/25/2011 16:12'!
recordAllParentValuesSeen
	self recordAllSeen: #values
! !

!WIdentityDictionary methodsFor: 'private' stamp: 'ael 9/25/2011 16:12'!
recordAllSeen: condition
	"condition is #keys or #values"
	"updateTrackers for an WIdentityDict is a tuple { reads. seenAllFromParent. writes. }"
	| thisWorld trackers |
	(thisWorld := self thisWorld) tracksReads ifFalse: [ ^self ].
	
	trackers := self updateTrackersForWorld: thisWorld createIfAbsent: true caching: true.
	trackers at: 2 put: condition.
! !

!WIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 6/24/2011 23:49'!
recordReadOf: assoc in: aDictOrNil
	aDictOrNil ifNotNil: [ aDictOrNil add: assoc ]
! !

!WIdentityDictionary methodsFor: 'enumerating' stamp: 'ael 6/21/2011 22:51'!
reject: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument. 
	Collect into a new collection like the receiver only those elements for 
	which aBlock evaluates to false. Answer the new collection."

	^self select: [:element | (aBlock value: element) == false]! !

!WIdentityDictionary methodsFor: 'removing' stamp: 'ael 6/21/2011 22:52'!
remove: oldObject 
	"Remove oldObject from the receiver's elements. Answer oldObject 
	unless no element is equal to oldObject, in which case, raise an error.
	ArrayedCollections cannot respond to this message."

	^ self remove: oldObject ifAbsent: [self errorNotFound: oldObject]! !

!WIdentityDictionary methodsFor: 'removing' stamp: 'ael 6/21/2011 22:15'!
remove: anObject ifAbsent: exceptionBlock

	self shouldNotImplement! !

!WIdentityDictionary methodsFor: 'removing' stamp: 'ael 6/23/2011 10:28'!
removeAll
	"remove all elements from this collection"
	self keys do: [ :key | self removeKey: key ]
! !

!WIdentityDictionary methodsFor: 'removing' stamp: 'ael 6/21/2011 22:52'!
removeAll: aCollection 
	"Remove each element of aCollection from the receiver. If successful for 
	each, answer aCollection. Otherwise create an error notification.
	ArrayedCollections cannot respond to this message."

	aCollection == self ifTrue: [^self removeAll].
	aCollection do: [:each | self remove: each].
	^ aCollection! !

!WIdentityDictionary methodsFor: 'removing' stamp: 'ael 6/21/2011 22:52'!
removeAllFoundIn: aCollection 
	"Remove each element of aCollection which is present in the receiver 
	from the receiver. Answer aCollection. No error is raised if an element
	isn't found. ArrayedCollections cannot respond to this message."

	aCollection do: [:each | self remove: each ifAbsent: []].
	^ aCollection! !

!WIdentityDictionary methodsFor: 'removing' stamp: 'ael 6/21/2011 22:52'!
removeAllSuchThat: aBlock 
	"Evaluate aBlock for each element and remove all that elements from
	the receiver for that aBlock evaluates to true.  Use a copy to enumerate 
	collections whose order changes when an element is removed (i.e. Sets)."

	self copy do: [:each | (aBlock value: each) ifTrue: [self remove: each]]! !

!WIdentityDictionary methodsFor: 'removing' stamp: 'ael 6/21/2011 22:15'!
removeKey: key 
	"Remove key from the receiver.
	If key is not in the receiver, notify an error."

	^ self removeKey: key ifAbsent: [self errorKeyNotFound: key]! !

!WIdentityDictionary methodsFor: 'removing' stamp: 'ael 10/27/2011 14:43'!
removeKey: key ifAbsent: aBlock 
	"Remove key (and its associated value) from the receiver. If key is not in 
	the receiver, answer the result of evaluating aBlock. Otherwise, answer 
	the value externally named by key."

	| deletion oldValue recordableValue reads thisWorld trackers writes dnk |
	(thisWorld := self thisWorld) isTopLevel
	  ifTrue: [ ^contents removeKey: key ifAbsent: aBlock ].
	
	"ok, so we're in a child world.  removal involves a (logged, but only if something was found in a higher world) lookup of the value being removed, followed by an overwrite into our Writes collection"
	"updateTrackers is a tuple { reads. seenAllFromParent. writes. }"
	trackers := self updateTrackersForWorld: thisWorld createIfAbsent: true caching: true.	
	reads := thisWorld tracksReads ifTrue: [ self iEnsureReadsRecordIn: trackers ] ifFalse: [ nil ].
	writes := self iEnsureWritesRecordIn: trackers.

	deletion := self deletionToken.
	dnk := self dnkToken.

	recordableValue := deletion.				"a value that cannot be found by a lookup"
	self lookForAssociationAt: key careAboutValue: true inHomeWorld: true ifPresentDo: [ :assoc :isHomeWorld |
		"we found something.  there are various possible cases to handle:
		  - if it came from a world above, log the read in our Reads dictionary
		  - if it was a delete (i.e., this key had already been removed), evaluate the ifAbsent block and return
			* otherwise record a deletion write at our level, and return the old value
			
		We may also receive a record with a DNK value, showing an inconsistency between a child world and one of its ancestors (see note in associationAt:ifAbsent:loggingReads:).  We treat this as if there were no value to remove."

		oldValue := assoc value.
		recordableValue := oldValue == dnk ifTrue: [ deletion ] ifFalse: [ oldValue ].

		isHomeWorld ifFalse: [ self recordReadOf: key->recordableValue in: reads ].

		recordableValue == deletion ifTrue: [ ^aBlock value ].
		].

	"we found an item and need to record a local deletion, or we found no reference to this key at all"
	recordableValue == deletion
	  ifTrue: [	"we found no reference.  record the absence as a read of Delete"
		self recordReadOf: key->deletion in: reads.
		^aBlock value
		].

	"record the deletion locally, and return the previous value"		
	writes add: key->deletion.
	^recordableValue
! !

!WIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 11/12/2011 20:19'!
roughSize
	"as a guess, return the size in the top-level world"
	^contents size max: 10
! !

!WIdentityDictionary methodsFor: 'enumerating' stamp: 'ael 6/21/2011 22:53'!
safeDo: aBlock

	^ self do: aBlock
! !

!WIdentityDictionary methodsFor: 'enumerating' stamp: 'ael 6/21/2011 22:15'!
select: aBlock 
	"Evaluate aBlock with each of my values as the argument. Collect into a new dictionary, only those associations for which aBlock evaluates to true."

	| newCollection |
	newCollection := self copyEmpty.
	self associationsDo: [ :each |
		(aBlock value: each value) ifTrue: [
			newCollection add: each copy ] ].
	^newCollection! !

!WIdentityDictionary methodsFor: 'enumerating' stamp: 'ael 6/21/2011 22:54'!
select: selectBlock thenCollect: collectBlock
	"Utility method to improve readability."

	^ (self select: selectBlock) collect: collectBlock! !

!WIdentityDictionary methodsFor: 'private' stamp: 'ael 9/25/2011 16:34'!
setContentsSize: n
	"Initialize to a collection size of n"

	contents := IdentityDictionary new: n
! !

!WIdentityDictionary methodsFor: 'siss-private' stamp: 'ael 9/25/2011 18:00'!
sissContentsInto: sexp context: dictionary
self flag: #aelToDo.
	self class
		sissContents: self
		into: sexp
		context: dictionary! !

!WIdentityDictionary methodsFor: 'accessing' stamp: 'ael 2/25/2012 02:18'!
size
	"world awareness: size is not cached, but any contribution from parent and above will cause a freeze of the keys we saw from there.  on subsequent size requests, we won't access parent worlds at all."
	| size tempDict thisWorld trackers |
	(thisWorld := self thisWorld) isTopLevel ifTrue: [ ^contents size ].

	"short cut if we know we've already seen everything from the parent"	
	"trackers is a tuple { reads. seenAllFromParent. writes. }"
	trackers := self updateTrackersForWorld: thisWorld createIfAbsent: false caching: true.
	(trackers notNil and: [ (trackers at: 2) notNil ])
	  ifTrue: [			"we have seen everything from the parent"
		"the number of entries is the number of keys for which we haven't recorded (or seen) a deletion.  we can find this by merging the reads and writes, then counting non-deletion values."
		tempDict := trackers at: 1.		"reads - may be nil"
		(trackers at: 3)		"writes - may be nil"
		  ifNotNil: [ :writes |
			tempDict
			  ifNil: [ tempDict := writes ]
			  ifNotNil: [
				tempDict := tempDict copy.		"protect the reads dictionary"
				writes keysAndValuesDo: [ :k :v | tempDict at: k put: v ].
				].
			].
		tempDict ifNil: [ ^0 ].
		^(tempDict values copyWithout: self deletionToken) size.
		].

	"haven't seen everything yet.  just iterate through the keys"
	size := 0.
	self keysDo: [ :k | size := size + 1].
	^size
! !

!WIdentityDictionary methodsFor: 'sorting' stamp: 'ael 6/21/2011 22:55'!
sorted
	"Return a new sequenceable collection which contains the same elements as self but its elements are sorted in ascending order using the #'<=' operator."

	^self sorted: nil! !

!WIdentityDictionary methodsFor: 'sorting' stamp: 'ael 6/21/2011 22:55'!
sorted: aSortBlockOrNil
	"Return a new sequenceable collection which contains the same elements as self but its elements are sorted by aSortBlockOrNil. The block should take two arguments and return true if the first element should preceed the second one. If aSortBlock is nil then <= is used for comparison."

	^self asArray sort: aSortBlockOrNil! !

!WIdentityDictionary methodsFor: 'printing' stamp: 'ael 6/21/2011 22:15'!
storeOn: aStream
	| noneYet |
	aStream nextPutAll: '(('.
	aStream nextPutAll: self class name.
	aStream nextPutAll: ' new)'.
	noneYet := true.
	self associationsDo: 
			[:each | 
			noneYet
				ifTrue: [noneYet := false]
				ifFalse: [aStream nextPut: $;].
			aStream nextPutAll: ' add: '.
			aStream store: each].
	noneYet ifFalse: [aStream nextPutAll: '; yourself'].
	aStream nextPut: $)! !

!WIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 5/30/2012 15:33'!
syncWithTopWorld
	"When a world-aware object is accessed for the first time in a deeply nested world, every access potentially ripples up through all worlds.  This method (invoked either as part of object creation, or when first accessing the object in a world chain where it has not been seen before) short-cuts the read-tracking mechanism to create a mirror of the object's top-level properties at the level of the accessing world.
	At the moment this is only implemented for WIdentityDictionary, where it serves as support for KSObject."
	| assocs reads thisWorld trackers |
	(thisWorld := WWorld thisWorld) isTopLevel ifTrue: [ ^self ].
	
	assocs := WTopLevelWorld default eval: [
		Array new: self size streamContents: [ :stream | self associationsDo: [ :each | stream nextPut: each ] ].
		].
	
	trackers := self updateTrackersForWorld: thisWorld createIfAbsent: true caching: true.

	(reads := trackers at: 1) ifNil: [ reads := trackers at: 1 put: (IdentityDictionary new: assocs size) ].
	assocs do: [ :assoc | reads add: assoc ].
	trackers at: 2 put: #values.		"make a note that we've seen everything"
! !

!WIdentityDictionary methodsFor: 'enumerating' stamp: 'ael 11/12/2011 17:15'!
uncheckedAt: key put: anObject 
	"for testing"
	^self at: key put: anObject
! !

!WIdentityDictionary methodsFor: 'enumerating' stamp: 'ael 6/21/2011 22:55'!
union: aCollection
	"Answer the set theoretic union of the receiver and aCollection, using the receiver's notion of equality and not side effecting the receiver at all."

	^ self copy addAll: aCollection; yourself

! !

!WIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 11/14/2011 21:21'!
unloggedSize
	| thisWorld |
	(thisWorld := self thisWorld) isTopLevel ifTrue: [ ^contents size ].

	"flattening doesn't log, so we can use that"
	^self worldFlattenedIn: thisWorld sizeOnly: true.
! !

!WIdentityDictionary methodsFor: 'accessing' stamp: 'ael 12/2/2011 02:46'!
values
	"Answer a Collection containing the receiver's values."
	"world awareness: don't ask the size, which can be a heavy operation"
	^(Array new: 10 streamContents: [ :stream |
		self valuesDo: [ :value | stream nextPut: value] ]) asArray! !

!WIdentityDictionary methodsFor: 'enumerating' stamp: 'ael 6/21/2011 22:15'!
valuesDo: aBlock 
	"Evaluate aBlock for each of the receiver's values."

	self associationsDo: [:association | aBlock value: association value]! !

!WIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 11/12/2011 16:05'!
worldFlattenIn: aWorld
	"our flattening primitive gives an array of associations.  go the extra step and turn the associations into a new WIdentityDictionary"
	| associations |
	associations := self worldFlattenedAssociationsIn: aWorld.	"fast, not recording reads (cf. self asArray)"
	^self class newAndDo: [ :inst | inst privateSetContents: (IdentityDictionary withAll: associations) ].
! !

!WIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 11/14/2011 21:20'!
worldFlattenedAssociationsIn: aWorld
	^self worldFlattenedIn: aWorld sizeOnly: false.
! !

!WIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 11/12/2011 11:50'!
worldFlattenedCopy
	"get a flattened copy of the receiver, keeping appropriate records of reads from parent worlds"
	self thisWorld isTopLevel ifTrue: [ ^self clone postCopy ].

	^self class withAll: self associations
! !

!WIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 11/14/2011 21:20'!
worldFlattenedIn: aWorld sizeOnly: returnSizeOnly
	"Two methods in one.  Using the flattening primitive, if available, find the associations in the receiver when seen from aWorld.  If returnSizeOnly is true, don't bother with the last step (copying the associations from the dictionary they were written into).
	If the primitive fails for some reason, try the equivalent Smalltalk method."
	| flattener tally assocs |
	flattener := IdentityDictionary new: (self roughSize * 4 // 3 + 1).	"try to avoid growing"
	tally := self primFlattenInWorld: aWorld into: flattener dnk: self dnkToken deletion: self deletionToken notFound: Object new valuesSym: #values assocClass: Association.

	tally ifNotNil: [		"the primitive worked"
		^returnSizeOnly
		  ifTrue: [ tally ]
		  ifFalse: [ flattener array copyFrom: 1 to: tally ].		"NB: as a dictionary it's now completely futzed"
		].
	
	"plan B: Smalltalk flattening"
	assocs := aWorld eval: [ self associationsSnapshotTrackingReads: false ].
	^returnSizeOnly
	  ifTrue: [ assocs size ]
	  ifFalse: [ assocs ]
! !

!WIdentityDictionary methodsFor: 'filter streaming' stamp: 'ael 6/21/2011 22:55'!
write: anObject 
	^ self add: anObject! !

!WIdentityDictionary methodsFor: 'world awareness' stamp: 'ael 11/1/2011 19:30'!
writesRecordIn: updateTrackers
	"updateTrackers for a WIdentityDict is a tuple { reads. seenAllFromParent. writes. }"
	^ updateTrackers at: 3
! !

WIdentityDictionary subclass: #WIdentitySet
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Worlds'!

!WIdentitySet commentStamp: 'ael 12/2/2011 02:26' prior: 0!
Awesomely naive implementation of a set: simply override add:, do: etc to use only the keys
!

!WIdentitySet class methodsFor: 'as yet unclassified' stamp: 'ael 12/2/2011 15:14'!
newFrom: aCollection 
	"Answer an instance of me containing the same values as aCollection."
	| newSet |
	newSet := self new: aCollection size.		"will be created in top-level world"
	newSet addAll: aCollection.
	^ newSet
! !

!WIdentitySet methodsFor: 'comparing' stamp: 'ael 12/2/2011 15:24'!
= aSet
	"Two sets are equal if
	 (a) they are the same 'kind' of thing.
	 (b) they have the same set of values."

	self == aSet ifTrue: [ ^ true ].
	(aSet isKindOf: WIdentitySet) ifFalse: [^false].
	self size = aSet size ifFalse: [^false].
	self do: [:obj| (aSet includes: obj) ifFalse: [ ^false ] ].
	^true

! !

!WIdentitySet methodsFor: 'adding' stamp: 'ael 2/25/2012 01:46'!
add: anObject
	"Include anObject as one of the receiver's elements.
	Since adding a key is much faster than checking whether we already had it, don't bother with the check.
	Anwser anObject."

	super at: anObject put: nil.
	^ anObject! !

!WIdentitySet methodsFor: 'adding' stamp: 'ael 12/1/2011 22:41'!
add: newObject withOccurrences: anInteger
	^self add: newObject
! !

!WIdentitySet methodsFor: 'adding' stamp: 'ael 12/1/2011 22:40'!
addAll: aCollection
	aCollection do: [ :anObject | self add: anObject ].
	^aCollection
! !

!WIdentitySet methodsFor: 'converting' stamp: 'ael 12/2/2011 02:58'!
asArray
	^WArray withAll: self keys
! !

!WIdentitySet methodsFor: 'converting' stamp: 'ael 12/2/2011 15:26'!
asIdentitySet
	^self
! !

!WIdentitySet methodsFor: 'converting' stamp: 'ael 12/2/2011 02:59'!
asOrderedCollection
	^WCollection withAll: self keys
	! !

!WIdentitySet methodsFor: 'converting' stamp: 'ael 12/2/2011 02:58'!
asSet
	^self
! !

!WIdentitySet methodsFor: 'adding' stamp: 'ael 12/2/2011 15:11'!
at: anObject put: aValue
	self errorNotIndexable
! !

!WIdentitySet methodsFor: 'enumerating' stamp: 'ael 12/2/2011 02:31'!
collect: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument.  
	Collect the resulting values into a collection like the receiver. Answer  
	the new collection."

	| newCollection |
	newCollection := self species new: self size.
	self do: [:each | newCollection add: (aBlock value: each)].
	^ newCollection! !

!WIdentitySet methodsFor: 'copying' stamp: 'ael 12/2/2011 02:52'!
copyWithout: oldElement 
	"Answer a copy of the receiver that does not contain any
	elements equal to oldElement."

	^ self reject: [:each | each == oldElement]
! !

!WIdentitySet methodsFor: 'enumerating' stamp: 'ael 12/2/2011 02:41'!
do: aBlock
	^self keysDo: aBlock! !

!WIdentitySet methodsFor: 'explorer' stamp: 'ael 12/2/2011 02:29'!
explorerContents 

	^Array new: self size streamContents: [ :stream |
		self do: [ :each |
			stream nextPut: (
				ObjectExplorerWrapper
					with: each
					name: (stream position + 1) printString
					model: self) ] ]! !

!WIdentitySet methodsFor: 'accessing' stamp: 'ael 12/2/2011 02:27'!
includes: anObject
	^self includesKey: anObject
! !

!WIdentitySet methodsFor: 'testing' stamp: 'ael 12/2/2011 15:30'!
includesAllOf: aCollection 
	"Answer whether all the elements of aCollection are in the receiver."
	aCollection do: [:elem | (self includes: elem) ifFalse: [^ false]].
	^ true! !

!WIdentitySet methodsFor: 'testing' stamp: 'ael 12/2/2011 15:30'!
includesAnyOf: aCollection 
	"Answer whether any element of aCollection is one of the receiver's elements."
	aCollection do: [:elem | (self includes: elem) ifTrue: [^ true]].
	^ false! !

!WIdentitySet methodsFor: 'testing' stamp: 'ael 12/2/2011 15:31'!
isDictionary
	^false
! !

!WIdentitySet methodsFor: 'testing' stamp: 'ael 12/2/2011 15:31'!
occurrencesOf: anObject 
	^ (self includes: anObject) ifTrue: [1] ifFalse: [0]! !

!WIdentitySet methodsFor: 'printing' stamp: 'ael 12/2/2011 02:50'!
printElementsOn: aStream 
	| size |
	aStream nextPut: $(.
	size := self size.
	size > 100
		ifTrue: [aStream nextPutAll: 'size '.
			size printOn: aStream]
		ifFalse: [self keysSortedSafely
				do: [:key | aStream print: key; space]].
	aStream nextPut: $)! !

!WIdentitySet methodsFor: 'private' stamp: 'ael 12/2/2011 15:40'!
privateAt: key put: anObject 
	^super at: key put: anObject
! !

!WIdentitySet methodsFor: 'removing' stamp: 'ael 12/2/2011 02:35'!
remove: anObject ifAbsent: aBlock
	self removeKey: anObject ifAbsent: [ ^aBlock value ].
	^anObject
! !

!WIdentitySet methodsFor: 'enumerating' stamp: 'ael 12/2/2011 03:37'!
select: aBlock 
	"Evaluate aBlock with each of my values as the argument."

	| newCollection |
	newCollection := self copyEmpty.
	self do: [ :each |
		(aBlock value: each) ifTrue: [
			newCollection add: each ] ].
	^newCollection! !

WIdentityDictionary subclass: #WSimpleOrderedCollection
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Worlds'!

!WSimpleOrderedCollection commentStamp: 'ael 11/1/2011 13:54' prior: 0!
An experiment!

!WSimpleOrderedCollection class methodsFor: 'instance creation' stamp: 'ael 11/1/2011 20:39'!
new
	^ self new: 10! !

!WSimpleOrderedCollection class methodsFor: 'instance creation' stamp: 'ael 11/1/2011 20:44'!
new: anInteger
	^self new: anInteger andDo: [ :x | ]
! !

!WSimpleOrderedCollection class methodsFor: 'stream creation' stamp: 'ael 8/30/2012 12:25'!
new: newSize streamContents: blockWithArg

	| stream inst result |
	result := WTopLevelWorld default eval: [
		inst := super basicNew.
		inst setContentsSize: newSize; initialize.

		stream := WriteStream on: inst.
		blockWithArg value: stream.
		stream position = newSize
			ifTrue: [ stream originalContents ]
			ifFalse: [ stream contents ]
		].
	^result
! !

!WSimpleOrderedCollection class methodsFor: 'instance creation' stamp: 'ael 11/1/2011 20:46'!
new: anInteger withAll: anObject
	^ self new: anInteger andDo: [ :inst | inst addAll: (Array new: anInteger withAll: anObject) ]! !

!WSimpleOrderedCollection class methodsFor: 'instance creation' stamp: 'ael 11/1/2011 20:39'!
newFrom: aCollection 
	"Answer an instance of me containing the same elements as aCollection."

	^ self new: aCollection size andDo: [ :inst | inst addAll: aCollection ]! !

!WSimpleOrderedCollection class methodsFor: 'instance creation' stamp: 'ael 11/1/2011 20:39'!
ofSize: n
	"Create a new collection of size n with nil as its elements.
	This method exists because OrderedCollection new: n creates an
	empty collection,  not one of size n."
	^ self new: n withAll: nil.
! !

!WSimpleOrderedCollection class methodsFor: 'benchmarks' stamp: 'ael 5/14/2013 23:25'!
simpleBenchmark
	"WSimpleOrderedCollection simpleBenchmark"
	| l |
	MessageTally spyOn: 
	[
	WWorld thisWorld "sprout" eval: [
	l := WSimpleOrderedCollection new..
	1 to: 100000 do: [:x |
		0 to: 9 do: [:i| l addFirst: i ].
		[ l isEmpty ] whileFalse: [ l removeFirst]
		].
	 ]
	]
! !

!WSimpleOrderedCollection class methodsFor: 'benchmarks' stamp: 'ael 5/15/2013 00:00'!
simpleBenchmark: n
	"WSimpleOrderedCollection simpleBenchmark: 100"
	| l runs |
	runs := 100.
	^(Time millisecondsToRun: [
		WWorld thisWorld sprout eval: [
			runs timesRepeat: [
				l := self new..
				1 to: n do: [:i| l addFirst: i ].
				[ l isEmpty ] whileFalse: [ l removeLast ]
				].
			 ]
		]) asFloat / runs.
! !

!WSimpleOrderedCollection class methodsFor: 'instance creation' stamp: 'ael 11/1/2011 20:39'!
streamContents: blockWithArg
	
	^self new: 100 streamContents: blockWithArg! !

!WSimpleOrderedCollection class methodsFor: 'instance creation' stamp: 'ael 3/8/2012 23:09'!
with: anObject 
	"Answer an instance of me containing anObject."

	^ self new: 10 andDo: [ :inst | inst add: anObject ].
! !

!WSimpleOrderedCollection class methodsFor: 'instance creation' stamp: 'ael 3/8/2012 23:09'!
with: firstObject with: secondObject 
	"Answer an instance of me containing the two arguments as elements."

	^ self new: 10 andDo: [ :inst | inst add: firstObject; add: secondObject ].
	! !

!WSimpleOrderedCollection class methodsFor: 'instance creation' stamp: 'ael 3/8/2012 23:10'!
with: firstObject with: secondObject with: thirdObject 
	"Answer an instance of me containing the three arguments as elements."

	^ self new: 10 andDo: [ :inst | inst add: firstObject; add: secondObject; add: thirdObject ]		! !

!WSimpleOrderedCollection class methodsFor: 'instance creation' stamp: 'ael 3/8/2012 23:10'!
with: firstObject with: secondObject with: thirdObject with: fourthObject 
	"Answer an instance of me, containing the four arguments as the elements."

	^ self new: 10 andDo: [ :inst | inst add: firstObject; add: secondObject; add: thirdObject; add: fourthObject ]
! !

!WSimpleOrderedCollection class methodsFor: 'instance creation' stamp: 'ael 3/8/2012 23:10'!
with: firstObject with: secondObject with: thirdObject with: fourthObject with: fifthObject
	"Answer an instance of me, containing the five arguments as the elements."

	^ self new: 10 andDo: [ :inst | inst add: firstObject; add: secondObject; add: thirdObject; add: fourthObject; add: fifthObject ]
! !

!WSimpleOrderedCollection class methodsFor: 'instance creation' stamp: 'ael 3/8/2012 23:10'!
with: firstObject with: secondObject with: thirdObject with: fourthObject with: fifthObject with: sixthObject
	"Answer an instance of me, containing the six arguments as the elements."

	^ self new: 10 andDo: [ :inst | inst add: firstObject; add: secondObject; add: thirdObject; add: fourthObject; add: fifthObject; add: sixthObject ]
! !

!WSimpleOrderedCollection class methodsFor: 'instance creation' stamp: 'ael 11/1/2011 20:39'!
withAll: aCollection
	"Create a new collection containing all the elements from aCollection."

	^ self new: aCollection size andDo: [ :inst | inst addAll: aCollection ]
! !

!WSimpleOrderedCollection methodsFor: 'adding' stamp: 'ael 11/1/2011 09:08'!
add: newObject

	^self addLast: newObject! !

!WSimpleOrderedCollection methodsFor: 'adding' stamp: 'ael 11/1/2011 09:29'!
add: newObject after: oldObject 
	"Add the argument, newObject, as an element of the receiver. Put it in 
	the sequence just succeeding oldObject. Answer newObject."
	
	| index |
	index := self indexOf: oldObject ifAbsent: [ ^self errorNotFound: oldObject ].
	self add: newObject afterIndex: index.
	^newObject! !

!WSimpleOrderedCollection methodsFor: 'adding' stamp: 'ael 11/1/2011 18:02'!
add: newObject afterIndex: index
	"Add the argument, newObject, as an element of the receiver. Put it in 
	the sequence just after index. Answer newObject."
	| size |
	size := self size.
	(index between: 0 and: size) ifFalse:[^self errorSubscriptBounds: index].
	size to: index+1 by: -1 do: [ :i | self privateAt: i + 1 put: (self at: i) ].
	^self privateAt: index+1 put: newObject.
! !

!WSimpleOrderedCollection methodsFor: 'adding' stamp: 'ael 11/1/2011 09:31'!
add: newObject before: oldObject 
	"Add the argument, newObject, as an element of the receiver. Put it in 
	the sequence just preceding oldObject. Answer newObject."
	
	| index |
	index := self indexOf: oldObject ifAbsent: [ ^self errorNotFound: oldObject ].
	^self add: newObject afterIndex: index-1.
! !

!WSimpleOrderedCollection methodsFor: 'adding' stamp: 'ael 11/1/2011 09:31'!
add: newObject beforeIndex: index 
	"Add the argument, newObject, as an element of the receiver. Put it in 
	the sequence just before index. Answer newObject."

	^self add: newObject afterIndex: index - 1.
! !

!WSimpleOrderedCollection methodsFor: 'adding' stamp: 'ael 11/1/2011 09:28'!
addAll: aCollection

	^self addAllLast: aCollection
! !

!WSimpleOrderedCollection methodsFor: 'adding' stamp: 'ael 11/1/2011 10:08'!
addAllFirst: anOrderedCollection 
	"Add each element of anOrderedCollection at the beginning of the 
	receiver. Answer anOrderedCollection."

	anOrderedCollection reverseDo: [:each | self addFirst: each].
	^anOrderedCollection! !

!WSimpleOrderedCollection methodsFor: 'adding' stamp: 'ael 11/1/2011 10:09'!
addAllFirstUnlessAlreadyPresent: anOrderedCollection 
	"Add each element of anOrderedCollection at the beginning of the receiver, preserving the order, but do not add any items that are already in the receiver.  Answer anOrderedCollection."

	anOrderedCollection reverseDo:
		[:each | (self includes: each) ifFalse: [self addFirst: each]].
	^ anOrderedCollection! !

!WSimpleOrderedCollection methodsFor: 'adding' stamp: 'ael 11/1/2011 09:29'!
addAllLast: col
	"Add a collection of objects to the end of the receiver. Answer the collection."
	col do: [ :item | self addLast: item ].
	^ col! !

!WSimpleOrderedCollection methodsFor: 'adding' stamp: 'ael 11/1/2011 13:38'!
addFirst: newObject 
	"Add newObject to the beginning of the receiver. Answer newObject."
	"Records updates for worlds"

	self size to: 1 by: -1 do: [ :index | self privateAt: index + 1 put: (self at: index) ].

	^self privateAt: 1 put: newObject.
! !

!WSimpleOrderedCollection methodsFor: 'adding' stamp: 'ael 11/1/2011 13:38'!
addLast: newObject 

	^self privateAt: self size + 1 put: newObject.
! !

!WSimpleOrderedCollection methodsFor: 'data classes' stamp: 'ael 11/1/2011 10:23'!
arrayClass
	^WArray
! !

!WSimpleOrderedCollection methodsFor: 'converting' stamp: 'ael 11/12/2011 20:00'!
asArray
	| array |
	array := Array new: self size.
	self keysAndValuesDo: [ :i :v | array at: i put: v ].		"random order, but who cares?"
	^array
! !

!WSimpleOrderedCollection methodsFor: 'converting' stamp: 'ael 11/1/2011 10:00'!
asOrderedCollection
	^self
! !

!WSimpleOrderedCollection methodsFor: 'enumerating' stamp: 'ael 11/1/2011 18:02'!
associationsDo: aBlock
	1 to: self unloggedSize do: [ :index | aBlock value: (self associationAt: index) ].
			! !

!WSimpleOrderedCollection methodsFor: 'adding' stamp: 'ael 11/1/2011 18:06'!
at: index ifAbsentPut: block
	"Return value at index, however, if value does not exist (nil or out of bounds) then add block's value at index (growing self if necessary)"
	"only record a read of size if we had to add"
	| v oldSize |
	oldSize := self unloggedSize.
	index <= oldSize ifTrue: [
		^ (v := self at: index)
			ifNotNil: [v]
			ifNil: [self at: index put: block value]
	].
	index - self size - 1 timesRepeat: [self add: nil].		"now log access to size"
	^ self privateAt: index put: block value! !

!WSimpleOrderedCollection methodsFor: 'adding' stamp: 'ael 11/1/2011 18:07'!
at: anInteger put: anObject 
	"Put anObject at element index anInteger. at:put: cannot be used to
	append, front or back, to an ordered collection; it is used by a
	knowledgeable client to replace an element."

	(anInteger < 1 or: [anInteger > self unloggedSize])
		ifTrue: [self errorNoSuchElement]
		ifFalse: [^super at: anInteger put: anObject]! !

!WSimpleOrderedCollection methodsFor: 'world awareness' stamp: 'ael 11/1/2011 21:04'!
basicContents
	"for testing"
	^self privateSortedAssociations collect: [ :assoc | assoc value ].
! !

!WSimpleOrderedCollection methodsFor: 'accessing' stamp: 'ael 11/1/2011 21:32'!
capacity
	^WTopLevelWorld default eval: [ contents capacity ]
! !

!WSimpleOrderedCollection methodsFor: 'enumerating' stamp: 'ael 11/1/2011 13:52'!
collect: aBlock 
	"Evaluate aBlock with each of my values as the argument.  Collect the resulting values into a collection that is like me. Answer with the new collection."
	
	| newCollection |
	newCollection := self species new.
	self do: [ :each | newCollection add: (aBlock value: each value) ].
	^newCollection! !

!WSimpleOrderedCollection methodsFor: 'enumerating' stamp: 'ael 11/1/2011 18:07'!
collect: aBlock from: fromIndex to: toIndex
	"Override superclass in order to use addLast:, not at:put:."
	"Records updates for worlds"
	| result element |
	(fromIndex < 1 or: [toIndex > self unloggedSize])
		ifTrue: [^self errorNoSuchElement].
	toIndex<fromIndex ifTrue: [ ^self error: 'negative range' ].

	result := self species new.
	fromIndex to: toIndex do:
		[:index |
			element := self at: index.
			result addLast: (aBlock value: element)].
	^ result
! !

!WSimpleOrderedCollection methodsFor: 'copying' stamp: 'ael 11/1/2011 09:37'!
copyFrom: startIndex to: endIndex 
	"Answer a copy of the receiver that contains elements from position
	startIndex to endIndex."

	| targetCollection |
	endIndex < startIndex ifTrue: [^self species new].
	targetCollection := self species new.
	startIndex to: endIndex do: [:index | targetCollection addLast: (self at: index)].
	^ targetCollection! !

!WSimpleOrderedCollection methodsFor: 'copying' stamp: 'ael 11/1/2011 10:18'!
copyReplaceFrom: start to: stop with: replacementCollection 
	"Answer a copy of the receiver with replacementCollection's elements in
	place of the receiver's start'th to stop'th elements. This does not expect
	a 1-1 map from replacementCollection to the start to stop elements, so it
	will do an insert or append."

	| newOrderedCollection delta startIndex stopIndex |
	"if start is less than 1, ignore stop and assume this is inserting at the front. 
	if start greater than self size, ignore stop and assume this is appending. 
	otherwise, it is replacing part of me and start and stop have to be within my 
	bounds. "
	delta := 0.
	startIndex := start.
	stopIndex := stop.
	start < 1
		ifTrue: [startIndex := stopIndex := 0]
		ifFalse: [startIndex > self size
				ifTrue: [startIndex := stopIndex := self size + 1]
				ifFalse: 
					[(stopIndex < (startIndex - 1) or: [stopIndex > self size])
						ifTrue: [self errorOutOfBounds].
					delta := stopIndex - startIndex + 1]].
	newOrderedCollection := 
		self species new: self size + replacementCollection size - delta.
	1 to: startIndex - 1 do: [:index | newOrderedCollection add: (self at: index)].
	1 to: replacementCollection size do: 
		[:index | newOrderedCollection add: (replacementCollection at: index)].
	stopIndex + 1 to: self size do: [:index | newOrderedCollection add: (self at: index)].
	^newOrderedCollection! !

!WSimpleOrderedCollection methodsFor: 'copying' stamp: 'ael 11/1/2011 09:38'!
copyUpTo: anElement 
	"Answer all elements up to but not including anObject. If there
	is no such object, answer a copy of the receiver."

	^ self first: (self indexOf: anElement ifAbsent: [^ self copy]) - 1! !

!WSimpleOrderedCollection methodsFor: 'copying' stamp: 'ael 11/1/2011 09:38'!
copyWith: newElement 
	"Answer a copy of the receiver that is 1 bigger than the receiver and 
	includes the argument, newElement, at the end."

	| newCollection |
	newCollection _ self copy.
	newCollection add: newElement.
	^newCollection! !

!WSimpleOrderedCollection methodsFor: 'copying' stamp: 'ael 11/1/2011 09:38'!
copyWithout: oldElement 
	"Answer a copy of the receiver that does not contain any
	elements equal to oldElement."

	^ self reject: [:each | each = oldElement]

"Examples:
	'fred the bear' copyWithout: $e
	#(2 3 4 5 5 6) copyWithout: 5
"! !

!WSimpleOrderedCollection methodsFor: 'enumerating' stamp: 'ael 11/1/2011 18:19'!
do: aBlock
	1 to: self unloggedSize do: [ :index | aBlock value: (self at: index) ].
! !

!WSimpleOrderedCollection methodsFor: 'private' stamp: 'ael 11/1/2011 13:32'!
errorNoSuchElement

	self error: 'attempt to index non-existent element in an ordered collection'! !

!WSimpleOrderedCollection methodsFor: 'accessing' stamp: 'ael 11/1/2011 10:25'!
fifth
	^self at: 5! !

!WSimpleOrderedCollection methodsFor: 'accessing' stamp: 'ael 11/1/2011 10:24'!
first
	^self at: 1! !

!WSimpleOrderedCollection methodsFor: 'accessing' stamp: 'ael 11/1/2011 09:37'!
first: n
	"Answer the first n elements of the receiver.
	Raise an error if there are not enough elements."

	^ self copyFrom: 1 to: n! !

!WSimpleOrderedCollection methodsFor: 'copying' stamp: 'ael 11/1/2011 20:48'!
forceTo: length paddingStartWith: elem 
	"Force the length of the collection to length, padding  
	the beginning of the result if necessary with elem.  
	Note that this makes a copy."
	
	"World-aware behaviour: only log access to size if length is off the end, because then size determines how many padding elements are inserted"
	| newCollection padLen size |
	newCollection := self species ofSize: length.
	size := self unloggedSize.
	padLen := length - size max: 0.
	padLen > 0 ifTrue: [ self size ].
	newCollection
		from: 1
		to: padLen
		put: elem.
	newCollection
		replaceFrom: padLen + 1
		to: ((padLen + size) min: length)
		with: self
		startingAt:  1.
	^ newCollection! !

!WSimpleOrderedCollection methodsFor: 'copying' stamp: 'ael 11/1/2011 20:18'!
forceTo: length paddingWith: elem
	"Force the length of the collection to length, padding
	if necessary with elem.  Note that this makes a copy."

	"World-aware behaviour: only log access to size if length is off the end, because then size determines how many padding elements are appended"
	| newCollection size |
	size := self unloggedSize.
	length > size ifTrue: [ self size ].		"our size affects the result"
	newCollection := self species new: length withAll: elem.
	newCollection replaceFrom: 1 to: (size min: length) with: self startingAt: 1.
	^ newCollection! !

!WSimpleOrderedCollection methodsFor: 'accessing' stamp: 'ael 11/1/2011 10:25'!
fourth
	^self at: 4! !

!WSimpleOrderedCollection methodsFor: 'accessing' stamp: 'ael 11/1/2011 20:49'!
from: startIndex to: endIndex put: anObject
	"Put anObject in all indexes between startIndex 
	and endIndex. Answer anObject"

	| written toWrite thisWrite |

	startIndex > endIndex ifTrue: [^self].
	self at: startIndex put: anObject.
	written := 1.
	toWrite := endIndex - startIndex + 1.
	[written < toWrite] whileTrue:
		[
			thisWrite := written min: toWrite - written.
			self 
				replaceFrom: startIndex + written
				to: startIndex + written + thisWrite - 1
				with: self startingAt: startIndex.
			written := written + thisWrite
		].
	^anObject! !

!WSimpleOrderedCollection methodsFor: 'private' stamp: 'ael 11/1/2011 21:21'!
grownBy: length
	"Answer a copy of receiver collection with size grown by length.  This is only sent when streaming onto a collection.  People shouldn't be doing that onto one of these collections.  But, just to humour the test case, we can grow a bit."
	10 timesRepeat: [ self add: nil ].
	^self

! !

!WSimpleOrderedCollection methodsFor: 'accessing' stamp: 'ael 11/1/2011 21:05'!
includes: anElement 
	"Return true if the element is in the collection"
	"World-awareness: if the item is found, only the element in question (the earliest such element in the list) is marked as having been read.  This guarantees that the serialisability check will only be approved if the parent world still includes at least this one instance of the sought value.  On the other hand, if the value is not found we mark all items, to avoid a situation where a decision made on the absence of the item would be incompatible with other changes in the parent world."
	self privateSortedAssociations
	  do: [ :assoc |
		assoc value = anElement
		  ifTrue: [
			self at: assoc key.		"ensure the read is recorded"
			^true
			]
		].

	"didn't find it.  go back through again, this time recording everything"
	self do: [ :x | ].
	^false
! !

!WSimpleOrderedCollection methodsFor: 'accessing' stamp: 'ael 11/1/2011 09:19'!
indexOf: x
	self withIndexDo: [ :elem :index | elem = x ifTrue: [^ index]].
	^ 0! !

!WSimpleOrderedCollection methodsFor: 'accessing' stamp: 'ael 11/1/2011 09:20'!
indexOf: anElement ifAbsent: exceptionBlock
	"Answer the index of the first occurence of anElement within the  
	receiver. If the receiver does not contain anElement, answer the 
	result of evaluating the argument, exceptionBlock."

	^ self indexOf: anElement startingAt: 1 ifAbsent: exceptionBlock! !

!WSimpleOrderedCollection methodsFor: 'accessing' stamp: 'ael 11/1/2011 18:19'!
indexOf: anElement startingAt: start ifAbsent: exceptionBlock
	start to: self unloggedSize do:
		[:index |
		(self at: index) = anElement ifTrue: [^ index]].

	^ exceptionBlock value! !

!WSimpleOrderedCollection methodsFor: 'enumerating' stamp: 'ael 11/1/2011 18:19'!
keysAndValuesDo: aBlock 
	1 to: self unloggedSize do: [:index | aBlock value: index value: (self at: index)].
! !

!WSimpleOrderedCollection methodsFor: 'accessing' stamp: 'ael 11/1/2011 09:21'!
last
	| s |
	s := self size.		"record access, because it determines which index we look at"
	s = 0 ifTrue: [^ self errorNoSuchElement].
	^ self at: s! !

!WSimpleOrderedCollection methodsFor: 'accessing' stamp: 'ael 11/1/2011 09:59'!
last: n
	"Answer the last n elements of the receiver.  
	Raise an error if there are not enough elements."

	| size |
	size := self size.		"record access, because it determines which indices we look at"
	^ self copyFrom: size - n + 1 to: size! !

!WSimpleOrderedCollection methodsFor: 'private' stamp: 'ael 11/1/2011 18:19'!
privateAt: anInteger put: anObject 
	"Put anObject at element index anInteger. at:put: cannot be used to
	append, front or back, to an ordered collection; it is used by a
	knowledgeable client to replace an element."

	(anInteger > 0 and: [ anInteger <= (self unloggedSize + 1)])
		ifTrue: [^super at: anInteger put: anObject].
	
	self errorNoSuchElement
! !

!WSimpleOrderedCollection methodsFor: 'private' stamp: 'ael 12/5/2011 23:19'!
privateSortedAssociations
	"return an OrderedCollection, because that's what the test cases expect"
	| assocs |
	assocs := self worldFlattenedAssociationsIn: self thisWorld.
	^(assocs sort: [ :x :y | x key < y key ]) asOrderedCollection
! !

!WSimpleOrderedCollection methodsFor: 'removing' stamp: 'ael 11/1/2011 09:23'!
remove: oldObject 
	"Remove oldObject from the receiver's elements. Answer oldObject 
	unless no element is equal to oldObject, in which case, raise an error.
	ArrayedCollections cannot respond to this message."

	^ self remove: oldObject ifAbsent: [self errorNotFound: oldObject]! !

!WSimpleOrderedCollection methodsFor: 'removing' stamp: 'ael 11/1/2011 09:23'!
remove: oldObject ifAbsent: absentBlock
	"Records updates for worlds
		- here we just record the elements that were seen when looking for oldObject
		- in the removeAt: method we'll record the deletion and any shifting"

	| index |
	index := self indexOf: oldObject ifAbsent: [ ^absentBlock value ].

	^self removeAt: index
! !

!WSimpleOrderedCollection methodsFor: 'removing' stamp: 'ael 11/1/2011 18:21'!
removeAt: index
	| removed oldSize |
	oldSize := self unloggedSize.
	removed := self removeKey: index.
	index+1 to: oldSize do: [ :i | self privateAt: i-1 put: (self at: i) ].
	index = oldSize ifFalse: [ self removeKey: oldSize ].
	^removed
	! !

!WSimpleOrderedCollection methodsFor: 'removing' stamp: 'ael 11/1/2011 09:25'!
removeFirst
	"Remove the first element of the receiver and answer it. If the receiver is 
	empty, create an error notification."
	self emptyCheck.
	^self removeAt: 1
! !

!WSimpleOrderedCollection methodsFor: 'removing' stamp: 'ael 11/1/2011 10:23'!
removeFirst: n 
	"Remove first n object into an array"
	| list |
	list := self arrayClass new: n.
	1
		to: n
		do:
			[ : i | list
				at: i
				put: self removeFirst ].
	^ list! !

!WSimpleOrderedCollection methodsFor: 'removing' stamp: 'ael 11/1/2011 09:25'!
removeLast
	self emptyCheck.
	^self removeAt: self size
! !

!WSimpleOrderedCollection methodsFor: 'removing' stamp: 'ael 11/1/2011 10:24'!
removeLast: n 
	"Remove last n object into an array with last in last position"
	| list |
	list := self arrayClass new: n.
	n
		to: 1
		by: -1
		do:
			[ : i | list
				at: i
				put: self removeLast ].
	^ list! !

!WSimpleOrderedCollection methodsFor: 'accessing' stamp: 'ael 11/1/2011 20:47'!
replaceFrom: start to: stop with: replacement startingAt: repStart 
	"This destructively replaces elements from start to stop in the receiver 
	starting at index, repStart, in the sequenceable collection, 
	replacementCollection. Answer the receiver. No range checks are 
	performed."

	| index repOff |
	repOff := repStart - start.
	index := start - 1.
	[(index := index + 1) <= stop]
		whileTrue: [self at: index put: (replacement at: repOff + index)]! !

!WSimpleOrderedCollection methodsFor: 'enumerating' stamp: 'ael 11/1/2011 09:26'!
reverseDo: aBlock 
	| size |
	size := self size.		"record read of size, because that determines where the accesses start"
	size to: 1 by: -1 do: [ :index | aBlock value: (self at: index) ]
! !

!WSimpleOrderedCollection methodsFor: 'copying' stamp: 'ael 11/1/2011 09:26'!
reversed
	"Answer a copy of the receiver with element order reversed.  "
	| newCol |
	newCol := self species new.
	self reverseDo:
		[:elem | newCol addLast: elem].
	^ newCol

"#(2 3 4 'fred') reversed"! !

!WSimpleOrderedCollection methodsFor: 'accessing' stamp: 'ael 11/1/2011 10:24'!
second
	^self at: 2! !

!WSimpleOrderedCollection methodsFor: 'accessing' stamp: 'ael 11/1/2011 10:25'!
sixth
	^self at: 6! !

!WSimpleOrderedCollection methodsFor: 'accessing' stamp: 'ael 11/1/2011 10:25'!
third
	^self at: 3! !

!WSimpleOrderedCollection methodsFor: 'adding' stamp: 'ael 11/12/2011 15:52'!
uncheckedAt: anInteger put: anObject 
	"for testing.
	Put anObject at element index anInteger, assuming it's within our range"
	^super at: anInteger put: anObject
! !

!WSimpleOrderedCollection methodsFor: 'enumerating' stamp: 'ael 11/1/2011 18:23'!
with: otherCollection collect: twoArgBlock 
	"Collect and return the result of evaluating twoArgBlock with 
	corresponding elements from this collection and otherCollection."
	| result size |
	size := self unloggedSize.
	otherCollection size = size ifFalse: [self error: 'otherCollection must be the same size'].
	result := self species new: size.
	1 to: size do:
		[:index | result addLast: (twoArgBlock value: (self at: index)
									value: (otherCollection at: index))].
	^ result! !

!WSimpleOrderedCollection methodsFor: 'enumerating' stamp: 'ael 11/1/2011 18:23'!
withIndexDo: aBlock
	1 to: self unloggedSize do: [ :index | aBlock value: (self at: index) value: index ].
	! !

!WSimpleOrderedCollection methodsFor: 'world awareness' stamp: 'ael 11/12/2011 20:09'!
worldFlattenedCopy
	"get a flattened copy of the receiver, keeping appropriate records of reads from parent worlds"
	self thisWorld isTopLevel ifTrue: [ ^self clone postCopy ].

	^self class privateNewFromIDict: (IdentityDictionary withAll: self associations)
! !

!WObjectWithCachedLookup class methodsFor: 'instance variables' stamp: 'ael 9/18/2011 15:09'!
firstRedirectedIV
	^ 2
! !

!WObjectWithCachedLookup methodsFor: 'world awareness' stamp: 'ael 10/27/2011 14:39'!
commitInto: parentWorld using: updateTrackers
	"updateTrackers is a tuple { orig. delta }.
	if the parent world is not top-level, we can save a bit of time by caching its trackers (especially since we won't be looking up the child's trackers after this point)"
	| delta val parentTrackers dnk |
	parentWorld isTopLevel ifTrue: [ ^super commitInto: parentWorld using: updateTrackers ].
	
	delta := updateTrackers at: 2.
	delta ifNotNil: [
		dnk := self dnkToken.
		parentTrackers := nil.
		self class firstCommittableIV to: self iSize do: [:ii |
			val := delta at: ii.
			val == dnk
			  ifFalse: [
				parentWorld
					valueAt: ii
					put: val
					for: self
					using: (parentTrackers ifNil: [ parentTrackers := self updateTrackersForWorld: parentWorld createIfAbsent: true caching: true ])
				]
			]
		]! !

!WObjectWithCachedLookup methodsFor: 'debugging' stamp: 'ael 5/21/2012 21:30'!
debugAllUpdateTrackerStates
	^self thisWorld withAllAncestors allButLast collect: [ :w | self debugUpdateTrackerStateIn: w ].
! !

!WObjectWithCachedLookup methodsFor: 'debugging' stamp: 'ael 5/21/2012 21:29'!
debugUpdateTrackerStateIn: aWorld
	| trackers |
	trackers := self updateTrackersForWorld: aWorld createIfAbsent: false caching: false.
	^trackers ifNotNil: [ trackers collect: [ :x | x clone ] ].
! !

!WObjectWithCachedLookup methodsFor: 'initialize-release' stamp: 'ael 10/30/2011 22:20'!
initLookupCache
	lookupCache := WeakArray new: 2.
! !

!WObjectWithCachedLookup methodsFor: 'initialize-release' stamp: 'ael 10/30/2011 22:20'!
initialize
	super initialize.
	self initLookupCache
! !

!WObjectWithCachedLookup methodsFor: 'accessing' stamp: 'ael 9/26/2011 12:35'!
instVarAt: index
	| thisWorld trackers |
	(thisWorld := self thisWorld) isTopLevel ifTrue: [ ^self basicValueAt: index ].
	
	"if we already have update trackers for this world in our cache, use that. 
	but don't force creation of trackers: in a world that does not cache reads,
	we may never need them."
	trackers := self updateTrackersForWorld: thisWorld createIfAbsent: false caching: true.
	^thisWorld valueAt: index for: self using: trackers.
	! !

!WObjectWithCachedLookup methodsFor: 'accessing' stamp: 'ael 9/30/2011 21:44'!
instVarAt: index put: value
	| thisWorld trackers |
	(thisWorld := self thisWorld) isTopLevel ifTrue: [ ^self basicValueAt: index put: value ].

	trackers := self updateTrackersForWorld: thisWorld createIfAbsent: true caching: true.
	^thisWorld valueAt: index put: value for: self using: trackers.
! !

!WObjectWithCachedLookup methodsFor: 'copying' stamp: 'ael 10/30/2011 22:20'!
postCopy
	"set up our own cache"
	self initLookupCache
! !

!WObjectWithCachedLookup methodsFor: 'exceptions' stamp: 'ael 12/6/2011 15:02'!
reportSerializationErrorAt: index oldValue: readValue newValue: currentValue
	^(WorldValueChanged new indexPrefix: 'IV' index: (self class allInstVarNames at: index) readValue: readValue currentValue: currentValue) signal! !

!WObjectWithCachedLookup methodsFor: 'world awareness' stamp: 'ael 12/1/2011 22:24'!
updateTrackersForWorld: aWorld createIfAbsent: create caching: useCache
	| cache trackers |
	useCache
	  ifTrue: [
		lookupCache ifNil: [ self initLookupCache ].		"we could force classes to use initialize, but..."
		cache := lookupCache.		"NB: not a redirected lookup"
		(cache at: 1) == aWorld
		  ifTrue: [
			"double-check that the cache is non-nil (given that it's only held weakly)"
			(trackers := cache at: 2) ifNotNil: [ ^trackers ].
			].
		cache at: 1 put: nil.		"be sure to remove any previous entry"
		].
	
	trackers := aWorld updateTrackersFor: self createIfAbsent: create.
	trackers ifNil: [ ^nil ].

	useCache
	  ifTrue: [
		cache at: 2 put: trackers.
		cache at: 1 put: aWorld.
		].
	^trackers
! !

!WObjectWithCachedLookup methodsFor: 'copying' stamp: 'ael 10/30/2011 22:30'!
veryDeepInner: deepCopier
	super veryDeepInner: deepCopier.

	self initLookupCache.
! !

!WObjectWithCachedLookup methodsFor: 'copying' stamp: 'ael 11/12/2011 11:50'!
worldFlattenedCopy
	"get a flattened copy of the receiver by looking up each value in turn - thus keeping appropriate records of reads from parent worlds"
	| clone thisWorld trackers |
	(thisWorld := self thisWorld) isTopLevel ifTrue: [ ^self clone postCopy ].

	clone := self initializedClone.
	trackers := self updateTrackersForWorld: thisWorld createIfAbsent: false caching: true.
	WTopLevelWorld default eval: [
		1 to: self iSize do: [ :index | clone basicValueAt: index put: (thisWorld valueAt: index for: self using: trackers) ].
		clone postFlattenedCopy.
		].
	^clone
! !

Object subclass: #WWorld
	instanceVariableNames: 'id parent updateTrackerDict tracksReads deepUpdateDict delegationDepth hasChildren'
	classVariableNames: 'FlattenBlockSize NextId UsePlugin'
	poolDictionaries: ''
	category: 'Worlds'!

!WWorld commentStamp: 'ael 3/1/2012 21:05' prior: 0!
WWorld initialize.

4 timesRepeat: [Smalltalk garbageCollect].
WTopLevelWorld instanceCount == 1 and: [
WWorld instanceCount == 0
]
!

WWorld subclass: #WTopLevelWorld
	instanceVariableNames: ''
	classVariableNames: 'TheOneAndOnlyTopLevelWorld'
	poolDictionaries: ''
	category: 'Worlds'!

!WTopLevelWorld class methodsFor: 'as yet unclassified' stamp: 'aw 4/15/2010 11:13'!
default

	^ TheOneAndOnlyTopLevelWorld ifNil: [TheOneAndOnlyTopLevelWorld := self new]! !

!WTopLevelWorld methodsFor: 'as yet unclassified' stamp: 'aw 3/31/2010 14:24'!
commit! !

!WTopLevelWorld methodsFor: 'copying' stamp: 'ael 7/12/2011 15:18'!
copy
	^self
! !

!WTopLevelWorld methodsFor: 'as yet unclassified' stamp: 'ael 7/2/2012 22:00'!
delegatedTrackerWorld
	self shouldNotImplement
! !

!WTopLevelWorld methodsFor: 'as yet unclassified' stamp: 'ael 7/2/2012 22:40'!
delegationDepth
	^0
! !

!WTopLevelWorld methodsFor: 'as yet unclassified' stamp: 'ael 6/27/2011 20:56'!
ensureReadsRecordFor: obj

	^ nil
! !

!WTopLevelWorld methodsFor: 'as yet unclassified' stamp: 'ael 6/16/2011 16:25'!
ensureUpdateTrackersFor: obj

	^ nil! !

!WTopLevelWorld methodsFor: 'as yet unclassified' stamp: 'ael 6/16/2011 16:34'!
ensureWriteTargetFor: obj

	^ obj! !

!WTopLevelWorld methodsFor: 'as yet unclassified' stamp: 'ael 1/13/2012 15:58'!
initialize
	self initializeID.
	tracksReads := false.
! !

!WTopLevelWorld methodsFor: 'as yet unclassified' stamp: 'ael 10/1/2011 11:32'!
isDefunct
	^false
! !

!WTopLevelWorld methodsFor: 'as yet unclassified' stamp: 'ael 6/25/2011 00:05'!
isTopLevel
	^true
! !

!WTopLevelWorld methodsFor: 'as yet unclassified' stamp: 'ael 9/26/2011 17:22'!
nonCachingInstVarAt: index for: obj writesSlot: writesSlot readsSlot: readsSlot

	^ obj basicValueAt: index
! !

!WTopLevelWorld methodsFor: 'copying' stamp: 'ael 7/13/2011 23:33'!
postCopyFor: clone from: original
	"nothing to do"! !

!WTopLevelWorld methodsFor: 'as yet unclassified' stamp: 'ael 7/9/2012 23:35'!
sprout

	| child |
	child := WWorld new parent: self.
	child deepUpdateDict: (WeakIdentityKeyDictionary new: 100).
	child delegationDepth: self delegationDepth + 1.
	^child
! !

!WTopLevelWorld methodsFor: 'as yet unclassified' stamp: 'ael 6/25/2011 00:10'!
tracksReads
	^false
! !

!WTopLevelWorld methodsFor: 'as yet unclassified' stamp: 'ael 6/16/2011 16:49'!
updateTrackersFor: obj

	^ nil! !

!WTopLevelWorld methodsFor: 'as yet unclassified' stamp: 'ael 10/3/2011 19:31'!
valueAt: index put: value for: obj
	"NB: must be evaluated in this world"
	^ obj basicValueAt: index put: value! !

!WTopLevelWorld methodsFor: 'as yet unclassified' stamp: 'ael 6/27/2011 20:57'!
writesRecordFor: obj

	^ obj! !

!WWorld class methodsFor: 'as yet unclassified' stamp: 'ael 3/2/2012 10:51'!
flattenBlockSize: size
	"FlattenBlockSize is a remnant from our investigation into tweaking the efficiency of a flattening operation by taking advantage of the CPU's memory cache.  When flattening a large array-like object (e.g., a WBitmap), accessing a block of several elements in sequence can be faster than random access interspersed with other memory accesses.  However, the results of our experiments showed that in most cases the 'sparse' algorithm (which doesn't try to group accesses) performs well enough.

	Valid values for FlattenBlockSize are:
		0: old bottom-up (leaf world to top world) breadth-first strategy (i.e., figure out all slots affected by a given world before moving to the next world)
		-1: (default) top-down depth-first strategy (i.e., figure out which world contributes the value for each slot before moving to the next slot), found to be best for objects in which no more than 20% of slots are typically overwritten at each child world
		n: also top-down depth-first, but dealing with n slots at a time
		  
	A little more documentation is available in the method comments in the WorldsPlugin package."

	FlattenBlockSize := size
! !

!WWorld class methodsFor: 'as yet unclassified' stamp: 'ael 3/6/2012 23:39'!
initialize
	"
		self initialize
	"
	NextId := 1.
	UsePlugin := true.
	self flattenBlockSize: -1.		"in general, use the flattening method for sparse per-world changes (see comment in flattenBlockSize:)" 
	Processor activeProcess currentWWorld: WTopLevelWorld default.! !

!WWorld class methodsFor: 'as yet unclassified' stamp: 'ael 11/2/2011 08:53'!
runWithoutPlugin: aBlock
	| prev |
	prev := UsePlugin.
	UsePlugin := false.
	[ aBlock value ]
	  ensure: [ UsePlugin := prev ]

! !

!WWorld class methodsFor: 'as yet unclassified' stamp: 'aw 4/15/2010 11:16'!
thisWorld

	^ Processor activeProcess currentWWorld! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 1/13/2012 23:28'!
allUpdates
	| updateDict newTrackers |
	updateDict := IdentityDictionary new.
	updateTrackerDict keysAndValuesDo: [ :object :trackers |
		(newTrackers := object iUpdatesIn: trackers)
		  ifNotNil: [ updateDict at: object put: newTrackers ]
		].
	^updateDict
! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 10/31/2011 20:22'!
ancestorCopyingValueAt: index for: obj using: updateTrackers proxySlot: proxySlot cacheRead: cacheRead
	"For some instance variables, such as the contents in WCollection and the updatedVector in LObject, we want special handling because the value is not a world-aware object and for efficiency we don't wish to make it so.  But we mustn't use the same object in a child world as in a parent (because then the child would wag the parent, and potentially all the child's siblings).  So, unless the variable is already held in the proxy slot in the updateTrackers, we record in the proxy a copy of the variable found from the ancestor chain, and optionally put in Reads a further copy that can be compared as part of the commit-time serialisability check.
	NB: this method assumes that updateTrackers is non-nil; the caller must ensure it is so"
	| proxy v readsHere |
"Transcript cr; show: 'request IV ', index printString."

	proxy := updateTrackers at: proxySlot.
	proxy
	  ifNotNil: [
		v := proxy at: index.
		v == obj dnkToken ifFalse: [ ^v ].		"already installed"
		].

	"look only in the proxy slot in parent worlds, and don't do any caching"	
	v := self lookupFor: obj at: index caching: false in: updateTrackers writesSlot: proxySlot readsSlot: 0.

	"assume we found a value.  store a copy in the reads, if we have been asked to do so"
	cacheRead
	  ifTrue: [
		readsHere := obj iEnsureReadsRecordIn: updateTrackers.
		readsHere at: index put: v copy.
		].

	"store another copy in the proxy, and return this for use by the caller"
	proxy ifNil: [ proxy := updateTrackers at: proxySlot put: obj iMakeDeltaOrOrig ].

	^proxy at: index put: v copy
! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 3/1/2012 21:43'!
assertCommitIsSerializableIn: target

	| obj |
	updateTrackerDict associations do: [ :objAndTrackers |
		(obj := objAndTrackers key)
		  ifNotNil: [ (obj isSerializableIn: target using: objAndTrackers value)
			  ifFalse:  [self error: 'commit aborted -- failed serializability check']]
		].! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 3/1/2012 21:49'!
commit
	"traditional commit, into the parent world"
	self eval: [
		self commitInto: parent.
		self initializeTrackers.
		].
! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 7/10/2012 00:24'!
commitInto: target
	"NB: always evaluated in the receiver world"
	| obj |
	self assertCommitIsSerializableIn: target.
	updateTrackerDict associations do: [ :objAndTrackers |
		(obj := objAndTrackers key) ifNotNil: [ 
			obj copyMissingOrigValuesInto: target using: objAndTrackers value ] ].
	updateTrackerDict associations do: [ :objAndTrackers  |
		(obj := objAndTrackers key) ifNotNil: [
			obj commitInto: target using: objAndTrackers value ] ].

	deepUpdateDict associations do: [ :objAndTrackers  |
		(obj := objAndTrackers key) ifNotNil: [
			obj commitInto: target using: objAndTrackers value ] ].
	target deepUpdateDict: deepUpdateDict.	
! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 5/15/2013 16:03'!
copyWithWritesOnly
	"create a new sibling world that has all the recorded writes, but none of the recorded reads, of the receiver.   This re-opens it to changes in the parent (when using non-snapshot sprouts).
	NB: a world created with this operation should never be used as the source of changes merged (or committed) into another world, because the serialisability check will no longer be meaningful.
	"
	| newWorld |
	newWorld := parent sprout.
	newWorld mergeUpdatesFrom: self.
	^newWorld
! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 7/9/2012 23:33'!
deepUpdateDict
	^deepUpdateDict
! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 7/9/2012 23:33'!
deepUpdateDict: d
	deepUpdateDict := d
! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 7/2/2012 22:39'!
delegationDepth
	^delegationDepth
! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 7/2/2012 22:39'!
delegationDepth: d
	delegationDepth := d
! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 6/27/2011 20:51'!
ensureReadsRecordFor: obj

	| updateTrackers |
	tracksReads ifFalse: [ ^nil ].

	updateTrackers := self ensureUpdateTrackersFor: obj.
	^ obj iEnsureReadsRecordIn: updateTrackers.
! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 6/23/2011 10:20'!
ensureUpdateTrackersFor: obj
	^ updateTrackerDict at: obj ifAbsentPut: [ 
"Transcript cr; show: 'creating trackers for ', obj oopString, ' in world ', self oopString."
		Array new: obj iUpdateTrackerSize withAll: nil ]! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 6/16/2011 16:34'!
ensureWriteTargetFor: obj

	| updateTrackers |
	updateTrackers := self ensureUpdateTrackersFor: obj.
	^ obj iEnsureWriteTargetIn: updateTrackers.
! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 5/2/2012 11:29'!
eval: aBlock
	"we send valueWithEnoughArguments: to handle the case where the supplied block 
	takes a dummy argument - as will be the case, for example, in a block representing
	a function compiled by KSObject"
	| oldWorld ans |
	oldWorld := Processor activeProcess currentWWorld.
	
	Processor activeProcess currentWWorld: self.
	[ ans := aBlock numArgs = 0
		ifTrue: [ aBlock value ]
		ifFalse: [ aBlock value: nil ]
	] ensure: [ Processor activeProcess currentWWorld: oldWorld ].
	^ ans! !

!WWorld methodsFor: 'debugging' stamp: 'ael 6/28/2011 09:47'!
findChangedObjects
	^IdentityDictionary withAll: (updateTrackerDict associations select: [ :a | a key notNil and: [ a key iChangedOrNewInWorld: self ] ])
! !

!WWorld methodsFor: 'debugging' stamp: 'ael 6/28/2011 09:46'!
findUnchangedObjects
	^IdentityDictionary withAll: (updateTrackerDict associations reject: [ :a | a key isNil or: [ a key iChangedOrNewInWorld: self ] ])
! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 3/6/2012 23:39'!
flatten: obj into: dst
	"dst is the object into which writes should be made (i.e., for objects with indirect contents - such as WArray - we are passed the contents object)"
	| dnk startIndex w worlds writesSlot |
	startIndex := obj class firstCommittableIV.
	dnk := obj dnkToken.
	UsePlugin
	  ifFalse: [ self nonPluginFlatten: obj into: dst startingAt: startIndex delegate: dnk ]
	  ifTrue: [
		writesSlot := obj class writesSlot.
		"choose which algorithm to use.  see comment in flattenBlockSize:"
		FlattenBlockSize = 0
		  ifTrue: [ self primFlatten: obj into: dst startingAt: startIndex writesSlot: writesSlot delegate: dnk ]
		  ifFalse: [
			worlds := OrderedCollection new.
			w := self.
			[ w isNil ] whileFalse: [ worlds addFirst: w. w := w parent ].	"1st is the top world"

			FlattenBlockSize = -1		"sparse"
			  ifTrue: [ obj primSparseFlattenInWorlds: worlds asArray to: dst startingAt: startIndex writesSlot: writesSlot delegate: dnk ]
			  ifFalse: [ obj primDepthFirstInBlocksFlattenInWorlds: worlds asArray to: dst startingAt: startIndex blockSize: FlattenBlockSize writesSlot: writesSlot delegate: dnk.
				].
			]
		]! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'aw 9/1/2009 15:01'!
id

	^ id! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 7/2/2012 21:53'!
initialize

	self initializeID.
	self initializeTrackers.
	tracksReads := true.
	hasChildren := false.
! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 10/3/2011 19:39'!
initializeID

	id := NextId.
	NextId := NextId + 1.
! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 3/1/2012 21:22'!
initializeTrackers
	updateTrackerDict := WeakIdentityKeyDictionary new: 100.
! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 6/25/2011 00:05'!
isTopLevel
	^false
! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 12/22/2011 21:50'!
lookupFor: obj at: index caching: cacheReads in: trackersOrNil writesSlot: writesSlot readsSlot: readsSlot
	| dnk readsHere updateTrackers v resultHolder |
	dnk := obj dnkToken.
	
	UsePlugin ifFalse: [ 
		^self nonPluginLookupFor: obj at: index caching: cacheReads in: trackersOrNil writesSlot: writesSlot readsSlot: readsSlot delegate: dnk.
		].
	
	resultHolder := #(nil nil).
	self primFor: obj at: index in: trackersOrNil writesSlot: writesSlot readsSlot: readsSlot delegate: dnk resultHolder: resultHolder.

	"if the value was found in a parent world, cache it in our Reads (if appropriate) and return it"
	v := resultHolder at: 1.
	v == dnk
	  ifTrue: [			"the primitive failed.  resort to non-primitive lookup"
		^self nonPluginLookupFor: obj at: index caching: cacheReads in: trackersOrNil writesSlot: writesSlot readsSlot: readsSlot delegate: dnk.
		].

	"position 2 of the resultHolder tells us whether the value came from an ancestor world.  check now, if we care about it"
	(tracksReads and: [ cacheReads and: [ resultHolder at: 2 ] ])
	  ifTrue: [	"we can record reads, but might not have the record yet"
		updateTrackers := trackersOrNil ifNil: [ self ensureUpdateTrackersFor: obj ].
		readsHere := obj iEnsureReadsRecordIn: updateTrackers.
		readsHere at: index put: v.		"bound to be overwriting a dnk"
		].

	^ v

	! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 1/13/2012 22:41'!
mergeUpdatesFrom: aWorld
	"simplest form of merge: all writes from the supplied world are incorporated without question"
	| oldTrackers transfers |
	transfers := aWorld allUpdates.
	self eval: [
		transfers keysAndValuesDo: [ :object :transferTrackers |
			oldTrackers := self ensureUpdateTrackersFor: object.
			object iMergeUpdatesFrom: transferTrackers into: oldTrackers.
			].
		].
! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 10/31/2011 20:12'!
nonCachingInstVarAt: index for: obj writesSlot: writesSlot readsSlot: readsSlot

	^ self lookupFor: obj at: index caching: false in: (self updateTrackersFor: obj) writesSlot: writesSlot readsSlot: readsSlot
		! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 11/2/2011 19:24'!
nonPluginFlatten: obj into: dst startingAt: startIndex delegate: delegate

	| d w dstSize numLeft v tmpDst |

	w := self.
	(dst class ~~ obj class or: [(dstSize := dst iSize) ~~ obj iSize]) ifTrue: [
		self error: 'destination object not the same class or size for flatten operation'].

	"rather than iterating multiple times through dst, a world-aware object, for speed we build up the values in an Array then copy them into dst in one shot at the end"
	tmpDst := Array new: dstSize withAll: delegate.
	numLeft := dstSize.
	[numLeft > 0] whileTrue: [
		w isTopLevel
		  ifTrue: [
			startIndex to: dstSize do: [:index |
				(tmpDst at: index) == delegate
				  ifTrue: [ tmpDst at: index put: (obj basicValueAt: index) ].
				].
			numLeft := 0.		"force us out of the loop"
			].
			
		d := w writesRecordFor: obj.
		d ifNotNil: [
			startIndex to: dstSize do: [:index |
				(numLeft > 0 and: [ (tmpDst at: index) == delegate ])
				  ifTrue: [
					v := d at: index.			"the writes record is always a plain object (non-WObject)"
					v ~~ delegate ifTrue: [
						numLeft := numLeft - 1.
						tmpDst at: index put: v].
					]
				]
			].
		w := w parent.
		].

	WTopLevelWorld default eval: [ startIndex to: dstSize do: [ :index | dst basicValueAt: index put: (tmpDst at: index) ] ].

	^ dst! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 11/2/2011 08:53'!
nonPluginLookupFor: obj at: index caching: cacheReads in: trackersOrNil writesSlot: writesSlot readsSlot: readsSlot delegate: dnk
	"trackersOrNil will only be nil if this world doesn't have trackers for the object yet"
	| w v readsHere writesHere updateTrackers reads trackers writes |

	readsHere := nil.
	trackersOrNil
	  ifNotNil: [		"we may already have a Writes or Reads record for this IV"
		"first look at the Writes"
		writesHere := trackersOrNil at: writesSlot.
		writesHere ifNotNil: [
			v := writesHere at: index.
			v == dnk ifFalse: [ ^ v ].
			].
		"then at the Reads, if there is a slot"
		readsSlot = 0
		  ifFalse: [
			readsHere := trackersOrNil at: readsSlot.
			readsHere ifNotNil: [
				v := readsHere at: index.
				v == dnk ifFalse: [ ^ v ].
				].
			].
		].

	"next search the ancestor worlds"	
	v := dnk.			"unless we find something"
	w := parent.
	[w notNil] whileTrue: [
		w isTopLevel
		  ifTrue: [ v := obj basicValueAt: index ]
		  ifFalse: [
			trackers := w updateTrackersFor: obj.
			trackers
			  ifNotNil: [
				"first look at the Writes"
				writes := trackers at: writesSlot.
				writes ifNotNil: [ v := writes at: index ].

				(v == dnk and: [ readsSlot ~= 0 ])
				  ifTrue: [ "not found yet, so try the Reads"
					reads := trackers at: readsSlot.
					reads ifNotNil: [ v := reads at: index ].
					].
				].
			].

		"if we found a value, cache it in our Reads (if appropriate) and return it"
		v == dnk ifFalse: [
			tracksReads & cacheReads
			  ifTrue: [	"we can record reads, but might not have the record yet"
				updateTrackers := trackersOrNil ifNil: [ self ensureUpdateTrackersFor: obj ].
				readsHere ifNil: [ readsHere := obj iEnsureReadsRecordIn: updateTrackers ].
				readsHere at: index put: v.		"bound to be overwriting a dnk"
				].
			^ v
			].

		"not found in this world; try its parent"
		w := w parent
		].

	self error: 'Instance variable not found'! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'aw 8/24/2009 13:55'!
parent

	^ parent! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'aw 4/5/2010 16:10'!
parent: p

	parent ifNotNil: [self error: 'a world''s parent pointer cannot be reset'].
	parent := p.! !

!WWorld methodsFor: 'copying' stamp: 'ael 5/16/2013 22:03'!
postCopy
	"NB: doesn't yet support worlds that are using deepUpdateDict"
	| newUTs oldTrackerDict |
	super postCopy.

	self initializeID.
	oldTrackerDict := updateTrackerDict.
	self initializeTrackers.
	oldTrackerDict keysAndValuesDo: [ :obj :oldUTs |
		obj 
		  ifNotNil: [
			newUTs := Array new: obj iUpdateTrackerSize withAll: nil.
			obj postCopyTrackers: oldUTs into: newUTs.
			updateTrackerDict at: obj put: newUTs.
			].
		].! !

!WWorld methodsFor: 'primitives' stamp: 'ael 11/9/2011 18:07'!
primFlatten: obj into: dst startingAt: startIndex writesSlot: writesSlot delegate: dnk
	<primitive: 'primFlattenObjInto' module: 'WorldsPlugin'>
	self nonPluginFlatten: obj into: dst startingAt: startIndex delegate: dnk! !

!WWorld methodsFor: 'primitives' stamp: 'ael 12/22/2011 21:49'!
primFor: obj at: index in: trackersOrNil writesSlot: writesSlot readsSlot: readsSlot delegate: dnk resultHolder: holder
	<primitive: 'primInstVarAt' module: 'WorldsPlugin'>
	holder at: 1 put: dnk		"tell the sender we failed"
! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'aw 8/24/2009 11:24'!
printOn: aStream

	self class printOn: aStream.
	aStream nextPut: $(; nextPutAll: id printString; nextPut: $).! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 10/30/2011 15:13'!
removeUpdateTrackersFor: obj
	updateTrackerDict removeKey: obj ifAbsent: [ ]
! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 12/5/2011 23:53'!
reparent: p
	"added by ael.  use with caution."
	parent := p.! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 5/15/2013 15:55'!
snapshotAndSprout
	"a potentially useful operation for some applications (though the current implementation is clumsy).

	The receiver is a world whose application state the user wants to replicate, so that the receiver and its new sibling start with the same state but can make their own independent modifications to it.

	The shared state is entrusted to a new world, snapshotWorld, that is inserted between the receiver and its existing parent to become the parent of both the receiver and the sibling.
	
	It might make more sense to split this operation into two: snapshotAndContinue to insert the snapshotWorld, then sprout (however many times you like) from that.
	"
	| snapshotWorld newTrackers |
	snapshotWorld := self class new parent: parent.
	parent := snapshotWorld.
	newTrackers := snapshotWorld updateTrackerDict.
	snapshotWorld updateTrackerDict: updateTrackerDict.
	updateTrackerDict := newTrackers.

	^snapshotWorld sprout.
! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'sn 9/10/2013 02:42'!
sprout
	| child newDeepUpdateDict obj parentDepth |

	child := self class new parent: self.
	parentDepth := self delegationDepth.
	child delegationDepth: parentDepth + 1.
	"If the receiver already had children, prepare a new deep-tracker dictionary to be used by the new child and any children it may acquire.  Note that this secondary chain of worlds won't see subsequent updates recorded in worlds at or above the receiver (see WDeepIdentityDictionary>>at:put:)"
	hasChildren
		ifTrue: [
			newDeepUpdateDict := WeakIdentityKeyDictionary new: 100.
			self deepUpdateDict associations do: [ :objAndTrackers |
				(obj := objAndTrackers key) ifNotNil: [
					newDeepUpdateDict at: obj put: (obj duplicateTrackers: objAndTrackers value forDepth: parentDepth).
					].
				].
			child deepUpdateDict: newDeepUpdateDict.
			]
		ifFalse: [
			child deepUpdateDict: self deepUpdateDict.
			hasChildren := true.
			].
	^child
! !

!WWorld methodsFor: 'debugging' stamp: 'ael 6/27/2011 21:07'!
trackerDictSize
	^updateTrackerDict size! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 6/24/2011 22:38'!
tracksReads
	^tracksReads
! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 6/24/2011 22:38'!
tracksReads: bool
	tracksReads := bool

! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 3/1/2012 21:28'!
updateTrackerDict
	^updateTrackerDict
	! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 3/1/2012 21:28'!
updateTrackerDict: d
	updateTrackerDict := d
	! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 9/25/2011 16:20'!
updateTrackersFor: obj

	^ self updateTrackersFor: obj createIfAbsent: false
! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 10/28/2011 15:41'!
updateTrackersFor: obj createIfAbsent: create
	^ updateTrackerDict at: obj ifAbsent: [
		create
		  ifTrue: [ updateTrackerDict at: obj put: (Array new: obj iUpdateTrackerSize withAll: nil) ]
		  ifFalse: [ nil ]
		]! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 10/31/2011 20:12'!
valueAt: index for: obj using: trackers
	"standard entry point for accessing values in objects that use the simple { reads. writes } style of update trackers.
	NB: trackers can be nil, meaning that this world doesn't have trackers for obj yet"
	^ self lookupFor: obj at: index caching: true in: trackers writesSlot: 2 readsSlot: 1
! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 9/25/2011 19:31'!
valueAt: index put: value for: obj
	"generalisation of instVarAt:put:for:, given that it may be an indexed object.
	write into whatever proxy object is recording writes for obj"
	^ (self ensureWriteTargetFor: obj) at: index put: value! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 10/2/2011 11:36'!
valueAt: index put: value for: obj using: updateTrackers
	"NB: updateTrackers must be non-nil"
	^ (obj iEnsureWriteTargetIn: updateTrackers) at: index put: value
! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 5/21/2012 16:05'!
withAllAncestors
		"return an ordered collection from the receiver up to and including the top-level world"
		| w worlds |
		worlds := OrderedCollection with: self.
		w := self.
		[ (w := w parent) notNil ] whileTrue: [ worlds add: w ].
		^worlds
! !

!WWorld methodsFor: 'as yet unclassified' stamp: 'ael 11/1/2011 19:28'!
writesRecordFor: obj
	| updateTrackers |
	updateTrackers := self updateTrackersFor: obj.
	^ updateTrackers ifNotNil: [ obj writesRecordIn: updateTrackers ]! !

Object subclass: #WorldSpecialObject
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Worlds'!

WorldSpecialObject class
	instanceVariableNames: 'theInstance'!

WorldSpecialObject class
	instanceVariableNames: 'theInstance'!

WorldSpecialObject subclass: #WorldDNKToken
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Worlds'!

!WorldDNKToken commentStamp: 'ael 6/25/2011 12:34' prior: 0!
WorldLookupDelegate initialize!

!WorldDNKToken methodsFor: 'as yet unclassified' stamp: 'ael 6/25/2011 12:32'!
identifier
	^'[don''t know]'
! !

WorldSpecialObject subclass: #WorldDeletionToken
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Worlds'!

!WorldDeletionToken commentStamp: 'ael 6/25/2011 12:33' prior: 0!
WorldDeletionToken initialize
!

!WorldDeletionToken methodsFor: 'as yet unclassified' stamp: 'ael 6/25/2011 15:52'!
identifier
	^'[not there]'
! !

!WorldDeletionToken methodsFor: 'as yet unclassified' stamp: 'ael 9/18/2011 16:05'!
isWorldDeletion
	^true
! !

WorldSpecialObject subclass: #WorldOverwriteToken
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Worlds'!

!WorldOverwriteToken commentStamp: 'ael 6/25/2011 12:34' prior: 0!
WorldOverwriteToken initialize!

!WorldOverwriteToken methodsFor: 'as yet unclassified' stamp: 'ael 6/25/2011 12:33'!
identifier
	^'[overwritten]'
! !

!WorldOverwriteToken methodsFor: 'as yet unclassified' stamp: 'ael 9/18/2011 16:05'!
isWorldOverwrite
	^true
! !

!WorldSpecialObject class methodsFor: 'as yet unclassified' stamp: 'ael 6/25/2011 12:27'!
initialize
	theInstance := self basicNew
! !

!WorldSpecialObject class methodsFor: 'as yet unclassified' stamp: 'ael 7/12/2011 16:44'!
new
	theInstance ifNil: [ self initialize ].
	^theInstance
! !

!WorldSpecialObject methodsFor: 'as yet unclassified' stamp: 'ael 6/25/2011 12:29'!
clone
	^self
! !

!WorldSpecialObject methodsFor: 'as yet unclassified' stamp: 'ael 6/25/2011 12:30'!
copy
	^self
! !

!WorldSpecialObject methodsFor: 'as yet unclassified' stamp: 'ael 6/25/2011 12:30'!
doesNotUnderstand: msg
	^self
! !

!WorldSpecialObject methodsFor: 'as yet unclassified' stamp: 'ael 6/25/2011 12:31'!
identifier
	^'[don''t know]'
! !

!WorldSpecialObject methodsFor: 'nil' stamp: 'ael 6/25/2011 20:51'!
isZero
	^false
! !

!WorldSpecialObject methodsFor: 'as yet unclassified' stamp: 'ael 6/25/2011 12:31'!
printOn: aStream
	aStream nextPutAll: self identifier! !

!SetTest methodsFor: '*worlds-tests-override' stamp: 'ael 12/2/2011 15:01'!
setUp
	empty := self targetClass new.
	full := self targetClass with: 5 with: #abc! !

!SetTest methodsFor: '*worlds-tests-override' stamp: 'md 4/16/2003 15:03'!
tearDown
	"I am called whenever your test ends. 
I am the place where you release the ressources"! !

!SetTest methodsFor: '*worlds-tests-override' stamp: 'GL 2/28/2006 08:26'!
testAdd
	empty add: 5.
	self assert: (empty includes: 5).! !

!SetTest methodsFor: '*worlds-tests-override' stamp: 'GL 3/16/2006 10:36'!
testAddWithOccurences

	empty add: 2 withOccurrences: 3.
	self assert: (empty includes: 2).
	self assert: ((empty occurrencesOf: 2) = 1).! !

!SetTest methodsFor: '*worlds-tests-override' stamp: 'ael 12/2/2011 15:01'!
testAsSet
	"could be moved in Array or Collection"

	| newFull |
	newFull := #(#abc 5) as: self targetClass.
	newFull add: 5.
	self assert: (newFull = full).! !

!SetTest methodsFor: '*worlds-tests-override' stamp: 'GL 2/28/2006 09:25'!
testAtRandom
	| rand |
	rand := Random new.
	full add: 3.
	full add: 2.
	full add: 4.
	full add: 1.
	self assert: (full includes: (full atRandom: rand)).
	! !

!SetTest methodsFor: '*worlds-tests-override' stamp: 'ael 12/2/2011 15:02'!
testCollect
	| newFull result |
	newFull := self targetClass withAll: (1 to: 10).
	result := newFull collect: [:each | each >= 1 ifTrue: [each] ifFalse: ['no']].
	self assert: (result = newFull).
	result := newFull collect: [:each | each >= 5 ifTrue: [each] ifFalse: ['no']].
	self assert: (result = ((self targetClass withAll: (5 to: 10)) add: 'no'; yourself)).! !

!SetTest methodsFor: '*worlds-tests-override' stamp: 'GL 2/28/2006 09:46'!
testCopy
	| newFull |
	full add: 3.
	full add: 2.
	newFull := full copy.
	self assert: (full size = newFull size).
	self assert: ((full select: [:each | (newFull includes: each) not]) isEmpty).
	self assert: ((newFull select: [:each | (full includes: each) not]) isEmpty).! !

!SetTest methodsFor: '*worlds-tests-override' stamp: 'ael 12/2/2011 15:02'!
testCopyWithout
	| newFull |
	full add: 3.
	full add: 2.
	newFull := full copyWithout: 3.
	self assert: (newFull size = (full size - 1)).
	self deny: (newFull includes: 3).
	self assert: ((newFull select: [:each | (full includes: each) not]) isEmpty).
	self assert: ((full select: [:each | (newFull includes: each) not]) = (self targetClass with: 3)).
	! !

!SetTest methodsFor: '*worlds-tests-override' stamp: 'ael 12/2/2011 15:02'!
testDo
	| newFull result |
	newFull := self targetClass withAll: (1 to: 5).
	result := 0.
	newFull do: [:each | result := (result + each)].
	self assert: (result = 15).! !

!SetTest methodsFor: '*worlds-tests-override' stamp: 'GL 2/28/2006 08:27'!
testGrow
	empty addAll: (1 to: 100).
	self assert: (empty size = 100).
			! !

!SetTest methodsFor: '*worlds-tests-override' stamp: 'GL 2/28/2006 08:23'!
testIllegal
	self 
		should: [empty at: 5] raise: TestResult error.
	self 
		should: [empty at: 5 put: #abc] raise: TestResult error.
			! !

!SetTest methodsFor: '*worlds-tests-override' stamp: 'GL 2/27/2006 21:42'!
testIncludes
	self assert: (full includes: 5).
	self assert: (full includes: #abc).
	self deny: (full includes: 3).
			! !

!SetTest methodsFor: '*worlds-tests-override' stamp: 'ael 12/2/2011 15:03'!
testIntersection
	| newFull col |
	full add: 3; add: 2.
	col := full intersection: full.
	self assert: (full = col).

	newFull := self targetClass with: 8 with: 9 with: #z.
	col := newFull intersection: full.
	self assert: (col isEmpty).
	
	newFull add: 5; add: #abc; add: 7.
	col := newFull intersection: full.
	self assert: ((full select: [:each | (newFull includes: each)]) = col).
	
	
	! !

!SetTest methodsFor: '*worlds-tests-override' stamp: 'ael 12/2/2011 15:20'!
testLike
self targetClass == Set ifFalse: [ ^self ].		"only supported by Set"
	self assert: ((full like: 5) = 5).
	self assert: ((full like: 8) isNil).! !

!SetTest methodsFor: '*worlds-tests-override' stamp: 'GL 2/28/2006 08:27'!
testOccurrences
	self assert: ((empty occurrencesOf: 0) = 0).
	self assert: ((full occurrencesOf: 5) = 1).
	full add: 5.
	self assert: ((full occurrencesOf: 5) = 1).! !

!SetTest methodsFor: '*worlds-tests-override' stamp: 'GL 3/8/2006 08:20'!
testRemove
	full remove: 5.
	self assert: (full includes: #abc).
	self deny: (full includes: 5).! !

!SetTest methodsFor: '*worlds-tests-override' stamp: 'nice 9/14/2009 21:08'!
testRemoveAll
	"Allows one to remove all elements of a collection" 
	
	| c1 c2 s2 |
	c1 := full.
	c2 := c1 copy.
	s2 := c2 size.
	
	c1 removeAll.
	
	self assert: c1 size = 0.
	self assert: c2 size = s2 description: 'the copy has not been modified'.! !

!SetTest methodsFor: '*worlds-tests-override' stamp: 'GL 3/8/2006 08:27'!
testRemoveIfAbsent
	| result1 result2  |
	result1 := true.
	result2 := true.
	full remove: 8 ifAbsent: [ result1 := false ].
	self assert: (result1 = false).
	full remove: 5 ifAbsent: [ result2 := false ].
	self assert: (result2 = true).
	
	
	! !

!SetTest methodsFor: '*worlds-tests-override' stamp: 'GL 2/28/2006 09:51'!
testSize
	self assert: (empty size = 0).
	self assert: (full size = 2).
	empty add: 2.
	empty add: 1.
	full add: 2.
	self assert: (empty size = 2).
	self assert: (full size = 3).
	empty remove: 2.
	self assert: (empty size = 1).! !

!SetTest methodsFor: '*worlds-tests-override' stamp: 'ael 12/2/2011 15:03'!
testUnion
	| newFull col newCol |
	full add: 3.
	full add: 2.
	col := full union: full.
	self assert: (full = col).

	newFull := self targetClass with: 8 with: 9 with: #z.
	col := newFull union: full.
	self assert: (col size = (full size + newFull size)).
	self assert: ((col select: [:each | (full includes: each) not]) = newFull).
	self assert: ((col select: [:each | (newFull includes: each) not]) = full).

	full add: 9.
	col := newFull union: full.
	newCol := newFull copy.
	newCol remove: 9.
	self assert: (col size = (full size + newFull size - 1)).
	self assert: ((col select: [:each | (full includes: each) not]) = newCol).
	newCol := full copy.
	newCol remove: 9.
	self assert: ((col select: [:each | (newFull includes: each) not]) = newCol).
	
	
	! !

SetTest subclass: #WIdentitySetTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Worlds-CollectionsTests'!

Error subclass: #WorldSerializationError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Worlds'!

!WorldSerializationError methodsFor: 'as yet unclassified' stamp: 'ael 6/17/2011 20:52'!
isResumable
	^true
! !

WorldSerializationError subclass: #WorldSizeChanged
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Worlds'!

WorldSerializationError subclass: #WorldValueChanged
	instanceVariableNames: 'index indexPrefix readValue currentValue'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Worlds'!

!WorldValueChanged methodsFor: 'as yet unclassified' stamp: 'ael 12/6/2011 14:55'!
indexPrefix: ip index: i readValue: rv currentValue: cv
	indexPrefix := ip.
	index := i.
	readValue := rv.
	currentValue := cv.
! !

!WorldValueChanged methodsFor: 'as yet unclassified' stamp: 'ael 12/6/2011 14:59'!
messageText
	^'Value read from ', indexPrefix, ' ', index asString, ' has changed'
! !

!WorldValueChanged methodsFor: 'as yet unclassified' stamp: 'ael 12/6/2011 14:47'!
readValue: rv currentValue: cv
	readValue := rv.
	currentValue := cv.
! !

Compiler subclass: #WCompiler
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Worlds-Compiler'!

!WCompiler methodsFor: 'as yet unclassified' stamp: 'ael 10/4/2011 15:44'!
compile: textOrStream in: aClass classified: aCategory notifying: aRequestor ifFail: failBlock

	^ self compile: textOrStream in: aClass classified: aCategory notifying: aRequestor ifFail: failBlock for: nil
! !

!WCompiler methodsFor: 'as yet unclassified' stamp: 'ael 10/4/2011 15:46'!
compile: textOrStream in: aClass classified: aCategory notifying: aRequestor ifFail: failBlock for: namespaceObject
	"Answer a MethodNode for the argument, textOrStream. If the 
	MethodNode can not be created, notify the argument, aRequestor; if 
	aRequestor is nil, evaluate failBlock instead. The MethodNode is the root 
	of a parse tree. It can be told to generate a CompiledMethod to be 
	installed in the method dictionary of the argument, aClass."
	self parserClass: WParser.
	self from: textOrStream
		class: aClass
		classified: aCategory 
		context: nil
		notifying: aRequestor.
	^self
		translate: sourceStream
		noPattern: false
		ifFail: failBlock for: namespaceObject! !

!WCompiler methodsFor: 'as yet unclassified' stamp: 'ael 10/4/2011 15:44'!
compile: textOrStream in: aClass notifying: aRequestor ifFail: failBlock

	^ self compile: textOrStream in: aClass classified: nil notifying: aRequestor ifFail: failBlock
! !

!WCompiler methodsFor: 'as yet unclassified' stamp: 'ael 10/4/2011 15:44'!
compile: textOrStream in: aClass notifying: aRequestor ifFail: failBlock for: namespaceObject

	^ self compile: textOrStream in: aClass classified: nil notifying: aRequestor ifFail: failBlock for: namespaceObject
! !

!WCompiler methodsFor: 'as yet unclassified' stamp: 'ael 10/24/2011 15:19'!
parserClass
	"needed for trunk versions"
	^parser ifNil: [self class parserClass] ifNotNil: [parser class]! !

!WCompiler methodsFor: 'as yet unclassified' stamp: 'sn 9/10/2013 13:11'!
translate: aStream noPattern: noPattern ifFail: failBlock for: namespaceObject

	^self parserClass new
		parse: aStream
		cue: cue 
		noPattern: noPattern
		ifFail: [^failBlock value]
		for: namespaceObject
	"^ self parserClass new
			parse: aStream
			class: class
			noPattern: noPattern
			context: context
			notifying: requestor
			ifFail: [^ failBlock value]
			for: namespaceObject."
! !

!OrderedCollection class methodsFor: '*worlds' stamp: 'ael 5/15/2013 00:15'!
simpleBenchmark: n
	"just for comparison with WCollection and WSimpleOrderedCollection implementations"
	"OrderedCollection simpleBenchmark: 1000"
	| l runs |
	runs := 100000.
	^(Time millisecondsToRun: [
		WWorld thisWorld sprout eval: [
			runs timesRepeat: [
				l := self new..
				1 to: n do: [:i| l addFirst: i ].
				[ l isEmpty ] whileFalse: [ l removeLast ]
				].
			 ]
		]) -
	(Time millisecondsToRun: [ WWorld thisWorld sprout eval: [ runs timesRepeat: [ 0 ]]])
	asFloat / runs.
! !

ArrayTest subclass: #WArrayTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Worlds-CollectionsTests'!

!WArrayTest methodsFor: 'as yet unclassified' stamp: 'ael 12/6/2011 15:12'!
testThreeLevels
	| w1 w2 w2Contents w2Reads dnk w1Contents w1Reads testArray |
	testArray := WArray new: 6.
	testArray
		at: 1 put: #a0;
		at: 2 put: #b0;
		at: 3 put: #c0;
		at: 4 put: #d0;
		at: 5 put: #e0;
		at: 6 put: #f0.

	dnk := testArray dnkToken.

	w1 := self thisWorld sprout.
	w2 := w1 sprout.
	self assert: (w2 eval: [ testArray includes: #c0 ]).
	w2Reads := w2 eval: [ testArray debugUpdateTrackerState at: 1 ].
	self assert: (w2Reads as: Array) = { #a0. #b0. #c0. dnk. dnk. dnk }.

	testArray at: 3 put: nil.
	self assert: (w2 eval: [ testArray includes: #c0 ]).		"should not be affected"

	testArray at: 1 put: #a0a.				"index 1 has been read in w2 but not w1"
	testArray at: 4 put: #d0a.				"index 4 has not been read yet in w2"

	self assert: (w1 eval: [ testArray lastIndexOf: #e0 ]) = 5.
	w1Reads := w1 eval: [ testArray debugUpdateTrackerState at: 1 ].
	self assert: (w1Reads as: Array) = { dnk. dnk. dnk. dnk. #e0. #f0 }.

	w2 eval: [ testArray at: 1; at: 4 ].
	w2Reads := w2 eval: [ testArray debugUpdateTrackerState at: 1 ].
	self assert: (w2Reads as: Array) = { #a0. #b0. #c0. #d0a. dnk. dnk }.

	w1 eval: [
		testArray
			at: 1 put: #a1;
			at: 5 put: #e1;
			at: 6 put: nil.
		].
	
	w2 eval: [
		testArray at: 6 put: #f2.						"NB: overwriting without reading"
		w2Contents := testArray as: Array.
		w2Reads := testArray debugUpdateTrackerState at: 1.
		].

	self assert: w2Contents = { #a0. #b0. #c0. #d0a. #e1. #f2 }.
	self assert: (w2Reads as: Array) = { #a0. #b0. #c0. #d0a. #e1. dnk }.	"#e1 read from w1"

	self should: [ w2 commit ] raise: WorldValueChanged.	"clash between #a0 and #a0a, #c0 and nil"

	w1 eval: [ testArray at: 1 put: #a0; at: 3 put: #c0 ].					"restore to the state seen in w2"

	self shouldnt: [ w2 commit ] raise: WorldValueChanged.

	w1Reads := w1 eval: [ testArray debugUpdateTrackerState at: 1 ].
	self assert: (w1Reads as: Array) = { dnk. #b0. dnk. #d0a. #e0. #f0 }.	"#a and #c values were not delivered from the top world"

	w1Contents := w1 eval: [ testArray as: Array ].
	self assert: w1Contents = { #a0. #b0. #c0. #d0a. #e1. #f2 }.

	self shouldnt: [ w1 commit ] raise: WorldSerializationError.

	self assert: (testArray as: Array) = { #a0. #b0. #c0. #d0a. #e1. #f2 }.	"#a and #c entries overwritten from w1"
! !

!WArrayTest methodsFor: 'as yet unclassified' stamp: 'ael 10/25/2011 16:40'!
thisWorld
	^WWorld thisWorld
! !

!ClassOrganizer methodsFor: '*worlds-compiler' stamp: 'yo 3/23/2010 13:59'!
classifySilently: element under: heading suppressIfDefault: aBoolean
	super classify: element under: heading suppressIfDefault: aBoolean! !

!TextEditor methodsFor: '*worlds-override' stamp: 'ael 4/6/2012 23:12'!
evaluateSelectionAndDo: aBlock
	"Treat the current selection as an expression; evaluate it and invoke aBlock with the result."
	| result rcvr ctxt ww |
	self lineSelectAndEmptyCheck: [^ nil].

	(model respondsTo: #doItReceiver) 
		ifTrue: [ rcvr := model doItReceiver.
				ctxt := model doItContext]
		ifFalse: [rcvr := ctxt := nil].

	ww := (model respondsTo: #doItWWorld)
		ifTrue: [ model doItWWorld ]
		ifFalse: [ WTopLevelWorld default ].
		
	result := [
		ww eval: [
			rcvr class evaluatorClass new 
				evaluate: self selectionAsStream
				in: ctxt
				to: rcvr
				notifying: self
				ifFail: [morph flash. ^ nil]
				logged: true.
			].
	] 
		on: OutOfScopeNotification 
		do: [ :ex | ex resume: true].
	^ww eval: [ aBlock value: result ]! !

!ObjectExplorerWrapper class methodsFor: '*worlds' stamp: 'ael 5/3/2012 14:25'!
new
	"hijack all wrapper creation"
	^WObjectExplorerWrapper basicNew initialize

! !

ObjectExplorerWrapper subclass: #WObjectExplorerWrapper
	instanceVariableNames: 'getWWorldSelector wworldModel'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Worlds-Tools'!

!WObjectExplorerWrapper methodsFor: 'worldiness' stamp: 'ael 5/3/2012 14:19'!
asString
	| explorerString string |
	explorerString := 
		[ self worldEval: [ item asExplorerString ]]
			on: Error 
			do: ['<error: ', item class name, ' in asExplorerString: evaluate "' , itemName , ' asExplorerString" to debug>'].
	string := (itemName ifNotNil: [itemName asString, ': '] ifNil: ['']) , explorerString.
	(string includes: Character cr)
		ifTrue: [^ string withSeparatorsCompacted].
	^ string! !

!WObjectExplorerWrapper methodsFor: 'worldiness' stamp: 'ael 5/3/2012 14:19'!
contents

	| wrappers |
	(item respondsTo: #explorerContents)
		ifTrue: [
			wrappers := self worldEval: [ item explorerContents ].
			wrappers do: [ :w | w parent: self; getWWorldFrom: wworldModel selector: getWWorldSelector ].
			^wrappers
			].
	"For all others, show named vars first, then indexed vars"
	^(item class allInstVarNames asOrderedCollection withIndexCollect: [:each :index |
		(self class
			with: (self worldEval: [item instVarAt: index])
			name: each
			model: item
			parent: self)
				getWWorldFrom: wworldModel selector: getWWorldSelector
		]) ,
	((1 to: item basicSize) collect: [:index |
		(self class
			with: (self worldEval: [item basicAt: index])
			name: index printString
			model: item
			parent: self)
				getWWorldFrom: wworldModel selector: getWWorldSelector
		])! !

!WObjectExplorerWrapper methodsFor: 'worldiness' stamp: 'ael 5/3/2012 14:20'!
getWWorldFrom: obj selector: selector
	wworldModel := obj.
	getWWorldSelector := selector
! !

!WObjectExplorerWrapper methodsFor: 'worldiness' stamp: 'ael 5/3/2012 14:20'!
hasContents

	^self worldEval: [ item hasContentsInExplorer ]
	
! !

!WObjectExplorerWrapper methodsFor: 'worldiness' stamp: 'ael 5/3/2012 14:20'!
icon
	"Answer a form to be used as icon"
	^ Preferences visualExplorer
		ifTrue: [ self worldEval: [item iconOrThumbnailOfSize: 16]]
		ifFalse: [nil]! !

!WObjectExplorerWrapper methodsFor: 'ksobjects-tools' stamp: 'ael 5/3/2012 14:21'!
isRoot
	^parent isNil
! !

!WObjectExplorerWrapper methodsFor: 'worldiness' stamp: 'ael 5/3/2012 14:20'!
itemInWorld: aWorld ifNone: failBlock
	"find the object represented by this wrapper in the specified world, based on knowing the parent (model) object and an item name/index"
	| tmpItem |
	self isRoot ifTrue: [ ^item ].			"root item never changes"
	[ | index |
		"if it's not a world-aware object, it won't mind which world the lookup happens in"
		aWorld eval: [
			(model class allInstVarNames includes: itemName)
				ifTrue: [ tmpItem := model instVarNamed: itemName ]
				ifFalse: [
					"if the item name is all digits, assume it's a numerical index"
					"if the access raises an error, just catch it below"
					itemName isAllDigits
						ifTrue: [
							index := itemName asNumber.
							tmpItem := (model isKindOf: WObject)
								ifFalse: [ model basicAt: index ]
								ifTrue: [ model basicValueAt: index ].
							]
						ifFalse: [ tmpItem := model at: itemName ].
					]
				]
	] on: Error do: [ :ex | ^failBlock value ].
	^tmpItem
! !

!WObjectExplorerWrapper methodsFor: 'ksobjects-tools' stamp: 'ael 5/3/2012 14:22'!
model
	"KSObjectInspector wants to know"
	^model
! !

!WObjectExplorerWrapper methodsFor: 'ksobjects-tools' stamp: 'ael 5/3/2012 14:22'!
pathFromRoot
	| path wrapper |
	path := OrderedCollection new.
	wrapper := self.
	[ wrapper isRoot ] whileFalse: [ path addFirst: wrapper itemName. wrapper := wrapper parent ].
	path addFirst: ''.		"the root"
	^path reduce: [ :a :b | a, '/', b ]

! !

!WObjectExplorerWrapper methodsFor: 'worldiness' stamp: 'ael 5/3/2012 14:20'!
refresh
	"find the latest value of the object represented by this wrapper"
	item := self itemInWorld: self worldFromModel ifNone: []
! !

!WObjectExplorerWrapper methodsFor: 'ksobjects-tools' stamp: 'ael 5/3/2012 14:22'!
replaceValueFromString: codeString ifFail: failBlock
	"the receiver represents a field within a (Smalltalk, not KScript) object.  give that field the supplied new value."
	| index result |
	"if it's not a world-aware object, it won't mind which world the change happens in"
	self worldEval: [
		result := model class evaluatorClass new
				evaluate: (ReadStream on: codeString)
				in: nil
				to: model
				notifying: nil
				ifFail: [ ^failBlock value ].
		
		(model class allInstVarNames includes: itemName)
			ifTrue: [ model instVarNamed: itemName put: result ] 
			ifFalse: [
				"if the item name is all digits, assume it's a numerical index"
				"if the access raises an error, just catch it below"
				itemName isAllDigits
					ifTrue: [
						index := itemName asNumber.
						(model isKindOf: WObject)
							ifFalse: [ model basicAt: index put: result ]
							ifTrue: [ model basicValueAt: index put: result ].
						]
					ifFalse: [ model at: itemName put: result ].
				]
		]
! !

!WObjectExplorerWrapper methodsFor: 'worldiness' stamp: 'ael 5/3/2012 14:21'!
worldEval: aBlock
	^self worldFromModel eval: aBlock! !

!WObjectExplorerWrapper methodsFor: 'worldiness' stamp: 'ael 5/3/2012 14:21'!
worldFromModel
	^getWWorldSelector ifNil: [ WTopLevelWorld default ] ifNotNil: [ wworldModel perform: getWWorldSelector ].
! !

!OrderedCollectionTest methodsFor: '*worlds-testsAdding-override' stamp: 'ael 6/17/2011 16:44'!
testAdd
	| l |
	l := #(1 2 3 4) as: self targetClass.
	l add: 88.
	self assert: (l =  (#(1 2 3 4 88) as: self targetClass)).
	l add: 99.
	self assert: (l =  (#(1 2 3 4 88 99) as: self targetClass)). 

! !

!OrderedCollectionTest methodsFor: '*worlds-testsAdding-override' stamp: 'ael 6/17/2011 17:39'!
testAddAfter

	| l |
	l := #(1 2 3 4)  as: self targetClass.
	l add: 88 after: 1.
	self assert: (l =  (#(1 88 2 3 4)  as: self targetClass)).
	l add: 99 after: 2.
	self assert: (l =  (#(1 88 2 99 3 4)  as: self targetClass)). 

! !

!OrderedCollectionTest methodsFor: '*worlds-testing-public methods-override' stamp: 'ael 6/17/2011 17:57'!
testAddAfterIndex
	"self run: #testAddAfterIndex"
	| l |
	l := #(1 2 3 4) as: self targetClass.
	l add: 77 afterIndex: 0.
	self assert: (l =  (#(77 1 2 3 4) as: self targetClass)).
	l add: 88 afterIndex: 2.
	self assert: (l =  (#(77 1 88 2 3 4) as: self targetClass)). 
	l add: 99 afterIndex: l size.
	self assert: (l =  (#(77 1 88 2 3 4 99) as: self targetClass)). 
	self should:[l add: 666 afterIndex: -1] raise: Error.
	self should:[l add: 666 afterIndex: l size+1] raise: Error.
	
	"Now make room by removing first two and last two elements,
	and see if the illegal bounds test still fails"
	(l first: 2) , (l last: 2) reversed do: [:e | l remove: e].
	self should: [l add: 666 afterIndex: -1] raise: Error.
	self should: [l add: 666 afterIndex: l size+1] raise: Error.! !

!OrderedCollectionTest methodsFor: '*worlds-testsAdding-override' stamp: 'ael 6/17/2011 17:40'!
testAddAll
	"Allows one to add each element of an orderedCollection at the end of another
	orderedCollection "
	"self run:#testAddAll" 
	
	| c1 c2 |
	c1 := #(1 2 3 4 ) as: self targetClass.
	c2 := #(5 6 7 8 9 ) as: self targetClass.
	c1 addAll: c2.
	self assert: c1 = (#(1 2 3 4 5 6 7 8 9) as: self targetClass)! !

!OrderedCollectionTest methodsFor: '*worlds-testsAdding-override' stamp: 'ael 6/17/2011 17:40'!
testAddAllFirst
	"Allows one to add each element of an orderedCollection at the beginning of another
	orderedCollection "
	"self run:#testAddAllFirst" 
	
	| c1 c2 |
	c1 := #(1 2 3 4 ) as: self targetClass.
	c2 := #(5 6 7 8 9 ) as: self targetClass.
	c2 addAllFirst: c1.
	self assert: c2 = (#(1 2 3 4 5 6 7 8 9) as: self targetClass)! !

!OrderedCollectionTest methodsFor: '*worlds-testsAdding-override' stamp: 'ael 6/17/2011 17:42'!
testAddAllFirstUnlessAlreadyPresent
	"Allows one to add each element of an orderedCollection at the beginning of
	another orderedCollection preserving the order but no duplicate element"
	"self run:#testAddAllFirstUnlessAlreadyPresent" 
	
	| c1 c2 c3 |
	c1 := #(1 2 3 4 ) as: self targetClass.
	c2 := #(5 6 7 8 9 ) as: self targetClass.
	c3 := #(0 1 ) as: self targetClass.
	c2 addAllFirstUnlessAlreadyPresent: c1.
	self assert: c2 = (#(1 2 3 4 5 6 7 8 9 ) as: self targetClass).
	c1 addAllFirstUnlessAlreadyPresent: c3.
	self deny: c1 = (#(0 1 1 2 3 4 ) as: self targetClass).
	self assert: c1 = (#(0 1 2 3 4 ) as: self targetClass).
	! !

!OrderedCollectionTest methodsFor: '*worlds-testsAdding-override' stamp: 'ael 6/17/2011 17:42'!
testAddAllLast
	"Allows one to add each element of an orderedCollection at the beginning of another
	orderedCollection "
	"self run:#testAddAllLast" 
	
	| c1 c2 |
	c1 := #(1 2 3 4 ) as: self targetClass.
	c2 := #(5 6 7 8 9 ) as: self targetClass.
	c1 addAllLast: c2.
	self assert: c1 = (#(1 2 3 4 5 6 7 8 9) as: self targetClass)! !

!OrderedCollectionTest methodsFor: '*worlds-testsAdding-override' stamp: 'ael 6/17/2011 17:43'!
testAddBefore

	| l |
	l := #(1 2 3 4) as: self targetClass.
	l add: 88 before: 1.
	self assert: (l =  (#(88 1 2 3 4) as: self targetClass)).
	l add: 99 before: 2.
	self assert: (l =  (#(88 1 99 2 3 4) as: self targetClass)). 

! !

!OrderedCollectionTest methodsFor: '*worlds-testsAdding-override' stamp: 'ael 6/17/2011 17:43'!
testAddBeforeAndRemove

	| l initialCollection |
	l := #(1 2 3 4) as: self targetClass.
	initialCollection := l shallowCopy.
	l add: 88 before: 1.
	self assert: (l =  (#(88 1 2 3 4) as: self targetClass)).
	l add: 99 before: 2.
	self assert: (l =  (#(88 1 99 2 3 4) as: self targetClass)). 
	l remove: 99.
	l remove: 88.
	self assert: l = initialCollection.

! !

!OrderedCollectionTest methodsFor: '*worlds-testing-public methods-override' stamp: 'ael 6/17/2011 17:57'!
testAddBeforeIndex
	"self run: #testAddBeforeIndex"
	| l |
	l := #(1 2 3 4) as: self targetClass.
	l add: 77 beforeIndex: 1.
	self assert: (l =  (#(77 1 2 3 4) as: self targetClass)).
	l add: 88 beforeIndex: 3.
	self assert: (l =  (#(77 1 88 2 3 4) as: self targetClass)). 
	l add: 99 beforeIndex: l size+1.
	self assert: (l =  (#(77 1 88 2 3 4 99) as: self targetClass)). 
	self should:[l add: 666 beforeIndex: 0] raise: Error.
	self should:[l add: 666 beforeIndex: l size+2] raise: Error.
	
	"Now make room by removing first two and last two elements,
	and see if the illegal bounds test still fails"
	(l first: 2) , (l last: 2) reversed do: [:e | l remove: e].
	self should:[l add: 666 beforeIndex: 0] raise: Error.
	self should:[l add: 666 beforeIndex: l size+2] raise: Error.

! !

!OrderedCollectionTest methodsFor: '*worlds-testsAdding-override' stamp: 'ael 6/17/2011 17:45'!
testAddDuplicateItem1
	| collection |
	collection := #('Jim' 'Mary' 'John' 'Andrew' ) as: self targetClass.
	collection add: 'John' before: 'John'.
	self
		assert: ((collection asBag occurrencesOf: 'John')
					= 2
				and: [(collection at: (collection indexOf: 'John')
							+ 1)
						= (collection
								at: (collection indexOf: 'John'))])! !

!OrderedCollectionTest methodsFor: '*worlds-testsAdding-override' stamp: 'ael 6/17/2011 17:45'!
testAddFirst
	| l |
	l := #(1 2 3 4) as: self targetClass.
	l addFirst: 88.
	self assert: (l =  (#(88 1 2 3 4) as: self targetClass)).
	l addFirst: 99.
	self assert: (l =  (#(99 88 1 2 3 4) as: self targetClass)). 

! !

!OrderedCollectionTest methodsFor: '*worlds-testsAdding-override' stamp: 'ael 6/17/2011 17:46'!
testAddItem1

   | collection size |
   collection := #('Jim' 'Mary' 'John' 'Andrew' ) as: self targetClass.
   size := collection size.
   collection add: 'James' before: 'Jim'.
   collection add: 'Margaret' before: 'Andrew'.
   self assert: size + 2 = collection size.
! !

!OrderedCollectionTest methodsFor: '*worlds-testsAdding-override' stamp: 'ael 6/17/2011 17:46'!
testAddItem2
	| collection |
	collection := #('Jim' 'Mary' 'John' 'Andrew' ) as: self targetClass.
	collection add: 'James' before: 'Jim'.
	collection add: 'Margaret' before: 'Andrew'.
	self assert: (collection indexOf: 'James')
			+ 1
			= (collection indexOf: 'Jim').
	self assert: (collection indexOf: 'Margaret')
			+ 1
			= (collection indexOf: 'Andrew')! !

!OrderedCollectionTest methodsFor: '*worlds-testsAdding-override' stamp: 'ael 6/17/2011 17:46'!
testAddLast
	| l |
	l := #(1 2 3 4) as: self targetClass.
	l addLast: 88.
	self assert: (l =  (#(1 2 3 4 88) as: self targetClass)).
	l addLast: 99.
	self assert: (l =  (#(1 2 3 4 88 99) as: self targetClass)). 

! !

!OrderedCollectionTest methodsFor: '*worlds-testsAccessing-override' stamp: 'ael 6/17/2011 17:48'!
testAt
	| collection |
	collection := #('Jim' 'Mary' 'John' 'Andrew' ) as: self targetClass.
	self assert: (collection at:1) = 'Jim'.
	self assert: (collection at:2) = 'Mary'! !

!OrderedCollectionTest methodsFor: '*worlds-testsAdding-override' stamp: 'ael 6/17/2011 17:47'!
testAtIfAbsentPut
	"Allows one to add an element at an index if no element exist at this index"
	"self run:#testAtIfAbsentPut" 
	
	| c |
	c := #(1 2 3 4 ) as: self targetClass.
	self
		shouldnt: [c at: 2 ifAbsentPut: [5]]
		raise: Error.
	self assert: c = (#(1 2 3 4 ) as: self targetClass).
	c at: 5 ifAbsentPut: [5].
	self assert: c = (#(1 2 3 4 5 ) as: self targetClass).
	c at: 7 ifAbsentPut: [7].
	self assert: c = (#(1 2 3 4 5 nil 7 ) as: self targetClass)! !

!OrderedCollectionTest methodsFor: '*worlds-testsAccessing-override' stamp: 'ael 6/17/2011 17:49'!
testAtPut
	"Allows one to replace an element but not at an off range index"
	"self run:#testAtPut"
	| c |
	c := #(1 2 3 4 ) as: self targetClass.
	c at: 2 put: 5.
	self assert: c = (#(1 5 3 4 ) as: self targetClass).
	self
		should: [c at: 5 put: 8]
		raise: Error.
	self deny: c = (#(1 5 3 4 8 ) as: self targetClass)! !

!OrderedCollectionTest methodsFor: '*worlds-testsAccessing-override' stamp: 'ael 11/1/2011 22:11'!
testCapacity
	"Allows one to check the current capacity of an Ordered collection"
	"self run:#testCapacity"
	
	| c1 c2 c3 |
	"WCollection does more checking wrt size"
	self targetClass == WCollection
	  ifTrue: [
		c1 := #(1 2 ) as: self targetClass.
		self assert: (c1 capacity =  2).
		c2 := self targetClass new: 10.
		c2 add: 3.
		self assert: (c2 capacity = 10).	
		c3 := self targetClass new.
		self deny: (c3 capacity =  0).
		].! !

!OrderedCollectionTest methodsFor: '*worlds-testsEnumerating-override' stamp: 'ael 6/17/2011 17:50'!
testCollect
	"Allows one to collect some element of a collection into another collection"
	"self run: #testCollect"
	 
	| c1 c2 res |
	c1 := #(-1 2 -3 4) as: self targetClass.
	c2 := #(1 2 3 4) as: self targetClass.
	res := c1 collect: [:each | each abs].
	self assert: (c2 = res).! !

!OrderedCollectionTest methodsFor: '*worlds-testsEnumerating-override' stamp: 'ael 6/17/2011 19:30'!
testCollectFromTo
	"Allows one to collect some element of a collection into another collection between a first index and an end index for the collect"
	"self run: #testCollectFromTo"
	
	| c1 res |
	c1 := #(-1 2 -3 4 -5 6 -7 8) as: self targetClass.
	res := c1 collect: [:each | each abs] from: 1 to: 3.
	self assert: (res = (#(1 2 3) as: self targetClass)).
	self should: [c1 collect: [:each | each abs] from: 10 to: 13] raise: Error.
	self should: [c1 collect: [:each | each abs] from: 5 to: 2] raise: Error.! !

!OrderedCollectionTest methodsFor: '*worlds-testsCopying-override' stamp: 'ael 6/17/2011 17:51'!
testCopyEmpty
	"Allows one to create a copy of the receiver that contains no elements"
	"self run:#testCopyEmpty"
	
	| c1 c2 |
	c1 := #(1 2 3 4 ) as: self targetClass.
	c2 := c1 copyEmpty.
	self assert: (c2 size = 0).! !

!OrderedCollectionTest methodsFor: '*worlds-testsCopying-override' stamp: 'ael 6/17/2011 17:52'!
testCopyFromTo
	"Allows one to create a copy of the receiver that contains elements from position start to end"
	"self run: #testCopyFromTo"
	
	| c1 c2 c3 | 
	c1 := #(1 2 3 4) as: self targetClass.
	c2 := (c1 copyFrom: 1 to: 2).
	self assert: c2 = (#(1 2) as: self targetClass).
	self should: [c1 copyFrom: 10 to: 20] raise: Error.
	
	c3 := c1 copyFrom: 4 to: 2.
	self assert: c3 isEmpty.
	
	self should: [c1 copyFrom: 4 to: 5 ] raise: Error.
	
	
	
! !

!OrderedCollectionTest methodsFor: '*worlds-testsCopying-override' stamp: 'ael 6/17/2011 17:52'!
testCopyReplaceFromToWith
	"Allows one to create a copy from the receiver which elements between start and end of the 	receiver being replace by 	element of the collection after with:"
	"self run:#testCopyReplaceFromToWith"

	| c1 c2 c3 c4 |
	c1 := #(1 2 3 4) as: self targetClass.
	c2 := #(5 6 7 8 9) as: self targetClass.
	c3 := (c2 copyReplaceFrom: 1 to: 2 with: c1).
	self assert: c3 = (#(1 2 3 4 7 8 9) as: self targetClass).
	self should: [c2 copyReplaceFrom: 3 to: 1 with: c1] raise: Error.
	
	c4 := (c2 copyReplaceFrom: 10 to: 25 with: c1).
	self assert: c4 = (#(5 6 7 8 9 1 2 3 4) as: self targetClass).
	
	! !

!OrderedCollectionTest methodsFor: '*worlds-testsCopying-override' stamp: 'ael 6/17/2011 17:52'!
testCopyWith
	"Allows one to create a copy of the receiver that contains the new element at the end"
	"self run: #testCopyWith"
	
	| c1 | 
	c1 := #(1 2 3 4) as: self targetClass.
	c1 := c1 copyWith: 6.
	self assert: c1 = (#(1 2 3 4 6) as: self targetClass).
	
	

	
	
	
! !

!OrderedCollectionTest methodsFor: '*worlds-testsCopying-override' stamp: 'ael 6/17/2011 17:54'!
testForceToPadding
	"This is a non regression bug for http://bugs.squeak.org/view.php?id=7296"
	
	| c1 c2 paddingElement |
	"source collection"
	c1 := #(5 3 7 2 ) as: self targetClass.
	paddingElement := nil.
	
	"force length to longer"
	self shouldnt: [c2 := c1 forceTo: 10 paddingWith: paddingElement] raise: Error.
	self assert: (c1 ~~ c2) description: 'forceTo:paddingWith: should modify a copy'.
	self assert: (c2 size = 10).
	self assert: (c2 copyFrom: 1 to: c1 size) = c1.
	self assert: ((c1 size + 1 to: c2 size) allSatisfy: [:i | (c2 at: i) = paddingElement]).
	
	"force length to shorter"
	self shouldnt: [c2 := c1 forceTo: 3 paddingWith: paddingElement] raise: Error.
	self assert: (c1 ~~ c2) description: 'forceTo:paddingWith: should modify a copy'.
	self assert: (c2 size = 3).
	self assert: (c2 copyFrom: 1 to: 3) = (c1 copyFrom: 1 to: 3).
	
	"SAME TEST BUT PADDING AT START FOLLOWING..."
	
	"force length to longer"
	self shouldnt: [c2 := c1 forceTo: 10 paddingStartWith: paddingElement] raise: Error.
	self assert: (c1 ~~ c2) description: 'forceTo:paddingStartWith: should modify a copy'.
	self assert: (c2 size = 10).
	self assert: (c2 copyFrom: c2 size - c1 size + 1 to: c2 size) = c1.
	self assert: ((1 to: c2 size - c1 size) allSatisfy: [:i | (c2 at: i) = paddingElement]).
	
	"force length to shorter"
	self shouldnt: [c2 := c1 forceTo: 3 paddingStartWith: paddingElement] raise: Error.
	self assert: (c1 ~~ c2) description: 'forceTo:paddingStartWith: should modify a copy'.
	self assert: (c2 size = 3).
	self assert: (c2 copyFrom: 1 to: 3) = (c1 copyFrom: 1 to: 3).! !

!OrderedCollectionTest methodsFor: '*worlds-testsEnumerating-override' stamp: 'ael 6/17/2011 17:50'!
testIndexOf
	| collection indices |
	collection := #('Jim' 'Mary' 'John' 'Andrew' ) as: self targetClass.
	indices := collection
				collect: [:item | collection indexOf: item].
	self assert: ((1 to: 4) as: self targetClass) = indices! !

!OrderedCollectionTest methodsFor: '*worlds-testsEnumerating-override' stamp: 'ael 6/17/2011 17:50'!
testIndexOfWithDuplicates

   | collection indices bagOfIndices |
   collection := #('Jim' 'Mary' 'John' 'Andrew' 'Mary' 'John' 'Jim' 'Micheal') as: self targetClass.
   indices := collection collect: [:item | collection indexOf: item].
   self assert: indices asSet size = collection asSet size.
   bagOfIndices := indices asBag.
   self assert: (indices asSet 
                    allSatisfy: [:index | (bagOfIndices occurrencesOf: index)
	                                       = (collection occurrencesOf: (collection at: index))]).

  "  indexOf:  returns the index of the first occurrence of an item.
     For an item with n occurrences, the index of its first occurrence
     is found  n  times. "! !

!OrderedCollectionTest methodsFor: '*worlds-testsEnumerating-override' stamp: 'ul 3/1/2011 08:21'!
testOverlappingPairsCollect

	| o |
	o := #(5 4 3 2 1) asOrderedCollection.
	self
		assert: #(9 7 5 3) asOrderedCollection
		equals: (o overlappingPairsCollect: [ :a :b | a + b ])! !

!OrderedCollectionTest methodsFor: '*worlds-testsRemoving-override' stamp: 'ael 6/17/2011 17:55'!
testRemoveAll
	"Allows one to remove all elements of a collection" 
	
	| c1 c2 s2 |
	c1 := #(2 3 4 6) as: self targetClass.
	c1 addAll: (1 to: 200).
	c2 := c1 copy.
	s2 := c2 size.
	
	c1 removeAll.
	
	self assert: c1 size = 0.
	self assert: c2 size = s2 description: 'the copy has not been modified'
	! !

!OrderedCollectionTest methodsFor: '*worlds-testsRemoving-override' stamp: 'ael 6/17/2011 17:55'!
testRemoveAllSuchThat
	| collection |
	collection := (1 to: 10) as: self targetClass.
	collection
		removeAllSuchThat: [:e | e even].
	self assert: collection = ((1 to: 10 by: 2) as: self targetClass)! !

!OrderedCollectionTest methodsFor: '*worlds-testsRemoving-override' stamp: 'ael 6/17/2011 17:55'!
testRemoveAt
	"Allows one to remove an element from a collection at an index"
	"self run:#testRemoveAt" 
	
	| c1 |
	c1 := #(2 3 4 6) as: self targetClass.
	c1 removeAt: 2.
	self assert: (c1 = (#(2 4 6) as: self targetClass)).
	self should: [c1 removeAt: 10] raise: Error.
	self should: [c1 removeAt: -1] raise: Error.
	! !

!OrderedCollectionTest methodsFor: '*worlds-testsRemoving-override' stamp: 'ael 6/17/2011 17:55'!
testRemoveFirst
	"Allows one to remove n element of a collection at the first"
	"self run:#testRemoveFirst" 
	
	| c1 |
	c1 := #(2 3 4 6) as: self targetClass.
	c1 removeFirst: 1.
	self assert: (c1 = (#(3 4 6) as: self targetClass)).
	c1 removeFirst: 2.
	self assert: (c1 = (#(6) as: self targetClass)).
	self should: [c1 removeFirst: 10] raise: Error.
	
	! !

!OrderedCollectionTest methodsFor: '*worlds-testsRemoving-override' stamp: 'ael 6/17/2011 17:56'!
testRemoveIfAbsent
	"Allows one to remove an element from a collection and to copy it in another collection."
	"If the element isn't in the first collection, the second collection copy the element after ifAbsent"
	"self run:#testRemoveIfAbsent"
	
	| c1 c2 |
	c1 := #(1 2 3 4) as: self targetClass.
	c2 := self targetClass new.
	
	c2 add: (c1 remove: 2 ifAbsent: [6]).
	self assert: (c1 = (#(1 3 4) as: self targetClass)).
	self assert: (c2 = (#(2) as: self targetClass)).
	
	c2 add: (c1 remove: 18 ifAbsent: [6]).
	self assert: (c1 = (#(1 3 4) as: self targetClass)).
	self assert: (c2 = (#(2 6) as: self targetClass)).! !

!OrderedCollectionTest methodsFor: '*worlds-testsRemoving-override' stamp: 'ael 6/17/2011 17:56'!
testRemoveLast
	"Allows one to remove n element of a collection at the end"
	"self run:#testRemoveLast" 
	
	| c1 |
	c1 := #(2 3 4 6) as: self targetClass.
	c1 removeLast: 1.
	self assert: (c1 = (#(2 3 4) as: self targetClass)).
	c1 removeLast: 2.
	self assert: (c1 = (#(2) as: self targetClass)).
	self should: [c1 removeLast: 10] raise: Error.! !

!OrderedCollectionTest methodsFor: '*worlds-testsCopying-override' stamp: 'ael 6/17/2011 17:54'!
testReversed
	| collection1 collection2 |
	collection1 := #('Jim' 'Mary' 'John' 'Andrew' ) as: self targetClass.
	collection2 := collection1 reversed.
	self assert: collection2 first = 'Andrew'.
	self assert: collection2 last = 'Jim'! !

!OrderedCollectionTest methodsFor: '*worlds-testsAccessing-override' stamp: 'ael 6/17/2011 17:50'!
testSize
	"Allows one to check the size of an Ordered collection"
	"self run:#testSize"
	
	| c1 c2 |
	c1 := #(1 2 ) as: self targetClass.
	self assert: (c1 size =  2).
	
	c2 := self targetClass new.
	self assert: (c2 size = 0)
	! !

!OrderedCollectionTest methodsFor: '*worlds-testStreaming-override' stamp: 'ael 6/17/2011 17:58'!
testStreamContents
	| oc |
	self shouldnt: [oc := self targetClass streamContents: [:stream | 1 to: 3 do: [:i | stream nextPut: i]]]
		raise: Error
		description: 'it should be possible to stream over an OrderedCollection'.
	self assert: oc class = self targetClass.
	self assert: oc = ((1 to: 3) as: self targetClass).! !

!OrderedCollectionTest methodsFor: '*worlds-testsEnumerating-override' stamp: 'ael 6/17/2011 17:51'!
testWithCollect
	"Allows one to collect some element of two collections into another collection with element corresponding to the condition in the blocks"
	"self run: #testWithCollect"
	
	| c1 c2 res |
	c1 := #(-1 2 -3 4 -5 6 -7 8) as: self targetClass.
	c2 := #(-9 10 -11 12 -13 14 -15 16) as: self targetClass.
	res := c1 with: c2 collect: [:each1 :each2 | each1 < each2
		ifTrue: [each1]
		ifFalse: [each2]].
	self assert: (res = (#(-9 2 -11 4 -13 6 -15 8) as: self targetClass)).
	! !

OrderedCollectionTest subclass: #WCollectionTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Worlds-CollectionsTests'!

!WCollectionTest class methodsFor: 'as yet unclassified' stamp: 'ael 7/12/2011 16:51'!
shouldInheritSelectors
	^true
! !

!WCollectionTest methodsFor: 'testsBasic' stamp: 'ael 9/25/2011 21:36'!
testAddRemove1
	| l w w1 w2 |
	l := #(a b) as: self targetClass.
	w := self thisWorld sprout.
	w eval: [ l addFirst: #z; remove: #a; remove: #b ].
	self shouldnt: [ w commit ] raise: WorldSerializationError.
	self assert: ( l = (#(z) as: self targetClass) ).
	
	l := #(a b) as: self targetClass.
	w1 := self thisWorld sprout.
	w2 := w1 sprout.
	w2 eval: [ l addFirst: #z; remove: #a; remove: #b ].
	self shouldnt: [ w2 commit ] raise: WorldSerializationError.
	self assert: (w1 eval: [ l basicContents ]) = #(z) asOrderedCollection.
	self assert: l = (#(a b) as: self targetClass).
	self shouldnt: [ w1 commit ] raise: WorldSerializationError.
	self assert: l = (#(z) as: self targetClass).

	
! !

!WCollectionTest methodsFor: 'testsBasic' stamp: 'ael 9/25/2011 21:36'!
testAddRemove2
	| l w w1 w2 |
	l := #(a b) as: self targetClass.
	w := self thisWorld sprout.
	w eval: [ l addFirst: #z; remove: #b ].
	self shouldnt: [ w commit ] raise: WorldSerializationError.
	self assert: ( l = (#(z a) as: self targetClass) ).
	
	"and now with nested worlds"
	l := #(a b) as: self targetClass.
	w1 := self thisWorld sprout.
	w2 := w1 sprout.
	w2 eval: [ l addFirst: #z; remove: #b ].
	self shouldnt: [ w2 commit ] raise: WorldSerializationError.
	self assert: (w1 eval: [ l basicContents ]) = #(z a) asOrderedCollection.
	self assert: l = (#(a b) as: self targetClass).
	self shouldnt: [ w1 commit ] raise: WorldSerializationError.
	self assert: l = (#(z a) as: self targetClass).

! !

!WCollectionTest methodsFor: 'testsBasic' stamp: 'ael 12/6/2011 15:14'!
testAddRemove3
	| l w w1 w2 |
	l := #(a b) as: self targetClass.
	w := self thisWorld sprout.
	w eval: [ l addFirst: #z; remove: #a ].
	l remove: #b; add: #c.
	"should fail, because both a and b were read but now b has gone"
	self should: [ w commit ] raise: WorldValueChanged.
	
	l := #(a b) as: self targetClass.
	w := self thisWorld sprout.
	w eval: [ l removeLast; add: #z ].
	l at: 1 put: #c.
	"should succeed, because only b was examined and it's still there"
	self shouldnt: [ w commit ] raise: WorldSerializationError.
	
	"and now with nested worlds"
	l := #(a b) as: self targetClass.
	w1 := self thisWorld sprout.
	w2 := w1 sprout.
	w2 eval: [ l removeLast; add: #z ].		"w2 has #(a z), has read b(2) and size"
	l at: 1 put: #c.							"top has #(c b); w1 has yet to access it"
	"should succeed, because contents b(2) read in w2 will still look the same from w1" 
	self shouldnt: [ w2 commit ] raise: WorldSerializationError.
	self assert: (w1 eval: [ l basicContents ]) = #(c z) asOrderedCollection.
	self assert: l = (#(c b) as: self targetClass).
	"should succeed, because nothing has been read in w1; just an overwrite from w2"
	self shouldnt: [ w1 commit ] raise: WorldSerializationError.

! !

!WCollectionTest methodsFor: 'testsBasic' stamp: 'ael 12/6/2011 15:14'!
testAddRemove4
	| l w w1 w2 |
	"replacing elements without reading them"
	l := #(a b) as: self targetClass.
	w := self thisWorld sprout.
	w eval: [ l at: 1 put: #d; at: 2 put: #e ].
	l add: #c.
	self shouldnt: [ w commit ] raise: WorldSerializationError.
	self assert: l = (#(d e c) as: self targetClass).
	
	"and now with nested worlds"
	l := #(a b) as: self targetClass.
	w1 := self thisWorld sprout.
	w2 := w1 sprout.
	w2 eval: [ l at: 1 put: #d; at: 2 put: #e ].		"w2 has #(d e)"
	l add: #c.									"top has #(a b c); w1 has yet to look"
	self assert: (w1 eval: [ l basicContents ]) = #(a b c) asOrderedCollection.
	self shouldnt: [ w2 commit ] raise: WorldSerializationError.
	self assert: (w1 eval: [ l basicContents ]) = #(d e c) asOrderedCollection.
	self assert: l = (#(a b c) as: self targetClass).
	self shouldnt: [ w1 commit ] raise: WorldSerializationError.
	self assert: l = (#(d e c) as: self targetClass).

	"and now with changes in two levels of world"
	l := #(a b) as: self targetClass.
	w1 := self thisWorld sprout.
	w2 := w1 sprout.
	w2 eval: [ l at: 1 put: #d; at: 2 put: #e ].		"w2 has #(d e), but has not read anything"
	l add: #c.									"top has #(a b c)"
	w1 eval: [ l add: #f ].						"w1 has #(a b c f), and has noted size = 3"
	self assert: (w1 eval: [ l basicContents ]) = #(a b c f) asOrderedCollection.
	self shouldnt: [ w2 commit ] raise: WorldSerializationError.	"nothing was read from w1"
	self assert: (w1 eval: [ l basicContents ]) = #(d e c f) asOrderedCollection.
	l add: #g.
	self assert: l = (#(a b c g) as: self targetClass).

	"WCollection does more checking wrt size"
	self targetClass == WCollection
	  ifTrue: [ self should: [ w1 commit ] raise: WorldSizeChanged.	"size in top has changed" ].
		
	l remove: #a.
	self shouldnt: [ w1 commit ] raise: WorldSerializationError.	"should be ok now"
	self assert: l = (#(d e g f) as: self targetClass).
! !

!WCollectionTest methodsFor: 'testsBasic' stamp: 'ael 12/6/2011 15:15'!
testAddRemove5
	| l w w1 w2 |
	"search for an item involves reading the list up to that point"
	l := #(a b c) as: self targetClass.
	w := self thisWorld sprout.
	w eval: [ l remove: #c ].				"reads #a, #b and #c"
	l at: 1 put: #e.
	self should: [ w commit ] raise: WorldValueChanged.

	"if you don't search, you aren't reading"
	l := #(a b c) as: self targetClass.
	w := self thisWorld sprout.
	w eval: [ l removeLast ].				"only reads #c"
	l at: 1 put: #e.
	self shouldnt: [ w commit ] raise: WorldSerializationError.
	self assert: l = (#(e b) as: self targetClass).

	"similarly..."
	l := #(a b c) as: self targetClass.
	w := self thisWorld sprout.
	w eval: [ l removeAt: 2; removeAt: 2 ].	"reads #b and #c (when shifting it)"
	l at: 1 put: #e.
	self shouldnt: [ w commit ] raise: WorldSerializationError.
	self assert: l = (#(e) as: self targetClass).

	"fails because a shifted element is marked as read" 
	l := #(a b c) as: self targetClass.
	w := self thisWorld sprout.
	w eval: [ l removeAt: 2 ].	"reads #b and #c (when shifting it)"
	l at: 3 put: #e.
	self should: [ w commit ] raise: WorldValueChanged.

	
	"and now with nested worlds"
	l := #(a b c) as: self targetClass.
	w1 := self thisWorld sprout.
	w2 := w1 sprout.
	w2 eval: [ l remove: #c ].				"reads #a, #b and #c; now w2 has #(a b)"
	l at: 1 put: #e.							"top has #(e b c); w1 has not yet read"
	self should: [ w2 commit ] raise: WorldValueChanged.	"expecting a@1 but sees e"
	w1 eval: [ l at: 1 put: #a ].				"read from top, then replace a"
	self shouldnt: [ w2 commit ] raise: WorldSerializationError.	"fixed?"
	self assert: (w1 eval: [ l basicContents ]) = #(a b) asOrderedCollection.
	self assert: l = (#(e b c) as: self targetClass).
	self shouldnt: [ w1 commit ] raise: WorldSerializationError.	"w1 hasn't read anything"
	self assert: l = (#(a b) as: self targetClass).

! !

!WCollectionTest methodsFor: 'testsBasic' stamp: 'ael 10/30/2011 22:51'!
testCopy
	| l m w w1 w2 |
	l := #(a b) as: self targetClass.
	w := self thisWorld sprout.
	w eval: [ l addFirst: #z. m := l copy. l remove: #a; remove: #b ].
	self shouldnt: [ w commit ] raise: WorldSerializationError.
	self assert: ( l = (#(z) as: self targetClass) ).
	self assert: ( m = (#(z a b) as: self targetClass) ).
	
	"now two levels deep"
	l := #(a b) as: self targetClass.
	w1 := self thisWorld sprout.
	w2 := w1 sprout.
	w2 eval: [ l addFirst: #z. m := l copy. l remove: #a; remove: #b ].
	self assert: m = (#(z a b) as: self targetClass).
	self shouldnt: [ w2 commit ] raise: WorldSerializationError.
	self assert: (w1 eval: [ l basicContents ]) = #(z) asOrderedCollection.
	self assert: (w1 eval: [ m basicContents ]) = #(z a b) asOrderedCollection.
	self assert: l = (#(a b) as: self targetClass).
	self shouldnt: [ w1 commit ] raise: WorldSerializationError.
	self assert: l = (#(z) as: self targetClass).
	self assert: m = (#(z a b) as: self targetClass).

	
! !

!WCollectionTest methodsFor: 'testsBasic' stamp: 'ael 10/31/2011 22:28'!
testFlatten
	"WCollectionTest new testFlatten"
	| w1 w2 c w3 |
	c := #(a b c) as: self targetClass.
	w1 := self thisWorld sprout.
	w2 := w1 sprout.
	w3 := w2 sprout.
	w1 eval: [ c at: 3 put: #u; add: #v; add: #w ].
	w3 eval: [ c at: 1 put: #e; at: 3 put: #g; removeLast; removeLast ].
	self assert: (c worldFlattenIn: self thisWorld) basicContents = #(a b c) asOrderedCollection.
	self assert: (c worldFlattenIn: w1) basicContents = #(a b u v w) asOrderedCollection.
	self assert: (c worldFlattenIn: w2) basicContents = #(a b u v w) asOrderedCollection.
	self assert: (c worldFlattenIn: w3) basicContents = #(e b g) asOrderedCollection.

	c := #(a b c d) as: self targetClass.
	w1 := self thisWorld sprout.
	w2 := w1 sprout.
	w3 := w2 sprout.
	w1 eval: [ c at: 3 put: #e ].
	w3 eval: [ c at: 4 put: #f ].
	self assert: (c worldFlattenIn: self thisWorld) basicContents = #(a b c d) asOrderedCollection.
	self assert: (c worldFlattenIn: w1) basicContents = #(a b e d) asOrderedCollection.
	self assert: (c worldFlattenIn: w2) basicContents = #(a b e d) asOrderedCollection.
	self assert: (c worldFlattenIn: w3) basicContents = #(a b e f) asOrderedCollection.
	w3 commit.
	self assert: (c worldFlattenIn: w2) basicContents = #(a b e f) asOrderedCollection.
	w2 commit.
	self assert: (c worldFlattenIn: w1) basicContents = #(a b e f) asOrderedCollection.
	w1 commit.
	self assert: c basicContents = #(a b e f) asOrderedCollection.
	
	c := #(a b c d) as: self targetClass.
	w1 := self thisWorld sprout.
	w2 := w1 sprout.
	w3 := w2 sprout.
	w1 eval: [ c at: 2 put: #e ].
	w3 eval: [ c removeLast; removeLast ].
	self assert: (c worldFlattenIn: self thisWorld) basicContents = #(a b c d) asOrderedCollection.
	self assert: (c worldFlattenIn: w1) basicContents = #(a e c d) asOrderedCollection.
	self assert: (c worldFlattenIn: w2) basicContents = #(a e c d) asOrderedCollection.
	self assert: (c worldFlattenIn: w3) basicContents = #(a e) asOrderedCollection.
	w3 commit.
	self assert: (c worldFlattenIn: w2) basicContents = #(a e) asOrderedCollection.
	w2 commit.
	self assert: (c worldFlattenIn: w1) basicContents = #(a e) asOrderedCollection.
	w1 commit.
	self assert: c basicContents = #(a e) asOrderedCollection.

	c := #(a b c d) as: self targetClass.
	w1 := self thisWorld sprout.
	w2 := w1 sprout.
	w3 := w2 sprout.
	w3 eval: [ c at: 2 put: #e ].
	w1 eval: [ c removeLast; removeLast ].
	self assert: (c worldFlattenIn: self thisWorld) basicContents = #(a b c d) asOrderedCollection.
	self assert: (c worldFlattenIn: w1) basicContents = #(a b) asOrderedCollection.
	self assert: (c worldFlattenIn: w2) basicContents = #(a b) asOrderedCollection.
	self assert: (c worldFlattenIn: w3) basicContents = #(a e) asOrderedCollection.
	w3 commit.
	self assert: (c worldFlattenIn: w2) basicContents = #(a e) asOrderedCollection.
	w2 commit.
	self assert: (c worldFlattenIn: w1) basicContents = #(a e) asOrderedCollection.
	w1 commit.
	self assert: c basicContents = #(a e) asOrderedCollection.

	c := #(a b) as: self targetClass.
	w1 := self thisWorld sprout.
	w2 := w1 sprout.
	w3 := w2 sprout.
	w1 eval: [ c at: 2 put: #c; add: #d ].
	w3 eval: [ c at: 3 put: #e; add: #f ].
	self assert: (c worldFlattenIn: self thisWorld) basicContents = #(a b) asOrderedCollection.
	self assert: (c worldFlattenIn: w1) basicContents = #(a c d) asOrderedCollection.
	self assert: (c worldFlattenIn: w2) basicContents = #(a c d) asOrderedCollection.
	self assert: (c worldFlattenIn: w3) basicContents = #(a c e f) asOrderedCollection.
	w3 commit.
	self assert: (c worldFlattenIn: w2) basicContents = #(a c e f) asOrderedCollection.
	w2 commit.
	self assert: (c worldFlattenIn: w1) basicContents = #(a c e f) asOrderedCollection.
	w1 commit.
	self assert: c basicContents = #(a c e f) asOrderedCollection.

	c := #(a b) as: self targetClass.
	w1 := self thisWorld sprout.
	w2 := w1 sprout.
	w3 := w2 sprout.
	w1 eval: [ c add: #x; removeLast; at: 2 put: #c ].
	w3 eval: [ c add: #e; add: #f ].
	self assert: (c worldFlattenIn: self thisWorld) basicContents = #(a b) asOrderedCollection.
	self assert: (c worldFlattenIn: w1) basicContents = #(a c) asOrderedCollection.
	self assert: (c worldFlattenIn: w2) basicContents = #(a c) asOrderedCollection.
	self assert: (c worldFlattenIn: w3) basicContents = #(a c e f) asOrderedCollection.
	w3 commit.
	self assert: (c worldFlattenIn: w2) basicContents = #(a c e f) asOrderedCollection.
	w2 commit.
	self assert: (c worldFlattenIn: w1) basicContents = #(a c e f) asOrderedCollection.
	w1 commit.
	self assert: c basicContents = #(a c e f) asOrderedCollection.
! !

!WCollectionTest methodsFor: 'testsBasic' stamp: 'ael 12/6/2011 15:17'!
testIncludes
	"if an includes: check fails, commit cannot go ahead if the parent includes that item."

	"first the case where the check succeeds"
	| l w w1 w2 |
	l := #(a b c c d) as: self targetClass.
	w := self thisWorld sprout.
	w eval: [ l includes: #c ].
	l at: 1 put: #e.
	self shouldnt: [ w commit ] raise: WorldSerializationError.

	"if the element where the value was found gets overwritten, commit should fail (even if the answer would still be the same).  our commit test is safe by being pessimistic."
	l := #(a b c c d) as: self targetClass.
	w := self thisWorld sprout.
	w eval: [ l includes: #c ].
	l at: 3 put: #e.
	self should: [ w commit ] raise: WorldValueChanged.

	"check what happens if the value is found at another element"
	l := #(a b c c d) as: self targetClass.
	w := self thisWorld sprout.
	w eval: [ l at: 3 put: #f.  l includes: #c ].
	l at: 3 put: #e.
	self shouldnt: [ w commit ] raise: WorldSerializationError.

	"if the item is not found, commit should fail even if the item later turns up"
	l := #(a b c c d) as: self targetClass.
	w := self thisWorld sprout.
	w eval: [ l includes: #e ].
	l at: 1 put: #e.
	self should: [ w commit ] raise: WorldValueChanged.

	"...as a side effect of which, any item change is going to be rejected (highly pessimistic)"
	l := #(a b c c d) as: self targetClass.
	w := self thisWorld sprout.
	w eval: [ l includes: #e ].
	l at: 1 put: #f.
	self should: [ w commit ] raise: WorldValueChanged.

	"all the same checks but with nested worlds..."

	l := #(a b c c d) as: self targetClass.
	w1 := self thisWorld sprout.
	w1 eval: [
		w2 := self thisWorld sprout.
		w2 eval: [ l includes: #c ].
		].
	l at: 1 put: #e.
	self shouldnt: [ w2 commit ] raise: WorldSerializationError.
	self shouldnt: [ w1 commit ] raise: WorldSerializationError.
	
	l := #(a b c c d) as: self targetClass.
	w1 := self thisWorld sprout.
	w1 eval: [
		w2 := self thisWorld sprout.
		w2 eval: [ l includes: #c ].
		].
	l at: 3 put: #e.
	self should: [ w2 commit ] raise: WorldValueChanged.

	l := #(a b c c d) as: self targetClass.
	w1 := self thisWorld sprout.
	w1 eval: [
		w2 := self thisWorld sprout.
		w2 eval: [ l at: 3 put: #f.  l includes: #c ].
		].
	l at: 3 put: #e.
	self shouldnt: [ w2 commit ] raise: WorldSerializationError.
	self assert: (w1 eval: [ l basicContents ]) = #(a b f c d) asOrderedCollection.
	self shouldnt: [ w1 commit ] raise: WorldSerializationError.
	self assert: (l basicContents = #(a b f c d) asOrderedCollection).

 	l := #(a b c c d) as: self targetClass.
	w1 := self thisWorld sprout.
	w1 eval: [
		w2 := self thisWorld sprout.
		w2 eval: [ l includes: #e ].		"false, so w2 has read everything"
		].
	l at: 1 put: #e.
	self should: [ w2 commit ] raise: WorldValueChanged.

 	l := #(a b c c d) as: self targetClass.
	w1 := self thisWorld sprout.
	w1 eval: [
		w2 := self thisWorld sprout.
		w2 eval: [ l includes: #e ].
		].
	l at: 1 put: #f.						"makes no difference, but..."
	self should: [ w2 commit ] raise: WorldValueChanged.	"sadly"
! !

!WCollectionTest methodsFor: 'testsBasic' stamp: 'ael 9/25/2011 21:38'!
testNoChange
	| l w w1 w2 |
	"first case: access collection contents, not size"
	l := #(a b c d) as: self targetClass.
	w := self thisWorld sprout.
	w eval: [ l at: 1; at: 2; at: 3; at: 4 ].			"access only"
	self shouldnt: [ w commit ] raise: Error.
	self assert: (l =  (#(a b c d) as: self targetClass)).

	"same but in a nested world"
	l := #(a b c d) as: self targetClass.
	w1 := self thisWorld sprout.
	w1 eval: [
		w2 := self thisWorld sprout.
		w2 eval: [l at: 1; at: 2; at: 3; at: 4 ].			"access only"
		].
	self shouldnt: [ w2 commit ] raise: Error.
	self shouldnt: [ w1 commit ] raise: Error.
	self assert: (l =  (#(a b c d) as: self targetClass)).

	"same but without committing the innermost world"
	l := #(a b c d) as: self targetClass.
	w1 := self thisWorld sprout.
	w1 eval: [
		w2 := self thisWorld sprout.
		w2 eval: [ l at: 1; at: 2; at: 3; at: 4 ].			"access only"
		].
	self shouldnt: [ w1 commit ] raise: Error.
	self assert: (l =  (#(a b c d) as: self targetClass)).

	"now access collection size only"
	l := #(a b c d) as: self targetClass.
	w := self thisWorld sprout.
	w eval: [ l size ].			"access the size"
	self shouldnt: [ w commit ] raise: Error.
	self assert: (l =  (#(a b c d) as: self targetClass)).

	"and now in a nested world"
	l := #(a b c d) as: self targetClass.
	w1 := self thisWorld sprout.
	w1 eval: [
		w2 := self thisWorld sprout.
		w2 eval: [ l size ].			"access size only"
		].
	self shouldnt: [ w2 commit ] raise: Error.
	self shouldnt: [ w1 commit ] raise: Error.
	self assert: (l =  (#(a b c d) as: self targetClass)).

	"without committing innermost world"
	l := #(a b c d) as: self targetClass.
	w1 := self thisWorld sprout.
	w1 eval: [
		w2 := self thisWorld sprout.
		w2 eval: [ l size ].			"access size only"
		].
	self shouldnt: [ w1 commit ] raise: Error.
	self assert: (l =  (#(a b c d) as: self targetClass)).
! !

!WCollectionTest methodsFor: 'testsBasic' stamp: 'ael 11/1/2011 21:10'!
testParentIndependence
	| l m w1 w2 k n |
	k := self targetClass new.
	l := #(a b) as: self targetClass.
	w1 := self thisWorld sprout.
	w1 tracksReads: false.

	w2 := w1 sprout.
	w2 eval: [
		k includes: #f.
		m := l copy.
		m at: 2 put: #c.
		n := self targetClass new.
		].

	self assert: ((w1 eval: [ k debugUpdateTrackerState ]) isNil).
	self assert: ((w1 eval: [ l debugUpdateTrackerState ]) isNil).
	self assert: ((w1 eval: [ m debugUpdateTrackerState ]) isNil).
	self assert: ((w1 eval: [ n debugUpdateTrackerState ]) isNil).

	w2 commit.
	
	self assert: ((w1 eval: [ k debugUpdateTrackerState ]) isNil).	"only size was accessed"
	self assert: ((w1 eval: [ l debugUpdateTrackerState ]) isNil).		"nothing accessed"

	"in w1 we've recorded a write to 2 only"
	self targetClass
	  caseOf: {
		[ WCollection ]
		  -> [ self assert: ((w1 eval: [ m debugUpdateTrackerState at: 3 ]) = #(2)) ].
		[ WSimpleOrderedCollection ]
		  -> [ self assert: (w1 eval: [ m debugUpdateTrackerState at: 3 ]) keys asArray = #(2)].
		}.
 
	self assert: ((w1 eval: [ n debugUpdateTrackerState ]) isNil).	"nothing accessed yet"

	n add: #p; add: #q.
	
	w1 eval: [ k add: #d.  m includes: #g. ].

	self assert: ((w1 eval: [ n size ]) = 2).
	self assert: (k size = 0).

	w1 commit.
! !

!WCollectionTest methodsFor: 'testsBasic' stamp: 'ael 12/6/2011 15:19'!
testSizeChange
	"commit should fail if the child has accessed size and the size in the parent has changed"
	| l w w1 w2 |
	"WCollection does more checking wrt size"
	self targetClass == WCollection ifFalse: [ ^self ].
	
	l := #() as: self targetClass.
	w := self thisWorld sprout.
	w eval: [ l size ].			"access size only"
	l add: #a..
	self should: [ w commit ] raise: WorldSizeChanged.

	"similarly..."
	l := #(a b c d) as: self targetClass.
	w := self thisWorld sprout.
	w eval: [ l size ].			"access size only"
	l removeLast.
	self should: [ w commit ] raise: WorldSizeChanged.

	"same but with nested worlds"
	l := #(a b c d) as: self targetClass.
	w1 := self thisWorld sprout.
	w1 eval: [
		w2 := self thisWorld sprout.
		w2 eval: [ l size ].			"access size only"
		].
	l add: #e.
	self should: [ w2 commit ] raise: WorldSizeChanged.
	l remove: #c.
	self shouldnt: [ w2 commit ] raise: WorldSerializationError.
	self shouldnt: [ w1 commit ] raise: WorldSerializationError.
	
	"commit should succeed if the parent size returns to where it was before"
	l := #(a b c d) as: self targetClass.
	w := self thisWorld sprout.
	w eval: [ l size ].			"access size only"
	l removeLast; add: #e.
	self shouldnt: [ w commit ] raise: WorldSerializationError.
	self assert: l = (#(a b c e) as: self targetClass).

	"even if parent size changes, commit should succeed if the child has not accessed size, and has only accessed elements that are still in the parent"
	l := #(a b c) as: self targetClass.
	w := self thisWorld sprout.
	w eval: [ l at: 1; at: 2 ].			"access two elements only"
	l removeLast.
	self shouldnt: [ w commit ] raise: WorldSerializationError.
	self assert: (l = (#(a b) as: self targetClass)).
	
	"commit should fail if the child has accessed members beyond the size in the parent at commit"
	l := #(a b c) as: self targetClass.
	w := self thisWorld sprout.
	w eval: [ l at: 1; at: 2; at: 3 ].			"access contents only"
	l removeLast.
	self should: [ w commit ] raise: WorldSerializationError.
! !

!WCollectionTest methodsFor: 'testsBasic' stamp: 'sn 9/10/2013 03:10'!
testValueOrIdentity
	| l w |
	"search for an item involves reading the list up to that point.  replace one with a different number."
	l := {1. 2. 3.} as: self targetClass.
	w := (self thisWorld) sprout.

	w eval: [ l remove: 3 ].
	l at: 1 put: (2 - 2).
	self should: [ w commit ] raise: WorldValueChanged.

	"replace with the same number, and everything should go ok"
	l := #(1 2 3) as: self targetClass.
	w := self thisWorld sprout.
	w eval: [ l remove: 3 ].
	l at: 1 put: (2 - 1).
	self shouldnt: [ w commit ] raise: WorldSerializationError.
	self assert: l = (#(1 2) as: self targetClass).

	"likewise with strings"
	l := #('foo' 'bar' 'baz') as: self targetClass.
	w := self thisWorld sprout.
	w eval: [ l remove: 'baz' ].
	l at: 1 put: ('foobar' first: 3).
	self shouldnt: [ w commit ] raise: WorldSerializationError.
	self assert: l = (#('foo' 'bar') as: self targetClass).

	"but for the moment collections are considered non-comparable"
	l := #(('foo' 'boggle') 'bar' 'baz') as: self targetClass.
	w := self thisWorld sprout.
	w eval: [ l remove: 'baz' ].
	l at: 1 put: ('foo boggle' findTokens: $ ) asArray.		"same contents, but a different object"
	self should: [ w commit ] raise: WorldValueChanged.
! !

!WCollectionTest methodsFor: 'testsBasic' stamp: 'ael 10/30/2011 22:57'!
testWorldCopy
	| l m w1 w2 |
	l := #(a b) as: self targetClass.
	w1 := self thisWorld sprout.
	w1 eval: [ l at: 1 put: #c ].
	self assert: ( l = (#(a b) as: self targetClass) ).

	w2 := w1 copy.
	
	w1 eval: [ m := l copy ].			"m is a flattened copy of l"
	self assert: (w1 eval: [ m basicContents ]) = #(c b) asOrderedCollection.

	w2 eval: [
		l at: 1 put: #d.
		m at: 2 put: #e.		"NB: m normally wouldn't be visible in w2 unless put into the top-level world by a commit of w1.  We're just pretending that something like that has happened." 
		].
	self assert: (w2 eval: [ m basicContents ]) = #(c e) asOrderedCollection.

	w1 commit.
	self assert: ( l = (#(c b) as: self targetClass) ).
	self assert: ( m = (#(c b) as: self targetClass) ).

	w2 commit.
	self assert: ( l = (#(d b) as: self targetClass) ).
	self assert: ( m = (#(c e) as: self targetClass) ).
! !

!WCollectionTest methodsFor: 'testsBasic' stamp: 'ael 2/29/2012 01:54'!
testWriteBeyondEnd
	| l w |
	"One of the conditions for serialisability is that all writes performed in the child world can be replicated on the collection in the parent world.  This is not true if there is any gap in writes beyond the end of the parent collection at its current length (e.g., current length is 2 and there's a write at 4).
	
	How this situation can arise:
		start with a long(ish) collection.
		in a child world, overwrite something near the end (but without checking the length).
		in the parent, shrink to a length shorter than, and non-contiguous with, the child write.
	
	For now, this serialisability check isn't implemented for WSimpleOrderedCollection"
	
	self targetClass == WSimpleOrderedCollection ifTrue: [ ^self ].
	
	l := #(a b c d) as: self targetClass.
	w := self thisWorld sprout.
	w eval: [ l at: 4 put: #e ].
	l removeLast: 2.
	self should: [ w commit ] raise: WorldSerializationError.	

	l := #(a b c d) as: self targetClass.
	w := self thisWorld sprout.
	w eval: [ l at: 4 put: #e ].
	l removeLast.
	self shouldnt: [ w commit ] raise: WorldSerializationError.	
! !

!WCollectionTest methodsFor: 'accessing' stamp: 'ael 9/25/2011 21:38'!
thisWorld
	^WWorld thisWorld! !

WCollectionTest subclass: #WSimpleOrderedCollectionTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Worlds-CollectionsTests'!

!Process methodsFor: '*worlds' stamp: 'ael 9/25/2011 21:34'!
currentWWorld

	^ currentWWorld ifNil: [WTopLevelWorld default]! !

!Process methodsFor: '*worlds' stamp: 'aw 4/5/2010 15:05'!
currentWWorld: w

	currentWWorld := w! !
WIdentityDictionaryTest initialize!
WObject initialize!
WWorld initialize!
WorldSpecialObject initialize!
